(***********************************************************************)
(*                                                                     *)
(*                                GALAX                                *)
(*                             XQuery Engine                           *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xquery_algebra_ast_util.ml,v 1.43 2006/05/16 23:33:59 simeon Exp $ *)

(* Module: Xquery_algebra_ast_util
   Description:
     This module implements some useful operations the XQuery Algebra
     AST.
*)

open Finfo

open Xquery_common_ast
open Xquery_algebra_ast

open Error


(*************************)
(* Algebra AST utilities *)
(*************************)

let fmkapattern apattern_desc fi =
    { papattern_desc = apattern_desc;
      papattern_loc = fi }

let fmkasequencetype c fi =
    { pasequencetype_desc = c;
      pasequencetype_loc  = fi }

let algop_mkop eval_code eval_sig op_name indep_expressions dep_expressions annot comp_annot e_handle expr_loc =
  { palgop_expr_name = op_name;
    palgop_expr_eval = ref eval_code;
    palgop_expr_eval_sig = eval_sig;
    annotation = annot;
    compile_annotations = comp_annot;
    psub_expression = indep_expressions;
    pdep_sub_expression = dep_expressions;
    palgop_expr_origin = e_handle;  (* Handle to the original expr *)
    palgop_expr_loc = expr_loc;}  

let algop_decl_mkop eval_code op_name indep_expressions dep_expressions annot expr_loc =
  { alg_decl_name 	  = op_name;
    alg_decl_eval 	  = ref eval_code;
    alg_decl_indep 	  = indep_expressions;
    alg_decl_dep   	  = dep_expressions;
    alg_decl_annotation   = annot;
    alg_decl_loc   	  = expr_loc }

let fmkalgop_function_decl desc fi =
  { palgop_function_decl_desc = desc;
    palgop_function_decl_loc = fi }

let fmkalgop_prolog fns vars indices =
  { palgop_prolog_functions = fns;
    palgop_prolog_vars      = vars;
    palgop_prolog_indices   = indices }

let fmkalgop_xmodule prolog statements = 
  { palgop_module_prolog = prolog;
    palgop_module_statements = statements }


(* A more specific function adding some "dummy" code *)

let dummy_code = None (* No code for now *)
let dummy_sig = None 

let aalgop_mkop op_name indep_expressions dep_expressions e_handle expr_loc =
  algop_mkop dummy_code dummy_sig op_name indep_expressions dep_expressions (ref None) e_handle expr_loc

let logical_aalgop_mkop op_name indep_expressions dep_expressions e_handle expr_loc =
  algop_mkop () dummy_sig op_name indep_expressions dep_expressions (ref None) e_handle expr_loc

let aalgop_decl_mkop op_name indep_expressions dep_expressions expr_loc =
  algop_decl_mkop dummy_code op_name indep_expressions dep_expressions (ref None) expr_loc

let logical_aalgop_decl_mkop op_name indep_expressions dep_expressions expr_loc =
  algop_decl_mkop () op_name indep_expressions dep_expressions (ref None) expr_loc

let replace_aalgop_name algop op_name =
  aalgop_mkop op_name algop.psub_expression algop.pdep_sub_expression None algop.palgop_expr_origin algop.palgop_expr_loc

let rec copy_algop op = 
  algop_mkop !(op.palgop_expr_eval) op.palgop_expr_eval_sig op.palgop_expr_name 
    (copy_subexpr op.psub_expression) (copy_subexpr op.pdep_sub_expression)
    op.annotation op.compile_annotations 
    op.palgop_expr_origin op.palgop_expr_loc

and copy_subexpr se = 
  match se with
      NoSub -> 
	NoSub
    | OneSub s ->
	OneSub (copy_algop s)
    | TwoSub (s0,s1) ->
	TwoSub ((copy_algop s0), (copy_algop s1))
    | ManySub sa ->
	ManySub (Array.map copy_algop sa)

(* Generate a fresh variable name in the 'fs:' namespace, and a
   corresponding variable algop. *)

let tempvar = ref 0

let gen_new_vname () = 
  let _ = tempvar := !tempvar + 1 in 
  "a"^(string_of_int !tempvar)

let mk_annotation fc bc af rt tfuc = 
    { use_counts             = fc;
      bound_usage_counts     = bc;
      accessed_fields        = af;
      returned_fields        = rt;
      tuple_field_use_counts = tfuc;
    } 

let variable_comp v = 
  Some (mk_annotation [(v,(1,Once))] [] [] [] ([], [], NoTable))
 
let gen_new_avar annot =
  let v = (Namespace_builtin.fs_prefix, Namespace_builtin.fs_uri, gen_new_vname()) in
  let v_op =
    algop_mkop dummy_code dummy_sig (AOEVar v) NoSub NoSub annot (variable_comp v) None Finfo.bogus
  in
  (v, v_op)

(* This is slightly abusive to return unit in both cases *)
let access_nosub x = 
  match x with
  | NoSub -> ()        
  | _ -> raise (Query (Malformed_Algebra_Expr("Incorrect Number of assumed sub expressions (NoSub)")))

let access_unitsub x =
  match x with
    | NoSub -> ()
    | _ -> raise (Query (Malformed_Algebra_Expr("Incorrect Number of assumed sub expressions (UnitSub)")))

let access_onesub x = 
  match x with
    | OneSub y -> y
    | _ -> raise (Query (Malformed_Algebra_Expr("Incorrect Number of assumed sub expressions (OneSub)")))

let access_twosub x = 
  match x with
    | TwoSub(x,y) -> (x,y) 
    | _ -> raise (Query (Malformed_Algebra_Expr ("Incorrect Number of assumed sub expressions (TwoSub)")))

let access_manysub x =
  match x with
    | ManySub y -> y
    | _ -> raise (Query (Malformed_Algebra_Expr ("Incorrect Number of assumed sub expressions (ManySub)")))

(*************************)
(* Operations on modules *)
(*************************)

let empty_prolog_plan =
    { palgop_prolog_functions = [];
      palgop_prolog_vars = [];
      palgop_prolog_indices = [] }

let merge_alg_prologs prolog1 prolog2 =
  { palgop_prolog_functions =
      prolog1.palgop_prolog_functions @ prolog2.palgop_prolog_functions;
    palgop_prolog_vars =
      prolog1.palgop_prolog_vars @ prolog2.palgop_prolog_vars;
    palgop_prolog_indices =
      prolog1.palgop_prolog_indices @ prolog2.palgop_prolog_indices }

let subexpr_to_list se = 
  match se with
    | NoSub         -> []
    | OneSub x      -> x :: []
    | TwoSub (x,y)  -> x :: y :: []
    | ManySub x     -> (Array.to_list x)

let get_group_names gd = 
  gd.group_by_names

let get_induced_group gd = 
  gd.induced_groups

let get_aggregate_name gd = 
  let odt, name = gd.aggregate_name in
    name

let get_aggregate_type gd = 
  let odt, name = gd.aggregate_name in
    odt

let get_aggregate_name_and_type gd =
  gd.aggregate_name

let get_valid_names gd =
  gd.must_be_valid

let mk_group_desc nms ind valid agg =
  { group_by_names = nms;
    induced_groups = ind;
    must_be_valid  = valid;
    aggregate_name = agg }

let split_main_module_plan xmod =
  (xmod.palgop_module_prolog,xmod.palgop_module_statements)


(**
 * Twig Pattern utils
 **)
let get_select_path p =
  let rec get_select_path_aux index =
    match p.(index).child_twig with
    | Some (a,ch) ->
	ch :: (get_select_path_aux ch)
    | None -> []
  in
  Array.of_list (get_select_path_aux 0)

let get_all_outputs_from_twig_pattern p =
  let get_output twignode =
    match twignode.out with 
    | Some field ->
	[field]
    | None -> 
	[]
  in
  let pattern_list = Array.to_list p in
  List.flatten (List.map get_output pattern_list)

let get_restored_outputs_from_twig_pattern p =
  let get_output twignode =
    match twignode.out with 
    | Some field ->
	if twignode.restore_out then [field] else []
    | None -> 
	[]
  in
  let pattern_list = Array.to_list p in
  List.flatten (List.map get_output pattern_list)


let get_node_tests_from_twig_pattern p =
  let get_nt n = 
    match n.node_test with
    | Some nt -> [nt]
    | None -> []
  in 
  List.flatten (List.map get_nt (Array.to_list p))

(* does not include predicates *)
let rec get_leaf_twig_node p index = 
  match p.(index).child_twig with
  | None -> p.(index)
  | Some (twig_type, twig_index) -> 
      get_leaf_twig_node p twig_index

(* does include predicates *)
let rec get_first_leaf_node_index p index =
  match p.(index).child_twig with
  | None -> 
      begin
	match  p.(index).pred_twigs with
	| (_, i) :: tl -> get_first_leaf_node_index p i
	| [] -> index
      end
  | Some (_, i) -> get_first_leaf_node_index p i

let increase_twig_index offset twignode =
  let _ = 
    match twignode.child_twig with
    | Some (twig_type, twig_index) ->
	twignode.child_twig <- Some (twig_type, twig_index + offset)
    | None -> ()
  in
  twignode.pred_twigs <- List.map 
      (fun (tt,ti) -> (tt, ti + offset)) 
      twignode.pred_twigs


let find_attach_point pattern vname =
  let index = ref (-1) in
  for i = 0 to Array.length pattern -1 do
    match pattern.(i).out with
    | Some o when o = vname -> index := i
    | _ -> ()
  done;
  if !index > 0 then
    !index
  else
     raise (Query (Malformed_Algebra_Expr ("Cannot find attachpoint for twig merge")))

let append_twigs p1 p2 vname =
  let leaf = p1.(find_attach_point p1 vname) in
(*  let _ = Array.iter (fun x -> x.out <- None) p1 in *)
  let _ = Array.iter (fun x -> x.restore_out <- false) p1 in 
  let offset = (Array.length p1) -1 in  

  (* 1. increase all indices in p2 with Array.length p1 *)
  let _ = Array.map (increase_twig_index offset) p2 in

  (* 2. add the list of corresponding indices to the twiglist of the leaf of p1 *)
  let _ =
    match leaf.child_twig with
    | None -> leaf.child_twig <- p2.(0).child_twig
    | Some _ -> raise (Query (Malformed_Algebra_Expr ("Attempt to add a second child twig (append_twigs).")))
  in
	
  (* 3. concatenate arrays p1 and p2 -- except for the root of p2 *) 
  let len =  (Array.length p2) -1 in
  let res = Array.append p1 (Array.sub p2 1 len) in
  res

let rec merge_twigs p1 p_list =
  match p_list with
  | (a2,p2) :: tl ->
      let i = find_attach_point p1 a2 in
      let offset = (Array.length p1) -1 in  
      let _ = Array.map (increase_twig_index offset) p2 in
      let _ = Array.map (fun tn -> tn.restore_out <- false) p2 in
      let p2_child_twig = 
	match p2.(0).child_twig with
	| Some n -> n
	| None -> raise (Query (Malformed_Algebra_Expr ("Attempt to merge an empty twig (merge_twigs).")))
      in
      let _ = p1.(i).pred_twigs <- p1.(i).pred_twigs @ [ p2_child_twig ] @ p2.(0).pred_twigs in
      let len =  (Array.length p2) -1 in
      let res = Array.append p1 (Array.sub p2 1 len) in
      merge_twigs res tl
  | [] -> p1


let is_leaf_node pattern node =
  match node.child_twig, node.pred_twigs with
  | None,[]  -> true
  | _    -> false

let is_root_node pattern node =
  match node.node_test with
  | Some _ -> false
  | None   -> true

let rec get_parent_node_index pattern node =
  begin
    let parent_index = ref (-1) in
    for i = 0 to (Array.length pattern) -1 do
      begin
	let match_twig (_, ind) =
	  if pattern.(ind) = node
	  then parent_index := i
	in
	(match pattern.(i).child_twig with
	| Some twig -> match_twig twig
	| None -> ();
	);
	let _ = List.map match_twig pattern.(i).pred_twigs in ()
      end
    done;
    !parent_index
  end

let get_child_node_indices pattern node =
  let ch =
    match node.child_twig with
    | Some (_, i) -> [i]
    | _ -> []
  in
  let get_ch_node (_, i) = i in
  ch @ List.map get_ch_node node.pred_twigs

let rec get_sub_node_indices pattern node =
  let children_indices = get_child_node_indices pattern node in
  let f lst child_index =
    let sub_indices = get_sub_node_indices pattern pattern.(child_index) in
    lst @ sub_indices
  in
  children_indices @ List.fold_left f [] children_indices
	

let replace_field_in_pattern p cur_name replacement_name =
  let changed = ref true in
  (* start at index one, since the first outfield is always empty *)
  for i = 1 to Array.length p -1 do
    match p.(i).out with
    | Some field 
      when field = cur_name ->
	begin
	  p.(i).out <- Some replacement_name;
	  changed := true
	end
    | _ -> ()
  done;
  !changed


(*
 * a treepattern is streamable when
 * - there are no branches (predicates)
 * - it contains only child, descendant or descentant-or-self twigs
 *)
let is_streamable_treepattern p =
  let is_streamable_twig_node node =
    if (List.length node.pred_twigs) = 0 then
      match node.child_twig with
      | Some (Descendant,_)
      | Some (Descendant_or_self, _)
      | Some (Child, _) 
      | None -> 
	  true
      | _ -> 
	  false
    else
      false
  in
  let fold_fun a b = a && b in
  Array.fold_left fold_fun true (Array.map is_streamable_twig_node p)

let is_single_step_twig p =
  Array.length p = 2

(* build axis array 
 * - Returns an array telling us for each corresponding 
 *   twig node how it was reached
 *)
let get_axis_array pattern =
  let axis_array = Array.create (Array.length pattern) Xquery_common_ast.Child  in

  let rec build_axis_array_aux i =
    let ch_t = match pattern.(i).child_twig with
               | Some t -> [t]
	       | None -> []
    in
    let twigs = ch_t @ pattern.(i).pred_twigs in
    List.iter
      (fun (ax,ind) ->  
	axis_array.(ind) <- ax; 
	build_axis_array_aux ind
      ) twigs
  in 

  build_axis_array_aux 0; 
  axis_array

let string_of_algop_expr_name algop_expr_name = 
  match algop_expr_name with
  | AOEIf 	-> "AOEIf"
  | AOELet _ 	-> "AOELet"
  | AOETypeswitch _ -> "AOETypeswitch"
  | AOEVar _ 	-> "AOEVar"
  | AOEScalar _ -> "AOEScalar"
  | AOESeq      -> "AOESeq"
  | AOEEmpty    -> "AOEEmpty"
  | AOEDocument -> "AOEDocument"
  | AOEPI _ 	-> "AOEPI"
  | AOEPIComputed  -> "AOEPIComputed"
  | AOEComment _   -> "AOEComment"
  | AOECommentComputed -> "AOECommentComputed"
  | AOEText _ 	    -> "AOEText"
  | AOETextComputed -> "AOETextComputed"
  | AOEElem _ 	    -> "AOEElem"
  | AOEAnyElem _    -> "AOEAnyElem"
  | AOEAttr _ 	    -> "AOEAttr"
  | AOEAnyAttr _    -> "AOEAnyAttr"
  | AOEError 	    -> "AOEError"
  | AOETreat _ 	    -> "AOETreat"
  | AOEValidate _   -> "AOEValidate"
  | AOECast _ 	    -> "AOECast"
  | AOECastable _   -> "AOECastable"
  | AOESome _ 	    -> "AOESome"
  | AOEEvery _ 	    -> "AOEEvery"
  | AOEInputTuple   -> "AOEInputTuple"
  | AOECallBuiltIn _-> "AOECallBuiltIn"
  | AOECallOverloaded _	 -> "AOECallOverloaded"
  | AOECallUserDefined _ -> "AOECallUserDefined"
  | AOEConvertSimple _ 	 -> "AOEConvertSimple"
  | AOEPromoteNumeric _  -> "AOEPromoteNumeric"
  | AOEUnsafePromoteNumeric _ -> "AOEUnsafePromoteNumeric"
  | AOEKeyRef _ 	-> "AOEKeyRef"
  | AOEExecute  	-> "AOEExecute"
  | AOECreateTuple _ 	-> "AOECreateTuple"
  | AOEAccessTuple _ 	-> "AOEAccessTuple"
  | AOEConcatTuples 	-> "AOEConcatTuples"
  | AOEMapFromItem _ 	-> "AOEMapFromItem"
  | AOEMapToItem    	-> "AOEMapToItem"
  | AOEMap           	-> "AOEMap"
  | AOENullMap _ 	-> "AOENullMap"
  | AOEMapIndex _ 	-> "AOEMapIndex"
  | AOEMapIndexStep _ 	-> "AOEMapIndexStep"
  | AOEMapConcat       	-> "AOEMapConcat"
  | AOEOuterMapConcat _ -> "AOEOuterMapConcat"
  | AOEMaterializeTable -> "AOEMaterializeTable"
  | AOEProduct          -> "AOEProduct"
  | AOESelect _ 	-> "AOESelect"
  | AOEJoin _ 	        -> "AOEJoin"
  | AOELeftOuterJoin _ 	-> "AOELeftOuterJoin"
  | AOEGroupBy _ 	-> "AOEGroupBy"
  | AOEOrderBy _ 	-> "AOEOrderBy"
  | AOECopy             -> "AOECopy"
  | AOEDelete           -> "AOEDelete"
  | AOEDetach           -> "AOEDetach"
  | AOEInsert _ 	-> "AOEInsert"
  | AOERename           -> "AOERename"
  | AOEReplace _ 	-> "AOEReplace"
  | AOESnap _ 	        -> "AOESnap"
  | AOEParse _ 	        -> "AOEParse"
  | AOETreeJoin _ 	-> "AOETreeJoin"
  | AOETupleTreePattern _-> "AOETupleTreePattern"
  | AOEDistinct _ 	-> "AOEDistinct"
  | AOEPrune _ 	        -> "AOEPrune"
  | AOEProject _        -> "AOEProject"
