(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xquery_core_ast_annotation.ml,v 1.11 2005/08/23 12:36:37 simeon Exp $ *)

(* Module: Xquery_core_ast_annotation
   Description:
     This module contains support for XQuery Core AST annotations.
 *)


open Error
open Xquery_type_core_ast

type xpath_sbdo_annot =
  | Xp_document_call
  | Xp_root_call
  | Xp_slash
  | Xp_axis_step
  | Xp_pred_axis_step
  | Xp_pos_pred_axis_step
  | Xp_ddo_call
  | Xp_flwor

type ddo_annot =
  | Needs_docorder
  | Needs_distinct
  | Needs_distinct_docorder
  | Needs_none

type free_var_annot = Xquery_common_ast.cvname list

(** If static typing is enabled, the core is annotated with types
resulting from the static type analysis **)
type type_annot = Xquery_type_core_ast.cxtype

type ast_annot =
    { mutable xpath_annot      : xpath_sbdo_annot option;(* Normalization *)
      mutable type_annot       : type_annot option;      (* Static typing *)
      mutable sbdo_optim_annot : ddo_annot option;       (* SBDO analysis *)
      mutable free_var_annot   : free_var_annot option;  (* free variable analysis *) }

(* Type annotations *)

let empty_ast_annot () = 
  { xpath_annot      = None;
    type_annot       = None;
    sbdo_optim_annot = None;
    free_var_annot   = None }

let annot_components ast_annot = 
  (ast_annot.type_annot,
   ast_annot.xpath_annot,
   ast_annot.sbdo_optim_annot,
   ast_annot.free_var_annot)

(** fixme: does this really make a copy of the annotation ?? **)
let copy_annot a =
  let (ta, xa, da, fva) = annot_components a in
  { type_annot       = ta;
    xpath_annot      = xa;
    sbdo_optim_annot = da;
    free_var_annot   = fva }

let set_type_annot ast_annot cxtype = 
  ast_annot.type_annot <- Some cxtype

let get_type_annot ast_annot = 
  match ast_annot.type_annot with
  | None ->
      raise (Query(Annotation_Error("In get_type_annot: Annotation does not contain type")))
  | Some cxtype -> cxtype


(* XPath annotations *)
let set_xpath_annot ast_annot xpannot =
  ast_annot.xpath_annot <- Some xpannot
  
let get_xpath_annot ast_annot =
  ast_annot.xpath_annot

let print_xpath_annot annot =
  match annot with
  | Xp_document_call -> "document_call"
  | Xp_root_call -> "root_call"
  | Xp_slash -> "slash"
  | Xp_axis_step -> "axis_step"
  | Xp_pred_axis_step -> "pred_axis_step"
  | Xp_pos_pred_axis_step -> "pos_pred_axis_step"
  | Xp_ddo_call -> "ddo_call"
  | Xp_flwor -> "XPath FLWOR"

(* DDO annotations *)
let set_ddo_annot ast_annot da =
  ast_annot.sbdo_optim_annot <- Some da
  
let get_ddo_annot ast_annot =
  ast_annot.sbdo_optim_annot
  
let print_ddo_annot annot =
  match annot with
      Needs_docorder -> "needs_order"
    | Needs_distinct -> "needs_distinct"
    | Needs_distinct_docorder -> "needs_distinct-docorder"
    | Needs_none -> "needs_none"

let set_free_var_annot ast_annot fv =
  ast_annot.free_var_annot <- Some fv

let get_free_var_annot ast_annot = 
  ast_annot.free_var_annot

let rec print_free_var_annot flist = 
  match flist with 
    | [] -> ""
    | x :: [] ->
	(Namespace_names.prefixed_string_of_rqname x) 
    | x :: rest -> 
	(Namespace_names.prefixed_string_of_rqname x) ^ "," ^ print_free_var_annot rest

(* Global setting of a given annotation *)

let set_annotation annot1 annot2 =
  annot1.type_annot <- annot2.type_annot;
  annot1.xpath_annot <- annot2.xpath_annot;
  annot1.sbdo_optim_annot <- annot2.sbdo_optim_annot;
  annot1.free_var_annot <- annot2.free_var_annot

