(***********************************************************************)
(*                                                                     *)
(*                                GALAX                                *)
(*                             XQuery Engine                           *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xquery_type_core_ast_util.ml,v 1.6 2006/03/02 17:40:49 mff Exp $ *)

(* Module: Xquery_type_core_ast_util
   Description:
     This module implements some useful operations the XQuery core
     type AST.
*)

open Error

open Namespace_symbols_util

open Xquery_common_ast
open Xquery_type_core_ast
open Xquery_type_core_ast_annotation

(**************************)
(* Core type AST creation *)
(**************************)

let fmkcxtype cxtype_desc finfo =
  { pcxtype_desc = cxtype_desc;
    pcxtype_simplified = false;
    pcxtype_loc  = finfo }

let fmkcxtype_builtin cxtype_desc =
  { pcxtype_desc = cxtype_desc;
    pcxtype_simplified = true;
    pcxtype_loc  = Finfo.bogus }

let fmkctype_decl sym deriv cxtype_opt = 
  { ctypedecl_name   = sym;
    ctypedecl_deriv  = deriv;
    ctypedecl_cxtype = cxtype_opt }

let rec equal_cxtypes t1 t2 = 
  match (t1.pcxtype_desc, t2.pcxtype_desc) with 
  | CDocument t1, CDocument t2 -> 
      equal_cxtypes t1 t2
  | CBound (t1, min1, max1), CBound (t2, min2, max2) ->
      equal_cxtypes t1 t2 && min1 = min2 && max1 = max2 
  | CSequence(u1, v1), CSequence(u2, v2) ->
      equal_cxtypes u1 u2 && equal_cxtypes v1 v2
  | CChoice(u1, v1), CChoice(u2, v2) ->
      equal_cxtypes u1 u2 && equal_cxtypes v1 v2
  | CInterleave(u1, v1), CInterleave(u2, v2) ->
      equal_cxtypes u1 u2 && equal_cxtypes v1 v2
  | _ -> t1.pcxtype_desc = t2.pcxtype_desc

(**************************)
(* Letter Mappings        *)
(**************************)
let check_caname_consistency caname (typeopt1) (typeopt2) =
  match (typeopt1,typeopt2) with
  | (None,None) -> ()
  | (Some ctname1, Some ctname2) ->
      if not(Namespace_symbols.symbol_equal ctname1 ctname2)
      then
	raise (Query (Schema_Internal ("Subtyping: [Attribute Declaration Constraint] not satisfied for "^(Namespace_symbols.symbol_prefix_string ctname1)^" and "^(Namespace_symbols.symbol_prefix_string ctname2))))
      else
	()
  | (Some ctname, _)
  | (_, Some ctname) ->
      raise (Query (Schema_Internal ("Subtyping: [Attribute Declaration Constraint] not satisfied for "^(Namespace_symbols.symbol_prefix_string ctname))))

let make_ca_letter_unit mapping (caname,typeopt) =
  if SQNameHashtbl.mem (snd mapping) caname
  then
    begin
      (* Why are we checking for sub-type consistency here? *)
      let (typeopt2,letternum) = SQNameHashtbl.find (snd mapping) caname in
      check_caname_consistency caname typeopt typeopt2
    end
  else
    begin
      incr (fst mapping);
      let letternum = !(fst mapping) in
if !Conf.debugging then print_string ("Map "^(Namespace_symbols.symbol_prefix_string caname)^"->"^(string_of_int letternum)^""); 
      SQNameHashtbl.add (snd mapping) caname (typeopt,letternum)
    end

let check_cename_consistency cename (typeopt1,nil1) (typeopt2,nil2) =
  if not(nil1 = nil2)
  then
    raise (Query (Schema_Internal "Subtyping: [Element Declaration Constraint] not satisfied for nil annotations"))
  else
    match (typeopt1,typeopt2) with
    | (None,None) -> ()
    | (Some ctname1, Some ctname2) ->
	if not(Namespace_symbols.symbol_equal ctname1 ctname2)
	then
	  raise (Query (Schema_Internal ("Subtyping: [Element Declaration Constraint] not satisfied for "^(Namespace_symbols.symbol_prefix_string ctname1)^" and "^(Namespace_symbols.symbol_prefix_string ctname2))))
	else
	  ()
    | (Some ctname, _)
    | (_, Some ctname) ->
	raise (Query (Schema_Internal ("Subtyping: [Element Declaration Constraint] not satisfied for "^(Namespace_symbols.symbol_prefix_string ctname))))

let make_ce_letter_unit mapping (cename,typeopt,nil) =
  if SQNameHashtbl.mem (snd mapping) cename
  then
    begin
      let (typeopt2,nil2,letternum) = SQNameHashtbl.find (snd mapping) cename in
      check_cename_consistency cename (typeopt,nil) (typeopt2,nil2)
    end
  else
    begin
      incr (fst mapping);
      let letternum = !(fst mapping) in
if !Conf.debugging then print_string ("Map "^(Namespace_symbols.symbol_prefix_string cename)^"->"^(string_of_int letternum)^""); 
      SQNameHashtbl.add (snd mapping) cename (typeopt,nil,letternum)
    end

let make_at_letter_unit mapping ctname =
  if SQNameHashtbl.mem (snd mapping) ctname
  then ()
  else
    begin
      incr (fst mapping);
      let letternum = !(fst mapping) in
if !Conf.debugging then print_string ("Map "^(Namespace_symbols.symbol_prefix_string ctname)^"->"^(string_of_int letternum)^"\n"); 
      SQNameHashtbl.add (snd mapping) ctname letternum
    end

let rec make_mappings mapping cxtype =
  match cxtype.pcxtype_desc with
  | CElementRef cename ->
(*      if not(Namespace_symbols.symbol_equal cename Namespace_symbols_builtin.wild_symbol) then  *)
	make_ce_letter_unit (elem_letter_map mapping) (cename,None,NonNillable)
  | CElementLocal (cename,nil,ctname) ->
(*      if not(Namespace_symbols.symbol_equal cename Namespace_symbols_builtin.wild_symbol) then  *)
	make_ce_letter_unit (elem_letter_map mapping) (cename,Some ctname,nil)
  | CAttributeRef caname ->
(*     if not(Namespace_symbols.symbol_equal caname Namespace_symbols_builtin.wild_symbol) then  *)
	make_ca_letter_unit (attr_letter_map mapping) (caname,None)
  | CAttributeLocal (caname,ctname) ->
(*      if not(Namespace_symbols.symbol_equal caname Namespace_symbols_builtin.wild_symbol) then  *)
	make_ca_letter_unit (attr_letter_map mapping) (caname,Some ctname)
  | CAtomicRef ctname ->
      make_at_letter_unit (type_letter_map mapping) ctname
  | CDocument m1 ->
      make_mappings mapping m1
  | CText
  | CPI
  | CComment
  | CEmpty
  | CNone -> ()
  | CBound (m1,_,_) ->
      make_mappings mapping m1
  | CSequence (m1,m2) ->
      make_mappings mapping m1;
      make_mappings mapping m2
  | CChoice (m1,m2) ->
      make_mappings mapping m1;
      make_mappings mapping m2
  | CInterleave (m1,m2) ->
      make_mappings mapping m1;
      make_mappings mapping m2

(* This function extracts cxtypes from a type declaration for the
   purpose of building letter mappings *)
let cxtypes_of_ctype_derivation rsym ctype_deriv =
  let (rsym_deriv_from, attr_content, children_content) = ctype_deriv in
  let child_types = 
    match children_content with
    | CComplexTypeRestriction (mxd, ty) -> [ty]
    | CComplexTypeExtension  (mxd, ty) -> [ty]
    | CAtomicTypeRestriction -> [fmkcxtype (CAtomicRef rsym) Finfo.bogus]
    | CSimpleTypeList sym -> [fmkcxtype (CAtomicRef sym) Finfo.bogus]
    | CSimpleTypeUnion sym_list -> 
	List.map (fun sym -> fmkcxtype (CAtomicRef sym) Finfo.bogus) sym_list
  in
  let attr_types = 
    match attr_content with 
    | None -> []
    | Some t -> [t]
  in
  attr_types @ child_types 

(* This function takes a schema and produces a letter mapping for
   every element, attribute, and atomic type name that occurs in the
   schema *)

(*
let build_name_letter_mappings mappings type_decls =
  (* Element & attribute declarations do not contain cxtypes, so we do nothing here *)
  SQNameHashtbl.iter 
    (fun rsym decl -> 
      let (_,deriv) = decl in (List.iter (fun t -> make_mappings mappings t) (cxtypes_of_ctype_derivation rsym deriv)))
    type_decls;
   mappings
*)

 (*******************)
 (* Schema creation *)
 (*******************)

let rqname_of_cxtype_decl decl = fst decl

let add_one_decl cxs (rqname,decl) =
  if SQNameHashtbl.mem cxs rqname
  then
    let (_,decl') = SQNameHashtbl.find cxs rqname in
     (* Using O'Caml structural equivalence to check whether two decls
	are the same. *)
    if (decl = decl') then ()
    else
      let declstr  = Namespace_symbols.symbol_prefix_string rqname in
      raise(Query(Schema ("Distinct declarations for " ^ declstr ^ "' conflict.")))
  else
    SQNameHashtbl.add cxs rqname (rqname,decl)

let add_one_type_decl cxs decl =
  if SQNameHashtbl.mem cxs decl.ctypedecl_name
  then
    let decl' = SQNameHashtbl.find cxs decl.ctypedecl_name in
     (* Using O'Caml structural equivalence to check whether two decls
	are the same. *)
    if (decl = decl') then ()
    else
      let declstr  = Namespace_symbols.symbol_prefix_string decl.ctypedecl_name in
      raise(Query(Schema ("Distinct declarations for " ^ declstr ^ "' conflict.")))
  else
    SQNameHashtbl.add cxs decl.ctypedecl_name decl

let dummy_cxschema = 
  { cxschema_letter_mappings = create_letter_mappings(); 
    cxschema_element_declarations = SQNameHashtbl.create 1;
    cxschema_attribute_declarations = SQNameHashtbl.create 1;
    cxschema_type_declarations = SQNameHashtbl.create 1; }

let fmkcxschema celem_declaration_list cattr_declaration_list ctype_declaration_list =
if !Conf.debugging then   print_string (">>>fmkcxschema\n");
  try
    begin
      let el = SQNameHashtbl.create 167 in
      let at = SQNameHashtbl.create 167 in
      let ty = SQNameHashtbl.create 167 in
      let mappings = create_letter_mappings() in
      List.iter (fun decl -> add_one_decl el decl) celem_declaration_list;
      List.iter (fun decl -> add_one_decl at decl) cattr_declaration_list;
      List.iter (fun decl -> add_one_type_decl ty decl) ctype_declaration_list;
      let s = 
	{ cxschema_letter_mappings = mappings; (* build_name_letter_mappings mappings ty; *)
	  cxschema_element_declarations = el;
	  cxschema_attribute_declarations = at;
	  cxschema_type_declarations = ty }
      in
if !Conf.debugging then       print_string ("<<<fmkcxschema\n");
      s
    end
  with
  | e ->
      begin
	eprintf_error "  " e;
	Format.fprintf (!Conf.glx_err_formatter) "@.";
	dummy_cxschema
      end

let merge_cxschema cx1 cx2 =
if !Conf.debugging then print_string (">>>merge_cxschema\n");
  let el = SQNameHashtbl.copy cx1.cxschema_element_declarations in
  let at = SQNameHashtbl.copy cx1.cxschema_attribute_declarations in
  let ty = SQNameHashtbl.copy cx1.cxschema_type_declarations in
  let mappings = copy_letter_mappings cx1.cxschema_letter_mappings in 
  begin
    SQNameHashtbl.iter (fun rqname decl -> add_one_decl el decl) cx2.cxschema_element_declarations; 
    SQNameHashtbl.iter (fun rqname decl -> add_one_decl at decl) cx2.cxschema_attribute_declarations; 
    SQNameHashtbl.iter (fun rqname decl -> add_one_type_decl ty decl) cx2.cxschema_type_declarations;
    let s = 
    { cxschema_letter_mappings = mappings; (* build_name_letter_mappings mappings ty;  *)
      cxschema_element_declarations = el;
      cxschema_attribute_declarations = at;
      cxschema_type_declarations = ty }
    in 
if !Conf.debugging then print_string ("<<<merge_cxschema\n");
    s
  end

let merge_cxschemas cxl =
if !Conf.debugging then print_string (">>>merge_cxschemas\n");
  let el = SQNameHashtbl.create 167 in
  let at = SQNameHashtbl.create 167 in
  let ty = SQNameHashtbl.create 167 in
  let mappings = create_letter_mappings() in
  begin
    List.iter 
      (fun cx2 ->
	SQNameHashtbl.iter (fun rqname decl -> add_one_decl el decl) cx2.cxschema_element_declarations;
	SQNameHashtbl.iter (fun rqname decl -> add_one_decl at decl) cx2.cxschema_attribute_declarations;
	SQNameHashtbl.iter (fun rqname decl -> add_one_type_decl ty decl) cx2.cxschema_type_declarations) 
      cxl;
    let s = 
    { cxschema_letter_mappings = mappings; (* build_name_letter_mappings mappings ty; *)
      cxschema_element_declarations = el;
      cxschema_attribute_declarations = at;
      cxschema_type_declarations = ty }
    in
if !Conf.debugging then print_string ("<<<merge_cxschemas\n");
    s
  end

let is_atomic_cxtype t =
    match t.pcxtype_desc with 
  | CAtomicRef      _ -> true
  | _ -> false

let is_node_cxtype t =
    match t.pcxtype_desc with 
  | CElementRef _
  | CAttributeRef _
  | CElementLocal _
  | CAttributeLocal _
  | CDocument       _
  | CText
  | CPI
  | CComment -> true
  | _ -> false

let is_document_cxtype t =
    match t.pcxtype_desc with 
  | CDocument _ -> true
  | _ -> false

let is_empty_cxtype cxtype =
  match cxtype.pcxtype_desc with
  | CEmpty -> true
  | _ -> false

let is_none_cxtype m =
  match m.pcxtype_desc with
  | CNone -> true
  | _ -> false

