(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: galax_http.ml,v 1.1 2006/01/17 23:03:03 simeon Exp $ *)

(* Module: Glx_http
   Description:
     This module contains some stuff used to hook up Galax to the http
     client provided by Trevor.
*)

open Http
open Neturl
open Error

type http_method =
  | File of string
  | Http of (string * int * string)
  | ExternalSource of (string * string * string option * string)

let registered_methods =
  ref []

let is_registered b =
  List.exists (fun x -> x=b) !registered_methods

let register_method b =
  if (is_registered b)
  then
    raise (Query (Toplevel_Error ("URL method : " ^ b ^ " already registered")))
  else
    registered_methods := b :: !registered_methods


(* fix_local worked with Http.decode_url, and fix_local2 works with decode_url2 below *)

let fix_local r =
  match r with
  | None -> "/"
  | Some x -> "/" ^ x

let fix_local2 r =
  match r with
  | None -> "/"
  | Some x -> x

let decodeaux s =
      try 
	let scheme = 
	  extract_url_scheme s in
	let newurl = 
	  if is_registered scheme then
	    url_of_string ip_url_syntax s
	  else
	    url_of_string (Hashtbl.find common_url_syntax scheme) s in
	let urlhost = 
	  try 
	    url_host newurl
	  with Not_found -> "" in
	let urlportopt = 
	  try 
	    Some(string_of_int (url_port newurl))
	  with Not_found -> None in
	let urlpath = 
	  try 
	    join_path (norm_path (url_path newurl)) ^
	    if is_registered scheme then
	  (* If it is an external registered source then the url fragment
	     needs to be appended on to the path *)
	      try
		"#" ^ url_fragment newurl 
	      with
	      | Not_found -> ""

	    else
	      ""
	  with Not_found -> "" in
	Some((url_scheme newurl), urlhost, urlportopt, Some urlpath)
      with _ -> None


let decode_url2 s =
  try
    if((String.compare (String.sub s 0 6) ("file:\\")) = 0) 
    then
      (
       (* Indicating a windows file uri *)
       (* Netstring does not parse them correctly so the old parsing*)
       (* routine has to be used *)
      Http.decode_url s
	)
    else
      decodeaux s
  with 
  | Invalid_argument _ -> decodeaux s
  | x -> raise x


let glx_decode_url s =
  match decode_url2 s with
  | None ->
      (File s)
  | Some ("file", host, port, s) ->
      if host = "" && port = None then
	let local = fix_local2 s in
	(File local)
      else
	raise (Query (URI_Error "File URL should not have a host or port number"))
  | (Some ("http", host, port, s)) ->
      let local = fix_local2 s in
      let new_port =
	match port with
	| None -> 80   (* Default port *)
	| Some portnum ->
	    begin
	      try
		(int_of_string portnum)
	      with
	      | _ ->
		  raise (Query (URI_Error (portnum ^ " is not a valid port number")))
	    end
      in 
      (Http (host, new_port, local))
  | (Some (me, host, port, s)) ->
      try
	if (is_registered me)
	then
	  let local = fix_local2 s in
	  ExternalSource (me,host,port,local)
	else
	  raise Not_found
      with
      | _ ->
	  raise (Query (URI_Error ("Method: " ^ me ^ " not supported in URL")))

(* Processing HTTP responses *)

let process_status status_line =
  let (version,code,phrase) =
    try
      let (http_version,resp1) = Gmisc.split_left_on_char status_line ' ' in
      let resp2 = Whitespace.remove_whitespace resp1 in
      let (status_string,resp3) =  Gmisc.split_left_on_char resp2 ' ' in
      let reason_phrase = Whitespace.remove_whitespace resp3 in
      (http_version, int_of_string status_string, reason_phrase)
    with
    | _ ->
	raise (Query (Protocol_Error "[HTTP: Could not process status line]"))
  in
  if (code < 100 || code >= 600) then 
    raise (Query (Protocol_Error "[HTTP: Unknown code in status line]"))
  else if (code >= 100 && code < 200) then
    raise (Query (Protocol_Error "[HTTP: Incomplete request]"))
  else if (code >= 300 && code < 400) then
    raise (Query (Protocol_Error "[HTTP: Cannot process redirection]"))
  else if (code >= 400 && code < 500) then
    raise (Query (Protocol_Error ("[HTTP: " ^ (string_of_int code) ^ " " ^ phrase ^ "]")))
  else if (code >= 500 && code < 600) then
    raise (Query (Protocol_Error ("[HTTP: " ^ (string_of_int code) ^ " " ^ phrase ^ "]")))
  else
    ()

let process_response result =

  let rec after_crlfcrlf start =
    let cr = String.index_from result start '\r' in
    if String.get result (cr+1) = '\n' &
      String.get result (cr+2) = '\r' &
      String.get result (cr+3) = '\n'
    then cr+4
    else after_crlfcrlf (cr+1) in
  try
    let x = after_crlfcrlf 0 in
    let header = String.sub result 0 x in
    let rec after_crlf start =
      let cr = String.index_from header start '\r' in
      if String.get result (cr+1) = '\n'
      then cr+2
      else after_crlf (cr+1) in
    let y = after_crlf 0 in
    let status_line = String.sub header 0 y in
    begin
      process_status status_line;
      String.sub result x (String.length result - x)
    end
  with
  | Query e -> raise (Query e)
  | _ -> raise (Query (Protocol_Error "[HTTP: Cannot process server response]"))

(* Sending http requests *)

let glx_http_get server port local =
  let result =
    try
      HTTP.get server port local
    with
    | _ -> raise (Query (Protocol_Error "[HTTP: Cannot process server response]"))
  in
  process_response result



let get_http_message h p s meth extra = 
  let result = HTTP.gen_request h p s meth extra in
  if !Conf.debugging then
    print_string ("HTTP CALL RESULT:\n"^result); flush stdout;
  let rec after_crlfcrlf start =
    let cr = String.index_from result start '\r' in
      if String.get result (cr+1) = '\n' &&
	String.get result (cr+2) = '\r' &&
				   String.get result (cr+3) = '\n'
      then cr+4
      else after_crlfcrlf (cr+1) in
    try
      let x = after_crlfcrlf 0 in
        String.sub result x (String.length result - x)
    with _ -> "ERROR"


(**
  *
  * @param uri the server address
  * @param meth HTTP method (usually POST or GET)
  * @param extra additional HTTP parameters (usually soapAction) <-- '\n' is 
  * automatically inserted after them.
  * @param request (SOAP request <-> message associated to the HTTP request)
  *
**)
let http_soap uri meth extra request =
  match glx_decode_url uri with
    | Http (host,port,local) ->
	get_http_message host port local meth 
	(Some ("Content-Type: text/xml; charset=utf-8\r\n" ^ extra,
	       request))
    | File local ->
	raise (Query (URI_Error ("Method: file not supported in SOAP server URL")))
    | ExternalSource (me,_,_,_) ->
	raise (Query (URI_Error ("Method: " ^ me ^ " not supported in SOAP server URL")))
	
	
