(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: galax_io.ml,v 1.7 2006/05/08 12:09:38 simeon Exp $ *)

open Error

open Pxp_ev_parser
open Pxp_types

(* Module: Galax_io
   Description:
     This module contains some basic I/O structures and operations.
*)


(****************************)
(* Galax I/O specifications *)
(****************************)

(* What kind of input ? *)

type input_spec =
  | File_Input of string
  | String_Input of string
  | Buffer_Input of Netbuffer.t
  | Http_Input of string
  | Channel_Input of in_channel

(* What kind of output ? *)

type output_spec =
  | File_Output of string
  | Buffer_Output of Buffer.t
  | Channel_Output of out_channel
  | Formatter_Output of Format.formatter


(*********************)
(* PXP Parsing hooks *)
(*********************)

let source_from_input_spec gio =
  match gio with
  | File_Input s ->
      from_file s
  | String_Input s ->
      from_string s
  | Buffer_Input nb ->
      let (in_channel,close) = Netchannels.create_input_netbuffer nb in
      close ();
      from_obj_channel in_channel
  | Http_Input uri ->
      begin
	match Galax_http.glx_decode_url uri with
	| Galax_http.Http (host,port,local) ->
	    (* Note: This is pretty inefficient here. - Jerome *)
	    let string_buffer = Galax_http.glx_http_get host port local in
	    from_string string_buffer
	| Galax_http.File f ->
	    from_file f
	| Galax_http.ExternalSource _ ->
	    raise (Query (Prototype "Cannot parse external source"))
      end
  | Channel_Input ic ->
      from_channel ic

let close_source s =
  let resolver =
    match s with
    | Entity (_,resolver)
    | ExtID (_,resolver)
    | XExtID (_,_,resolver) -> resolver
  in
  resolver#close_in

type pull_handle = source
type pxp_stream = (unit -> Pxp_types.event option) * pull_handle

let pull_parser_from_input_spec gio =
  let config = Galax_pxp.glx_default_config in
  let entry = Galax_pxp.glx_default_entry in
  let source = source_from_input_spec gio in
  let entity_manager = create_entity_manager config source in
  (create_pull_parser config entry entity_manager,source)

let close_pull_parser source =
  close_source source

let dtd_from_input_spec gio =
  let config = Galax_pxp.glx_default_config in
  let source = source_from_input_spec gio in
  Pxp_dtd_parser.parse_dtd_entity config source

let ff = "file://"

let uri_string_of_gio gio =
  match gio with
  | File_Input s ->
      if Filename.is_relative s
      then
	Some
	  (Gmisc.rename_dir (Filename.concat ff (Filename.concat (Sys.getcwd ()) s)))
      else
	Some (ff ^ s)
  | String_Input s -> None
  | Buffer_Input _ -> None
  | Http_Input uri -> Some uri
  | Channel_Input _ -> None

