(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies and AT&T Labs.            *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: code_fn.ml,v 1.14 2006/05/16 23:33:59 simeon Exp $ *)

(* Module: Code_fn
   Description:
     This module contains evaluation code for F&O.
*)

open Error

open Namespace_names
open Namespace_util
open Namespace_builtin

open Decimal
open Datatypes
open Datatypes_util

open Dm_types
open Dm_atomic
open Dm
open Dm_util

open Physical_item
open Physical_item_util

open Gmisc

open Physical_value
open Physical_list
open Physical_xml_value
open Physical_value_util
open Physical_util

open Xquery_common_ast

open Compile_context
open Code_selection_context
open Execution_context 
open Algebra_type


(***************************************)
(* Signature of all built-in functions *)
(***************************************)

type bltin = algebra_context -> item Cursor.cursor array -> item Cursor.cursor
type bltin_mat = Algebra_type.item_list_nary_to_item_list_code


(***************************)
(* Dummy built-in function *)
(***************************)

let _DUMMY fn comp_ctxt alg_ctxt n =
  raise (Query (Prototype ("Function " ^ (fn)^ " not supported in "
			   ^ Conf.system ^ " version " ^ Conf.version)))

(****************************)
(* F&O Section 2. Accessors *)
(****************************)

(* F&O Section 2.` fn:node-name *)

let _fn_node_name comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1)
  then
    Cursor.cursor_empty()
  else
    let node = get_singleton_node p1 in
    match node#node_name() with 
    | Some nn ->
	_atomic_value nn
    | None ->
	Cursor.cursor_empty()

let _fn_nilled comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1)
  then
    Cursor.cursor_empty()
  else
    let node = get_singleton_node p1 in
    match node#node_kind() with
    | ElementNodeKind ->
	Cursor.cursor_of_singleton (_boolean ((node#getElementNode())#nilled()))
    | _ -> Cursor.cursor_empty()

(* F&O Section 2.2 fn:string *)

let _fn_string comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let string_result =
    if (Cursor.cursor_is_empty p1)
    then ""
    else
      let i = get_item p1 in
      if (isNode i) then 
	(getNode i)#string_value()
      else 
	(getAtomicValue i)#erase_atomic_value()
  in
  Cursor.cursor_of_singleton (_string string_result)

(* F&O Section 2.3 fn:data *)

(* NOTE: Group operator replicates this functionality and can rewrite
         fn:datas away. If the semantic of fn:data changes this needs
         to change as well.  Currently done in
         Group_code.get_atomic_content
  - chris

  NOTE: THIS SEEMS OBSOLETE??! CHECK WITH CHRIS - JEROME 08/02/2005
*)
let _fn_data comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let items_cursor = p1 in
  let get_atomic_content item =
    if (isNode item)
    then
      Cursor.cursor_map (fun a -> (Item_Atomic a)) ((getNode item)#typed_value())
    else
      Cursor.cursor_of_singleton item
  in (Cursor.cursor_map_concat get_atomic_content items_cursor)

(* F&O Section 2.4 fn:base-uri *)

let _fn_base_uri_one comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if Cursor.cursor_is_empty p1
  then
    Cursor.cursor_empty()
  else
    let node = get_singleton_node p1 in
    match (!(node#base_uri())) with
    | None -> Cursor.cursor_empty()
    | Some astr -> _atomic_value astr

(************************)
(* F&O Section 4. Trace *)
(************************)

let _fn_trace comp_ctxt =
  let proc_ctxt = processing_context_from_compile_context comp_ctxt in
  (fun alg_ctxt n ->
    let (ic, s) = Args.get_array_param2  n in
    (* Note:
       Trace requires materialization of the input cursor as it must
       both print it on file, and return is as the result of the
       function call.
       - Jerome *)
    let il = Cursor.list_of_cursor "Cs_code_fn._fn_trace" ic in
    let s' = get_string s in
    let _ =
      begin
	print_string s';
	Serialization.fserialize_datamodel proc_ctxt Format.std_formatter (Cursor.cursor_of_list il);
	flush stdout
      end
    in
    Cursor.cursor_of_list il)

(*****************************************************)
(* F&O Section 5. Functions and Operators on Numbers *)
(*****************************************************)

(* GENERIC POLYMORPHIC FUNCIONS
  
   unwrap_fn : unwraps arguments
   op_fn : applies operators
   wrap_fn : wraps results

Stricter version:
  if (Cursor.cursor_is_empty p1) && (Cursor.cursor_is_empty p2)
  then
    Cursor.cursor_empty()
  else if (Cursor.cursor_is_empty p1)
  then
    let _ = unwrap_fn1 p2 in
    Cursor.cursor_empty()
  else if (Cursor.cursor_is_empty p2)
  then
    let _ = unwrap_fn2 p1 in
    Cursor.cursor_empty()
  else


*)

let _fn_binary_op_uneven (unwrap_fn1, unwrap_fn2, binop_fn, wrap_fn, (zero_test1,zero_test2)) comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  if (Cursor.cursor_is_empty p1)
  then Cursor.cursor_empty()
  else if (Cursor.cursor_is_empty p2)
  then Cursor.cursor_empty()
  else
    let a1 = zero_test1 (unwrap_fn1 p1) in
    let a2 = zero_test2 (unwrap_fn2 p2) in
    let (i1,i2) = (a1, a2) in
    Cursor.cursor_of_singleton (wrap_fn (binop_fn i1 i2))

let _fn_binary_op (unwrap_fn, binop_fn, wrap_fn, (zero_test1,zero_test2)) comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven (unwrap_fn, unwrap_fn, binop_fn, wrap_fn, (zero_test1,zero_test2)) comp_ctxt alg_ctxt n

let _fn_unary_op (unwrap_fn, unop_fn, wrap_fn) comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1) then Cursor.cursor_empty()
  else Cursor.cursor_of_singleton (wrap_fn (unop_fn (unwrap_fn p1)))

let _fn_compare_op (unwrap_fn, compare_fn) comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let (i1,i2) = (unwrap_fn p1, unwrap_fn p2) in
  Cursor.cursor_of_singleton (_boolean (compare_fn i1 i2))

let _fn_min_poly (cursor_fn, compare_fn, constructor_fn) comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allVals = cursor_fn p1 in
  match (Cursor.cursor_peek allVals) with
  | None ->
      Cursor.cursor_empty()
  | Some start_min ->
      let min =
	Cursor.cursor_fold_left (fun x0 x1 -> if (compare_fn x0 x1) = -1 then x0 else x1) start_min allVals
      in
      Cursor.cursor_of_singleton (constructor_fn min)

let _fn_max_poly (cursor_fn, compare_fn, constructor_fn)  comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allVals = cursor_fn p1 in
  match (Cursor.cursor_peek allVals) with
  | None ->
      Cursor.cursor_empty()
  | Some start_max ->
      let max =
	Cursor.cursor_fold_left (fun x0 x1 -> if (compare_fn x0 x1) = 1 then x0 else x1) start_max allVals
      in
      Cursor.cursor_of_singleton (constructor_fn max)

(* xsd:int *)

let raise_zero () = raise (Query (Wrong_Args "Division by zero"))
let raise_nan () = raise (Query (Wrong_Args "NaN in idiv"))
let raise_infinity () = raise (Query (Wrong_Args "Infinity in idiv"))

let def_zero_test = ((fun x -> x), (fun x -> x))
let integer_zero_test =
  ((fun x -> x),
   (fun x -> if (_integer_eq _integer_zero x) then raise_zero() else x))
let decimal_zero_test =
  ((fun x -> x),
   (fun x -> if (_decimal_eq _decimal_zero x) then raise_zero() else x))
let float_nan_zero_test =
  ((fun x ->
    if x == nan then raise_nan()
    else if (x = infinity) or (x = -.infinity) then raise_infinity()
    else x),
   (fun x ->
     if x == nan then raise_nan()
     else if (x = 0.0) then raise_zero() else x))

let _op_hexBinary_equal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_hexBinary, hexBinary_equal) comp_ctxt alg_ctxt n 

let _op_base64Binary_equal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_base64Binary, base64Binary_equal) comp_ctxt alg_ctxt n 

let _op_hexBinary_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_hexBinary, (fun x1 x2 -> not(hexBinary_equal x1 x2))) comp_ctxt alg_ctxt n 

let _op_base64Binary_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_base64Binary, (fun x1 x2 -> not(base64Binary_equal x1 x2))) comp_ctxt alg_ctxt n 

(* xsd:integer *)
let _fn_integer_add comp_ctxt alg_ctxt n =
  _fn_binary_op (get_integer, _integer_add, _integer, def_zero_test) comp_ctxt alg_ctxt n

let _fn_integer_sub comp_ctxt alg_ctxt n =
  _fn_binary_op (get_integer, _integer_sub, _integer, def_zero_test) comp_ctxt alg_ctxt n

let _fn_integer_mult comp_ctxt alg_ctxt n =
  _fn_binary_op (get_integer, _integer_mult, _integer, def_zero_test) comp_ctxt alg_ctxt n

let _fn_integer_div comp_ctxt alg_ctxt n =
  _fn_binary_op (get_integer, _integer_div, _decimal, integer_zero_test) comp_ctxt alg_ctxt n

let _fn_integer_idiv comp_ctxt alg_ctxt n =
  _fn_binary_op (get_integer, _integer_idiv, _integer, integer_zero_test) comp_ctxt alg_ctxt n

let _fn_integer_mod comp_ctxt alg_ctxt n =
  _fn_binary_op (get_integer, _integer_mod, _integer, def_zero_test) comp_ctxt alg_ctxt n

let _fn_integer_unary_minus comp_ctxt alg_ctxt n =
  _fn_unary_op (get_integer, _integer_unary_minus, _integer) comp_ctxt alg_ctxt n

let _fn_integer_equal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_integer, _integer_eq) comp_ctxt alg_ctxt n 

let _fn_integer_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_integer, fun i1 i2 -> not(_integer_eq i1 i2)) comp_ctxt alg_ctxt n 

let _fn_integer_lt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_integer, _integer_lt) comp_ctxt alg_ctxt n 

let _fn_integer_gt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_integer, _integer_gt) comp_ctxt alg_ctxt n 

let _fn_integer_le comp_ctxt alg_ctxt n =
  _fn_compare_op(get_integer, _integer_le) comp_ctxt alg_ctxt n 

let _fn_integer_ge comp_ctxt alg_ctxt n =
  _fn_compare_op(get_integer, _integer_ge) comp_ctxt alg_ctxt n 

(* unary plus is a no-op *)
let _fn_integer_unary_plus comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  p1 

(* xsd:decimal *)
let _fn_decimal_add comp_ctxt alg_ctxt n =
  _fn_binary_op (get_decimal, _decimal_add, _decimal, def_zero_test) comp_ctxt alg_ctxt n

let _fn_decimal_sub comp_ctxt alg_ctxt n =
  _fn_binary_op (get_decimal, _decimal_sub, _decimal, def_zero_test) comp_ctxt alg_ctxt n

let _fn_decimal_mult comp_ctxt alg_ctxt n =
  _fn_binary_op (get_decimal, _decimal_mult, _decimal, def_zero_test) comp_ctxt alg_ctxt n

let _fn_decimal_div comp_ctxt alg_ctxt n =
  _fn_binary_op (get_decimal, _decimal_div, _decimal, decimal_zero_test) comp_ctxt alg_ctxt n

let _fn_decimal_idiv comp_ctxt alg_ctxt n =
  _fn_binary_op (get_decimal, _decimal_idiv, _integer, decimal_zero_test) comp_ctxt alg_ctxt n

let _fn_decimal_mod comp_ctxt alg_ctxt n =
  _fn_binary_op (get_decimal, _decimal_mod, _decimal, def_zero_test) comp_ctxt alg_ctxt n

let _fn_decimal_unary_minus comp_ctxt alg_ctxt n =
  _fn_unary_op (get_decimal, _decimal_unary_minus, _decimal) comp_ctxt alg_ctxt n

let _fn_decimal_equal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_decimal, _decimal_eq) comp_ctxt alg_ctxt n 

let _fn_decimal_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_decimal, fun i1 i2 -> not (_decimal_eq i1 i2)) comp_ctxt alg_ctxt n 

let _fn_decimal_lt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_decimal, _decimal_lt) comp_ctxt alg_ctxt n 

let _fn_decimal_gt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_decimal, _decimal_gt) comp_ctxt alg_ctxt n 

let _fn_decimal_le comp_ctxt alg_ctxt n =
  _fn_compare_op(get_decimal, _decimal_le) comp_ctxt alg_ctxt n 

let _fn_decimal_ge comp_ctxt alg_ctxt n =
  _fn_compare_op(get_decimal, _decimal_ge) comp_ctxt alg_ctxt n 

(* unary plus is a no-op *)
let _fn_decimal_unary_plus comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  p1 

(* xsd:float *)
let _fn_float_add comp_ctxt alg_ctxt n =
  _fn_binary_op (get_float, (fun i1 i2 -> i1 +. i2), _float, def_zero_test) comp_ctxt alg_ctxt n

let _fn_float_sub comp_ctxt alg_ctxt n =
  _fn_binary_op (get_float, (fun i1 i2 -> i1 -. i2), _float, def_zero_test) comp_ctxt alg_ctxt n

let _fn_float_mult comp_ctxt alg_ctxt n =
  _fn_binary_op (get_float, (fun i1 i2 -> i1 *. i2), _float, def_zero_test) comp_ctxt alg_ctxt n

let _fn_float_div comp_ctxt alg_ctxt n =
  _fn_binary_op (get_float, (fun i1 i2 -> i1 /. i2), _float, def_zero_test) comp_ctxt alg_ctxt n

let _fn_float_idiv comp_ctxt alg_ctxt n =
  _fn_binary_op (get_float, _float_idiv, _integer, float_nan_zero_test) comp_ctxt alg_ctxt n

let _fn_float_mod comp_ctxt alg_ctxt n =
  _fn_binary_op (get_float, _float_mod, _float, def_zero_test) comp_ctxt alg_ctxt n

let _fn_float_unary_minus comp_ctxt alg_ctxt n =
  _fn_unary_op (get_float, (fun x -> (-.(x))), _float) comp_ctxt alg_ctxt n

let _fn_float_equal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_float, fun i1 i2 -> i1 = i2) comp_ctxt alg_ctxt n 

(* NB! Must use not(i1 = i2).  i1 != i2 checks for physical equivalence. *)
let _fn_float_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_float, fun i1 i2 -> not(i1 = i2)) comp_ctxt alg_ctxt n 

let _fn_float_lt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_float, fun i1 i2 -> i1 < i2) comp_ctxt alg_ctxt n 

let _fn_float_gt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_float, fun i1 i2 -> i1 > i2) comp_ctxt alg_ctxt n 

let _fn_float_le comp_ctxt alg_ctxt n =
  _fn_compare_op(get_float, fun i1 i2 -> i1 <= i2) comp_ctxt alg_ctxt n 

let _fn_float_ge comp_ctxt alg_ctxt n =
  _fn_compare_op(get_float, fun i1 i2 -> i1 >= i2) comp_ctxt alg_ctxt n 

(* unary plus is a no-op *)
let _fn_float_unary_plus comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1  n in
  p1 

(* xsd:double *)
let _fn_double_add comp_ctxt alg_ctxt n =
  _fn_binary_op (get_double, (fun i1 i2 -> i1 +. i2), _double, def_zero_test) comp_ctxt alg_ctxt n

let _fn_double_sub comp_ctxt alg_ctxt n =
  _fn_binary_op (get_double, (fun i1 i2 -> i1 -. i2), _double, def_zero_test) comp_ctxt alg_ctxt n

let _fn_double_mult comp_ctxt alg_ctxt n =
  _fn_binary_op (get_double, (fun i1 i2 -> i1 *. i2), _double, def_zero_test) comp_ctxt alg_ctxt n

let _fn_double_div comp_ctxt alg_ctxt n =
  _fn_binary_op (get_double, (fun i1 i2 -> i1 /. i2), _double, def_zero_test) comp_ctxt alg_ctxt n

let _fn_double_idiv comp_ctxt alg_ctxt n =
  _fn_binary_op (get_double, _float_idiv, _integer, float_nan_zero_test) comp_ctxt alg_ctxt n

let _fn_double_mod comp_ctxt alg_ctxt n =
  _fn_binary_op (get_double, _float_mod, _double, def_zero_test) comp_ctxt alg_ctxt n

let _fn_double_unary_minus comp_ctxt alg_ctxt n =
  _fn_unary_op (get_double, (fun x -> (-.(x))), _double) comp_ctxt alg_ctxt n

let _fn_double_equal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_double, fun i1 i2 -> i1 = i2) comp_ctxt alg_ctxt n 

(* NB! Must use not(i1 = i2).  i1 != i2 checks for physical equivalence. *)
let _fn_double_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_double, fun i1 i2 -> not(i1 = i2)) comp_ctxt alg_ctxt n 

let _fn_double_lt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_double, fun i1 i2 -> i1 < i2) comp_ctxt alg_ctxt n 

let _fn_double_gt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_double, fun i1 i2 -> i1 > i2) comp_ctxt alg_ctxt n 

let _fn_double_le comp_ctxt alg_ctxt n =
  _fn_compare_op(get_double, fun i1 i2 -> i1 <= i2) comp_ctxt alg_ctxt n 

let _fn_double_ge comp_ctxt alg_ctxt n =
  _fn_compare_op(get_double, fun i1 i2 -> i1 >= i2) comp_ctxt alg_ctxt n 

(* unary plus is a no-op *)
let _fn_double_unary_plus comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1  n in
  p1 

(* F&O Section 6.4.1 fn:abs *)
let _fn_abs_double comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1  n in
  let f  = get_double s in
  Cursor.cursor_of_singleton (_double (abs_float f))

let _fn_abs_float comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1 n in
  let f  = get_float  s in
  Cursor.cursor_of_singleton (_float (abs_float f))

let _fn_abs_decimal comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1 n in
  let f  = get_decimal  s in
  Cursor.cursor_of_singleton (_decimal (Num.abs_num f))

let _fn_abs_integer comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1 n in
  let i = get_integer  s in
  Cursor.cursor_of_singleton (_integer (Big_int.abs_big_int i))

(* F&O Section 6.4.2 fn:ceiling *)

let _fn_ceiling_double comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1 n in
  let f  = get_double s in
  let f' = ceil f in
  Cursor.cursor_of_singleton (_double f')

let _fn_ceiling_float comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1 n in
  let f  = get_float s in
  let f' = ceil f in
  Cursor.cursor_of_singleton (_float f')

let _fn_ceiling_decimal comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1 n in
  let f  = get_decimal s in
  let f' = Num.ceiling_num f in
  Cursor.cursor_of_singleton (_decimal f')

let _fn_ceiling_integer comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1 n in
  let _ = get_integer s in
  s

(* F&O Section 6.4.3 fn:floor *)
let _fn_floor_double comp_ctxt alg_ctxt n =
  let s  = Args.get_array_param1 n in
  let f  = get_double s in
  let f' = floor f in
  Cursor.cursor_of_singleton (_double f')

let _fn_floor_float comp_ctxt alg_ctxt n =
  let s  = Args.get_array_param1 n in
  let f  = get_float s in
  let f' = floor f in
  Cursor.cursor_of_singleton (_float f')

let _fn_floor_decimal comp_ctxt alg_ctxt n =
  let s  = Args.get_array_param1 n in
  let f  = get_decimal s in
  let f' = Num.floor_num f in
  Cursor.cursor_of_singleton (_decimal f')

let _fn_floor_integer comp_ctxt alg_ctxt n =
  let s  = Args.get_array_param1 n in
  let _ = get_integer s in 
  s

(* F&O Section 6.4.4 fn:round *)

let _fn_round_double comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1  n in
  let f  = get_double s in
  let f' = floor (f +. 0.5) in
  Cursor.cursor_of_singleton (_double f')

let _fn_round_float comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1  n in
  let f  = get_float s in
  let f' = floor (f +. 0.5) in
  Cursor.cursor_of_singleton (_float f')

let _fn_round_decimal comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1  n in
  let f  = get_decimal s in
  let f' = Num.floor_num (_decimal_add f _decimal_onehalf) in
  Cursor.cursor_of_singleton (_decimal f')

let _fn_round_integer comp_ctxt alg_ctxt n = 
  let s  = Args.get_array_param1 n in
  let _  = get_integer s in
  s

(* F&O Section 6.4.4 fn:round-half-to-even *)

let _fn_round_half_to_even_double comp_ctxt alg_ctxt n =
  let s,p = Args.get_array_param2 n in
  let f   = get_double s in
  let pr  = 10.0 ** (_float_of_integer (get_integer p)) in
  let bef = f *. pr in
  let flo = floor (bef +. 0.5) in
  let cei = ceil (bef -. 0.5) in
  let f'  =
    if (flo = cei)
    then flo
    else if (mod_float flo 2.0) = 0.0 then flo else cei
  in
  let aft = f' /. pr in
  Cursor.cursor_of_singleton (_double aft)

let _fn_round_half_to_even_float comp_ctxt alg_ctxt n =
  let s,p = Args.get_array_param2 n in
  let f   = get_float s in
  let pr  = 10.0 ** (_float_of_integer (get_integer p)) in
  let bef = f *. pr in
  let flo = floor (bef +. 0.5) in
  let cei = ceil (bef -. 0.5) in
  let f'  =
    if (flo = cei)
    then flo
    else if (mod_float flo 2.0) = 0.0 then flo else cei
  in
  let aft = f' /. pr in
  Cursor.cursor_of_singleton (_float aft)

let _fn_round_half_to_even_decimal comp_ctxt alg_ctxt n =
  let s,p = Args.get_array_param2 n in
  let f   = get_decimal s in
  let pr  = _decimal_power _decimal_ten (_decimal_of_integer (get_integer p)) in
  let bef = _decimal_mult f pr in
  let flo = _decimal_floor (_decimal_add bef _decimal_onehalf) in
  let cei = _decimal_ceil (_decimal_sub bef _decimal_onehalf) in
  let f'  =
    if (_decimal_eq flo cei)
    then flo
    else if _decimal_eq (_decimal_mod flo _decimal_two) _decimal_zero
    then flo
    else cei
  in
  let aft = _decimal_div f' pr in
  Cursor.cursor_of_singleton (_decimal aft)

let _fn_round_half_to_even_integer comp_ctxt alg_ctxt n =
  let s,p = Args.get_array_param2 n in
  let _   = get_integer s in
  s


(***************************************)
(* F&O Section 7. Functions on Strings *)
(***************************************)

(* F&O Section 7.3 equality and comparison of strings *)

(** Gives the unboxed string value of p1, or "" if p1 is passed as ().
   All of the xquery functions on strings require an empty sequence to be
   treated as an empty string. This function is used in place of the get_string
   function in order to do so.
   @param comp_ctxt alg_ctxt The algebra context
   @param p1 The string to be unboxed, or the empty sequence
   @return The string value of p1 or else "" if p1 is empty *)

let normalize_optional_string p1 =
  if (Cursor.cursor_is_empty p1)
  then ""
  else get_string p1 

(** Unboxes a dayTimeDuration in order to represent a timezone. If timezone is
   passed as the empty sequence, then it gets the timezone provided in the
   algebra context, which is the specified behavior for many operations on dates
   and times dealing with timezones
   @param comp_ctxt alg_ctxt The algebra context
   @param tz1 The timezone
   @return The unboxed timezone or the timezone provided by the algebra context if
   no timezone is provided *)
let get_optional_timezone comp_ctxt alg_ctxt tz1 =
  if (Cursor.cursor_is_empty tz1)
  then None
  else Some (get_dayTimeDuration tz1)


let _fn_compare comp_ctxt alg_ctxt n = 
  let compop i1 i2 = Decimal._integer_of_int (String.compare i1 i2) in
  _fn_binary_op (get_string,compop,_integer,def_zero_test) comp_ctxt alg_ctxt n

let _op_string_equal comp_ctxt alg_ctxt n =
  _fn_compare_op(normalize_optional_string, string_equal) comp_ctxt alg_ctxt n 

let _op_string_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op(normalize_optional_string, fun i1 i2 -> not (string_equal i1 i2)) comp_ctxt alg_ctxt n 

let _op_string_lt comp_ctxt alg_ctxt n =
  _fn_compare_op(normalize_optional_string, string_lt) comp_ctxt alg_ctxt n 

let _op_string_le comp_ctxt alg_ctxt n =
  _fn_compare_op(normalize_optional_string, string_lteq) comp_ctxt alg_ctxt n 

let _op_string_gt comp_ctxt alg_ctxt n =
  _fn_compare_op(normalize_optional_string, string_gt) comp_ctxt alg_ctxt n 

let _op_string_ge comp_ctxt alg_ctxt n =
  _fn_compare_op(normalize_optional_string, string_gteq) comp_ctxt alg_ctxt n 

(* F&O Section 7.4.1 fn:concat *)

let _fn_concat comp_ctxt alg_ctxt n =
  let (s1,s2) = Args.get_array_param2 n in
  let a1 = get_optional_atomic s1 in
  let a2 = get_optional_atomic s2 in
  let s1 =
    match a1 with
    | None -> ""
    | Some at -> (at#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_string ATString)#getAtomicString()
  in
  let s2 =
    match a2 with
    | None -> ""
    | Some at -> (at#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_string ATString)#getAtomicString()
  in
  Cursor.cursor_of_singleton (_string (s1 ^ s2))

(* F&O Section 7.4.2 fn:string-join *)

let _fn_string_join comp_ctxt alg_ctxt n =
  let (s1, s2) = (Args.get_array_param2 n) in
  (* Note:
       fn:string-join is more efficient using the String.concat
       function, which only does one string allocation. This means we
       first do a materialization to a list.
     - Jerome *)
  let l1 = Cursor.list_of_cursor "Cs_code_fn._fn_string_join" (get_string_cursor s1) in
  let concatenated_string = String.concat (get_string s2) l1 in
  Cursor.cursor_of_singleton (_string concatenated_string)

(* F&O Section 7.4.3 fn:substring *)

(** Returns the portion of the value of $sourceString beginning at the
  position indicated by the value of $startingLoc and continuing for
  the number of characters indicated by the value of $length. The
  characters returned do not extend beyond $sourceString. If
  $startingLoc is zero or negative, only those characters in positions
  greater than zero are returned.

  The three argument version of the function
  returns the characters in $sourceString whose position $p obeys:

    fn:round($startingLoc) <= $p < fn:round($startingLoc) + fn:round($length)

  The two argument version of the function assumes that$length is
  infinite and returns the characters in $sourceString whose position
  $p obeys:

    fn:round($startingLoc) <= $p < fn:round(INF)

   @param comp_ctxt alg_ctxt The algebra context, unused in this function
   @param n The arguments
   @returns A cursor containing the substring 
*)
let _fn_substring comp_ctxt alg_ctxt n =
  let (str, start, len) = Args.get_array_param3 n in
  if Cursor.cursor_is_empty str then 
    Cursor.cursor_of_singleton (_string "")
  else
      (* The three argument version of the function
	 returns the characters in $sourceString whose position $p obeys: 
	 fn:round($startingLoc) <= $p < fn:round($startingLoc) + fn:round($length)

	 The two argument version of the function assumes that$length is
	 infinite and returns the characters in $sourceString whose position
	 $p obeys:

	 fn:round($startingLoc) <= $p < fn:round(INF)
      *)
    begin
      let str' = normalize_optional_string str in
      let startpos' = (floor ((get_double start) +. 0.5)) in
      let endpos' = startpos' +. (floor ((get_double len) +. 0.5)) in
      (* substring indexes start at position 1 in XQuery, but 0 in Caml. - Jerome *)
      let startpos'' = if startpos' > 0.0 then int_of_float(startpos') - 1 else 0 in
      let len' = 
	(if endpos' = nan then 0
 	else if endpos' = infinity then (String.length str')
 	else int_of_float(endpos') - 1) - startpos''
      in
      let sub =
	try
          String.sub str' startpos'' len'
	with
	| Invalid_argument _ -> ""
      in
      Cursor.cursor_of_singleton (_string sub)
    end


(* F&O Section 7.4.4 fn:string-length *)

let _fn_string_length comp_ctxt alg_ctxt n =
  let s  = Args.get_array_param1 n in
  if Cursor.cursor_is_empty s then 
    (*Cursor.cursor_empty()*)
    Cursor.cursor_of_singleton(_integer(Decimal._integer_of_int 0))
  else
    begin
      let s' = normalize_optional_string s in
      let i = String.length s' in
      Cursor.cursor_of_singleton (_integer (Decimal._integer_of_int i))
    end

(* F&O Section 7.4.5 fn:normalize-space *)

let _fn_normalize_space comp_ctxt alg_ctxt n =
  let s  = Args.get_array_param1 n in
  let str = 
    if Cursor.cursor_is_empty s then ""
    else
      String.concat " "
	(List.filter (fun s -> not (s = ""))
	   (List.map Whitespace.remove_whitespace 
	      (Whitespace.whitespace_separate (normalize_optional_string s))))
  in Cursor.cursor_of_singleton (_string str)

(* F&O Section 7.4.6 fn:normalize-unicode *)

let _fn_normalize_unicode comp_ctxt alg_ctxt n =
  let p1,p2 = Args.get_array_param2 n in
  let s1 = get_optional_string p1 in
  let norm_fun =
    let norm = String.uppercase (Whitespace.remove_whitespace (get_string p2)) in
    match norm with
    | "" -> (fun x -> x)
    | "NFC" -> Galax_camomile.nfc
    | "NFD" -> Galax_camomile.nfd
    | "NFKC" -> Galax_camomile.nfkc
    | "NFKD" -> Galax_camomile.nfkd
    | _ -> raise (Query (Wrong_Args ("Unicode normalization " ^norm^ " unknown")))
  in
  let ns =
    match s1 with
    | None -> ""
    | Some s1 ->
	norm_fun s1
  in
  Cursor.cursor_of_singleton (_string ns)

(* F&O Section 7.4.7 fn:upper-case *)

let _fn_upper_case comp_ctxt alg_ctxt n =
  let s   = Args.get_array_param1 n in
  let string_result =
  if Cursor.cursor_is_empty s then 
    ""
  else
    let s'  = normalize_optional_string s in
    String.uppercase s' 
  in
  Cursor.cursor_of_singleton (_string string_result)

(* F&O Section 7.4.8 fn:lower-case *)

let _fn_lower_case comp_ctxt alg_ctxt n =
  let s   = Args.get_array_param1 n in
  let string_result =
  if Cursor.cursor_is_empty s then 
    ""
  else
    let s'  = normalize_optional_string s in
    String.lowercase s'
  in
  Cursor.cursor_of_singleton (_string string_result)

(* F&O Section 7.4.9 fn:translate *)

let _translate_aux str mapStr transStr =
  let char_stream = Stream.of_string str in
  let buffer = Buffer.create 0 in
  let new_char c = Buffer.add_char buffer c in
  let translate_char c =
    try
      let index = String.index mapStr c in
      begin
	try
	  new_char (String.get transStr index)
	with
	| Invalid_argument _ ->
	    ()
      end
    with
    | Not_found ->
	new_char c
  in
  while not(Stream.peek char_stream = None) do
    translate_char (Stream.next char_stream)
  done;
  Buffer.contents buffer
    
let _fn_translate comp_ctxt alg_ctxt n =
  let (str, mapStr, transStr) = Args.get_array_param3 n in
  let string_result =
    if Cursor.cursor_is_empty str then 
      ""
    else
      begin
	let str'       = normalize_optional_string  str      in
	let mapStr'    = normalize_optional_string  mapStr   in
	let transStr'  = normalize_optional_string  transStr in
	_translate_aux str' mapStr' transStr'
    end
  in Cursor.cursor_of_singleton (_string string_result)

(* F&O Section 7.4.10 fn:encode-for-uri *)

let _fn_encode_for_uri comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let s = normalize_optional_string p1 in
  let ns = AnyURI.encode_string_for_uri s in
  Cursor.cursor_of_singleton (_string ns)

(* F&O Section 7.4.11 fn:iri-to-uri *)

let _fn_iri_to_uri comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let s = normalize_optional_string p1 in
  let ns = AnyURI.encode_iri_to_uri s in
  Cursor.cursor_of_singleton (_string ns)

(* F&O Section 7.4.12 fn:escape-html-uri *)

let _fn_escape_html_uri comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let s = normalize_optional_string p1 in
  let ns = AnyURI.encode_html_uri s in
  Cursor.cursor_of_singleton (_string ns)

(* F&O Section 7.2.1 fn:codepoints-to-string *)
let _fn_codepts_to_string comp_ctxt alg_ctxt n =
  let p1   = Args.get_array_param1 n in
  let ptlist = Cursor.list_of_cursor "" (get_integer_cursor p1) in
  match ptlist with
  | [] -> Cursor.cursor_empty()
  | _ -> 
      let c = List.map _int_of_integer ptlist in
      let s = Galax_camomile.utf8_string_of_code_points c in
      Cursor.cursor_of_singleton (_string s)

(* F&O Section 7.3.3 fn:codepoint-equal *)
let _fn_codepoint_equal comp_ctxt alg_ctxt n =
  let compop x y = ((Galax_camomile.utf8_codepoint_compare x y)= 0) in
  _fn_binary_op (get_string,compop,_boolean,def_zero_test) comp_ctxt alg_ctxt n
  

(* F&O Section 7.2.2 fn:string-to-codepoints *)

let _fn_string_to_codepts comp_ctxt alg_ctxt n =
  let s   = Args.get_array_param1 n in
  if Cursor.cursor_is_empty s then 
    Cursor.cursor_empty()
  else
    let s'  = get_string s in
    if (s' = "") then Cursor.cursor_empty()
    else
      begin
	try
	  let codept_list = ref [] in
	  Camomile.UTF8.iter (fun uc -> let ic = Camomile.UChar.code uc in codept_list := (!codept_list) @ [ic]) s';
	  Cursor.cursor_of_list (List.map (fun i -> _integer (Decimal._integer_of_int i)) (!codept_list))
	with 
	| Camomile.UChar.Out_of_range -> raise(Query(Unicode_Error("Unicode character cannot be represented by positive integer.")))
      end
(* F&O Section 7.5.1 fn:contains *)

let _fn_contains comp_ctxt alg_ctxt n =
  let (s1, s2) = Args.get_array_param2  n in
  begin
    let s1' = normalize_optional_string s1 in
    let s2' = normalize_optional_string s2 in
    let b = Regularexp.matches s1' s2' "" in
      Cursor.cursor_of_singleton (_boolean b)
  end

(* F&O Section 7.5.2 fn:starts-with *)

let _fn_starts_with comp_ctxt alg_ctxt n = 
  let (s1, s2) = Args.get_array_param2 n in
  begin
      let s1' = normalize_optional_string s1 in
      let s2' = normalize_optional_string s2 in
      let lens2' = String.length s2' in 
      let b = 
	try 
	  let s' = if (s1' = "") && (s2' = "") then s2' else String.sub s1' 0 lens2' in (s' = s2')
	with _ -> false
      in
      Cursor.cursor_of_singleton (_boolean b)
    end

(* F&O Section 7.5.3 fn:ends-with *)
let _fn_ends_with comp_ctxt alg_ctxt n = 
  let (s1, s2) = Args.get_array_param2 n in
  begin
    let s1' = normalize_optional_string s1 in
    let s2' = normalize_optional_string s2 in
    let lens1' = String.length s1' in
    let lens2' = String.length s2' in 
    let pos = lens1' - lens2' in 
    let b = 
      try 
	let s' = if (s1' = "") && (s2' = "") then s2' else String.sub s1' pos lens2' in (s' = s2')
      with _ -> false
    in
    Cursor.cursor_of_singleton (_boolean b)
  end

(* F&O Section 7.5.4 fn:substring-before *)

(** Gives the substring that occurs before the given character in a given string.
   Use the regular expression library to tokenize the string using the given character
   as a delimeter, and then take the head of the returned list as the substring to be returned
   @param comp_ctxt alg_ctxt The algebra context
   @param n The arguments
   @return A cursor containing the substring before the given character *)
let _fn_substring_before comp_ctxt alg_ctxt n =
  let (str, substr) = Args.get_array_param2 n in
  begin
    let str' = normalize_optional_string str in
    let substr' = normalize_optional_string substr in
    if (substr' = "") then 
      Cursor.cursor_of_singleton (_string "")
    else
      let rex = Str.regexp (Str.quote substr') in
      let split = (Str.bounded_split_delim rex str' 2) in
      let resstr = 
	if (List.length split < 2) then ""
	else List.hd split
      in Cursor.cursor_of_singleton (_string resstr)
  end
    
(* F&O Section 7.5.5 fn:substring-after *)

(** Since the delimeter can only be one character based on the definition in
   F&O, the Str.bounded_split function can be used to split the string into
   two separate strings at the character specified. 
   @param comp_ctxt alg_ctxt The algebra context (not used in this case)
   @param n The arguments
   @return A cursor containing the substring after the given chracater *)
let _fn_substring_after comp_ctxt alg_ctxt n =
  let (str, substr) = Args.get_array_param2 n in
  begin
    let str' = normalize_optional_string str in
    let substr' = normalize_optional_string substr in
    if (substr' = "") then 
      Cursor.cursor_of_singleton (_string str')
    else
      let rex = Str.regexp (Str.quote substr') in
      let split = (Str.bounded_split_delim rex str' 2) in
      let resstr = 
	if (List.length split < 2) then ""
	else List.hd (List.tl split)
      in Cursor.cursor_of_singleton (_string resstr)
  end

(* glx:string-pad : previously was fn:string-pad *)
 let _glx_string_pad comp_ctxt alg_ctxt n  =
   let (str, d) = Args.get_array_param2 n in
     if Cursor.cursor_is_empty str then
       Cursor.cursor_empty()
     else
   let str'   = normalize_optional_string str in
   let d'     = _int_of_integer (get_integer d) in
   let str'' = (String.make d' (String.get str' 0)) ^ str' in
   Cursor.cursor_of_singleton (_string str'')

(* F&0 Section 7.6 String Functions that use Pattern Matching *)

(** The matches, replace, and tokenize functions all use 
   the Regularexp library located in the datatypes folder *)

let _fn_matches comp_ctxt alg_ctxt n = 
  let (input, pattern, flags) = Args.get_array_param3 n in
  begin
    let input' = normalize_optional_string input in
    let pattern' = normalize_optional_string pattern in
    let flags' = normalize_optional_string flags in
    let res = Regularexp.matches input' pattern' flags' in
    Cursor.cursor_of_singleton (_boolean res)
  end

let _fn_replace comp_ctxt alg_ctxt n =
  let (input, pattern, reppattern, flags) = Args.get_array_param4 n in
  begin
    let input' = normalize_optional_string input in
    let pattern' = normalize_optional_string pattern in
    let reppatern' = normalize_optional_string reppattern in
    let flags' = normalize_optional_string flags in
    let res = Regularexp.replace input' pattern' reppatern' flags' in
    Cursor.cursor_of_singleton (_string res)
  end

let _fn_tokenize comp_ctxt alg_ctxt n =
  let (input, pattern, flags) = Args.get_array_param3 n in
  begin
    let input' = normalize_optional_string input in
    let pattern' = normalize_optional_string pattern in
    let flags' = normalize_optional_string flags in
    let res = Regularexp.tokenize input' pattern' flags' in
    Cursor.cursor_of_list (List.map _string res)
  end
    
(******************************************************)
(* F&O Section 8. Functions and Operators on Booleans *)
(******************************************************)

(* F&O Section 8.1.1 fn:true *)

let _fn_true comp_ctxt alg_ctxt n =
  Cursor.cursor_of_singleton (_boolean true)


(* F&O Section 8.1.2 fn:false *)

let _fn_false comp_ctxt alg_ctxt n =
  Cursor.cursor_of_singleton (_boolean false)

(* F&O Section 8.3.1 fn:not *)

let _fn_not comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let b1 = get_boolean p1 in
  Cursor.cursor_of_singleton (_boolean (not b1))

let _op_boolean_equal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_boolean, fun i1 i2 -> i1 = i2) comp_ctxt alg_ctxt n 

let _op_boolean_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op(get_boolean, fun i1 i2 -> i1 != i2) comp_ctxt alg_ctxt n 

let _op_boolean_lt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_boolean, fun i1 i2 -> i1 < i2) comp_ctxt alg_ctxt n 

let _op_boolean_gt comp_ctxt alg_ctxt n =
  _fn_compare_op(get_boolean, fun i1 i2 -> i1 > i2) comp_ctxt alg_ctxt n 

let _op_boolean_le comp_ctxt alg_ctxt n =
  _fn_compare_op(get_boolean, fun i1 i2 -> i1 <= i2) comp_ctxt alg_ctxt n 

let _op_boolean_ge comp_ctxt alg_ctxt n =
  _fn_compare_op(get_boolean, fun i1 i2 -> i1 >= i2) comp_ctxt alg_ctxt n 

(*******************************************************)
(* F&O Section 9. Functions on Dates, Times, Durations *)
(*******************************************************)
let _op_date_equal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_date, date_equal (Some tz)) comp_ctxt alg_ctxt n 

let _op_date_nequal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_date, (fun x y -> not(date_equal (Some tz) x y))) comp_ctxt alg_ctxt n 

let _op_date_lt comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_date, date_lt (Some tz)) comp_ctxt alg_ctxt n 

let _op_date_le comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_date, date_lteq (Some tz)) comp_ctxt alg_ctxt n 

let _op_date_gt comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_date, date_gt (Some tz)) comp_ctxt alg_ctxt n 

let _op_date_ge comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_date, date_gteq (Some tz)) comp_ctxt alg_ctxt n 

let _op_gYearMonth_equal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gYearMonth, gYearMonth_equal (Some tz)) comp_ctxt alg_ctxt n 
let _op_gYearMonth_nequal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gYearMonth, (fun x y -> not(gYearMonth_equal (Some tz) x y))) comp_ctxt alg_ctxt n 

let _op_gYear_equal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gYear, gYear_equal (Some tz)) comp_ctxt alg_ctxt n 
let _op_gYear_nequal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gYear, (fun x y -> not(gYear_equal (Some tz) x y))) comp_ctxt alg_ctxt n 

let _op_gMonthDay_equal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gMonthDay, gMonthDay_equal (Some tz)) comp_ctxt alg_ctxt n 
let _op_gMonthDay_nequal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gMonthDay, (fun x y -> not(gMonthDay_equal (Some tz) x y))) comp_ctxt alg_ctxt n 

let _op_gDay_equal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gDay, gDay_equal (Some tz)) comp_ctxt alg_ctxt n 
let _op_gDay_nequal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gDay, (fun x y -> not(gDay_equal (Some tz) x y))) comp_ctxt alg_ctxt n 

let _op_gMonth_equal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gMonth, gMonth_equal (Some tz)) comp_ctxt alg_ctxt n 
let _op_gMonth_nequal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_gMonth, (fun x y -> not(gMonth_equal (Some tz) x y))) comp_ctxt alg_ctxt n 

let _op_add_yearMonthDuration_to_date comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_date, get_yearMonthDuration, DateTime.add_yearMonthDuration_to_date, _date, def_zero_test) comp_ctxt alg_ctxt n

let _op_add_dayTimeDuration_to_date comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_date, get_dayTimeDuration, DateTime.add_dayTimeDuration_to_date, _date, def_zero_test) comp_ctxt alg_ctxt n

let _op_subtract_yearMonthDuration_from_date comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_date, get_yearMonthDuration, DateTime.subtract_yearMonthDuration_from_date, _date, def_zero_test) comp_ctxt alg_ctxt n

let _op_subtract_dayTimeDuration_from_date comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_date, get_dayTimeDuration, DateTime.subtract_dayTimeDuration_from_date, _date, def_zero_test) comp_ctxt alg_ctxt n

let _op_time_equal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_time, time_equal (Some tz)) comp_ctxt alg_ctxt n 

let _op_time_nequal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_time, (fun x y -> not(time_equal (Some tz) x y))) comp_ctxt alg_ctxt n 

let _op_time_lt comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_time, time_lt (Some tz)) comp_ctxt alg_ctxt n 

(***** LEFT OFF HERE *********)
let _op_time_le comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_time, time_lteq (Some tz)) comp_ctxt alg_ctxt n 

let _op_time_gt comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_time, time_gt (Some tz)) comp_ctxt alg_ctxt n 

let _op_time_ge comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op(get_time, time_gteq (Some tz)) comp_ctxt alg_ctxt n 

let _op_add_dayTimeDuration_to_time comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_time, get_dayTimeDuration, DateTime.add_dayTimeDuration_to_time, _time, def_zero_test) comp_ctxt alg_ctxt n

let _op_subtract_dayTimeDuration_from_time comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_time, get_dayTimeDuration, DateTime.subtract_dayTimeDuration_from_time, _time, def_zero_test) comp_ctxt alg_ctxt n

let _op_subtract_dateTimes comp_ctxt alg_ctxt n =
  _fn_binary_op
    (get_dateTime, DateTime.subtract_dateTimes, _dayTimeDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_subtract_dates comp_ctxt alg_ctxt n =
  _fn_binary_op
    (get_date, DateTime.subtract_dates, _dayTimeDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_subtract_times comp_ctxt alg_ctxt n =
  _fn_binary_op
    (get_time, DateTime.subtract_times, _dayTimeDuration, def_zero_test) comp_ctxt alg_ctxt n

let _fn_min_date comp_ctxt alg_ctxt n =
  _fn_min_poly(get_date_cursor, (DateTime.date_compare None), _date) comp_ctxt alg_ctxt n

let _fn_min_time comp_ctxt alg_ctxt n =
  _fn_min_poly(get_time_cursor, (DateTime.time_compare None), _time) comp_ctxt alg_ctxt n

let _fn_min_dateTime comp_ctxt alg_ctxt n =
  _fn_min_poly(get_dateTime_cursor, (DateTime.dateTime_compare None), _dateTime) comp_ctxt alg_ctxt n

let _fn_max_date comp_ctxt alg_ctxt n =
  _fn_max_poly(get_date_cursor, (DateTime.date_compare None), _date) comp_ctxt alg_ctxt n

let _fn_max_time comp_ctxt alg_ctxt n =
  _fn_max_poly(get_time_cursor, (DateTime.time_compare None), _time) comp_ctxt alg_ctxt n

let _fn_max_dateTime comp_ctxt alg_ctxt n =
  _fn_max_poly(get_dateTime_cursor, (DateTime.dateTime_compare None), _dateTime) comp_ctxt alg_ctxt n

let _op_dateTime_equal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op (get_dateTime, dateTime_equal (Some tz)) comp_ctxt alg_ctxt n

let _op_dateTime_nequal comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op (get_dateTime, (fun x y -> not(dateTime_equal (Some tz) x y))) comp_ctxt alg_ctxt n

let _op_dateTime_lt comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op (get_dateTime, dateTime_lt (Some tz)) comp_ctxt alg_ctxt n

let _op_dateTime_le comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op (get_dateTime, dateTime_lteq (Some tz)) comp_ctxt alg_ctxt n

let _op_dateTime_gt comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op (get_dateTime, dateTime_gt (Some tz)) comp_ctxt alg_ctxt n

let _op_dateTime_ge comp_ctxt alg_ctxt n =
  let tz = (get_timezone alg_ctxt)#getAtomicDayTimeDuration() in 
  _fn_compare_op (get_dateTime, dateTime_gteq (Some tz)) comp_ctxt alg_ctxt n

let _op_add_dayTimeDuration_to_dateTime comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_dateTime, get_dayTimeDuration, DateTime.add_dayTimeDuration_to_dateTime, _dateTime, def_zero_test) comp_ctxt alg_ctxt n

let _op_add_yearMonthDuration_to_dateTime comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_dateTime, get_yearMonthDuration, DateTime.add_yearMonthDuration_to_dateTime, _dateTime, def_zero_test) comp_ctxt alg_ctxt n

let _op_subtract_dayTimeDuration_from_dateTime comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_dateTime, get_dayTimeDuration, DateTime.subtract_dayTimeDuration_from_dateTime, _dateTime, def_zero_test) comp_ctxt alg_ctxt n

let _op_subtract_yearMonthDuration_from_dateTime comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_dateTime, get_yearMonthDuration, DateTime.subtract_yearMonthDuration_from_dateTime, _dateTime, def_zero_test) comp_ctxt alg_ctxt n

let _op_yearMonthDuration_equal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_yearMonthDuration, yearMonthDuration_equal) comp_ctxt alg_ctxt n

let _op_yearMonthDuration_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_yearMonthDuration, (fun x y -> not (yearMonthDuration_equal x y))) comp_ctxt alg_ctxt n

let _op_yearMonthDuration_lt comp_ctxt alg_ctxt n =
  _fn_compare_op (get_yearMonthDuration, yearMonthDuration_lt) comp_ctxt alg_ctxt n

let _op_yearMonthDuration_le comp_ctxt alg_ctxt n =
  _fn_compare_op (get_yearMonthDuration, yearMonthDuration_lteq) comp_ctxt alg_ctxt n

let _op_yearMonthDuration_gt comp_ctxt alg_ctxt n =
  _fn_compare_op (get_yearMonthDuration, yearMonthDuration_gt) comp_ctxt alg_ctxt n

let _op_yearMonthDuration_ge comp_ctxt alg_ctxt n =
  _fn_compare_op (get_yearMonthDuration, yearMonthDuration_gteq) comp_ctxt alg_ctxt n

let _op_subtract_yearMonthDurations comp_ctxt alg_ctxt n =
  _fn_binary_op
    (get_yearMonthDuration, DateTime.subtract_yearMonthDurations, _yearMonthDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_add_yearMonthDurations comp_ctxt alg_ctxt n =
  _fn_binary_op
    (get_yearMonthDuration, DateTime.add_yearMonthDurations, _yearMonthDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_multiply_yearMonthDuration comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_yearMonthDuration, get_double, DateTime.multiply_yearMonthDuration, _yearMonthDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_divide_yearMonthDuration comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_yearMonthDuration, get_double, DateTime.divide_yearMonthDuration, _yearMonthDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_divide_yearMonthDuration_by_yearMonthDuration comp_ctxt alg_ctxt n = 
  _fn_binary_op
    (get_yearMonthDuration, DateTime.divide_yearMonthDuration_by_yearMonthDuration, _decimal, def_zero_test) comp_ctxt alg_ctxt n

let _fn_sum_yearMonthDuration comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let all_durs = get_yearMonthDuration_cursor p1 in
  match (Cursor.cursor_peek all_durs) with
  | None ->
      p2
  | Some _ ->
      let ymzero = DateTime.mkyearMonthDuration (0, 0) in
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> DateTime.add_yearMonthDurations x0 x1) ymzero all_durs in
      Cursor.cursor_of_singleton (_yearMonthDuration sum)

let _fn_min_yearMonthDuration comp_ctxt alg_ctxt n =
  _fn_min_poly (get_yearMonthDuration_cursor, DateTime.yearMonthDuration_compare, _yearMonthDuration) comp_ctxt alg_ctxt n

let _fn_max_yearMonthDuration comp_ctxt alg_ctxt n =
  _fn_max_poly (get_yearMonthDuration_cursor, DateTime.yearMonthDuration_compare, _yearMonthDuration) comp_ctxt alg_ctxt n

let _fn_avg_yearMonthDuration comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allyearMonthDurations = get_yearMonthDuration_cursor p1 in
  match (Cursor.cursor_peek allyearMonthDurations) with
  | None ->
      Cursor.cursor_empty()
  | Some _ ->
      let length = ref 0.0 in
      let ymzero = DateTime.mkyearMonthDuration (0, 0) in
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> length := !length +. 1.0; DateTime.add_yearMonthDurations x0 x1) ymzero allyearMonthDurations in
      Cursor.cursor_of_singleton (_yearMonthDuration (DateTime.divide_yearMonthDuration sum (!length)))

(* xs:dayTimeDurations *)
let _op_dayTimeDuration_equal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_dayTimeDuration, dayTimeDuration_equal) comp_ctxt alg_ctxt n

let _op_dayTimeDuration_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_dayTimeDuration, (fun x y -> not(dayTimeDuration_equal x y))) comp_ctxt alg_ctxt n

let _op_dayTimeDuration_lt comp_ctxt alg_ctxt n =
  _fn_compare_op (get_dayTimeDuration, dayTimeDuration_lt) comp_ctxt alg_ctxt n

let _op_dayTimeDuration_le comp_ctxt alg_ctxt n =
  _fn_compare_op (get_dayTimeDuration, dayTimeDuration_lteq) comp_ctxt alg_ctxt n

let _op_dayTimeDuration_gt comp_ctxt alg_ctxt n =
  _fn_compare_op (get_dayTimeDuration, dayTimeDuration_gt) comp_ctxt alg_ctxt n

let _op_dayTimeDuration_ge comp_ctxt alg_ctxt n =
  _fn_compare_op (get_dayTimeDuration, dayTimeDuration_gteq) comp_ctxt alg_ctxt n

let _op_subtract_dayTimeDurations comp_ctxt alg_ctxt n =
  _fn_binary_op
    (get_dayTimeDuration, DateTime.subtract_dayTimeDurations, _dayTimeDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_add_dayTimeDurations comp_ctxt alg_ctxt n =
  _fn_binary_op
    (get_dayTimeDuration, DateTime.add_dayTimeDurations, _dayTimeDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_multiply_dayTimeDuration comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_dayTimeDuration, get_double, DateTime.multiply_dayTimeDuration, _dayTimeDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_divide_dayTimeDuration comp_ctxt alg_ctxt n =
  _fn_binary_op_uneven
    (get_dayTimeDuration, get_double, DateTime.divide_dayTimeDuration, _dayTimeDuration, def_zero_test) comp_ctxt alg_ctxt n

let _op_divide_dayTimeDuration_by_dayTimeDuration comp_ctxt alg_ctxt n = 
  _fn_binary_op
    (get_dayTimeDuration, DateTime.divide_dayTimeDuration_by_dayTimeDuration, _decimal, def_zero_test) comp_ctxt alg_ctxt n

(* xs:duration *)

let _op_duration_equal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_duration, duration_equal) comp_ctxt alg_ctxt n

let _op_duration_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_duration, (fun x y -> not (duration_equal x y))) comp_ctxt alg_ctxt n


let _fn_sum_dayTimeDuration comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let all_durs = get_dayTimeDuration_cursor p1 in
  match (Cursor.cursor_peek all_durs) with
  | None ->
      p2
  | Some _ ->
      let dtzero = DateTime.mkdayTimeDuration (0, 0, 0, Decimal._decimal_zero) in
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> DateTime.add_dayTimeDurations x0 x1) dtzero all_durs in
      Cursor.cursor_of_singleton (_dayTimeDuration sum)

let _fn_min_dayTimeDuration comp_ctxt alg_ctxt n =
  _fn_min_poly (get_dayTimeDuration_cursor, DateTime.dayTimeDuration_compare, _dayTimeDuration) comp_ctxt alg_ctxt n

let _fn_max_dayTimeDuration comp_ctxt alg_ctxt n =
  _fn_max_poly (get_dayTimeDuration_cursor, DateTime.dayTimeDuration_compare, _dayTimeDuration) comp_ctxt alg_ctxt n

let _fn_avg_dayTimeDuration comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let alldayTimeDurations = get_dayTimeDuration_cursor p1 in
  match (Cursor.cursor_peek alldayTimeDurations) with
  | None ->
      Cursor.cursor_empty()
  | Some _ ->
      let length = ref 0.0 in
      let dtzero = DateTime.mkdayTimeDuration (0, 0, 0, Decimal._decimal_zero) in
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> length := !length +. 1.0; DateTime.add_dayTimeDurations x0 x1) dtzero alldayTimeDurations in
      Cursor.cursor_of_singleton (_dayTimeDuration (DateTime.divide_dayTimeDuration sum (!length)))

(*****************************************)
(* F&O Section 9.4. Date/Time Extraction *)
(*****************************************)

let _fn_years_from_duration comp_ctxt alg_ctxt n =
  _fn_unary_op (get_yearMonthDuration, compose Decimal._integer_of_int DateTime.years_from_duration, _integer) comp_ctxt alg_ctxt n

let _fn_months_from_duration comp_ctxt alg_ctxt n =
  _fn_unary_op (get_yearMonthDuration, compose Decimal._integer_of_int DateTime.months_from_duration, _integer) comp_ctxt alg_ctxt n

let _fn_days_from_duration comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dayTimeDuration, compose Decimal._integer_of_int DateTime.days_from_duration, _integer) comp_ctxt alg_ctxt n

let _fn_hours_from_duration comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dayTimeDuration, compose Decimal._integer_of_int DateTime.hours_from_duration, _integer) comp_ctxt alg_ctxt n

let _fn_minutes_from_duration comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dayTimeDuration, compose Decimal._integer_of_int DateTime.minutes_from_duration, _integer) comp_ctxt alg_ctxt n

let _fn_seconds_from_duration comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dayTimeDuration, DateTime.seconds_from_duration, _decimal) comp_ctxt alg_ctxt n

let _fn_year_from_dateTime comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dateTime, compose Decimal._integer_of_int DateTime.year_from_dateTime, _integer) comp_ctxt alg_ctxt n

let _fn_month_from_dateTime comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dateTime, compose Decimal._integer_of_int DateTime.month_from_dateTime, _integer) comp_ctxt alg_ctxt n

let _fn_day_from_dateTime comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dateTime, compose Decimal._integer_of_int DateTime.day_from_dateTime, _integer) comp_ctxt alg_ctxt n

let _fn_hours_from_dateTime comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dateTime, compose Decimal._integer_of_int DateTime.hours_from_dateTime, _integer) comp_ctxt alg_ctxt n

let _fn_minutes_from_dateTime comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dateTime, compose Decimal._integer_of_int DateTime.minutes_from_dateTime, _integer) comp_ctxt alg_ctxt n

let _fn_seconds_from_dateTime comp_ctxt alg_ctxt n =
  _fn_unary_op (get_dateTime, DateTime.seconds_from_dateTime, _decimal) comp_ctxt alg_ctxt n

let _fn_timezone_from_dateTime comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1)
  then Cursor.cursor_empty()
  else
    let d = get_dateTime p1 in
    match DateTime.opt_timezone_from_dateTime d with
    | None -> Cursor.cursor_empty()
    | Some tz -> Cursor.cursor_of_singleton (_dayTimeDuration tz)

let _fn_year_from_date comp_ctxt alg_ctxt n =
  _fn_unary_op (get_date, compose Decimal._integer_of_int DateTime.year_from_date, _integer) comp_ctxt alg_ctxt n

let _fn_month_from_date comp_ctxt alg_ctxt n =
  _fn_unary_op (get_date, compose Decimal._integer_of_int DateTime.month_from_date, _integer) comp_ctxt alg_ctxt n

let _fn_day_from_date comp_ctxt alg_ctxt n =
  _fn_unary_op (get_date, compose Decimal._integer_of_int DateTime.day_from_date, _integer) comp_ctxt alg_ctxt n

let _fn_timezone_from_date comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1)
  then Cursor.cursor_empty()
  else
    let d = get_date p1 in
    match DateTime.opt_timezone_from_date d with
    | None -> Cursor.cursor_empty()
    | Some tz -> Cursor.cursor_of_singleton (_dayTimeDuration tz)

let _fn_hours_from_time comp_ctxt alg_ctxt n =
  _fn_unary_op (get_time, compose Decimal._integer_of_int DateTime.hours_from_time, _integer) comp_ctxt alg_ctxt n

let _fn_minutes_from_time comp_ctxt alg_ctxt n =
  _fn_unary_op (get_time, compose Decimal._integer_of_int DateTime.minutes_from_time, _integer) comp_ctxt alg_ctxt n

let _fn_seconds_from_time comp_ctxt alg_ctxt n =
  _fn_unary_op (get_time, DateTime.seconds_from_time, _decimal) comp_ctxt alg_ctxt n

let _fn_timezone_from_time comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1)
  then Cursor.cursor_empty()
  else
    let d = get_time p1 in
    match DateTime.opt_timezone_from_time d with
    | None -> Cursor.cursor_empty()
    | Some tz -> Cursor.cursor_of_singleton (_dayTimeDuration tz)

(* F&O Section 9.6 Timezone adjustments *)
let _fn_adjust_time_to_timezone comp_ctxt alg_ctxt n =
  let (t1, tz1) = Args.get_array_param2 n in
  if (Cursor.cursor_is_empty t1)
  then Cursor.cursor_empty()
  else
    let time = get_time t1 in
    let timezone = get_optional_timezone comp_ctxt alg_ctxt tz1 in
    Cursor.cursor_of_singleton(_time (DateTime.adjust_time_to_timezone time timezone))
  
let _fn_adjust_date_to_timezone comp_ctxt alg_ctxt n =
  let (d1, tz1) = Args.get_array_param2 n in
  if (Cursor.cursor_is_empty d1)
  then Cursor.cursor_empty()
  else
    let date = get_date d1 in
    let timezone = get_optional_timezone comp_ctxt alg_ctxt tz1 in
    Cursor.cursor_of_singleton (_date (DateTime.adjust_date_to_timezone date timezone))

let _fn_adjust_dateTime_to_timezone comp_ctxt alg_ctxt n = 
  let (dt1, tz1) = Args.get_array_param2 n in
  let (d1, tz1) = Args.get_array_param2 n in
  if (Cursor.cursor_is_empty dt1)
  then Cursor.cursor_empty()
  else
    let dateTime = get_dateTime dt1 in
    let timezone = get_optional_timezone comp_ctxt alg_ctxt tz1 in
    Cursor.cursor_of_singleton (_dateTime (DateTime.adjust_dateTime_to_timezone dateTime timezone))

(***************************************)
(* F&O Section 11. Functions on QNames *)
(***************************************)

(* F&O Section 11.1.2 fn:QName *)

let _fn_resolve_qname comp_ctxt alg_ctxt n =
  let (p1, p2) = Args.get_array_param2 n in
  let s1 = get_optional_string p1 in
  match s1 with
  | None -> Cursor.cursor_empty()
  | Some s1 ->
      let e = (get_singleton_node p2)#getElementNode() in
      let nsenv = e#namespace_environment() in
      let qn = ((new atomicQName (qname_of_untyped nsenv s1)) :> atomicValue) in
      _atomic_value qn

let get_prefix p =
  match fst p with
  | NSDefaultElementPrefix -> ""
  | NSPrefix ncname -> ncname
  | NSDefaultFunctionPrefix
  | NSWildcardPrefix -> raise (Query (Internal_Error "Wrong kind of prefix in fn:in-scope-prefix"))

let _fn_in_scope_prefixes comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let e = (get_singleton_node p1)#getElementNode() in
  let nsenv = e#namespace_environment() in
  let bindings = Namespace_context.flatten_bindings nsenv in
  let bindings = List.filter (fun (x,_) -> match x with NSDefaultFunctionPrefix | NSWildcardPrefix -> false | _ -> true) bindings in
  let prefixes = List.map (fun x -> _string (get_prefix x)) bindings in
  Cursor.cursor_of_list prefixes

let _fn_qname comp_ctxt alg_ctxt n =
  let (str1, str2) = Args.get_array_param2 n in
  let uri = normalize_optional_string str1 in
  let (prefix,localname) = uqname_element_of_raw_string (get_string str2) in
  if (uri = "") && (not(prefix = NSDefaultElementPrefix))
  then
    raise (Query (Wrong_Args "QName has a prefix in fn:QName"));
  let qname = Namespace_symbols.anon_symbol (prefix,NSUri uri, localname) in
  Cursor.cursor_of_singleton (_QName qname)


(* F&O Section 11.2.1 equality on QNames *)

let _op_QName_equal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_QName, qname_equal) comp_ctxt alg_ctxt n

let _op_QName_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_QName, (fun q1 q2 -> not(qname_equal q1 q2))) comp_ctxt alg_ctxt n

(* F&O Section 11.2.2 fn:get-local-name-from-QName *)

let _fn_local_name_from_qname comp_ctxt alg_ctxt n =
  let funqname qname =
    let (_,_,localname) = Namespace_symbols.anon_name qname in localname
  in
  _fn_unary_op (get_QName, funqname, _string) comp_ctxt alg_ctxt n

let _fn_prefix_from_qname comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1) then Cursor.cursor_empty()
  else
    let qn = get_QName p1 in
    let (prefix,_,_) = Namespace_symbols.anon_name qn in
    match prefix with
    | NSDefaultElementPrefix
    | NSDefaultFunctionPrefix
    | NSWildcardPrefix -> Cursor.cursor_empty()
    | NSPrefix ncname -> Cursor.cursor_of_singleton (_string ncname)

(* F&O Section 11.2.3 fn:get-namespace-uri-from-QName *)

let _fn_namespace_from_qname comp_ctxt alg_ctxt n =
  _fn_unary_op (get_QName, 
		(fun qname -> 
		  let (_,uri,_) = Namespace_symbols.anon_name qname in
		  match uri with
		  | NSUri urit -> urit
		  | _ -> raise (Query (Wrong_Args "Qname contains a wildcard URI"))),
		_string) comp_ctxt alg_ctxt n

let _fn_namespace_uri_for_prefix comp_ctxt alg_ctxt n = 
  let (p1, p2) = Args.get_array_param2 n in
  let prefix = normalize_optional_string p1 in
  let e = (get_singleton_node p2)#getElementNode() in
  let nsenv = e#namespace_environment() in
  let bindings = Namespace_context.flatten_bindings nsenv in
  let actual_prefix =
    match prefix with
    | "" -> NSDefaultElementPrefix
    | p -> NSPrefix p
  in
  begin
    try
      match List.assoc actual_prefix bindings with
      | NSUri u ->
	  Cursor.cursor_of_singleton (_anyURI (AnyURI._uri_of_string u))
      | _ -> Cursor.cursor_empty()
    with
    | Not_found -> Cursor.cursor_empty()
  end
  

(******************************************************)
(* F&O Section 11. Functions and Operators for anyURI *)
(******************************************************)

(* F&O Section 11.1 fn:resolve-uri *)

let _fn_resolve_uri comp_ctxt alg_ctxt n = 
  let proc_ctxt = processing_context_from_compile_context comp_ctxt in
  let (uri1, uri2) = Args.get_array_param2 n in
  let relative_uri_opt = get_optional_string uri1 in
  let base_uri_string = get_string uri2 in
  let base_uri =
    let pre_base_uri = AnyURI._uri_of_string base_uri_string in
    let static_base_uri  = Processing_context.get_base_uri proc_ctxt in
    begin
      match static_base_uri with
      | None -> pre_base_uri
      | Some stat_base_uri ->
	  AnyURI._uri_resolve stat_base_uri pre_base_uri
    end
  in
  match relative_uri_opt with
  | None -> Cursor.cursor_empty()
  | Some relative_uri ->
      let fixed_relative_uri =
	try
	  if (relative_uri.[0] = '/') || (base_uri_string.[(String.length base_uri_string)-1] = '/')
	  then
	    relative_uri
	  else
	    "/" ^ relative_uri
	with
	| _ -> relative_uri
      in
      let absolute_uri =
	AnyURI._uri_resolve base_uri (AnyURI._uri_of_string fixed_relative_uri)
      in
      Cursor.cursor_of_singleton (_string (AnyURI._string_of_uri absolute_uri))

let _op_anyURI_equal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_anyURI, anyURI_equal) comp_ctxt alg_ctxt n

let _op_anyURI_nequal comp_ctxt alg_ctxt n =
  _fn_compare_op (get_anyURI, (fun q1 q2 -> not(anyURI_equal q1 q2))) comp_ctxt alg_ctxt n

(****************************************************)
(* F&O Section 14. Functions and Operators on Nodes *)
(****************************************************)

(* F&O Section 14.1.1 fn:name *)

let _fn_name comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let name_string =
    if (Cursor.cursor_is_empty p1) then 
      ""
    else
      let node = get_singleton_node p1 in
      (get_node_qname node)
  in
  Cursor.cursor_of_singleton (_string name_string)


(* F&O Section 14.1.2 fn:local-name *)

let _fn_local_name comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let local_name_string =
    if (Cursor.cursor_is_empty p1) then 
      ""
    else
      let node = get_singleton_node p1 in
      match get_node_name node with 
      | Some (prefix,uri,local_name_string) ->
	  local_name_string
      | None ->
	  ""
  in Cursor.cursor_of_singleton (_string local_name_string)


(* F&O Section 14.1.3 fn:namespace-uri *)

let _fn_namespace_uri comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let local_name_string =
    if (Cursor.cursor_is_empty p1) then 
      ""
    else
      let node = get_singleton_node p1 in
      match get_node_name node with 
      | Some (prefix,NSUri uri,local_name_string) ->
	  uri
      | Some (prefix,NSWildcardUri,local_name_string) ->
	  raise (Query (Wrong_Args "Wildcard namespace should not appear in node"))
      | None ->
	  ""
  in Cursor.cursor_of_singleton (_string local_name_string)


(* F&O Section 14.1.5 fn:lang *)

let _fn_lang comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let textlang = normalize_optional_string p1 in
  let node = get_singleton_node p2 in
  let oactuallang = node#node_lang() in
  let result_bool =
    match oactuallang with
    | None -> false
    | Some actuallang ->
	if (String.lowercase actuallang) = (String.lowercase textlang)
	then
	  true
	else
	  let comparelang =
	    try
	      fst (Gmisc.split_left_on_char actuallang '-')
	    with
	    | _ -> actuallang
	  in
	  (String.lowercase comparelang) = (String.lowercase textlang)
  in
  Cursor.cursor_of_singleton (_boolean result_bool)

(* F&O Section 14.1.6 op:is-same-node *)

let _op_is_same_node comp_ctxt alg_ctxt n =
  _fn_binary_op (get_singleton_node, node_equal, _boolean, def_zero_test) comp_ctxt alg_ctxt n

(* F&O Section 14.1.7 op:node-before *)

let _fn_node_before comp_ctxt alg_ctxt n =
  _fn_binary_op (get_singleton_node, node_precedes, _boolean, def_zero_test) comp_ctxt alg_ctxt n

(* F&O Section 14.1.8 op:node-after *)

let _fn_node_after comp_ctxt alg_ctxt n =
  _fn_binary_op (get_singleton_node, node_follows, _boolean, def_zero_test) comp_ctxt alg_ctxt n

(* F&O Section 14.1.9 fn:root *)

let rec get_root (n : node) =
  let parent = n#parent None in
  match parent with
  | None -> (Item_Node n)
  | Some p -> get_root p

let _fn_root comp_ctxt alg_ctxt n =
  _fn_unary_op (get_singleton_node, get_root, fun i -> i) comp_ctxt alg_ctxt n

(* F&O Section 14.1.14 fn:number *)

let _fn_number comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let opt_item = Cursor.cursor_get_optional p1 in
  let new_base_value =
    match opt_item with
      (* If the input is the empty sequence, return NaN *)
    | None ->
	new atomicDouble nan
    | Some item ->
	let base_value = get_atomic item in
	try
	  base_value#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_double ATDouble
	with
	| _ ->
	    new atomicDouble nan
  in
  _atomic_value new_base_value
  

(********************************************************)
(* F&O Section 15. Functions and Operators on Sequences *)
(********************************************************)

(* F&O Section 15.1.1 fn:zero-or-one *)

let _fn_zero_or_one comp_ctxt alg_ctxt n =
  let items = Args.get_array_param1 n in
  if Cursor.cursor_is_optional items
  then items
  else 
   raise (Query (Wrong_Args ("Function [fn:zero-or-one]: not a singleton item or ()")))

(* F&O Section 15.1.2 fn:one-or-more *)

let _fn_one_or_more comp_ctxt alg_ctxt n =
  let items = Args.get_array_param1 n in
  if Cursor.cursor_is_empty items
  then 
   raise (Query (Wrong_Args ("Function [fn:one-or-more]: not a singleton item or more")))
  else 
    items

(* F&O Section 15.1.3 fn:exactly-one *)

let _fn_exactly_one comp_ctxt alg_ctxt n =
  let items = Args.get_array_param1 n in
  if Cursor.cursor_is_singleton items
  then items
  else 
   raise (Query (Wrong_Args ("Function [fn:exactly-one]: not a singleton item")))

(* F&O Section 15.1.1

   fn:boolean($arg as item()* ) as xs:boolean

      Summary: Computes the effective boolean value of the sequence $arg. 

    * If $arg is the empty sequence, fn:boolean returns false.
    * If $arg is a singleton value of type xs:boolean or a derived from xs:boolean, fn:boolean returns $arg.
    * If $arg is a singleton value of type xs:string or a type derived
      from xs:string or xs:untypedAtomic, fn:boolean returns false if
      the operand value has zero length; otherwise it returns true.
    * If $arg is a singleton value of any numeric type or a type
      derived from a numeric type, fn:boolean returns false if the
      operand value is NaN or is numerically equal to zero; otherwise
      it returns true.
    * If $arg is a sequence whose first item is a node, fn:boolean returns true.
    * In all other cases, fn:boolean raises a type error [err:FORG0006].
*)

let empty_string  = new atomicString ""
let empty_untyped = new atomicUntyped ""
let zero_integer  = new atomicInteger _integer_zero
let zero_decimal  = new atomicDecimal _decimal_zero
let zero_float    = new atomicFloat 0.0
let zero_double   = new atomicDouble 0.0

let _fn_boolean comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let effective_boolean_value =
    if (Cursor.cursor_is_empty p1) then  false
    else if (Cursor.cursor_is_singleton p1) then 
      let item = get_item p1 in
      if (isAtomicValue item) then
	let a = getAtomicValue item in
	match a#getAtomicValueKind() with
	| ATBoolean -> a#getAtomicBoolean()
	| ATString  -> not (a#atomic_value_eq empty_string)
	| ATAnyURI  -> not ((a#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_string ATString)#atomic_value_eq empty_string)
	| ATUntypedAtomic -> not (a#atomic_value_eq empty_untyped)
	| ATInteger -> not (a#atomic_value_eq zero_integer)
	| ATDecimal -> not (a#atomic_value_eq zero_decimal)
	| ATFloat   -> not ((a#atomic_value_eq zero_float) || ((a#erase_atomic_value()) = "NaN"))
	| ATDouble  -> not ((a#atomic_value_eq zero_double) || ((a#erase_atomic_value()) = "NaN"))
	| _ -> 
	    raise (Query(Type_Error("err:FORG0006: Argument to fn:boolean,"^(a#string_value())^", not string, anyURI, numeric, or node")))
      else
	true
    else
      match (Cursor.cursor_peek p1) with
      | None ->
	  raise (Query(Internal_Error ("Empty cursor in fn:boolean")))
      | Some n ->
	  if not(isAtomicValue n)
	  then
	    true
	  else
	    raise (Query(Type_Error("err:FORG0006: Argument to fn:boolean is a sequence of more than one item")))
  in
  Cursor.cursor_of_singleton (_boolean effective_boolean_value)

(* F&O Section 15.1.6 fn:index-of *)

(* Note:
     The function index-of atomizes its arguments and applies value
     equality to identify index of the second atomic-value argument in
     the fnrst atomic-value-sequence argument.
   -mff *)

let _fn_index_of comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let seq = get_atomic_cursor p1 in
  let item = get_singleton_atomic p2 in
  let eq_fn (idx, idxlist) av =
    (* converts xs:untypedAtomic to xs:string, to get the 'eq' semantics *)
    let item =
      match item#getAtomicValueKind() with
      | ATUntypedAtomic -> (item#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_string ATString)
      | _ -> item
    in
    let av =
      match av#getAtomicValueKind() with
      | ATUntypedAtomic -> (av#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_string ATString)
      | _ -> av
    in
    if (item#atomic_value_eq av) then (idx+1, idxlist@[idx])
    else (idx+1, idxlist)
  in
  let (idx', idxlist) = Cursor.cursor_fold_left eq_fn (1, []) seq
  in _integer_cursor (Cursor.cursor_map (fun i -> _integer_of_int i) (Cursor.cursor_of_list idxlist))


(* F&O Section 15.1.7 fn:empty *)

let _fn_empty comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  Cursor.cursor_of_singleton (_boolean (Cursor.cursor_is_empty p1))

(* F&O Section 15.1.8 fn:exists *)

let _fn_exists comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  Cursor.cursor_of_singleton (_boolean (not(Cursor.cursor_is_empty p1)))


(* Functions that support removal of duplicate nodes *)

(* Note:
     The following is now implemented directly over a cursor.  This
     allows it to scale to large sequences of items.
   - Jerome 04/13/2004
*)

let remove_sorted_node_dup (nl : node Cursor.cursor) =
  let rec remove_sorted_node_dup_fun x =
    match Cursor.cursor_peek nl with
    | None -> None
    | Some n1 ->
	begin
	  Cursor.cursor_junk nl;
	  match Cursor.cursor_peek nl with
	  | None -> Some n1
	  | Some n2 ->
	      if (node_equal n1 n2)
	      then
		remove_sorted_node_dup_fun x
	      else
		Some n1
	end
  in
  Cursor.cursor_of_function remove_sorted_node_dup_fun

(* F&O Section 15.1.9 fn:distinct-values *)

let valid_promotion t1 t2 pt =
  (t1 = pt) || (t2 = pt)

let dup_match ht promoted orig_type1 =
  let one_match x =
    let promoted_type = x#getAtomicValueKind() in
    if AtomicValueHash.mem ht x
    then
      let orig_type2 = AtomicValueHash.find ht x in
      valid_promotion orig_type1 orig_type2 promoted_type
    else
      false
  in
  (List.exists one_match promoted)

let remove_atomic_dup nsenv l =
  let ht = AtomicValueHash.create 1 in
  let hasnan = ref false in
  (* They keep their types.. *)
  let dup_remove (v:Physical_value.item) =
    if is_nan v
    then
      if !hasnan then false else begin hasnan := true; true end
    else
    let av = getAtomicValue v in
    let original_type = av#getAtomicValueKind () in
    let promoted = Cs_util.promote_atomicValue_to_all nsenv av in
    if dup_match ht promoted original_type
    then false
    else
      begin
	List.iter (fun x -> AtomicValueHash.add ht x original_type) promoted;
	true
      end
  in
  List.filter dup_remove l

let _fn_distinct_value comp_ctxt alg_ctxt n =
  let norm_ctxt = Compile_context.norm_context_from_compile_context comp_ctxt in
  let nsenv = Norm_context.nsenv_from_norm_context norm_ctxt in
  let p1 = Args.get_array_param1 n in
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._fn_distinct_value" p1 in
  let res = remove_atomic_dup nsenv pl1
  in (Cursor.cursor_of_list res)

(* F&O Section 15.1.10 fn:insert-before *)

let rec apply_insert n target inserts =
  if n <= 1
  then
    List.append inserts target
  else
    match target with
    | [] ->
	inserts
    | node :: target' ->
	node :: (apply_insert (n-1) target' inserts)

let _fn_insert_before comp_ctxt alg_ctxt n =
  let (p1,p2,p3) = Args.get_array_param3 n in
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._fn_insert_before" p1
  and pl3 = Cursor.list_of_cursor "Cs_code_fn._fn_insert_before" p3 in
  let i2 = _int_of_integer (get_integer p2) in
  Cursor.cursor_of_list (apply_insert i2 pl1 pl3)


(* F&O Section 15.1.11 fn:remove *)

let rec apply_remove n target =
  if n <= 0
  then
    target
  else
    match target with
    | [] ->
	[]
    | node :: target' ->
	if n = 1
	then
	  target'
	else
	  node :: (apply_remove (n-1) target')

let _fn_remove comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._fn_remove" p1 in
  let i2 = _int_of_integer (get_integer p2) in
  Cursor.cursor_of_list (apply_remove i2 pl1)

(* F&O Section 15.1.12 fn:reverse *)
let _fn_reverse comp_ctxt alg_ctxt n =
  let (p1) = Args.get_array_param1 n in
  Cursor.cursor_of_list (Cursor.rev_list_of_cursor p1)

(* F&O Section 15.1.13 fn:subsequence *)

let _fn_subsequence_2 comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let i2 = _int_of_integer (get_integer p2) in
  Cursor.cursor_subsequence2 p1 i2

let _fn_subsequence_3 comp_ctxt alg_ctxt n =
  let (p1,p2,p3) = Args.get_array_param3 n in
  let i2 = _int_of_integer(get_integer p2) in
  let i3 = _int_of_integer(get_integer p3) in
  Cursor.cursor_subsequence3 p1 i2 i3

(* fn:unordeded is the identity function for now - Jerome *)

let _fn_unordered comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  p1


(* Note:
     fs:first and fs:last are built-in versions of
     fn:subsequence(Expr, 1/fs:last, 1)
*)

let _fs_first comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  Cursor.cursor_first p1

let _fs_last comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  Cursor.cursor_last p1


(* F&O Section 15.2.1 fn:deep-equal *)

let not_pi_or_comment n =
  match n#node_kind() with
  | ProcessingInstructionNodeKind
  | CommentNodeKind ->
      false
  | _ ->
      true

let same_name n1 n2 =
  match (n1, n2) with
  | (None, Some _)
  | (Some _, None) ->
      false
  | (None,None) ->
      true
  | (Some name1, Some name2) ->
      name1#atomic_value_eq name2

let compare_node_names n1 n2 =
  match (n1#node_name(), n2#node_name()) with
  | (None, Some _) ->
      -1
  | (Some _, None) ->
      1
  | (None,None) ->
      0
  | (Some name1, Some name2) ->
      name1#atomic_value_compare name2

let sort_attributes al =
  List.sort compare_node_names al

let rec item_cursor_equal il1 il2 =
  try
    Cursor.cursor_for_all2 item_equal il1 il2
  with
  | _ ->
      false

and node_cursor_equal nl1 nl2 =
  try
    Cursor.cursor_for_all2 node_deep_equal nl1 nl2
  with
  | _ ->
      false

and atomic_value_sequence_equal avl1 avl2 =
  try
    Cursor.cursor_for_all2 atomic_value_equal avl1 avl2
  with
  | _ ->
      false

and item_equal (i1 : item) (i2 : item) =
  match (item_kind i1,item_kind i2) with
  | (AtomicValueKind,AtomicValueKind) ->
      let av1 = getAtomicValue i1
      and av2 = getAtomicValue i2 in
      (av1#atomic_value_eq av2)
  | (NodeKind,NodeKind) ->
      node_deep_equal (getNode i1) (getNode i2)
  | _ ->
      false

and atomic_value_equal av1 av2 =
  (av1#atomic_value_eq av2)

and node_deep_equal n1 n2 =
  let name1 = n1#node_name()
  and name2 = n2#node_name() in
  if not(same_name name1 name2)
  then
    false
  else
    match (n1#node_kind(),n2#node_kind()) with
    | (DocumentNodeKind, DocumentNodeKind) ->
	let dn1 = n1#getDocumentNode ()
	and dn2 = n2#getDocumentNode () in
	let nl1 = Cursor.cursor_filter not_pi_or_comment (dn1#children None)
	and nl2 = Cursor.cursor_filter not_pi_or_comment (dn2#children None) in
	node_cursor_equal nl1 nl2
    | (ElementNodeKind, ElementNodeKind) ->
	let en1 = n1#getElementNode ()
	and en2 = n2#getElementNode () in
	let anl1 = Cursor.cursor_map (fun n -> (n :> node)) (Cursor.cursor_of_list (sort_attributes (Cursor.list_of_cursor "Cs_code_fn._fn_deep_equal" (en1#attributes None))))
	and anl2 = Cursor.cursor_map (fun n -> (n :> node)) (Cursor.cursor_of_list (sort_attributes (Cursor.list_of_cursor "Cs_code_fn._fn_deep_equal" (en2#attributes None)))) in
	if not(node_cursor_equal anl1 anl2)
	then
	  false
	else
	  let nl1 = Cursor.cursor_filter not_pi_or_comment (en1#children None)
	  and nl2 = Cursor.cursor_filter not_pi_or_comment (en2#children None) in
	  node_cursor_equal nl1 nl2
    | (AttributeNodeKind, AttributeNodeKind) ->
	let av1 = n1#typed_value()
	and av2 = n2#typed_value() in
	atomic_value_sequence_equal av1 av2
    | (TextNodeKind, TextNodeKind)
    | (ProcessingInstructionNodeKind, ProcessingInstructionNodeKind)
    | (CommentNodeKind, CommentNodeKind) ->
	n1#string_value() = n2#string_value()
    | _ ->
	false

let _fn_deep_equal comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  Cursor.cursor_of_singleton (_boolean (item_cursor_equal p1 p2))


let stringval i =
  if (isNode i) then 
    (getNode i)#string_value()
  else
    (getAtomicValue i)#erase_atomic_value()

module DeepValueHashType = 
 struct 
   type t = item
   let equal x y = item_equal x y

   let hash i = Hashtbl.hash (stringval i)
 end

module DeepValueHash = Hashtbl.Make(DeepValueHashType)


let remove_deep_dup nsenv l =
  let ht = DeepValueHash.create 1 in    
  (* They keep their types.. *)
  let dup_remove (v:Physical_value.item) = 
    if DeepValueHash.mem ht v then
      false
    else
      begin
	DeepValueHash.add ht v ();
	true
      end
  in
  List.filter dup_remove l
      
let _glx_deep_distinct comp_ctxt alg_ctxt n =
  let norm_ctxt = Compile_context.norm_context_from_compile_context comp_ctxt in
  let nsenv = Norm_context.nsenv_from_norm_context norm_ctxt in

  let p1 = Args.get_array_param1 n in
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._fn_deep_distinct" p1 in
  let res = remove_deep_dup nsenv pl1
  in (Cursor.cursor_of_list res)


(* F&O Section 15.2.2 op:union *)
(* 

  From XQuery spec:

  All these operators eliminate duplicate nodes from their result
  sequences based on node identity. If ordering mode is ordered, the
  resulting sequence is returned in document order; otherwise it is
  returned in implementation-dependent order.

  Since Galax does not yet support ordering mode, these functions
  enforce doc order.

  *** NB: We should probably change the normalization of op:{union,
  intersect, except} so that the SBDO operator is explicit.  Then if
  we know the inputs are already in document order, we can use faster
  sorted-merge implementations for each operator and also avoid the
  explicit SBDO.  These rewritings should be added to the rewriting
  rules.

*)

(* Sorting by document order and duplicate removal *)

let item_node_compare n1 n2 = 
  node_compare n1 n2

let _op_union comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let sorted_list = List.sort item_node_compare (List.append (Cursor.list_of_cursor "Cs_code_op_union" (get_node_cursor p1)) (Cursor.list_of_cursor "Cs_code_op_union" (get_node_cursor p2))) in
  _node_cursor (remove_sorted_node_dup (Cursor.cursor_of_list sorted_list))

(* F&O Section 15.2.3 op:intersect *)

let compare eq x y =
  eq x y

let rec apply_intersect eq nf1 nf2 =
  match nf1 with
  | [] -> []
  | node :: nf1' ->
      let rest_intersect = apply_intersect eq nf1' nf2 in
      if (List.exists (compare eq node) nf2)
      then
	node :: rest_intersect
      else
	rest_intersect

let _op_intersect comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let nf1 = get_node_cursor p1 in
  let nf2 = get_node_cursor p2 in
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._op_intersect" nf1 in
  let pl2 = Cursor.list_of_cursor "Cs_code_fn._op_intersect" nf2 in
  let sorted_list = List.sort item_node_compare (apply_intersect (fun i1 i2 -> node_equal i1 i2) pl1 pl2)
  in _node_cursor(remove_sorted_node_dup(Cursor.cursor_of_list sorted_list))


(* F&O Section 15.2.4 op:except *)

let rec apply_except eq nf1 nf2 =
  match nf1 with
  | [] -> []
  | node :: nf1' ->
      let rest_except = apply_except eq nf1' nf2 in
      if (List.exists (compare eq node) nf2)
      then
	rest_except
      else
	node :: rest_except

let _op_except comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let nf1 = get_node_cursor p1 in 
  let nf2 = get_node_cursor p2 in 
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._op_except" nf1 in
  let pl2 = Cursor.list_of_cursor "Cs_code_fn._op_except" nf2 in
  let sorted_list = List.sort item_node_compare (apply_except (fun i1 i2 -> node_equal i1 i2) pl1 pl2)
  in _node_cursor(remove_sorted_node_dup(Cursor.cursor_of_list sorted_list))


(********************************)
(* F&O Section 15.3 Aggregate functions *)
(********************************)

(* F&O Section 15.3.1 fn:count *)
let _fn_count comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  Cursor.cursor_of_singleton (_integer (Decimal._integer_of_int (Cursor.cursor_length p1)))

(* F&O Section 15.3.2 fn:avg *)
(* F&O Section 15.3.3 fn:max *)
(* F&O Section 15.3.4 fn:min *)
(* F&O Section 15.3.5 fn:sum *)
(* Monomorphic version implemented in order:
   xsd:integer, xsd:decimal, xsd:float, xsd:double *)

(* xsd:integer *)

let _fn_avg_integer comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allintegers = get_integer_cursor p1 in
  match (Cursor.cursor_peek allintegers) with
  | None ->
      Cursor.cursor_empty()
  | Some _ ->
      let length = ref 0 in
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> incr length ; _integer_add x0 x1) _integer_zero allintegers in
      Cursor.cursor_of_singleton (_decimal (_integer_div sum (_integer_of_int !length)))

let _fn_max_integer comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allintegers = get_integer_cursor p1 in
  match (Cursor.cursor_peek allintegers) with
  | None ->
      Cursor.cursor_empty()
  | Some start_max ->
      let max =
	Cursor.cursor_fold_left (fun x0 x1 -> if _integer_le x0 x1 then x1 else x0) start_max allintegers
      in
      Cursor.cursor_of_singleton (_integer max)

let _fn_min_integer comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allintegers = get_integer_cursor p1 in
  match (Cursor.cursor_peek allintegers) with
  | None ->
      Cursor.cursor_empty()
  | Some start_min ->
      let min =
	Cursor.cursor_fold_left (fun x0 x1 -> if _integer_lt x0 x1 then x0 else x1) start_min allintegers
      in
      Cursor.cursor_of_singleton (_integer min)

let _fn_sum_integer comp_ctxt alg_ctxt n =
  let (p1, p2) = Args.get_array_param2 n in
  let allintegers = get_integer_cursor p1 in
  match (Cursor.cursor_peek allintegers) with
  | None ->
      p2
  | Some _ ->
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> _integer_add x0 x1) _integer_zero allintegers in
      Cursor.cursor_of_singleton (_integer sum)

(* xsd:decimal *)

let _fn_avg_decimal comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let alldecimals = get_decimal_cursor p1 in
  match (Cursor.cursor_peek alldecimals) with
  | None ->
      Cursor.cursor_empty()
  | Some _ ->
      let length = ref 0 in
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> incr length ; _decimal_add x0 x1) _decimal_zero alldecimals in
      Cursor.cursor_of_singleton (_decimal (_decimal_div sum (_decimal_of_int !length)))

let _fn_max_decimal comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let alldecimals = get_decimal_cursor p1 in
  match (Cursor.cursor_peek alldecimals) with
  | None ->
      Cursor.cursor_empty()
  | Some start_max ->
      let max =
	Cursor.cursor_fold_left (fun x0 x1 -> if _decimal_le x0 x1 then x1 else x0) start_max alldecimals
      in
      Cursor.cursor_of_singleton (_decimal max)

let _fn_min_decimal comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let alldecimals = get_decimal_cursor p1 in
  match (Cursor.cursor_peek alldecimals) with
  | None ->
      Cursor.cursor_empty()
  | Some start_min ->
      let min =
	Cursor.cursor_fold_left (fun x0 x1 -> if _decimal_lt x0 x1 then x0 else x1) start_min alldecimals
      in
      Cursor.cursor_of_singleton (_decimal min)

let _fn_sum_decimal comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let alldecimals = get_decimal_cursor p1 in
  match (Cursor.cursor_peek alldecimals) with
  | None ->
      p2
  | Some _ ->
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> _decimal_add x0 x1) _decimal_zero alldecimals in
      Cursor.cursor_of_singleton (_decimal sum)

(* xsd:float *)

let _fn_avg_float comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allfloats = get_float_cursor p1 in
  match (Cursor.cursor_peek allfloats) with
  | None ->
      Cursor.cursor_empty()
  | Some _ ->
      let length = ref 0 in
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> incr length ; x0 +. x1) 0.0 allfloats in
      let sum = _cast_double_to_float sum in
      Cursor.cursor_of_singleton (_float (sum/.(float_of_int !length)))

let _fn_max_float comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allfloats = get_float_cursor p1 in
  match (Cursor.cursor_peek allfloats) with
  | None ->
      Cursor.cursor_empty()
  | Some start_max ->
      let max = Cursor.cursor_fold_left (fun x0 x1 -> if x0 <= x1 then x1 else x0) start_max allfloats in
      Cursor.cursor_of_singleton (_float max)

let _fn_min_float comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let allfloats = get_float_cursor p1 in
  match (Cursor.cursor_peek allfloats) with
  | None ->
      Cursor.cursor_empty()
  | Some start_min ->
      let min = Cursor.cursor_fold_left (fun x0 x1 -> if x0 < x1 then x0 else x1) start_min allfloats in
      Cursor.cursor_of_singleton (_float min)

let _fn_sum_float comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let allfloats = get_float_cursor p1 in
  match (Cursor.cursor_peek allfloats) with
  | None ->
      p2
  | Some _ ->
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> x0 +. x1) 0.0 allfloats in
      Cursor.cursor_of_singleton (_float sum)

(* xsd:double *)

let _fn_avg_double comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let alldoubles = get_double_cursor p1 in
  match (Cursor.cursor_peek alldoubles) with
  | None ->
      Cursor.cursor_empty()
  | Some _ ->
      let length = ref 0 in
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> incr length ; x0 +. x1) 0.0 alldoubles in
      Cursor.cursor_of_singleton (_double (sum/.(float_of_int !length)))

let _fn_max_double comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let alldoubles = get_double_cursor p1 in
  match (Cursor.cursor_peek alldoubles) with
  | None ->
      Cursor.cursor_empty()
  | Some start_max ->
      let max = Cursor.cursor_fold_left (fun x0 x1 -> if x0 <= x1 then x1 else x0) start_max alldoubles in
      Cursor.cursor_of_singleton (_double max)

let _fn_min_double comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let alldoubles = get_double_cursor p1 in
  match (Cursor.cursor_peek alldoubles) with
  | None ->
      Cursor.cursor_empty()
  | Some start_min ->
      let min = Cursor.cursor_fold_left (fun x0 x1 -> if x0 < x1 then x0 else x1) start_min alldoubles in
      Cursor.cursor_of_singleton (_double min)

let _fn_sum_double comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let alldoubles = get_double_cursor p1 in
  match (Cursor.cursor_peek alldoubles) with
  | None ->
      p2
  | Some _ ->
      let sum = Cursor.cursor_fold_left (fun x0 x1 -> x0 +. x1) 0.0 alldoubles in
      Cursor.cursor_of_singleton (_double sum)

(* F&O Section 15.4.1 op:to *)

let build_integer_cursor i1 i2 =
  let current_index = ref i1 in
  let build_integer_cursor_fun x =
    if Decimal._integer_gt !current_index i2
    then
      None
    else
      begin
	let i = !current_index in
	current_index := Decimal._integer_add i (_integer_of_int 1);
	Some i
      end
  in
  Cursor.cursor_of_function build_integer_cursor_fun

let _op_to comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  if Cursor.cursor_is_empty p1 || Cursor.cursor_is_empty p2
  then
    Cursor.cursor_empty()
  else
    let i1 = get_integer p1
    and i2 = get_integer p2 in
    _integer_cursor (build_integer_cursor i1 i2)


(* F&O Section 15.4.4 fn:doc *)

(* This function builds an XML stream differently for various kind of
   URIs.  - Jerome *)

let _fn_doc comp_ctxt =
  let proc_ctxt = processing_context_from_compile_context comp_ctxt in
  let alive_documents = alive_documents_from_compile_context comp_ctxt in
  let base_uri = Processing_context.get_base_uri proc_ctxt in
  (fun alg_ctxt n ->
    let p1 = Args.get_array_param1 n in
    if (Cursor.cursor_is_empty p1)
    then Cursor.cursor_empty()
    else
      let uri_string = get_string p1 in
      let absolute_uri_string =
	match base_uri with
	| None -> uri_string
	| Some base_uri ->
	    let uri = AnyURI._uri_of_string uri_string in
	    let absolute_uri = AnyURI._uri_resolve base_uri uri in
	    AnyURI._string_of_uri absolute_uri
      in
      (* check for name indices in code context *)
      
      let retrieve_document_function =
	Fn_doc.lookup_doc_function absolute_uri_string
      in
      Cursor.cursor_of_list (retrieve_document_function alive_documents proc_ctxt)
  )

(************************************************)
(* F&O Section 16. Functions on Algebra Context *)
(************************************************)
let _fn_current_datetime comp_ctxt alg_ctxt n =
  Cursor.cursor_of_singleton (Item_Atomic (get_current_datetime alg_ctxt))

let _fn_current_date comp_ctxt alg_ctxt n =
  Cursor.cursor_of_singleton (_date(DateTime.date_from_dateTime((get_current_datetime alg_ctxt)#getAtomicDateTime())))

let _fn_current_time comp_ctxt alg_ctxt n =
  Cursor.cursor_of_singleton (_time(DateTime.time_from_dateTime((get_current_datetime alg_ctxt)#getAtomicDateTime())))

let _fn_implicit_timezone comp_ctxt alg_ctxt n =
  Cursor.cursor_of_singleton (Item_Atomic (get_timezone alg_ctxt))

let _fn_static_base_uri comp_ctxt alg_ctxt n =
  let proc_ctxt = processing_context_from_compile_context comp_ctxt in
  match Processing_context.get_base_uri proc_ctxt with
  | None -> Cursor.cursor_empty()
  | Some base_uri ->
      Cursor.cursor_of_singleton (_anyURI base_uri)

let _fn_default_collation comp_ctxt alg_ctxt n =
  let proc_ctxt = processing_context_from_compile_context comp_ctxt in
  let uri = proc_ctxt.Processing_context.default_collation in
  Cursor.cursor_of_singleton (_anyURI uri)

let _fn_document_uri comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if Cursor.cursor_is_empty p1
  then
    Cursor.cursor_empty()
  else
    let node = get_singleton_node p1 in
    match node#node_kind() with
    | DocumentNodeKind ->
	begin
	  match ((node#getDocumentNode())#document_uri()) with
	  | None -> Cursor.cursor_empty()
	  | Some astr -> _atomic_value astr
	end
    | _ -> Cursor.cursor_empty()

let _fn_dateTime comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let d = get_date p1 in
  let t = get_time p2 in
  Cursor.cursor_of_singleton (_dateTime (DateTime.dateTime_from_date_and_time d t))

(************************)
(* Comparison functions *)
(************************)

(* comparison operations on atomic values *)

let _op_equal_left_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()
let _op_equal_right_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()

let _op_notequal_left_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()
let _op_notequal_right_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()

let _op_lteq_left_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()
let _op_lteq_right_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()

let _op_gteq_left_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()
let _op_gteq_right_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()

let _op_lt_left_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()
let _op_lt_right_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()

let _op_gt_left_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()
let _op_gt_right_empty comp_ctxt alg_ctxt n =
  Cursor.cursor_empty()

(******************************)
(* Formal Semantics functions *)
(******************************)

let _fs_docorder comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in 
  let node_cursor = get_node_cursor p1 in
  let node_list = Cursor.list_of_cursor "Cs_code_fn._fs_docorder" node_cursor in
  _node_list (List.sort item_node_compare node_list)

let _fs_docorder_or_atomic_sequence comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  (* Peek at fnrst item in sequence, then dispatch to fs_distinct_docorder or assert the sequence is atomic()* *)
  let oitem = Cursor.cursor_peek p1 in
  match oitem with 
  | None -> Cursor.cursor_empty()
  | Some item -> 
      if (item_kind item) = AtomicValueKind
      then _atomic_cursor (get_atomic_cursor p1)
      else _fs_docorder comp_ctxt alg_ctxt (Array.of_list [p1])

(* Note:
     The following fs:distinct function is an optimization for
     fs:distinct-docorder, which is only applied when we can infer
     that the input is already sorted by document order. This is why
     this is using the remove_sorted_node_dup operation.

     Note that the fs:distinct function is implemented as a cursor
     operation, while the fs:distinct-docorder function requires
     materialization, as it relies on sorting.

   - Jerome and Mary  09/03/2004
 *)
let _fs_distinct comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let node_cursor = (get_node_cursor p1) in
  _node_cursor (remove_sorted_node_dup node_cursor)

let _fs_distinct_or_atomic_sequence comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  (* Peek at fnrst item in sequence, then dispatch to fs_distinct_docorder or assert the sequence is atomic()* *)
  let oitem = Cursor.cursor_peek p1 in
  match oitem with 
  | None -> Cursor.cursor_empty()
  | Some item -> 
      if (item_kind item) = AtomicValueKind
      then _atomic_cursor (get_atomic_cursor p1)
      else _fs_distinct comp_ctxt alg_ctxt (Array.of_list [p1])

let _fs_distinct_docorder_aux comp_ctxt alg_ctxt p1 =
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._fn_distinct_docorder" (get_node_cursor p1) in
  let sorted_list = List.sort item_node_compare pl1 in
  _node_cursor (remove_sorted_node_dup (Cursor.cursor_of_list sorted_list))

let _fs_distinct_docorder comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  _fs_distinct_docorder_aux comp_ctxt alg_ctxt p1

let _fs_distinct_docorder_or_atomic_sequence comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  (* Peek at fnrst item in sequence, then dispatch to fs_distinct_docorder or assert the sequence is atomic()* *)
  let oitem = Cursor.cursor_peek p1 in
  match oitem with
  | None -> Cursor.cursor_empty()
  | Some item ->
      if (item_kind item) = AtomicValueKind
      then _atomic_cursor (get_atomic_cursor p1)
      else _fs_distinct_docorder_aux comp_ctxt alg_ctxt p1

(* Asserts that argument is sequence of nodes *)

let _fs_node_sequence comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in p1

let _fs_node_sequence_or_atomic_sequence comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  (* Peek at fnrst item in sequence, then dispatch to fs_distinct_docorder or assert the sequence is atomic()* *)
  let oitem = Cursor.cursor_peek p1 in
  match oitem with
  | None -> Cursor.cursor_empty()
  | Some item ->
      if (item_kind item) = AtomicValueKind
      then _atomic_cursor (get_atomic_cursor p1)
      else _node_cursor (get_node_cursor p1)

(* Element/Attribute construction *)

let _fs_item_sequence_to_untypedAtomic comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let items = Cursor.list_of_cursor "Cs_code_fn._fs_item_sequence_to_untypedAtomic" p1 in
  let get_atomic_content item =
    if (isNode item)
    then
      List.map (fun a -> a#string_value()) (Cursor.list_of_cursor "Cs_code_fn._fs_item_sequence_to_untypedAtomic" ((getNode item)#typed_value()))
    else
      [string_value item]
  in
  Cursor.cursor_of_singleton (_untyped (String.concat " " (List.concat (List.map get_atomic_content items))))

let _fs_item_sequence_to_untypedAtomic_optional comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1) then Cursor.cursor_empty()
  else
  let items = Cursor.list_of_cursor "Cs_code_fn._fs_item_sequence_to_untypedAtomic_optional" p1 in
  let get_atomic_content item =
    if (isNode item)
    then
      List.map (fun a -> a#string_value()) (Cursor.list_of_cursor "Cs_code_fn._fs_item_sequence_to_untypedAtomic_optional" ((getNode item)#typed_value()))
    else
      [string_value item]
  in
  Cursor.cursor_of_singleton (_untyped (String.concat " " (List.concat (List.map get_atomic_content items))))

(* fs:untyped-to-double is used in the definition of arithmetic operators *)
let _fs_untyped_to_double comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1) then Cursor.cursor_empty()
  else
    let a1 = get_singleton_atomic p1 in
      match a1#getAtomicValueKind() with
      | ATUntypedAtomic -> _atomic_value (a1#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_double ATDouble)
      | _ -> p1

(* fs:untyped-to-integer is used in the definition of idiv *)
let _fs_untyped_to_integer comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1) then Cursor.cursor_empty()
  else
    let a1 = get_singleton_atomic p1 in
      match a1#getAtomicValueKind() with
      | ATUntypedAtomic -> _atomic_value(a1#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_integer ATInteger)
      | _ -> p1

(* fs:untyped-to-string is used in the definition of value comparisons *)
let _fs_untyped_to_string comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1) then Cursor.cursor_empty()
  else
    let a1 = get_singleton_atomic p1 in
      match a1#getAtomicValueKind() with
      | ATUntypedAtomic -> _atomic_value(a1#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_string ATString)
      | _ -> p1

(* fs:untyped-to-any is used in the definition of general comparison
   operators. 

   What we implement:

   If fnrst arg is untypedAtomic and
     the second is untypedAtomic or string then we cast untypedAtomic to string.
     the second is numeric then we cast untypedAtomic to double.
     otherwise we cast untypedAtomic to the type of the other argument

*)
let _fs_untyped_to_any comp_ctxt alg_ctxt n =
  let (p1, p2) = Args.get_array_param2 n in
  let a1 = get_singleton_atomic p1 in
  let a2 = get_singleton_atomic p2 in
  match a1#getAtomicValueKind() with
  | ATUntypedAtomic -> 
      begin 
        let a2k = a2#getAtomicValueKind() in
	match a2k with
	| ATUntypedAtomic 
	| ATString ->
	    _atomic_value(a1#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_string ATString)
	| ATInteger 
	| ATDecimal
	| ATFloat
	| ATDouble ->
	    _atomic_value(a1#cast_to Namespace_context.default_xquery_nsenv Namespace_symbols_builtin.xs_double ATDouble)
	| _ ->
	    let a2t = a2#atomic_type() in
	    _atomic_value(a1#cast_to Namespace_context.default_xquery_nsenv a2t a2k)
      end
  | _ -> p1


(*******************)
(* Galax functions *)
(*******************)

(* Galax function glx:union-values *)

let _glx_union_values comp_ctxt alg_ctxt n =
  let norm_ctxt = Compile_context.norm_context_from_compile_context comp_ctxt in
  let nsenv = Norm_context.nsenv_from_norm_context norm_ctxt in
  let (p1,p2) = Args.get_array_param2 n in
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._glx_union_values" p1 in
  let pl2 = Cursor.list_of_cursor "Cs_code_fn._glx_union_values" p2 in
  let big_seq = pl1 @ pl2 in
  let res = remove_atomic_dup nsenv big_seq in
  (Cursor.cursor_of_list res)

(* Galax function glx:intersect-values *)

let _glx_intersect_values comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let nf1 = get_atomic_cursor p1 in
  let nf2 = get_atomic_cursor p2 in
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._glx_intersect_values" nf1 in
  let pl2 = Cursor.list_of_cursor "Cs_code_fn._glx_intersect_values" nf2 in
  _atomic_list (apply_intersect (fun i1 i2 -> i1#atomic_value_eq(i2)) pl1 pl2)

(* Galax function glx:except-values *)

let _glx_except_values comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  let nf1 = get_atomic_cursor p1 in
  let nf2 = get_atomic_cursor p2 in
  let pl1 = Cursor.list_of_cursor "Cs_code_fn._glx_except_values" nf1 in
  let pl2 = Cursor.list_of_cursor "Cs_code_fn._glx_except_values" nf2 in
  _atomic_list (apply_except (fun i1 i2 -> i1#atomic_value_eq(i2)) pl1 pl2)

(* Galax function glx:print-string *)

let _glx_print_string comp_ctxt alg_ctxt n =
  let s1 = Args.get_array_param1 n in
  let s1' = get_string s1 in
  let _ =
    begin
      print_string s1';
      flush stdout
    end
  in
  Cursor.cursor_empty()

(* Galax function glx:print-item *)

let _glx_print_item comp_ctxt =
  let proc_ctxt = processing_context_from_compile_context comp_ctxt in
  (fun alg_ctxt n ->
    let i1 = Args.get_array_param1 n in
    begin
      Serialization.fserialize_datamodel proc_ctxt Format.std_formatter i1;
      flush stdout;
      Cursor.cursor_empty()
    end)

(* Galax function glx:save-document *)

let _glx_save_document comp_ctxt =
  let proc_ctxt = processing_context_from_compile_context comp_ctxt in
  (fun alg_ctxt n ->
    let (output_file_param, document_param) = Args.get_array_param2 n in
    let output_file_name = get_string output_file_param in
    let output_channel = open_out output_file_name in
    let output_formatter = Format.formatter_of_out_channel output_channel in
    begin
      Serialization.fserialize_datamodel proc_ctxt output_formatter document_param;
      close_out output_channel;
      Cursor.cursor_empty()
    end)

(* Galax function glx:print-string-err *)

let _glx_print_string_err comp_ctxt alg_ctxt n =
  let s1 = Args.get_array_param1 n in
  let s1' = get_string s1 in
  let _ =
(*    match !Conf.glx_stderr with
    | None -> ()
    | Some oc ->
*)
	begin
	  output_string !Conf.glx_stderr s1';
	  flush !Conf.glx_stderr
	end
  in
  Cursor.cursor_empty()

(* Galax function glx:print-item-err *)

let _glx_print_item_err comp_ctxt =
  let proc_ctxt = processing_context_from_compile_context comp_ctxt in
  (fun alg_ctxt n ->
    let i1 = Args.get_array_param1 n in
    let _ =
      begin
	let formatter = Format.formatter_of_out_channel !Conf.glx_stderr in
	Serialization.fserialize_datamodel proc_ctxt formatter i1;
	flush !Conf.glx_stderr
      end
    in
    Cursor.cursor_empty())

(* Galax function glx:get-lang *)

let _glx_get_lang comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let node = get_singleton_node p1 in
  let olang = node#node_lang() in
  match olang with
  | None -> Cursor.cursor_empty()
  | Some lang -> _atomic_value (new atomicString lang)

(* Galax function glx:exponent *)

let _glx_exponent comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let (i1,i2) = (get_double p1, get_double p2) in
  Cursor.cursor_of_singleton (_double(i1 ** i2))

(* Galax function glx:sqrt *)

let _glx_sqrt comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(sqrt i1))

(* Galax function glx:exp *)

let _glx_exp comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(exp i1))

(* Galax function glx:log *)

let _glx_log comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(log i1))

(* Galax function glx:log10 *)

let _glx_log10 comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(log10 i1))

(* Galax function glx:cos *)

let _glx_cos comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(cos i1))

(* Galax function glx:sin *)

let _glx_sin comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(sin i1))

(* Galax function glx:tan *)

let _glx_tan comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(tan i1))

(* Galax function glx:acos *)

let _glx_acos comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(acos i1))

(* Galax function glx:asin *)

let _glx_asin comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(asin i1))

(* Galax function glx:atan *)

let _glx_atan comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(atan i1))

(* Galax function glx:atan2 *)

let _glx_atan2 comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let (i1,i2) = (get_double p1, get_double p2) in
  Cursor.cursor_of_singleton (_double(atan2 i1 i2))

(* Galax function glx:cosh *)

let _glx_cosh comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(cosh i1))

(* Galax function glx:sinh *)

let _glx_sinh comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(sinh i1))

(* Galax function glx:tanh *)

let _glx_tanh comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (Cursor.cursor_is_empty p1) || (Cursor.cursor_is_empty p2) then Cursor.cursor_empty()
  else *)
  let i1 = get_double p1 in
  Cursor.cursor_of_singleton (_double(tanh i1))

let _glx_soap_call comp_ctxt alg_ctxt =
  (fun  n ->
    let (p1,p2,p3,p4) = Args.get_array_param4 n in
    (* server URI *)
    let uri_server = get_anyURI p1 in
    let string_server = AnyURI._string_of_uri uri_server in
    (* HTTP method ("GET" or "POST") *)
    let http_method = get_string p2 in
    (* soapAction (usually)*)
    let extra_params = get_string p3 in
    (* soap message *)
    let request = Serialization.bserialize_datamodel (Processing_context.default_processing_context()) p4 in
    if !Conf.debugging then
      print_string ("SOAP REQUEST:\n"^request); flush stdout;
    let response = Galax_http.http_soap string_server http_method extra_params request in
    if !Conf.debugging then
      print_string ("SOAP RESPONSE:\n"^response); flush stdout;
    let _, xml_stream = Streaming_parse.open_xml_stream_from_io (Galax_io.String_Input response) in

    (* First resolve namespaces *)
    let resolved_xml_stream = Streaming_ops.resolve_xml_stream xml_stream in
    (* Then apply type annotations *)
    let typed_xml_stream = Streaming_ops.typed_of_resolved_xml_stream resolved_xml_stream in
    (* Finally, load the document *)
    let nodeid_context = Nodeid_context.default_nodeid_context () in
    Cursor.cursor_of_list (Physical_load.load_xml_document_from_typed_stream nodeid_context typed_xml_stream))

let _glx_file_exists comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  if (Cursor.cursor_is_empty p1)
  then Cursor.cursor_empty()
  else
    let file_string = get_string p1 in
    Cursor.cursor_of_singleton(_boolean(Sys.file_exists(file_string)))

(* Galax function glx:stem *)

let _glx_stem comp_ctxt alg_ctxt n =
  let (p1,p2) = Args.get_array_param2 n in
  if (Cursor.cursor_is_empty p1)
  then Cursor.cursor_empty()
  else
    let in_word = get_string p1 in
    let case = get_string p2 in
    let s = Stemmer.stem in_word case 
    in Cursor.cursor_of_singleton(_string(s))

(* Added accessor function for pre-order -- Philippe *)
let _glx_get_order comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let item = Cursor.cursor_peek p1 in
  match item with
  | None ->
      raise (Query (Wrong_Args ("Function [glx:pre_order]: expected node(), got ()")))
  | Some i ->
      let ord = (get_node i)#docorder() in
      let (docid, pre) = Nodeid.big_int_pair_of_docorder ord in
      Cursor.cursor_of_singleton (_integer pre)
        
let _glx_get_docid comp_ctxt alg_ctxt n =
  let p1 = Args.get_array_param1 n in
  let item = Cursor.cursor_peek p1 in
  match item with
  | None ->
      raise (Query (Wrong_Args ("Function [glx:pre_order]: expected node(), got ()")))
  | Some i ->
      let ord = (get_node i)#docorder() in
      let (docid, pre) = Nodeid.big_int_pair_of_docorder ord in
      Cursor.cursor_of_singleton (_integer docid)
        
(*******************************)
(* Table of built-in functions *)
(*******************************)

let fun_table = [
  (fn_unique_ID, 1),  (_DUMMY (Namespace_names.prefixed_string_of_rqname fn_unique_ID)) ;   (* Dummy *)
  (* F&O Section 2. Accessors *)
  (fn_node_name, 1),  _fn_node_name; 
  (fn_nilled, 1),     _fn_nilled; 
  (fn_string, 1),     _fn_string;
  (fn_data, 1),       _fn_data;
  (fn_base_uri, 1),   _fn_base_uri_one; 
  (fn_boolean, 1),    _fn_boolean;

  (* F&O Section 5. Functions and Operators on Numbers *)
  (* Note: arithmetic functions are monomorphic in Galax for now... *)

  (op_integer_add, 2),      _fn_integer_add;
  (op_integer_subtract, 2), _fn_integer_sub;
  (op_integer_multiply, 2), _fn_integer_mult;
  (op_integer_divide, 2),   _fn_integer_div;
  (op_integer_idivide, 2),  _fn_integer_idiv;
  (op_integer_mod, 2),      _fn_integer_mod;

  (op_integer_unary_plus, 1),  _fn_integer_unary_plus;
  (op_integer_unary_minus, 1), _fn_integer_unary_minus;

  (op_integer_equal, 2),  _fn_integer_equal;
  (op_integer_nequal, 2), _fn_integer_nequal;
  (op_integer_lt, 2), 	  _fn_integer_lt;
  (op_integer_gt, 2), 	  _fn_integer_gt;
  (op_integer_le, 2), 	  _fn_integer_le;
  (op_integer_ge, 2), 	  _fn_integer_ge;

  (op_decimal_add, 2),      _fn_decimal_add;
  (op_decimal_subtract, 2), _fn_decimal_sub;
  (op_decimal_multiply, 2), _fn_decimal_mult;
  (op_decimal_divide, 2),   _fn_decimal_div;
  (op_decimal_idivide, 2),  _fn_decimal_idiv;
  (op_decimal_mod, 2),      _fn_decimal_mod;

  (op_decimal_unary_plus, 1),  _fn_decimal_unary_plus;
  (op_decimal_unary_minus, 1), _fn_decimal_unary_minus;

  (op_decimal_equal, 2),  _fn_decimal_equal;
  (op_decimal_nequal, 2), _fn_decimal_nequal;
  (op_decimal_lt, 2), 	  _fn_decimal_lt;
  (op_decimal_gt, 2), 	  _fn_decimal_gt;
  (op_decimal_le, 2), 	  _fn_decimal_le;
  (op_decimal_ge, 2), 	  _fn_decimal_ge;

  (op_float_add, 2),      _fn_float_add;
  (op_float_subtract, 2), _fn_float_sub;
  (op_float_multiply, 2), _fn_float_mult;
  (op_float_divide, 2),   _fn_float_div;
  (op_float_idivide, 2),  _fn_float_idiv;
  (op_float_mod, 2),      _fn_float_mod;

  (op_float_unary_plus, 1),  _fn_float_unary_plus;
  (op_float_unary_minus, 1), _fn_float_unary_minus;

  (op_float_equal, 2),  _fn_float_equal;
  (op_float_nequal, 2), _fn_float_nequal;
  (op_float_lt, 2), 	_fn_float_lt;
  (op_float_gt, 2), 	_fn_float_gt;
  (op_float_le, 2), 	_fn_float_le;
  (op_float_ge, 2), 	_fn_float_ge;

  (op_double_add, 2),      _fn_double_add;
  (op_double_subtract, 2), _fn_double_sub;
  (op_double_multiply, 2), _fn_double_mult;
  (op_double_divide, 2),   _fn_double_div;
  (op_double_idivide, 2),  _fn_double_idiv;
  (op_double_mod, 2),      _fn_double_mod;

  (op_double_unary_plus, 1),  _fn_double_unary_plus;
  (op_double_unary_minus, 1), _fn_double_unary_minus;

  (op_double_equal, 2),  _fn_double_equal;
  (op_double_nequal, 2), _fn_double_nequal;
  (op_double_lt, 2), 	 _fn_double_lt;
  (op_double_gt, 2), 	 _fn_double_gt;
  (op_double_le, 2), 	 _fn_double_le;
  (op_double_ge, 2), 	 _fn_double_ge;

  (fn_floor_double, 1),  _fn_floor_double;
  (fn_floor_float, 1),   _fn_floor_float;
  (fn_floor_decimal, 1), _fn_floor_decimal;
  (fn_floor_integer, 1), _fn_floor_integer;

  (fn_abs_double, 1), _fn_abs_double;
  (fn_abs_float, 1), _fn_abs_float;
  (fn_abs_decimal, 1), _fn_abs_decimal;
  (fn_abs_integer, 1), _fn_abs_integer;

  (fn_ceiling_double, 1), _fn_ceiling_double;
  (fn_ceiling_float, 1), _fn_ceiling_float;
  (fn_ceiling_decimal, 1), _fn_ceiling_decimal;
  (fn_ceiling_integer, 1), _fn_ceiling_integer;

  (fn_round_double, 1),  _fn_round_double;
  (fn_round_float, 1),   _fn_round_float;
  (fn_round_decimal, 1), _fn_round_decimal;
  (fn_round_integer, 1), _fn_round_integer;

  (fn_round_half_to_even_double, 2),  _fn_round_half_to_even_double;
  (fn_round_half_to_even_float, 2),   _fn_round_half_to_even_float;
  (fn_round_half_to_even_decimal, 2), _fn_round_half_to_even_decimal;
  (fn_round_half_to_even_integer, 2), _fn_round_half_to_even_integer;

  (* F&O Section 4. Trace *)

  (fn_trace, 2), _fn_trace;

  (* F&O Section 6. Functions on Strings *)

  (* Note:
       (**) indicates the functions which have been partially
       implemented.
     BChoi *)

  (fn_concat, 2),             _fn_concat;
  (fn_string_join, 2),        _fn_string_join;
  (fn_starts_with, 2),        _fn_starts_with;                         (**)
  (fn_ends_with, 2),          _fn_ends_with;                           (**)
  (fn_contains, 2),           _fn_contains;
  (fn_substring, 3),          _fn_substring;                           (**)
  (fn_string_length, 1),      _fn_string_length;
  (fn_substring_before, 2),   _fn_substring_before;
  (fn_substring_after, 2),    _fn_substring_after;
  (fn_normalize_space, 1),    _fn_normalize_space;
  (fn_normalize_unicode, 2),  _fn_normalize_unicode;
  (fn_upper_case, 1),         _fn_upper_case;
  (fn_lower_case, 1),         _fn_lower_case;
  (fn_translate, 3),          _fn_translate;
  (fn_encode_for_uri, 1),     _fn_encode_for_uri;
  (fn_iri_to_uri, 1),         _fn_iri_to_uri;
  (fn_escape_html_uri, 1),    _fn_escape_html_uri;
  (fn_codepoints_to_string, 1), _fn_codepts_to_string;
  (fn_string_to_codepoints, 1), _fn_string_to_codepts;
  (fn_codepoint_equal, 2),      _fn_codepoint_equal;

  (fn_compare, 2),            _fn_compare;
  (op_string_equal, 2),       _op_string_equal;
  (op_string_nequal, 2),      _op_string_nequal;
  (op_string_lt, 2),          _op_string_lt;
  (op_string_gt, 2),          _op_string_gt;
  (op_string_le, 2),          _op_string_le;
  (op_string_ge, 2),          _op_string_ge;
  
  (fn_matches, 3),            _fn_matches;
  (fn_replace, 4),            _fn_replace;
  (fn_tokenize, 3),           _fn_tokenize;

  (* F&O Section 7. Functions and Operators on Booleans *)
  (fn_true, 0),  _fn_true;
  (fn_false, 0), _fn_false;
  (fn_not, 1),   _fn_not;
  (op_boolean_equal, 2), _op_boolean_equal;
  (op_boolean_nequal, 2), _op_boolean_nequal;
  (op_boolean_lt, 2), _op_boolean_lt;
  (op_boolean_gt, 2), _op_boolean_gt;
  (op_boolean_le, 2), _op_boolean_le;
  (op_boolean_ge, 2), _op_boolean_ge;

  (* F&O Section 9. Functions on Dates, Times, Durations *)
  (op_date_equal, 2),       _op_date_equal;
  (op_date_nequal, 2),      _op_date_nequal;
  (op_date_lt, 2), _op_date_lt;
  (op_date_le, 2), _op_date_le;
  (op_date_gt, 2), _op_date_gt;
  (op_date_ge, 2), _op_date_ge;

  (op_gYearMonth_equal, 2),       _op_gYearMonth_equal;
  (op_gYearMonth_nequal, 2),      _op_gYearMonth_nequal;

  (op_gYear_equal, 2),       _op_gYear_equal;
  (op_gYear_nequal, 2),      _op_gYear_nequal;

  (op_gMonthDay_equal, 2),       _op_gMonthDay_equal;
  (op_gMonthDay_nequal, 2),      _op_gMonthDay_nequal;

  (op_gDay_equal, 2),       _op_gDay_equal;
  (op_gDay_nequal, 2),      _op_gDay_nequal;

  (op_gMonth_equal, 2),       _op_gMonth_equal;
  (op_gMonth_nequal, 2),      _op_gMonth_nequal;

  (op_add_yearMonthDuration_to_date, 2), _op_add_yearMonthDuration_to_date; 
  (op_add_dayTimeDuration_to_date, 2), _op_add_dayTimeDuration_to_date;
  (op_subtract_yearMonthDuration_from_date, 2), _op_subtract_yearMonthDuration_from_date;
  (op_subtract_dayTimeDuration_from_date, 2), _op_subtract_dayTimeDuration_from_date;
  (fn_min_date, 1),       _fn_min_date;
  (fn_max_date, 1),       _fn_max_date; 
 
  (op_time_equal, 2), _op_time_equal;
  (op_time_nequal, 2), _op_time_nequal;
  (op_time_lt, 2), _op_time_lt;
  (op_time_le, 2), _op_time_le;
  (op_time_gt, 2), _op_time_gt;
  (op_time_ge, 2), _op_time_ge;
  (op_add_dayTimeDuration_to_time, 2), _op_add_dayTimeDuration_to_time;
  (op_subtract_dayTimeDuration_from_time, 2), _op_subtract_dayTimeDuration_from_time;
  (op_subtract_dateTimes, 2), _op_subtract_dateTimes;
  (op_subtract_dates, 2), _op_subtract_dates;
  (op_subtract_times, 2), _op_subtract_times;
  (fn_min_time, 1), _fn_min_time;
  (fn_max_time, 1), _fn_max_time;

  (op_dateTime_equal, 2),   _op_dateTime_equal;
  (op_dateTime_nequal, 2),  _op_dateTime_nequal;
  (op_dateTime_lt, 2), _op_dateTime_lt;
  (op_dateTime_le, 2), _op_dateTime_le;
  (op_dateTime_gt, 2), _op_dateTime_gt;
  (op_dateTime_ge, 2), _op_dateTime_ge;
  (op_add_yearMonthDuration_to_dateTime, 2), _op_add_yearMonthDuration_to_dateTime; 
  (op_add_dayTimeDuration_to_dateTime, 2), _op_add_dayTimeDuration_to_dateTime;
  (op_subtract_yearMonthDuration_from_dateTime, 2), _op_subtract_yearMonthDuration_from_dateTime;
  (op_subtract_dayTimeDuration_from_dateTime, 2), _op_subtract_dayTimeDuration_from_dateTime;
  (fn_min_dateTime, 1),       _fn_min_dateTime;
  (fn_max_dateTime, 1),       _fn_max_dateTime;

  (op_yearMonthDuration_equal, 2),       _op_yearMonthDuration_equal;
  (op_yearMonthDuration_nequal, 2),       _op_yearMonthDuration_nequal;
  (op_yearMonthDuration_lt, 2), _op_yearMonthDuration_lt;
  (op_yearMonthDuration_le, 2), _op_yearMonthDuration_le;
  (op_yearMonthDuration_gt, 2), _op_yearMonthDuration_gt;
  (op_yearMonthDuration_ge, 2), _op_yearMonthDuration_ge;
  (op_add_yearMonthDurations, 2), _op_add_yearMonthDurations;
  (op_subtract_yearMonthDurations, 2), _op_subtract_yearMonthDurations;
  (op_multiply_yearMonthDuration, 2), _op_multiply_yearMonthDuration;
  (op_divide_yearMonthDuration, 2), _op_divide_yearMonthDuration;
  (op_divide_yearMonthDuration_by_yearMonthDuration, 2), _op_divide_yearMonthDuration_by_yearMonthDuration;
  (fn_min_yearMonthDuration, 1),       _fn_min_yearMonthDuration;
  (fn_max_yearMonthDuration, 1),       _fn_max_yearMonthDuration;
  (fn_avg_yearMonthDuration, 1),       _fn_avg_yearMonthDuration;
  (fn_sum_yearMonthDuration, 2),       _fn_sum_yearMonthDuration;

  (op_dayTimeDuration_equal, 2),       _op_dayTimeDuration_equal;
  (op_dayTimeDuration_nequal, 2),       _op_dayTimeDuration_nequal;
  (op_dayTimeDuration_lt, 2), _op_dayTimeDuration_lt;
  (op_dayTimeDuration_le, 2), _op_dayTimeDuration_le;
  (op_dayTimeDuration_gt, 2), _op_dayTimeDuration_gt;
  (op_dayTimeDuration_ge, 2), _op_dayTimeDuration_ge;
  (op_add_dayTimeDurations, 2), _op_add_dayTimeDurations;
  (op_subtract_dayTimeDurations, 2), _op_subtract_dayTimeDurations;
  (op_multiply_dayTimeDuration, 2), _op_multiply_dayTimeDuration;
  (op_divide_dayTimeDuration, 2), _op_divide_dayTimeDuration;
  (op_divide_dayTimeDuration_by_dayTimeDuration, 2), _op_divide_dayTimeDuration_by_dayTimeDuration;

  (op_duration_equal, 2),       _op_duration_equal;
  (op_duration_nequal, 2),       _op_duration_nequal;

  (fn_min_dayTimeDuration, 1),       _fn_min_dayTimeDuration;
  (fn_max_dayTimeDuration, 1),       _fn_max_dayTimeDuration;
  (fn_avg_dayTimeDuration, 1),       _fn_avg_dayTimeDuration;
  (fn_sum_dayTimeDuration, 2),       _fn_sum_dayTimeDuration;

  (* F&O Section 9.4, Extractions of Dates/Times *)
  (fn_years_from_duration, 1), _fn_years_from_duration;
  (fn_months_from_duration, 1), _fn_months_from_duration;
  (fn_days_from_duration, 1), _fn_days_from_duration;
  (fn_hours_from_duration, 1), _fn_hours_from_duration;
  (fn_minutes_from_duration, 1), _fn_minutes_from_duration;
  (fn_seconds_from_duration, 1), _fn_seconds_from_duration;
  (fn_year_from_dateTime, 1), _fn_year_from_dateTime;
  (fn_month_from_dateTime, 1), _fn_month_from_dateTime;
  (fn_day_from_dateTime, 1), _fn_day_from_dateTime;
  (fn_hours_from_dateTime, 1), _fn_hours_from_dateTime;
  (fn_minutes_from_dateTime, 1), _fn_minutes_from_dateTime;
  (fn_seconds_from_dateTime, 1), _fn_seconds_from_dateTime;
  (fn_timezone_from_dateTime, 1), _fn_timezone_from_dateTime;
  (fn_year_from_date, 1), _fn_year_from_date;
  (fn_month_from_date, 1), _fn_month_from_date;
  (fn_day_from_date, 1), _fn_day_from_date;
  (fn_timezone_from_date, 1), _fn_timezone_from_date;
  (fn_hours_from_time, 1), _fn_hours_from_time;
  (fn_minutes_from_time, 1), _fn_minutes_from_time;
  (fn_seconds_from_time, 1), _fn_seconds_from_time;
  (fn_timezone_from_time, 1), _fn_timezone_from_time;
  (fn_adjust_time_to_timezone, 2), _fn_adjust_time_to_timezone;
  (fn_adjust_date_to_timezone, 2), _fn_adjust_date_to_timezone;
  (fn_adjust_dateTime_to_timezone, 2), _fn_adjust_dateTime_to_timezone;

  (* F&O Section 10. Functions on QNames *)
  (fn_resolve_QName, 2), _fn_resolve_qname;
  (fn_in_scope_prefixes, 1), _fn_in_scope_prefixes;
  (fn_QName, 2), _fn_qname;
  (fn_local_name_from_QName, 1), _fn_local_name_from_qname; 
  (fn_prefix_from_QName, 1), _fn_prefix_from_qname; 
  (fn_namespace_uri_from_QName, 1), _fn_namespace_from_qname;
  (fn_namespace_uri_for_prefix, 2), _fn_namespace_uri_for_prefix;
  (op_QName_equal, 2), _op_QName_equal;
  (op_QName_nequal, 2), _op_QName_nequal;

  (op_hexBinary_equal, 2), _op_hexBinary_equal;
  (op_base64Binary_equal, 2), _op_base64Binary_equal;

  (op_hexBinary_nequal, 2), _op_hexBinary_nequal;
  (op_base64Binary_nequal, 2), _op_base64Binary_nequal;

  (* F&O Section 11. Functions and Operators for anyURI *)
  (fn_resolve_uri, 2), _fn_resolve_uri;
  (op_anyURI_equal, 2), _op_anyURI_equal;
  (op_anyURI_nequal, 2), _op_anyURI_nequal;

  (* F&O Section 13. Functions and Operators on Nodes *)
  (fn_name, 1), _fn_name;
  (fn_local_name, 1), _fn_local_name;
  (fn_namespace_uri, 1), _fn_namespace_uri;
  (fn_lang, 2), _fn_lang;
  (op_is_same_node, 2), _op_is_same_node;
  (op_node_before, 2), _fn_node_before;
  (op_node_after, 2), _fn_node_after;
  (fn_root, 1), _fn_root;
  (fn_number,1), _fn_number;

  (* F&O Section 15. Functions and Operators on Sequences *)
  (fn_exactly_one, 1),           _fn_exactly_one;
  (fn_zero_or_one, 1),           _fn_zero_or_one;
  (fn_one_or_more, 1),           _fn_one_or_more;

  (op_to, 2), _op_to;

  (fn_index_of, 2), _fn_index_of;
  (fn_empty, 1), _fn_empty;
  (fn_exists, 1), _fn_exists;
  (fn_distinct_values, 1), _fn_distinct_value;
  (fn_insert_before, 3), _fn_insert_before;
  (fn_remove, 2), _fn_remove;
  (fn_reverse, 1), _fn_reverse;
  (fn_subsequence, 2), _fn_subsequence_2;
  (fn_subsequence, 3), _fn_subsequence_3;
  (fn_unordered,1), _fn_unordered;

  (fn_deep_equal, 2), _fn_deep_equal;
  (op_union, 2), _op_union;
  (op_intersect, 2), _op_intersect;
  (op_except, 2), _op_except;

  (fn_count, 1), _fn_count;

  (fn_avg_integer, 1), _fn_avg_integer;
  (fn_max_integer, 1), _fn_max_integer;
  (fn_min_integer, 1), _fn_min_integer;
  (fn_sum_integer, 2), _fn_sum_integer;

  (fn_avg_decimal, 1), _fn_avg_decimal;
  (fn_max_decimal, 1), _fn_max_decimal;
  (fn_min_decimal, 1), _fn_min_decimal;
  (fn_sum_decimal, 2), _fn_sum_decimal;

  (fn_avg_float, 1), _fn_avg_float;
  (fn_max_float, 1), _fn_max_float;
  (fn_min_float, 1), _fn_min_float;
  (fn_sum_float, 2), _fn_sum_float;

  (fn_avg_double, 1), _fn_avg_double;
  (fn_max_double, 1), _fn_max_double;
  (fn_min_double, 1), _fn_min_double;
  (fn_sum_double, 2), _fn_sum_double;

  (fn_doc, 1), _fn_doc;

  (* F&O Section 16 : Functions on Dynamic Context *)

  (* Comparison functions *)
  (fn_current_datetime, 0), _fn_current_datetime;
  (fn_current_date, 0), _fn_current_date;
  (fn_current_time, 0), _fn_current_time;
  (fn_implicit_timezone, 0), _fn_implicit_timezone; 
  (fn_static_base_uri, 0), _fn_static_base_uri; 
  (fn_default_collation, 0), _fn_default_collation; 
  (fn_document_uri, 1), _fn_document_uri; 
  (fn_dateTime, 2), _fn_dateTime; 

  (* Note: The following functions are effectively polymorphic over atomic values. *)

  (op_equal_left_empty, 2), _op_equal_left_empty;
  (op_equal_right_empty, 2), _op_equal_right_empty;
  (op_nequal_left_empty, 2), _op_notequal_left_empty;
  (op_nequal_right_empty, 2), _op_notequal_right_empty;
  (op_le_left_empty, 2), _op_lteq_left_empty;
  (op_le_right_empty, 2), _op_lteq_right_empty;
  (op_ge_left_empty, 2), _op_gteq_left_empty;
  (op_ge_right_empty, 2), _op_gteq_right_empty;
  (op_gt_left_empty, 2), _op_gt_left_empty;
  (op_gt_right_empty, 2), _op_gt_right_empty;
  (op_lt_left_empty, 2), _op_lt_left_empty;
  (op_lt_right_empty, 2), _op_lt_right_empty;

  (* Formal Semantics functions *)

  (fs_distinct_docorder, 1), _fs_distinct_docorder;
  (fs_distinct_docorder_or_atomic_sequence, 1), _fs_distinct_docorder_or_atomic_sequence;
  (fs_distinct, 1), _fs_distinct;
  (fs_distinct_or_atomic_sequence, 1), _fs_distinct_or_atomic_sequence;
  (fs_docorder, 1), _fs_docorder;
  (fs_docorder_or_atomic_sequence, 1), _fs_docorder_or_atomic_sequence;
  (fs_node_sequence, 1), _fs_node_sequence;
  (fs_node_sequence_or_atomic_sequence, 1), _fs_node_sequence_or_atomic_sequence;
  (fs_item_sequence_to_untypedAtomic, 1), _fs_item_sequence_to_untypedAtomic;
  (fs_item_sequence_to_untypedAtomic_optional, 1), _fs_item_sequence_to_untypedAtomic_optional;

  (fs_untyped_to_double, 1), _fs_untyped_to_double;
  (fs_untyped_to_integer, 1), _fs_untyped_to_integer;
  (fs_untyped_to_string, 1), _fs_untyped_to_string;
  (fs_untyped_to_any, 2), _fs_untyped_to_any;

  (fs_first, 1), _fs_first;
  (fs_last_fn, 1), _fs_last;

  (* Galax functions *)

  (* Note:
       Those functions are not part of the F&O spec. They are added in
       Galax for user's convenience.
     _ Jerome *)
  
  (glx_string_pad, 3),        _glx_string_pad;
  (glx_union_values, 2), _glx_union_values;
  (glx_intersect_values, 2), _glx_intersect_values;
  (glx_except_values, 2), _glx_except_values;

  (glx_print_string, 1), _glx_print_string;
  (glx_print_item, 1), _glx_print_item;

  (glx_print_string_err, 1), _glx_print_string_err;
  (glx_print_item_err, 1), _glx_print_item_err;

  (glx_save_document, 2), _glx_save_document;

  (glx_get_lang, 1), _glx_get_lang;

  (glx_exponent, 2), _glx_exponent;
  (glx_sqrt, 1), _glx_sqrt;
  (glx_exp, 1), _glx_exp;
  (glx_log, 1), _glx_log;
  (glx_log10, 1), _glx_log10;
  (glx_cos, 1), _glx_cos;
  (glx_sin, 1), _glx_sin;
  (glx_tan, 1), _glx_tan;
  (glx_acos, 1), _glx_acos;
  (glx_asin, 1), _glx_asin;
  (glx_atan, 1), _glx_atan;
  (glx_atan2, 2), _glx_atan2;
  (glx_cosh, 1), _glx_cosh;
  (glx_sinh, 1), _glx_sinh;
  (glx_tanh, 1), _glx_tanh;

  (glx_soap_call, 4), _glx_soap_call;
  (glx_file_exists, 1), _glx_file_exists;
  (glx_deep_distinct, 1), _glx_deep_distinct;
  (glx_stem, 2), _glx_stem;

  (glx_get_order, 1), _glx_get_order;
  (glx_get_docid, 1), _glx_get_docid;
]

(* creation of the table for built-in functions *)

let bltin_fctns : (Algebra_type.alg_compile_context -> bltin) RQNameIntHashtbl.t = RQNameIntHashtbl.create 137
let internal_add_bltin_fctn ((cfname,arity) as cfname_arity, f) = 
  (* all internal builtins except trace are side-effect free, so register them *)
  if not (fn_trace = cfname) then 
    Factorize_side_effect.register_side_effect_free_function cfname
  ;
  Compile_context.register_builtin cfname_arity;
  RQNameIntHashtbl.add bltin_fctns cfname_arity f

let add_bltin_fctn ((rfname, i), f) = 
(*  let g = (fun comp_ctxt -> f) in *)
  internal_add_bltin_fctn ((rfname, i), f)

let lookup_bltin_fctn (cfname,arity) comp_ctxt =
  try
    (RQNameIntHashtbl.find bltin_fctns (cfname,arity)) comp_ctxt
  with _ ->
    raise (Query (Unknown ("Builtin function "
			   ^ (prefixed_string_of_rqname cfname)
			   ^ " with arity "
			   ^ (string_of_int arity)
			   ^ " not found")))

let _ = 
  try
    List.iter internal_add_bltin_fctn fun_table
  with exn -> eprintf_error " " exn


