(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies and AT&T Labs.            *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* Module: Code_util_materialize
   Description:
   This module contains code to handle the materialization of tuple cursors into
   an array of physical xml value arrays (the individual tuples)

   Additionally it returns a function which stores one of these tuples without names
   into the "current" tuple slot destructively.
*)

open Datatypes
open Datatypes_util
open Dm_atomic
open Dm

open Error
open Array
open Code_selection_context
open Physical_sequence

open Xquery_algebra_ast_annotation_util

type sequence_index = int
(***************************************)
(* Local type for materialization code *)
(***************************************)

type ('init,'return,'handle) materialize_handle = {
  init        : 'init   -> 'handle;
  add_tuple   : 'handle -> Algebra_type.eval_fun -> 
		            Execution_context.algebra_context -> 
		            Physical_value.dom_value array -> unit;
  get_return  : 'handle -> 'return;
  reset_table : 'handle -> unit}

let generic_materialize_tuple_cursor mh code_ctxt c_annot init_params = 
  (* This gets all the names that need to be there in the return. *)
  let needed_names = Array.of_list (get_returned_fields c_annot) in

  (* For each name build retrieve and restore code *)
  let retrieve_code = Array.map (build_retrieve_dom_tuple_code code_ctxt) needed_names in
  let restore_code  = Array.map (build_create_dom_tuple_code code_ctxt) needed_names in

  let n_functions   = Array.length restore_code in
  let make_tuple () = Array.map (fun f -> f ()) retrieve_code in

  (* Possible forget about name association since these tuples are not accessed any
     where... *)
    (* We should use arrays here instead *)
  let store_tuple (vl:Physical_value.dom_value array) =
    for i = 0 to n_functions - 1 do
      restore_code.(i) vl.(i)
    done
  in


  (* take in a cursor -> and then get all its stuff *)
  let tbl        = mh.init init_params in
  let add_fn     = mh.add_tuple tbl    in

  let materialize_fun =
    (fun eval alg_ctxt tc ->                      
       let materializing = ref 0 in
	 try
	   mh.reset_table tbl;
	   while true do
	     (* This cursor is just called for its side-effects *)
	     Cursor.cursor_next tc;
	     incr materializing;
	     add_fn eval alg_ctxt (make_tuple ())
	   done;
	   mh.get_return tbl
	 with Stream.Failure ->
	   mh.get_return tbl
    ) 
  in
    
    materialize_fun, store_tuple, needed_names



(*********************************)
(* Generic materialize array ref *)
(*********************************)

let generic_materialize_array mh code_ctxt array_restore_fn c_annot init_params = 
  (* This gets all the names that need to be there in the return. *)
  let needed_names = Array.of_list (get_returned_fields c_annot) in

  (* For each name build retrieve and restore code *)
  let retrieve_code = Array.map (build_retrieve_dom_tuple_code code_ctxt) needed_names in
  let restore_code  = Array.map (build_create_dom_tuple_code code_ctxt) needed_names in

  let n_functions   = Array.length restore_code in
  let make_tuple () = Array.map (fun f -> f ()) retrieve_code in

  (* Possible forget about name association since these tuples are not accessed any
     where... *)
  (* We should use arrays here instead *)
  let store_tuple (vl:Physical_value.dom_value array) =
    for i = 0 to n_functions - 1 do
      restore_code.(i) vl.(i)
    done
  in

  (* take in a cursor -> and then get all its stuff *)
  let tbl        = mh.init init_params in
  let add_fn     = mh.add_tuple tbl    in

  let materialize_fun =
    (fun eval alg_ctxt array_to_materialize ->                      
       let materializing = ref 0 in
       let array_length  = Array.length array_to_materialize in 

	 mh.reset_table tbl;

	 while !materializing < array_length do
	   array_restore_fn array_to_materialize.(!materializing);
	   add_fn eval alg_ctxt (make_tuple ());
	   incr materializing
	 done;

	 mh.get_return tbl
    ) 
  in
  materialize_fun, store_tuple, needed_names


(*************************************************************************************)
(*************************************************************************************)
(****************************** SPECIFIC MATERIALIZATION *****************************)
(*************************************************************************************)
(*************************************************************************************)

(******************)
(* Specific array *)
(******************)

type array_materialize_fun  = 
    Algebra_type.eval_fun -> Execution_context.algebra_context ->
    Physical_value.tuple_unit Cursor.cursor -> Physical_value.dom_value array array

type restore_function = Physical_value.dom_value array -> unit

let initial_size    =  1024
let dummy_value     =  Array.create 1 (materialized_of_list [])
type array_handle   =
    { mutable table          : Physical_value.dom_value array array;
      mutable current_size   : int;
      n_items        : int ref}

let array_init_table () =  
  { table = Array.create initial_size dummy_value;
    current_size = initial_size;
    n_items      = ref 0;}

let array_reset tbl =
  tbl.n_items := 0

(* Seperate out this structure for materialization
   and optimize it. Could even put in cardinality estimates in the stream
   Such estimates would be known when we compiled the original trees into  
   tuples. 
 *)
let array_add_tuple tbl eval alg_ctxt t =
  let n_items_ref = tbl.n_items in
    try
  (* If it is time to grow the table *)
  if ((!n_items_ref + 1) >= tbl.current_size) then
    begin 
      let new_size  = tbl.current_size * 2 in (* Multiplicative instead? *)
      let old_table = tbl.table in
      tbl.table <- Array.create new_size tbl.table.(0); (* There has to be a better way to do this *)
      tbl.current_size <- new_size;

      (* Copy over the items *)
      Array.blit old_table 0 tbl.table 0 !n_items_ref;

    end
      ;
      (* Here the array is sure to have the proper size *)

      tbl.table.(!n_items_ref) <- t;
      incr n_items_ref

    with Invalid_argument x ->
      begin
	raise (Query (Code_Selection ("Materialization Out of bounds")))
      end

let array_helper_get_items tbl = !(tbl.n_items)
let array_get_return       tbl = 
  
  (* We need to trim this table for the return now, to keep out invalid values *)
  let n_items   = array_helper_get_items tbl in
  let ret_array = Array.create n_items tbl.table.(0) in (* need something better here *)
    Array.blit tbl.table 0 ret_array 0 n_items;
    ret_array

let array_materialize_handler = {
  init        = array_init_table;
  add_tuple   = array_add_tuple;
  get_return  = array_get_return;
  reset_table = array_reset;
}

(*********************)
(* Hash Table Arrays *)
(*********************)

type hash_materialize_fun    =
    Algebra_type.eval_fun -> Execution_context.algebra_context 
      -> Physical_value.dom_value array array
      -> int AtomicValueHash.t

type string_hash_table_handle = {
  predicate_branch     : Code_util_predicates.predicate_branch;
  ht                   : int (* the sequence order/reference into the array *) AtomicValueHash.t;
  seq_order            : int ref;
  nsenv                : Namespace_context.nsenv;  
}

let hash_init (pb,nsenv) = 
  { predicate_branch     = pb;
    ht                   = AtomicValueHash.create 1;
    seq_order            = ref 0;
    nsenv                = nsenv;
  }


(* Adding tuple:
   if there are untypedAtomics in the 
   stream, then we handle their promotion according
   to fs:untyped-to-any (assuming all options)

   o If the operand can be cast -> a double then 
     we hash and sore it as a double
   
   o The operand is cast to a string -> 
     and hashed.

   --- both are done ---

   this involves calling promote_to_higest_numeric op
   
   Also all numerics are promoted before they are hashed
   to type double.
   
*)
let hash_add_tuple 
  (hh: string_hash_table_handle) 
  (eval: Algebra_type.eval_fun)
  (alg_ctxt: Execution_context.algebra_context) current_tup =

  (**********************)
  (* Wow, this is nasty *)
  (**********************)

  let cur_store = !(hh.seq_order) in 

  (******************************)
  (* Get the sequence of values to which the some_expr would have been bound,
     if there is no some expr -> this is None -> we are all done. *)
  let value_seq = Code_util_predicates.evaluate_predicate_branch hh.predicate_branch eval alg_ctxt hh.nsenv in        
    List.iter
      (fun key -> AtomicValueHash.add hh.ht key cur_store)
      value_seq ;
  
    incr hh.seq_order

let hash_get_return mh = mh.ht
let hash_reset mh = AtomicValueHash.clear mh.ht 

let hash_materialize_handler () = 
  { init        = hash_init;
    add_tuple   = hash_add_tuple;
    get_return  = hash_get_return;
    reset_table = hash_reset }


(********************)
(* Sort materialize *)
(********************)

type sort_array_materialize_fun  = 
    Algebra_type.eval_fun -> Execution_context.algebra_context 
      -> (Physical_value.dom_value array array)
	-> Code_util_ridlist.rid Dm_atomic_btree.btree (* in sorted order *)

type sort_array_handle   =
    { mutable bulk_loader    : Code_util_ridlist.rid Dm_atomic_btree.bulk_loader;
      predicate_branch       : Code_util_predicates.predicate_branch;
      nsenv                  : Namespace_context.nsenv;
      seq_order              : int ref }

let sort_array_init_table (pb,nsenv) =
  let compare i i' = i - i' in
  { bulk_loader          = Dm_atomic_btree.init compare;
    predicate_branch     = pb;
    nsenv                = nsenv;
    seq_order            = ref 0 }

(* Seperate out this structure for materialization
   and optimize it. Could even put in cardinality estimates in the stream
   Such estimates would be known when we compiled the original trees into  
   tuples. 
 *)
let sort_array_add_tuple tbl eval alg_ctxt t =
  let value_seq =
    Code_util_predicates.evaluate_predicate_branch
      tbl.predicate_branch
      eval
      alg_ctxt
      tbl.nsenv
  in
  let index = !(tbl.seq_order) in
  Dm_atomic_btree.bulk_add tbl.bulk_loader (List.map (fun x -> (x,index)) value_seq);
  incr tbl.seq_order

let sort_array_get_return tbl =
  Dm_atomic_btree.finalize tbl.bulk_loader

let sort_array_reset_table tbl =
  Dm_atomic_btree.reset tbl.bulk_loader

let sort_array_materialize_handler () = {
  init        = sort_array_init_table;
  add_tuple   = sort_array_add_tuple;
  get_return  = sort_array_get_return;
  reset_table = sort_array_reset_table;
}


(**********************)
(* Exported Functions *)
(**********************)

let materialize_cursor_to_array code_ctxt c_annot () =
  generic_materialize_tuple_cursor
    array_materialize_handler code_ctxt c_annot ()

let materialize_array_to_sorted_array_index code_ctxt c_annot restore_fn init =  
  let (x,_,_) =
    generic_materialize_array
      (sort_array_materialize_handler ()) code_ctxt restore_fn c_annot init
  in x

let materialize_array_to_hash code_ctxt c_annot restore_fn init =
  generic_materialize_array
    (hash_materialize_handler ()) code_ctxt restore_fn c_annot init

