(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: code_selection_context.mli,v 1.19 2006/02/07 17:29:33 stark Exp $ *)

(* Code Selection Context. 
   Notes on usage:

   The code selection context uses the typical caml method to handle scoping (copies and carrying around the context).
 
   This module handles setting up contexts for modules and functions and compiling variable references to
   array lookups.
  
*)

open Xquery_common_ast
open Namespace_names

open Physical_value

open Processing_context
open Typing_context
open Logical_algebra_types
open Algebra_type
open Variable_context_manager
open Tuple_context_manager

(******************************)
(* The code selection context *)
(******************************)

type code_selection_context

type fname = cfname * int

(* Default code selection context *)

val default_code_selection_context : Processing_context.processing_context -> code_selection_context

(* Replace the compilation context *)

val replace_compile_context_in_code_selection_context :
    alg_compile_context -> code_selection_context -> code_selection_context

val replace_code_type_context_in_code_selection_context :
    Code_typing_context.code_type_context -> code_selection_context -> code_selection_context

(* Compilation steps - Can only compile a function once (or it creates two records) *)

val enter_statement_context : code_selection_context -> code_selection_context
val exit_statement_context  : code_selection_context -> unit 
val enter_function_context  : code_selection_context -> fname -> Namespace_names.rqname array -> code_selection_context
val exit_function_context   : code_selection_context -> fname -> unit

val enter_prolog_context    : code_selection_context -> code_selection_context
val exit_prolog_context     : code_selection_context -> unit

val enter_scope             : code_selection_context -> code_selection_context
val exit_scope              : bool -> code_selection_context -> code_selection_context -> code_selection_context
val restore_variables       : code_selection_context -> code_selection_context -> code_selection_context
(* On creation *)
val store_annotation              : code_selection_context -> Xquery_algebra_ast.free_variable_desc option -> code_selection_context
val store_global_annotation : code_selection_context -> Xquery_algebra_ast.free_variable_desc option -> code_selection_context

(* annotation is not optional for retrieve *)
val retrieve_annotation           : code_selection_context -> Xquery_algebra_ast.free_variable_desc 
val is_valid_annotation           : code_selection_context -> bool
val retrieve_global_annotation               : code_selection_context -> Xquery_algebra_ast.free_variable_desc 
val is_valid_global_annotation               : code_selection_context -> bool

(************)
(* Variable *)
(************)

val add_variable_to_current_context : code_selection_context -> cvname -> code_selection_context
val add_variable_with_ref           : code_selection_context -> cvname -> variable_ref -> code_selection_context
val get_variable_reference          : code_selection_context -> cvname -> variable_ref

val merge_context                   : code_selection_context -> code_selection_context -> code_selection_context 

(*********)
(* Tuple *)
(*********)

val add_tuple_reference             : code_selection_context -> cvname -> code_selection_context
val get_tuple_reference             : code_selection_context -> cvname -> tuple_ref

(*******************************)
(* Accessors of other contexts *)
(*******************************)

val cxschema_from_code_selection_context :
    code_selection_context -> Xquery_type_core_ast.cxschema

val compile_context_from_code_selection_context : code_selection_context -> logical_alg_compile_context

val annotated_compile_context_from_code_selection_context :
    code_selection_context -> alg_compile_context

val static_context_from_code_selection_context :
    code_selection_context -> static_context

val processing_context_from_code_selection_context :
    code_selection_context -> processing_context

val code_type_context_from_code_selection_context :
    code_selection_context -> Code_typing_context.code_type_context

(**********************)
(* External Variables *)
(**********************)

val add_external_variable_value     : code_selection_context -> cvname -> item list -> code_selection_context
val declare_external_variable       : code_selection_context -> cvname -> code_selection_context
val get_external_variable_fn_value  : code_selection_context -> cvname -> (unit -> item list) ref

(*********************)
(*** Code builders ***)
(*********************)

val build_current_insert_code :
    code_selection_context -> cvname -> (physical_value -> unit)

val build_current_retrieve_code :
    code_selection_context -> cvname -> (unit -> physical_value)

val build_retrieve_tuple_code : 
    code_selection_context -> crname -> (unit -> xml_value)

val build_retrieve_dom_tuple_code : 
    code_selection_context -> crname -> (unit -> dom_value)

val build_create_tuple_code  :
    code_selection_context -> crname -> (xml_value -> unit)

val build_create_tuple_code_unsafe  :
    code_selection_context -> crname -> (xml_value -> unit)

val build_create_dom_tuple_code  :
    code_selection_context -> crname -> (dom_value -> unit)

val build_create_dom_tuple_code_unsafe  :
    code_selection_context -> crname -> (dom_value -> unit)

(* Since entrance and exit code have the same signature, this will hopefully
   prevent confusion in its use.  *)
type cs_function_code =
    { parameter_insertion_code : (physical_value -> unit) array;
      entrance_code  : (unit -> unit);
      exit_code      : (unit -> unit) }

val build_function_code :
    code_selection_context -> fname -> cvname array -> cs_function_code

val copy_code_selection_context : code_selection_context -> code_selection_context

(* Type for the code_selection main functions, which has to be passed
   as parameter in some parts of the code selection, notably for
   predicate handling in joins *)

type code_selection_function =
    code_selection_context -> algop_expr -> code_selection_context


(*************************)
(* Operations on indices *)
(*************************)

val create_new_name_index :
    code_selection_context -> Namespace_symbols.relem_symbol -> code_selection_context
(* Note the int is for the number of BTrees for the index *)

(* Access to name indices *)
val get_name_index :
    code_selection_context -> Namespace_symbols.relem_symbol -> Physical_name_index.name_index option

val get_name_index_handler :
    code_selection_context -> Namespace_symbols.relem_symbol -> Physical_name_index.name_index_handler option

val get_all_name_indices :
    code_selection_context -> Physical_name_index.name_indices

val get_all_name_indices_handler :
    code_selection_context -> Physical_name_index.name_indices_handler


