(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies and AT&T Labs.            *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* Module: Cs_code
   Description:
    This module contains the actual caml code that is used for evaluation
    of the algebra. 
*)

(* $Id: cs_code.ml,v 1.63 2006/05/16 23:33:59 simeon Exp $ *)

open Error
open Namespace_names

open Xquery_core_ast
open Xquery_common_ast
open Xquery_algebra_ast
open Algebra_type

open Schema_util

open Norm_context
open Typing_context
open Compile_context
open Code_selection_context
open Execution_context

open Datatypes

open Dm_types
open Dm_atomic

open Physical_value
open Physical_item
open Physical_sequence
open Physical_value_util 
open Physical_table

open Code_util
open Code_util_xpath
open Code_util_matching

open Cs_code_update
open Cs_util

open Cs_code_materialize (* should be in util *)
open Optimization_predicates

(* CEScalar *)
let build_default_cescalar_code code_ctxt s =
  let atomic_value = 
    match s with
    | IntegerLiteral i -> (new atomicInteger i)
    | DecimalLiteral d -> (new atomicDecimal d)
    | DoubleLiteral d -> (new atomicDouble d)
    | StringLiteral s -> (new atomicString s)
    | BooleanLiteral b -> (new atomicBoolean b)
    | URILiteral u -> (new atomicAnyURI u)
  in
  let dm_value = [Item_Atomic atomic_value] in
  (fun alg_ctxt () -> dm_value)

(* CEText *)
let build_default_cetext_code code_ctxt t =
  (fun alg_ctxt () -> 
     let dmv = (sequence_of_singleton (Item_Atomic (new atomicUntyped t))) in
     let sdmv = Physical_xml_value.sax_value_of_mat_value dmv in
     (Streaming_constructors.text_constructor sdmv))

let build_default_cetextcomputed_code code_ctxt =
  (fun alg_ctxt dmv ->
    Streaming_constructors.text_constructor dmv)

(* CEPI *)
let build_default_cepi_code code_ctxt target pi_content =
  (fun alg_ctxt () ->
    let dmv = sequence_of_singleton (Item_Atomic (new atomicUntyped pi_content)) in
    let sdmv = Physical_xml_value.sax_value_of_mat_value dmv in
    Streaming_constructors.pi_constructor false target sdmv)

(* CEAnyPI *)
let build_default_cepicomputed_code code_ctxt  =
  (fun alg_ctxt pv1 pv2 ->
    let stat_ctxt = static_context_from_code_selection_context code_ctxt in
    let norm_ctxt = norm_context_from_stat_context stat_ctxt in
    let nsenv     = nsenv_from_norm_context norm_ctxt in
    let target = ((Physical_util.get_singleton_atomic pv1)#cast_to nsenv Namespace_symbols_builtin.xs_string ATString)#getAtomicString() in
    Streaming_constructors.pi_constructor true target pv2)

(* CEComment *)
let build_default_cecomment_code code_ctxt c =
  (fun alg_ctxt () ->
    let dmv = sequence_of_singleton  (Item_Atomic (new atomicUntyped c)) in
    let sdmv = Physical_xml_value.sax_value_of_mat_value dmv in
    Streaming_constructors.comment_constructor sdmv)

let build_default_cecommentcomputed_code code_ctxt =
  (fun alg_ctxt dmv ->
     Streaming_constructors.comment_constructor dmv)

(* CEDocument *)
let build_default_document_code code_ctxt =
  let proc_ctxt = processing_context_from_code_selection_context code_ctxt in
  let base_uri = Processing_context.get_base_uri proc_ctxt in
  let base_uri_item = Dm_atomic_util.uri_dm_of_uri base_uri in
  (fun alg_ctxt content_stream ->
    Streaming_constructors.document_constructor base_uri_item content_stream)

(* Special functions *)

let build_default_convert_simple_code code_ctxt atomic_type =
  let bt = Datatypes_util.symbol_of_primitive_type atomic_type in
  (fun alg_ctxt p1 ->
   (* fs:convert-simple-operand is used in the definition of function
      calls that take sequences of atomic values as arguments.  The
      function converts each occurrence of an xs:untypedAtomic
      argument to the target type.  *)
    let a1l = Physical_util.get_atomic_cursor p1 in
    Cursor.cursor_map (fun a1 -> 
      match a1#getAtomicValueKind() with
      | ATUntypedAtomic ->
	  (Item_Atomic (a1#cast_to Namespace_context.default_xquery_nsenv bt atomic_type))
      | _ -> (Item_Atomic a1))
      a1l)

let build_default_promote_numeric_code code_ctxt atomic_type =
  let comp_ctxt = annotated_compile_context_from_code_selection_context code_ctxt in
  begin
    match atomic_type with
    | ATInteger 
    | ATDecimal 
    | ATFloat   
    | ATDouble  -> ()
    | _ -> raise(Query(Cast_Error("Attempt to promote to non-numeric type")))
  end;
  let bt = Datatypes_util.symbol_of_primitive_type atomic_type in
  (fun alg_ctxt p1 ->
    (* The algebra promotion functions take an input sequence of
       atomic values and attempt to promote each atomic value in the
       input sequence to the target type. *)
    let _fs_promote_atomic_value_to_type comp_ctxt alg_ctxt a1 = 
      let at = a1#getAtomicValueKind() in
      let atl = Datatypes_util.can_be_promoted_to at in
      if (List.mem atomic_type atl) then
	if (at = atomic_type)
	then (Item_Atomic a1)
	else (Item_Atomic (a1#cast_to Namespace_context.default_xquery_nsenv bt atomic_type))
      else
	raise (Query(Cast_Error("Cannot promote "^(Datatypes_util.string_of_atomic_type at)^" to "^(Datatypes_util.string_of_atomic_type atomic_type))))
    in
    let a1l = Physical_util.get_atomic_cursor p1 in 
    Cursor.cursor_map (_fs_promote_atomic_value_to_type comp_ctxt alg_ctxt) a1l
      (* fs:convert-simple-operand is used in the definition of function
	 calls that take sequences of atomic values as arguments.  The
	 function converts each occurrence of an xs:untypedAtomic
	 argument to the target type.  *)
  )

let build_default_unsafe_promote_numeric_code code_ctxt atomic_type =
  let bt = Datatypes_util.symbol_of_primitive_type atomic_type in
  (fun alg_ctxt p1 ->
    (* fs:unsafe-promote-to-numeric is the same as
       fs:promote-to-numeric but is only applied when static typing
       guarantees that every atomic value in the input sequence can be
       promoted to the target type.  *)
    let a1l = Physical_util.get_atomic_cursor p1 in 
    let a1l' = Cursor.cursor_map (
      fun a1 ->
	let at = a1#getAtomicValueKind() in
	if (at = atomic_type) then (Item_Atomic a1)
	else (Item_Atomic (a1#cast_to Namespace_context.default_xquery_nsenv bt atomic_type))
     ) a1l in
    a1l')


(* CEKeyRef *)
let build_default_cekeyref_code code_ctxt kn =
  (fun alg_ctxt kv ->
    let keyval =
      try
	if (Cursor.cursor_is_empty kv)
	then 
	  ""
	else
	  let i = Physical_util.get_item kv in
	  match item_kind i with
	  | NodeKind -> (getNode i)#string_value()
	  | AtomicValueKind -> (getAtomicValue i)#erase_atomic_value()
      with
      | _ ->
	  raise (Query (Code_Selection "key definition: key value not an item"))
    in
    key_from_algebra_context alg_ctxt (kn,keyval)
  )

(* CESeq *)
let build_default_ceseq_code code_ctxt =
  (fun alg_ctxt ae1 ae2 -> ae1 @ ae2)

(*
let build_default_ceseq_code code_ctxt = 
  (fun alg_ctxt ae1 ae2 -> Cursor.cursor_append ae1 ae2)
*)

(* CEEmpty *)
let build_default_ceempty_code code_ctxt = 
  (fun alg_ctxt () -> [])

(* CEElem *)
let build_default_ceelem_code code_ctxt sym nsenv =
  let proc_ctxt = processing_context_from_code_selection_context code_ctxt in
  let base_uri = Processing_context.get_base_uri proc_ctxt in
  let base_uri_item = Dm_atomic_util.uri_dm_of_uri base_uri in
  (fun alg_ctxt content_streams ->
    let merged_stream = Cursor.cursor_array_fold content_streams in
    Streaming_constructors.element_constructor base_uri_item sym nsenv merged_stream)

(* CEAnyElem *)
let build_default_ceanyelem_code code_ctxt nsenv = 
  let proc_ctxt = processing_context_from_code_selection_context code_ctxt in
  let base_uri = Processing_context.get_base_uri proc_ctxt in
  let base_uri_item = Dm_atomic_util.uri_dm_of_uri base_uri in
  (fun alg_ctxt pv1 content_stream -> 
    let sym = Physical_util.get_QName pv1 in
    Streaming_constructors.element_constructor base_uri_item sym nsenv content_stream)

(* CEAttr *)
let build_default_ceattr_code code_ctxt sym = 
  (fun alg_ctxt content_streams ->
      let merged_stream = Cursor.cursor_array_fold content_streams in
	Streaming_constructors.attribute_constructor sym merged_stream)

(* CEAnyAttr *)
let build_default_ceanyattr_code code_ctxt nsenv =
  (fun alg_ctxt pv1 content_stream ->
    let sym = Physical_util.get_QName pv1 in
    Streaming_constructors.attribute_constructor sym content_stream)

(* CEError *)
let build_default_ceerror_code code_ctxt =
  (fun alg_ctxt v ->
    let proc_ctxt = processing_context_from_code_selection_context code_ctxt in
    let vl =  Cursor.list_of_cursor "Cs_code.build_default_ceerror_code" (Cursor.cursor_array_fold v) in 
    Fn_error.raise_error vl
  )

(* CETreat *)
   (* JS: no need to perform validation there
      because strong typing makes sure this will succeed *)
   (* JS: on the other hand, wouldn't validation annotate the data
      with appropriate type information ? *)
let build_default_cetreat_code code_ctxt cdt =
  let stat_ctxt = static_context_from_code_selection_context code_ctxt in
  (fun alg_ctxt c ->
    dynamic_type_check stat_ctxt cdt c)

(* CEValidate *)

let build_default_cevalidate_strict_code code_ctxt =
  let stat_ctxt = static_context_from_code_selection_context code_ctxt in
  let norm_ctxt = norm_context_from_stat_context stat_ctxt in
  let cxschema = cxschema_from_norm_context norm_ctxt in
  (fun alg_ctxt input_stream ->
    (* 1. Input must be a SAX Stream *)
    let erased_input_stream = Streaming_ops.erase_xml_stream input_stream in
    
    (* 2. Perform validation *) 
    Schema_validation.validate cxschema erased_input_stream	
  )

(* CECast 
   NB: Note that the namespace environment is a static parameter of the algebraic operator. *)
let build_default_cecast_code code_ctxt adt nsenv =
  let stat_ctxt = static_context_from_code_selection_context code_ctxt in
  let norm_ctxt = norm_context_from_stat_context stat_ctxt in
  let cxschema = cxschema_from_norm_context norm_ctxt in
  let (base_type,opt_flag) = Schema_norm_util.get_optional_atomic_type_asequencetype cxschema adt in
  let atomic_type = Schema_judge.atomic_type_of_typename cxschema base_type in
  (fun alg_ctxt e ->
    let e = Cursor.cursor_of_list e in
    if (Cursor.cursor_is_empty e && opt_flag)
    then []
    else
      let base_value =
	try
	  Physical_util.get_singleton_atomic e
	with
	| _ ->
	    raise (Query (Expr_Error ("Cast not applied to a single atomic value")))
      in
      let new_base_value = base_value#cast_to nsenv base_type atomic_type in
      [Item_Atomic new_base_value]
  )

(* CECastable:
   NB: Note that the namespace environment is a static parameter of the algebraic operator.  *)
let build_default_cecastable_code code_ctxt adt nsenv =
  let stat_ctxt = static_context_from_code_selection_context code_ctxt in
  let norm_ctxt = norm_context_from_stat_context stat_ctxt in
  let cxschema = cxschema_from_norm_context norm_ctxt in
  let (base_type,opt_flag) = Schema_norm_util.get_optional_atomic_type_asequencetype cxschema adt in
  let atomic_type = Schema_judge.atomic_type_of_typename cxschema base_type in
  (fun alg_ctxt e ->
    let e = Cursor.cursor_of_list e in
    if (Cursor.cursor_is_empty e && opt_flag)
    then []
    else
      let boo =
	try
	  let base_value =
	    try
	      Physical_util.get_singleton_atomic e
	    with
	    | _ ->
		raise (Query (Expr_Error ("Cast not applied to a single atomic value")))
	  in
	  let _ = base_value#cast_to nsenv base_type atomic_type in true
	with
	| _ -> false
      in
      [Physical_item_util._boolean boo])

(* CESome *)
let build_default_cesome_code code_ctxt odt v =
  let stat_ctxt = static_context_from_code_selection_context code_ctxt in
  let add_fn = build_add_var_xml_value_unsafe_allowed code_ctxt v in
  (fun e2 eval alg_ctxt input_cursor ->
    let eval_some x =       
      dynamic_opttype_check_item stat_ctxt odt x;
      let item_sequence = sequence_of_singleton x in
      add_fn item_sequence;
      Physical_util.get_boolean	(item_cursor_of_physical_value (eval alg_ctxt e2))
    in
    let b = Cursor.cursor_exists eval_some input_cursor in
    [Physical_item_util._boolean b])

(* CEEvery *)
let build_default_ceevery_code code_ctxt odt v =
  let stat_ctxt = static_context_from_code_selection_context code_ctxt in
  let add_fn = build_add_var_xml_value_unsafe_allowed code_ctxt v in
  (fun e2 eval alg_ctxt input_cursor ->
    (* Note:
       We may want to allow e1 to return a table, and v to range
       over tuples at future point.
       
       This is materializing a list for now. We may want to use a
       stream instead, but we have to know what to do with type
       assertions first.
       - Jerome
     *)
    let eval_every x =
      dynamic_opttype_check_item stat_ctxt odt x;
      let item_sequence = sequence_of_singleton x in
      add_fn item_sequence;
      Physical_util.get_boolean (item_cursor_of_physical_value (eval alg_ctxt e2))
    in
    let b = Cursor.cursor_for_all eval_every input_cursor in
    [Physical_item_util._boolean b])

(* Tuple construction *)

let build_default_cecreatetuple_code code_ctxt names_and_types =
  let stat_ctxt = static_context_from_code_selection_context code_ctxt in
  let add_type_assertion index f = f

(* JEROME SCREAM: What do we do with type assertions now that tuples
can contain streams ???
    match names_and_types.(index) with
      | _, None -> f
      | _, Some dt ->
	  (fun input_sequence ->
	     let input_cursor = Physical_sequence.cursor_of_sequence input_sequence in
	     let type_checked_cursor = dynamic_type_check stat_ctxt dt input_cursor in
	       f (Physical_sequence.streamed_of_cursor type_checked_cursor)
	  ) *)
  in
  let names              = Array.map (fun (name,dt) -> name) names_and_types in
    (* Here we build the insert code for each name *)  
  let insert_code_array  = Array.map (build_create_tuple_code code_ctxt) names in
  let typed_assert_code_array = Array.mapi add_type_assertion insert_code_array in
  let n_names            = Array.length insert_code_array in
  let returned_value     = empty_tuple in

    (fun alg_ctxt arg_list ->
       if (Array.length arg_list) != n_names then
	 raise (Query (Code_Selection ("Tuple Slot - Tuple Element mismatch")))
       else
	 for i = 0 to n_names - 1 do		
(* JEROME SCREAM: Not doing materialization here anymore !!
	   insert_code_array.(i) (materialize_sequence arg_list.(i))
*)
	   insert_code_array.(i) arg_list.(i)
	 done;
       returned_value
    )

(* Tuple field access *)
let build_default_ceaccesstuple_code code_ctxt rname =
  let retrieve_code = build_retrieve_tuple_code code_ctxt rname in
  (fun (alg_ctxt:algebra_context) () -> retrieve_code ())

(* Tuple concatenation *)
(* Concatentation does not do anything currently (since tuples are compiled) *)
let build_default_tuple_concat_code code_ctxt =
    (fun alg_ctxt tup1 tup2 ->  empty_tuple )
      (*  concat_tuples tup1 tup2) *)


(* Item-to-tuple map *)

let build_default_item_to_tuple_map_code code_ctxt v =
  let add_fn = build_add_var_item_list code_ctxt v  in
  (fun (return_exp:Algebra_type.algop_expr) eval alg_ctxt input_cursor ->     
    let eval_for single_item =
      add_fn [single_item];
      tuple_cursor_of_physical_value (eval alg_ctxt return_exp)
    in
    Cursor.cursor_map_concat eval_for input_cursor)

(* Tuple-to-item map *)

let build_default_tuple_to_item_map_code code_ctxt =
  (fun return_exp eval alg_ctxt input_cursor ->     
    let eval_for tup =
      item_cursor_of_physical_value (eval alg_ctxt return_exp)
    in
    Cursor.cursor_map_concat eval_for input_cursor)


let build_default_tuple_map_code code_ctxt =
  (fun return_exp eval alg_ctxt input_cursor ->
    let eval_for tup =
      tuple_cursor_of_physical_value (eval alg_ctxt return_exp)
    in
    Cursor.cursor_map_concat eval_for input_cursor)

(* Tuple index map *)
let build_default_tuple_map_index_code code_ctxt var =
  let restore_fn = build_create_tuple_code code_ctxt var in
    (fun alg_ctxt input_cursor ->
       (* If input_cursor is empty, we want to make sure that
	  the variable is invalid *)
       if Cursor.cursor_is_empty input_cursor then
	 begin
	   restore_fn empty_sequence;  
	   input_cursor
	 end
       else
	 begin
	   let index = ref 1 in
	   let eval_index tup =
	     let dm_index = Physical_xml_value.xml_value_of_mat_value (materialized_of_list ([Physical_item_util._integer (Decimal._integer_of_int !index)])) in
	       incr index;
	       (* Add in the tuple that gives the index ... *)
	       restore_fn dm_index;	
	       tup
	   in
	     Cursor.cursor_map eval_index input_cursor
	 end
    )




(* Tuple product *)
let build_default_tuple_product_code code_ctxt =
  let annot = retrieve_annotation code_ctxt in (* this needs to be of the dependent code *) 
    (* Need to retrieve_annotation for the left hand side *)
  let materialize_fun, restore_fn, needed_names = materialize_cursor_to_array code_ctxt annot () in
  
  let our_cursor m = 
    let offset = ref 0 in
    let len = Array.length m in    
      (fun () ->
	 let res = 
	   if !offset < len then	   
	     begin
	       restore_fn m.(!offset);
	       empty_tuple_opt
	     end
	   else
	     None
	 in
	   incr offset;
	   res)
  in
  (fun () eval alg_ctxt left_side_cursor right_side_cursor ->
     (* Materialize the right hand side of the product *)
     let materialized_right = materialize_fun eval alg_ctxt right_side_cursor in
       (* Now for each tuple in the left hand side, evaluate the concatenation of the 
	  two together *)
     let eval_concat tup1 =       
	 (* Should return a cursor that "concatenates" each tuple of the left
	    with one on the right *)
	 Cursor.cursor_of_function (our_cursor materialized_right)
     in
     Cursor.cursor_map_concat eval_concat left_side_cursor)


let build_default_nullmap_code code_ctxt tn = 
  let restore_fn     = build_create_tuple_code code_ctxt tn in
  (fun alg_ctxt input_cursor ->
    if Cursor.cursor_is_empty input_cursor then
      restore_fn empty_sequence
    else
      restore_fn non_empty_sequence;
    input_cursor 
  )

(* Tuple select *)
let build_default_tuple_select_code code_ctxt pd =
  (fun pred_exprs eval alg_ctxt input_cursor1 ->          
    let eval_pred tup =
      let eval_predicate pred_expr =
	 (Physical_util.get_boolean)
	    (item_cursor_of_physical_value 
	       (eval alg_ctxt pred_expr))
      in
      let rec eval_predicate_desc desc =
	match desc with
	  | SimpleConjunct (start_index, end_index) ->
	      let rec eval_conj cur_index end_index =
		if cur_index <= end_index then
		  ((eval_predicate pred_exprs.(cur_index))
		    &&  eval_conj (cur_index + 1) end_index)
		else
		  true
	      in
		eval_conj start_index end_index
	  | ComplexConjunct (l,r) ->
	      (eval_predicate_desc l)
		&& (eval_predicate_desc r)
	  | Disjunct (l,r) ->
	      (eval_predicate_desc l) 
	      || (eval_predicate_desc r)
      in
	eval_predicate_desc pd
    in
    Cursor.cursor_filter eval_pred input_cursor1)


(* Order by *)
(* Maybe should make actual tuples array based *)
let build_default_tuple_order_by_code code_ctxt stable sk_esk_list = 
  (* The following function needs to be called to get the value of the
     sorting expression. *)
  let annot = retrieve_annotation code_ctxt in
  let materialize_fun, restore_fn, needed_names = materialize_cursor_to_array code_ctxt annot () in

   (* Note:
       All of the evaluations of e2 must yield values (or nodes) of
       the same type.
     - Jerome *)
  let eval_e2 (eval:eval_fun)
      (alg_ctxt:algebra_context) 
      (dep:algop_expr) =
    (* Evaluates the sortspec expression *)
    (* This would be nice to get rid of... *)
    item_cursor_of_physical_value
      (eval alg_ctxt dep)
  in

  (* Here is the local comparison function *)

  (* Note:
       This applies both evaluation of the sorting expression, and
       actual comparison for each resulting value taking all the
       necessary parameters into account.
     - Jerome *)

    let compare_function eval alg_ctxt deps input_value1 input_value2 =
      let deps = Array.to_list deps in
      let rec compare_fun deps sk_esk_list =
	match (deps,sk_esk_list) with
	| [],[] -> 0
	| (dep::deps',(sk,esk) :: sk_esk_list') ->
	    let eval_e2' = (fun x ->
	      restore_fn x; (* setup evaluation tuple fields *)
	      eval_e2 eval alg_ctxt dep) in
	    let comp1 =	      
	      (forest_compare sk esk 
		 (eval_e2' input_value1)
		 (eval_e2' input_value2))
	    in
	    if comp1 = 0
	    then
	      compare_fun deps' sk_esk_list'
	    else
		comp1
	| _ ->
	    raise (Query (Code_Selection "Incompatible number of sub-exprs and order by modifiers in one order by expression"))
      in
      compare_fun deps sk_esk_list
    in

    (* The sort function *)
    let sort_function =
      match stable with
      | Stable ->
	  Array.stable_sort
      | NonStable ->
	  Array.sort
    in

    (* Should be a common operation.. *)
    let our_cursor m = 
      let offset = ref 0 in
      let len = Array.length m in

      (fun () ->
	 let res = 
	   if !offset < len then	   
	     begin
	       restore_fn m.(!offset);
	       empty_tuple_opt
	     end
	   else
	     None
	 in
	   incr offset;
	   res)
    in
    
    (fun (deps:algop_expr array) 
	 (eval:eval_fun)
	 (alg_ctxt: algebra_context) 
	 input_table ->     
	   let materialized = materialize_fun eval alg_ctxt input_table in
	     (* Now sorts it in place *)
	   sort_function (compare_function eval alg_ctxt deps) materialized;
	      Cursor.cursor_of_function (our_cursor materialized)
    )
				   
(*******************************************)
(* We return the empty typle, because tuples
   have little (nothing now) to do with the
   return, they are in-memory assignments *)
(*******************************************)
let build_default_input_tuple_code code_ctxt =
  (fun alg_ctxt () -> empty_tuple )


(***************)
(* Concat maps *)
(***************)
(* Tuple Concat:
   ConcatMap{ Expr_1 ++ input }( E_2 )   *)
(* tuple concat is really handled by the allocation, 
   not the function itself. The annotations decide what is returned
   so these functions really just iterate through the cursors
   with the correct cardinality.   
*)
let build_default_tuple_map_concat_code code_ctxt = 
    (fun left eval alg_ctxt input_cursor -> 
       let eval_fun tup = 
	 tuple_cursor_of_physical_value (eval alg_ctxt left) 
       in
	 Cursor.cursor_map_concat eval_fun input_cursor
    )

(******************************************************)
(* NOTE: There is a potential problem with this code. *)
(* It does *NOT* Properly destroy the contents of the
   dependent expression tuple. This is because these
   tuples have not been allocated at this point in code
   selection - although they are in scope. This is due
   to tuple scoping being the reverse of variable
   scoping. It should not present a problem as no operator
   accesses them - but, it could be a problem for 
   debugging. I would not suggest relying on this 
   behavior

   This should eventually be fixed                    *)
(******************************************************)
let build_default_outer_tuple_map_concat_code code_ctxt null_name = 
  let restore_fn = build_create_tuple_code code_ctxt null_name in
    (fun dep eval alg_ctxt input_cursor -> 
       let eval_fun tup = 
	 let ret = tuple_cursor_of_physical_value (eval alg_ctxt dep) in
	   if Cursor.cursor_is_empty ret then
	     begin
	       (* set to empty 
		  ((returned_fields dep) - (returned_fields input_cursor))  *)
	       restore_fn empty_sequence;
	       (* Physical_table.table_empty () *)
	       table_of_singleton empty_tuple
	     end
	   else
	     begin
	       restore_fn non_empty_sequence;
	       ret
	     end
       in
	 Cursor.cursor_map_concat eval_fun input_cursor
    )

     
let build_default_item_tuple_map_concat_code code_ctxt vn = 
  let add_fn = build_add_var_item_list code_ctxt vn in    
    (fun left eval alg_ctxt input_cursor -> 
       let eval_fun single_item = 
	 add_fn [single_item];
	 tuple_cursor_of_physical_value 
	   (eval alg_ctxt left)
       in
	 Cursor.cursor_map_concat eval_fun input_cursor
    )

(*****************************)
(* Update-related operations *)
(*****************************)

let build_default_materialize_table_code code_ctxt =
  let annot = retrieve_annotation code_ctxt in
  let materialize_fun, restore_fn, needed_names = materialize_cursor_to_array code_ctxt annot () in
    (* *REALLY* Should be a common operation.. *)
  let our_cursor m = 
    let offset = ref 0 in
    let len = Array.length m in

      (fun () ->
	 let res = 
	   if !offset < len then	   
	     begin
	       restore_fn m.(!offset);
	       empty_tuple_opt
	     end
	   else
	     None
	 in
	   incr offset;
	   res)
  in
  (fun () eval alg_ctxt input_cursor ->
    let materialized = materialize_fun eval alg_ctxt input_cursor in
    Cursor.cursor_of_function (our_cursor materialized)
  )

let build_default_copy_code code_ctxt =
  (fun alg_ctxt ae1 -> ae1)

(***********************)
(* Pruning over tuples *)
(***********************)
 
let prune_descendant input_cursor retrieve_code code_ctxt =
  let buffer = ref None in

  let filter next_tuple =
    let filter_aux b = 
      let next_node  = getNode (Physical_util.get_item (cursor_of_sequence (retrieve_code ()))) in    
      match snd (b#docorder()) with
      | Nodeid.PrePostInt (docid1, pre1, post1) ->
	  begin
	    match snd (next_node#docorder()) with
	    | Nodeid.PrePostInt (docid2, pre2, post2) ->
		if (docid1 = docid2) && (pre1 < pre2) && (post1 > post2) 
		then
		  false
		else
		  let _ = buffer := Some next_node in
		  true
	    | _ -> true
	  end
      | _ -> true
    in
    match !buffer with
    | Some b -> filter_aux b
    | None -> 
        let node = getNode (Physical_util.get_item (cursor_of_sequence (retrieve_code ()))) in
        buffer := Some node;
	filter_aux node

  in
    Cursor.cursor_filter filter input_cursor

(*  Note: semantics of Prune op requires input list to be SBDO, w/o duplicates *)
let build_default_prune_code code_ctxt axis fieldname =
  let retrieve_code = build_retrieve_mat_tuple_code code_ctxt fieldname in
  (fun alg_ctxt (input_cursor : tuple_unit Cursor.cursor) -> 
    match axis with
    | Descendant
    | Descendant_or_self ->
	prune_descendant input_cursor retrieve_code code_ctxt
    | _ ->
	input_cursor)
