(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: cs_code_selection_top.ml,v 1.14 2006/05/08 12:09:38 simeon Exp $ *)

(* Module: Cs_code_selection_top
   Description:
     This module is the entry-point to code-selection.
*)

open Error 

open Xquery_common_ast
open Xquery_algebra_ast
open Xquery_algebra_ast_util


open Algebra_type
open Cs_util_coercion

open Code_group_order
open Code_update
open Cs_code_top
open Cs_code_selection_expr

open Compile_context
open Code_selection_context

open Code_typing_context 
open Cs_code_typing_top

(**************)
(* Statements *)
(**************)

let default_code_selection_statement code_ctxt ae =
  (* Physical typing sub-phase *)
  let ctc'       = Cs_code_typing_top.code_typing_statement code_ctxt ae in
  let code_ctxt = store_global_annotation code_ctxt ae.compile_annotations in
  let code_ctxt = replace_code_type_context_in_code_selection_context ctc' code_ctxt in
  let code_ctxt = enter_statement_context code_ctxt in
  let code_ctxt = default_code_selection code_ctxt ae in
  let _ = exit_statement_context code_ctxt in
  code_ctxt


(***********************)
(* Prolog declarations *)
(***********************)

(* Code Selection for User Defined Functions:
   - Walk the function and allocate fresh variables
   (after allocating some for the params )
   - Allocate a new variable context for the function and associate it *)

let default_code_selection_function code_ctxt fn_decl  =
  let (((name, arity) as fname), _, fbod) = fn_decl.palgop_function_decl_desc in
  let (vname_list,alg_expr_ref,_) = fbod in
  let vc = enter_function_context code_ctxt (name,arity) vname_list in

  let alg_expr = !alg_expr_ref in (* De-ref it *)
    (* Here we should reset the counter for functions *)
  let vc = store_global_annotation vc alg_expr.compile_annotations in
  let vc = default_code_selection vc alg_expr in (* Decorate it*)
    alg_expr_ref := alg_expr;
    let comp_ctxt = annotated_compile_context_from_code_selection_context code_ctxt in 
    if Compile_context.mem_function_from_compile_context comp_ctxt fname 
    then Compile_context.update_function_from_compile_context comp_ctxt fname !alg_expr_ref
    else () (* Should this be an error *);
    begin
      exit_function_context vc (name,arity);
      code_ctxt
    end

let single_op_default_code_selection_decl code_ctxt algop =
  (* The algebraic operator contains thele location that should be
     correlated with errors, so we catch any exceptions here and
     rewrap the exceptions with thele location. *)
  try
  (* Modify the code *)
    let eval_code, csc =
      match algop.alg_decl_name with
      | AOEVarDecl (ocdt,vname) ->
	  let code_ctxt = add_variable_to_current_context code_ctxt vname in
	  let fn = build_default_var_decl_code code_ctxt ocdt vname in
	  let _ = access_nosub algop.alg_decl_dep in 
	  
	  (coerce_nodep_prolog fn coerce_unary_item_cursor_to_algebra_context), code_ctxt
	    
      | AOEVarDeclExternal (ocdt,vname) ->
	  let code_ctxt = declare_external_variable code_ctxt vname in 
	    (*	let code_ctxt =  add_variable_to_current_context code_ctxt vname in *)
	  let fn = build_default_var_decl_external_code code_ctxt ocdt vname in
	  let _ = access_nosub algop.alg_decl_dep in 
	  (coerce_nodep_prolog fn coerce_unit_to_algebra_context), code_ctxt

      | AOEValueIndexDecl str ->
	  let code_ctxt = add_variable_to_current_context code_ctxt Xquery_common_ast.fs_dot in
	  let fn = build_default_key_decl_code code_ctxt str in
	  let dep = access_onesub algop.alg_decl_dep in 
	  (coerce_onedep_prolog fn dep coerce_unary_item_cursor_to_algebra_context), code_ctxt
      | AOENameIndexDecl ename ->
	  let code_ctxt = create_new_name_index code_ctxt ename in
	  let fn = build_default_name_index_decl_code code_ctxt in
	  let _ = access_nosub algop.alg_decl_dep in 
	  (coerce_nodep_prolog fn coerce_unit_to_algebra_context), code_ctxt
    in
    algop.alg_decl_eval := eval_code;
    csc
  with
  | exn -> raise (Error.error_with_file_location algop.alg_decl_loc exn)
    
let default_code_selection_decl code_ctxt algop =
  let sub_exprs = algop.alg_decl_indep in
  let dep_sub_exprs = algop.alg_decl_dep in

  (* Selects the code for the subexpressions *)
  let code_ctxt = sub_expr_default_code_selection code_ctxt sub_exprs in

  let code_ctxt = single_op_default_code_selection_decl code_ctxt algop in    
  (* Selects the code for the dependant subexpressions *)
  let code_ctxt = sub_expr_default_code_selection code_ctxt dep_sub_exprs in
  (* Selects the code for the current expression *)
  code_ctxt


(**********)
(* Prolog *)
(**********)

let default_code_selection_prolog code_ctxt prolog =
  let code_ctxt = enter_prolog_context code_ctxt in
  let comp_ctxt = annotated_compile_context_from_code_selection_context code_ctxt in 
  let prolog',comp_ctxt = Cs_annotate.annotate_prolog_with_bindings comp_ctxt prolog in

  (* I'm not happy having physical typing wedged here, but I think it
     has to follow the "annotation" phase, which takes a logical
     algebraic op and returns a dummy physical operator that is
     elaborated during physical typing and code selection. -Mary *)

  (* Physical typing sub-phase *)
  let ctc' = Cs_code_typing_top.code_typing_prolog code_ctxt prolog' in
  let code_ctxt = replace_code_type_context_in_code_selection_context ctc' code_ctxt in
  (* Functions currently should not modify the code selection context
     -> this is because each of them should be starting an independent
     variable context *)
  let code_ctxt =
    List.fold_left default_code_selection_decl code_ctxt prolog'.palgop_prolog_indices
  in
  let code_ctxt =
    List.fold_left default_code_selection_decl code_ctxt prolog'.palgop_prolog_vars
  in
  (* Functions may refer to variables *)
  let code_ctxt =
    List.fold_left default_code_selection_function code_ctxt prolog'.palgop_prolog_functions
  in
  let _ = exit_prolog_context code_ctxt in
    code_ctxt, prolog'


(***********)
(* Modules *)
(***********)

let default_code_selection_module code_ctxt cm = 
  let code_ctxt,prolog  = default_code_selection_prolog code_ctxt cm.palgop_module_prolog in
  let comp_ctxt  = annotated_compile_context_from_code_selection_context code_ctxt in 
  (* Physical annotation sub-phase *)
  let statements = List.map (Cs_annotate.annotate_statement comp_ctxt) cm.palgop_module_statements in
  let code_ctxt  = List.fold_left default_code_selection_statement code_ctxt statements  in
    code_ctxt, (fmkalgop_xmodule prolog statements)


(*************)
(* Top-level *)
(*************)

let code_selection_statement code_ctxt ae =
  let comp_ctxt  = annotated_compile_context_from_code_selection_context code_ctxt in 
  (* Physical annotation sub-phase *)
  let ae = Cs_annotate.annotate_statement comp_ctxt ae in
  let _  = default_code_selection_statement code_ctxt ae 
  in ae

let code_selection_prolog code_ctxt prolog =
  default_code_selection_prolog code_ctxt prolog

let code_selection_module code_ctxt cm =
  default_code_selection_module code_ctxt cm

