(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: compile_top.ml,v 1.20 2006/01/10 22:13:49 mff Exp $ *)

(* Module: Compile_top
   Description:
     This module compiles an XQuery toplevel statement or module into
     the ALGEBRA.
*)

open Xquery_core_ast
open Xquery_core_ast_util

open Xquery_algebra_ast
open Xquery_algebra_ast_util

open Compile_context

open Compile_util
open Compile_annotate
open Compile_expr

open Error

(*************)
(* Statement *)
(*************)

let compile_cexpr_annotate compile_ctxt cexpr = 
  let aop = compile_cexpr compile_ctxt cexpr in
    annotate_algebraic_expression aop;
    aop

let compile_statement compile_ctxt cstatement =
  compile_cexpr_annotate compile_ctxt cstatement

(***********************)
(* Prolog declarations *)
(***********************)

(* Function definition *)

let compile_cfunction_body compile_ctxt cfunction_body =
  match cfunction_body with
  | CEFunctionBltIn ->
      raise (Query (Compilation "Built-in function declarations should have been filtered before compilation"))
  | CEFunctionUser cexpr1 ->
      let ao1 = compile_cexpr_annotate compile_ctxt cexpr1 in
      (ref ao1)

(* In here we add the function definition to the compilation context*)

let compile_cfunction_signature compile_ctxt (input_types,output_type) =
  List.map (compile_ctype compile_ctxt) input_types,
  compile_ctype compile_ctxt output_type

let compile_cfunction_def_desc compile_ctxt cfunction_def_desc =
  match cfunction_def_desc with
  | ((cfname,arity),vname_list,cfunction_signature,cfunction_body) ->
      let vname_array = Array.of_list vname_list in
      let (input_types,output_type) = compile_cfunction_signature compile_ctxt cfunction_signature in
      let compiled_body = (vname_array,compile_cfunction_body compile_ctxt cfunction_body,Some output_type) in
      add_function_to_compile_context compile_ctxt (cfname,arity) compiled_body;
      ((cfname,arity),(input_types,output_type), compiled_body)

let compile_cfunction_def compile_context cfunction_def =
  let cfunction_def_desc = cfunction_def.pcfunction_def_desc in
  let fi                 = cfunction_def.pcfunction_def_loc in
  let fn_desc            =
    compile_cfunction_def_desc compile_context cfunction_def_desc
  in
  fmkalgop_function_decl fn_desc fi

(* Global variable declaration *)

let compile_cvar_decl compile_ctxt cvar_decl =
  let cvar_decl_desc = cvar_decl.pcvar_decl_desc in
  let fi             = cvar_decl.pcvar_decl_loc in
    match cvar_decl_desc with
      | (vname,ocdt,Some cexpr1) ->
	  let ao1 = compile_cexpr_annotate compile_ctxt cexpr1 in
	  let oadt = compile_opt_ctype compile_ctxt ocdt in
	  let op_name = AOEVarDecl (oadt, vname) in
	  let indep = OneSub ao1 in
	  let dep = NoSub in
	    logical_aalgop_decl_mkop op_name indep dep fi 

      | (vname,ocdt,None) ->
	  let oadt = compile_opt_ctype compile_ctxt ocdt in
	  let op_name = AOEVarDeclExternal (oadt, vname) in
	  let indep = NoSub in
	  let dep = NoSub in
	    logical_aalgop_decl_mkop op_name indep dep fi

(* Key definitions *)

let compile_cindex_def compile_ctxt cindex_def =
  let fi = cindex_def.pcindex_def_loc in
    match cindex_def.pcindex_def_desc with
      | CValueIndex (str,cexpr1,cexpr2) ->
	  let ao1 = compile_cexpr_annotate compile_ctxt cexpr1 in
	  let ao2 = compile_cexpr_annotate compile_ctxt cexpr2 in
	  let op_name = AOEValueIndexDecl str in
	  let indep = OneSub ao1 in
	  let dep = OneSub ao2 in
	    logical_aalgop_decl_mkop op_name indep dep fi

      | CNameIndex cename ->
	  let relem_sym = Namespace_symbols.relem_symbol cename in
	  let op_name = AOENameIndexDecl relem_sym in
	  let indep = NoSub in
	  let dep = NoSub in
	    logical_aalgop_decl_mkop op_name indep dep fi

(**********)
(* Module *)
(**********)

let is_non_built_in_function f =
  match f.pcfunction_def_desc with
  | (_,_,_,CEFunctionUser _) -> true
  | _ -> false

let compile_prolog compile_context cprolog =
  let non_built_in_functions =
    List.filter is_non_built_in_function cprolog.pcprolog_functions
  in
  let prolog =
    { palgop_prolog_functions = List.map (compile_cfunction_def compile_context) non_built_in_functions;
      palgop_prolog_vars      = List.map (compile_cvar_decl compile_context) cprolog.pcprolog_vars;
      palgop_prolog_indices   = List.map (compile_cindex_def compile_context) cprolog.pcprolog_indices }
  in
  (compile_context,prolog)

let compile_xmodule compile_context cxmodule =
  let (compile_context',prolog') = compile_prolog compile_context cxmodule.pcmodule_prolog in
  let xmod =
    { palgop_module_prolog     = prolog';
      palgop_module_statements = List.map (compile_statement compile_context') cxmodule.pcmodule_statements }
  in
  (compile_context',xmod)

