(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies and AT&T Labs.            *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: evaluation_expr.ml,v 1.12 2006/03/23 06:34:26 stark Exp $ *)

open Algebra_type
open Xquery_algebra_ast
open Error

(* Module: Evaluation_expr
   Description:
     Execute the code associated to each algebraic operation
*)


(* NOTE:

   The reason we have this first step that instantiate the dependant
   expressions is because those may have changed in the course of
   algebraic rewriting.

   A better strategy would be to change the evaluation to instantiate
   all of the code in the AST *before* starting actual evaluation.

   - J & C *)


let has_delta_update op = 
  let {has_delta_update=hdu} = op.annotation in
    hdu

let has_two_deltas_sexprs sexpr = 
  match sexpr with 
    | NoSub  
    | OneSub _ -> false
    | TwoSub (op1,op2) ->	
	(has_delta_update op1) &&
	(has_delta_update op2)

    | ManySub ops ->
	let count_fold cur op =
	  if has_delta_update op 
	  then (cur+1)
	  else cur
	in
	let counts = Array.fold_left count_fold 0 ops in
	  counts >= 2 
	    
let update_allocate_holder_if_needed alg_ctxt subexpr = 
  match Execution_context.get_current_snap_semantic alg_ctxt with
    | Xquery_common_ast.Snap_Ordered_Deterministic   ->
	if has_two_deltas_sexprs subexpr
	then Execution_context.allocate_update_holder alg_ctxt
	else alg_ctxt

    | Xquery_common_ast.Snap_Unordered_Deterministic 
    | Xquery_common_ast.Snap_Nondeterministic        -> 
	(* Here we do not need to allocate holders because it is done else where *)
	alg_ctxt


(****************************************)
(* Physical typing debug code - Michael *)
(****************************************)

open Xquery_physical_type_ast
open Print_xquery_physical_type
open Physical_value

let counter_check_physical_type algop pv =
  let string_of_physical_value pv =
    match pv with
      | PXMLValue (SaxValue _) -> "PXMLValue (SaxValue _)"
      | PXMLValue (DomValue (CSeq _)) -> "PXMLValue (DomValue (CSeq _))"
      | PXMLValue (DomValue (LSeq _)) -> "PXMLValue (DomValue (LSeq _))"
      | PTable _ -> "PTable _"
  in
  let raise_type_mismatch pt =
    let physop = Cs_util.get_physical_opname algop in
      raise (Query (Prototype ("Physical type mismatch -  " ^ (string_of_physical_type pt) ^ " vs. " ^ (string_of_physical_value pv)
			       ^ " at " ^ (string_of_physop_expr_name physop) ^ "! This should have been compiled out (evaluation_expr)!")))
  in
  let counter_check_physical_sax_type pt =
    match pv with
      | PXMLValue (SaxValue _) -> ()
      | _ -> raise_type_mismatch pt
  in
  let counter_check_physical_dom_cursor_type pt =
    match pv with
      | PXMLValue (DomValue (CSeq _)) -> ()
      | _ -> raise_type_mismatch pt
  in
  let counter_check_physical_dom_list_type pt =
    match pv with
      | PXMLValue (DomValue (LSeq _)) -> ()
      | _ -> raise_type_mismatch pt
  in
  let counter_check_physical_table_type pt =
    match pv with
      | PTable _ -> ()
      | _ -> raise_type_mismatch pt
  in
    match algop.palgop_expr_eval_sig with
      | Some (_, _, pt) ->
	  begin
	    match pt with
	      | PT_Table _ -> counter_check_physical_table_type pt
	      | PT_XML (PT_Sax PT_Stream)-> counter_check_physical_sax_type pt
	      | PT_XML (PT_Sax PT_Discarded)-> raise_type_mismatch pt
	      | PT_XML (PT_Dom PT_CursorSeq) -> counter_check_physical_dom_cursor_type pt
	      | PT_XML (PT_Dom PT_ListSeq) -> counter_check_physical_dom_list_type pt
	  end
      | None -> ()
	  
(* Should be compiled out! - Michael *)
let bcounter_check_physical_type = false
  
let rec algebra_execute alg_ctxt (alg_term: algop_expr) =

  (* The algebraic operator contains the file location that should be
     correlated with errors, so we catch any dynamic exceptions here
     and rewrap the exceptions with the file location. *)
  let fi = alg_term.palgop_expr_loc in 
    try
      let result =
	(* let algebra_execute_with_input = algebra_execute alg_ctxt in *)
	(* 2. evaluate the non-dependant sub-expressions, then call the code *)
	(* 1. Instiantiate higher-order operations *)
	let actual_code =
	  match !(alg_term.palgop_expr_eval) with      
	    | NoDep f -> f
	    | SomeDep f -> f algebra_execute
	in
	let sub_alg_terms = alg_term.psub_expression in
	  match (sub_alg_terms, actual_code) with
	    | (NoSub,AOECUnit f) ->
		f alg_ctxt ()
	    | (OneSub alg1,AOECUnary f) ->
		let ac1    = update_allocate_holder_if_needed alg_ctxt sub_alg_terms in 
		let input1 = algebra_execute ac1 alg1 in
		  f alg_ctxt input1
	    | (TwoSub (alg1,alg2),AOECBinary f) ->	
		let ac1    = update_allocate_holder_if_needed alg_ctxt sub_alg_terms in 
		let ac2    = update_allocate_holder_if_needed alg_ctxt sub_alg_terms in 
		let input1 = algebra_execute ac1 alg1 in
		let input2 = algebra_execute ac2 alg2 in
		  f alg_ctxt input1 input2
	    | (ManySub algs , AOECMany f) ->
		let acs    = Array.map (fun _ -> update_allocate_holder_if_needed alg_ctxt sub_alg_terms) algs in 
		let inputs = Array.mapi (fun i alg -> algebra_execute acs.(i) alg) algs in
		  f alg_ctxt inputs
	    | _ ->
		raise (Query (Malformed_Algebra_Expr "Number of input sub-algebraic operations does not match code arity"))
      in
	
	(*****************************************)
	(* Physical typing debug code - Michael *)
	(****************************************)

	if (bcounter_check_physical_type)
	then (counter_check_physical_type alg_term result);

	result
    with
      | exn -> raise (error_with_file_location fi exn)
