open Galax
open Galax_io
open Dm
open Dm_functions

(* 
   Example 1 

   Evaluate a query contained in a string.
   Serialize result to standard output

*)
let example1 prepared_prolog =

  try
    (* Evaluate global variables *)
    let pc = procctxt_from_prepared_prolog prepared_prolog in
    let query_input = String_Input "for $s in $report//section[section.title = \"Procedure\"] return ($s//incision)[2]/instrument" in
    let res = eval_statement prepared_prolog query_input in  
    serialize pc (Formatter_Output Format.std_formatter) res
  with
  | exn -> Error.printf_error "Example 1:" exn

(* 
   Example 2

   Import a main module 
   Evaluate the first statement in that module
   Serialize result to standard output.

*)
let example2 compiled_prolog =

  try
    (* Load main module *)
    let (compiled_prolog', stmts) = import_main_module false compiled_prolog (File_Input "../docs/seq.xq") in
    (* Evaluate global variables *)
    let prepared_prolog = eval_prolog compiled_prolog' None in 
    let pc = procctxt_from_prepared_prolog prepared_prolog in
    serialize pc (Formatter_Output Format.std_formatter) (eval_compiled_statement prepared_prolog (List.nth stmts 0))
  with
    (* Any errors raised by Galax are printed to stderr *)
  | exn -> Error.printf_error "Example 2" exn

(* 
   Example 3

   Load a second library module from a file    
   Bind external variables to values
   Prepare/evaluate the library
   Evaluate a query function contained in the library module by applying it to the bound variables
   Serialize result to a string & print. 

*)
let example3 pc compiled_prolog xmldoc =
  try
    (* Load main module *)
(* 
print_string ("before import_library\n");
    let compiled_prolog = import_library_module compiled_prolog (File_Input "../docs/xq-example1.xq") in
print_string ("before import_main\n");
*)
    let (compiled_prolog, stmts) = import_main_module false compiled_prolog (File_Input "../docs/xq-example2.xq") in

    (* Evaluate global variables *)
    let xmlint = to_atomicValue_item(atomicInteger (Big_int.big_int_of_int 12)) in
    let xmlstr = to_atomicValue_item(atomicString "a string") in
    let exc = build_external_context pc None None [("v", [xmlint]); ("x",[xmlstr]); ("y", [xmlint])] in
    let prepared_prolog = eval_prolog compiled_prolog (Some exc) in 
    let pc = procctxt_from_prepared_prolog prepared_prolog in
    let res = eval_statement prepared_prolog (String_Input "xq-example1:test2($y,$x)") in
    (* Serialize the result to a string *)
    print_string(serialize_to_string pc res)
  with
    (* Any errors raised by Galax are printed to stderr *)
  | exn -> Error.printf_error "Example 3" exn

(* 
   Example 4

   Evaluate a query contained in a string.
   Serialize result to standard output

*)
let example4 prepared_prolog =

  try
    (* Evaluate global variables *)
    let pc = procctxt_from_prepared_prolog prepared_prolog in
    let query_input = String_Input "<foo1:a xmlns:foo1='http://www.foo.org/'/>" in
    let res = eval_statement prepared_prolog query_input in  
    serialize pc (Formatter_Output Format.std_formatter) res
  with

    (* Any errors raised by Galax are printed to stderr *)
  | exn -> Error.printf_error "Example 4:" exn


(* 
    Load a document 
    Load a library module from a file    
    Prepare/evaluate the prolog 
    Evaluate each example using the prepared or compiled prolog
*)
let _ =
  try 
    let pc = Processing_context.default_processing_context() in

    let doc_input = File_Input "../docs/report.xml" in
    let xmldoc = load_document pc doc_input in

    let compiled_prolog = load_standard_library pc in
    let library_input = File_Input "../docs/seq_context.xq" in
    let compiled_prolog = import_library_module compiled_prolog library_input in
    let prepared_prolog = eval_prolog compiled_prolog None in 

    example1 prepared_prolog; 
    example2 compiled_prolog;
    example3 pc compiled_prolog xmldoc; 
    example4 prepared_prolog
  with
  | exn -> Error.printf_error "" exn

