(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: factorize_side_effect.ml,v 1.8 2006/05/08 12:09:39 simeon Exp $ *)

(* Module: Factorize_side_effect
   Description:
   This module is a judgement for side-effects.
*)

open Xquery_common_ast
open Xquery_core_ast
open Xquery_core_ast_util


(* We need a set of those functions which are side-effect free *)
module Function_Name = struct
  type t = Xquery_common_ast.cfname
  let compare = compare
end

module Function_Name_Set = Set.Make(Function_Name)

let base_set = ref (Function_Name_Set.empty)

let register_side_effect_free_function cfname = 
  base_set := Function_Name_Set.add cfname !base_set

let global_is_side_effect_free_function cfname = 
  Function_Name_Set.mem cfname !base_set

type env = { side_effect_functions: Function_Name_Set.t} 
let is_side_effect_free_function env cfname =
  Function_Name_Set.mem cfname env.side_effect_functions 

let empty_env () = { side_effect_functions = !base_set } 

let rec has_side_effect (env: env) (ce: acexpr) =
    match ce.pcexpr_desc with
      | CEFLWOR (fl_clauses, where_clause, order_by_clause, return_clause)  ->
	  has_side_effect_flwor env fl_clauses where_clause order_by_clause return_clause

      | CEIf (cexpr1, cexpr2, cexpr3) ->
	  (has_side_effect env cexpr1) ||
	    (has_side_effect env cexpr2) ||
	    (has_side_effect env cexpr3) 
      | CETypeswitch (cexpr, cases) ->
	  let handle_branch b (pat, name, cexpr) =
	    b || (has_side_effect env cexpr)
	  in
	    List.fold_left handle_branch 
	      (has_side_effect env cexpr) cases
	      
      | CESome (odt, vname, cexpr1, cexpr2) ->
	  (has_side_effect env cexpr1) ||
	    (has_side_effect env cexpr2)

      | CEEvery (odt, vname, cexpr1, cexpr2) ->
	  (has_side_effect env cexpr1) ||
	    (has_side_effect env cexpr2) 

      | CEUnordered cexpr ->
	  (has_side_effect env cexpr)
      | CEOrdered cexpr ->
	  (has_side_effect env cexpr)

      | CEVar vname -> 
	  false
      | CEScalar _ -> 
	  false
      | CEProtoValue _ -> 
	  false

      | CEText _ ->
	  false

      | CETextComputed cexpr ->
	  (has_side_effect env cexpr)

      | CEPI _ ->
	  false

      | CEPIComputed (cexpr1,cexpr2) ->
	  (has_side_effect env cexpr1) ||
	    (has_side_effect env cexpr2) 

      | CEComment _ ->
	  false

      | CECommentComputed cexpr1 ->
	  (has_side_effect env cexpr1) 
      | CEDocument cexpr1 ->
	  (has_side_effect env cexpr1) 

      | CECall (fname, arguments, sign) ->
	  (not (is_side_effect_free_function env fname)) ||
	    (has_side_effect_list env arguments)

      | CEOverloadedCall (fname, arguments, sigs) ->
	  (* No builtins have side-effects and only builtins are overloaded *)
	  has_side_effect_list env arguments 

      | CEKeyRef (kname, arg) ->
	  has_side_effect env arg

      | CESeq (cexpr1, cexpr2) ->
	  (has_side_effect env cexpr1) ||
	    (has_side_effect env cexpr2) 

      | CEEmpty -> false

      | CEElem (relem_symbol, nsenv, cexprlist) ->
	  has_side_effect_list env cexprlist

      | CEAnyElem (cexpr1, nsenv, cexpr2) ->  
	  (has_side_effect env cexpr1) ||
	    (has_side_effect env cexpr2) 

      | CEAttr (rattr_symbol, cexprlist)->
	  has_side_effect_list env cexprlist

      | CEAnyAttr (cexpr1, nsenv, cexpr2) ->
	  (has_side_effect env cexpr1) ||
	    (has_side_effect env cexpr2) 

      | CEError cexpr_list ->
	  has_side_effect_list env cexpr_list

      | CETreat (cexpr, model) ->

	  has_side_effect env cexpr
      | CECast (cexpr, nsenv, model) ->
	  has_side_effect env cexpr

      | CECastable (cexpr, nsenv, model) ->
	  has_side_effect env cexpr

      | CEValidate (vmode,cexpr) ->
	  has_side_effect env cexpr

      | CEForwardAxis (v,axis, cnode_test) -> false
      | CEReverseAxis (v,axis, cnode_test) -> false

      | CECopy ce1 ->
	  has_side_effect env ce1

      | CEDelete ce1 -> true	  
      | CEDetach ce1 -> true
      | CEInsert (cexpr, cinsert_location) -> true
      | CERename (ce1, ce2) -> true
      | CEReplace (vof, ce1, ce2) -> true
      | CESnap (sm,ce) -> true
      | CEExecute (host, port, cexpr1) -> true

and has_side_effect_flwor env fls wc ob rc =  
  let cfl_map env b cfl =
    b || 
      (match cfl with
	 | CELET(odt, vn, cexpr) ->
	     has_side_effect env cexpr
	 | CEFOR(odt, vn, vn2, cexpr) ->
	     has_side_effect env cexpr)
  in
  let has_side_effect_where_clause env wc = 
    match wc with
      | None -> false
      | Some cexpr -> has_side_effect env cexpr
  in
  let has_side_effect_orderby_clause env ob = 
    match ob with 
      | None -> false
      | Some (sk, ospecs, osig) ->
	  let ospecs_fold b (cexpr,sk,esk) =
	    b || (has_side_effect env cexpr)
	  in
	    List.fold_left ospecs_fold false ospecs
  in
    
    (List.fold_left (cfl_map env) false fls)  ||
      (has_side_effect_where_clause env wc)   ||
      (has_side_effect_orderby_clause env ob) ||
      (has_side_effect env rc)
      
and has_side_effect_list env cexprs = 
  List.fold_left (fun b v -> b || (has_side_effect env v))
    false cexprs

let has_side_effect_judge ce = 
  has_side_effect (empty_env ()) ce 
