(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: galax.ml,v 1.59 2006/03/24 19:01:47 mff Exp $ *)

(* Module: Galax
   Description:
     This module contains the Galax Caml user API.
*)

open Format

open Error
open Namespace_context

open Xquery_common_ast
open Xquery_ast
open Xquery_core_ast
open Xquery_algebra_ast
open Print_top

open Physical_item
open Physical_sequence
open Physical_value_util

open Processing_context
open Monitoring_context
open Procmod_phases
open Monitor

(******************)
(* Galapi context *)
(******************)

(* Everything you wanted to know about context, but were afraid to ask *)

type external_context = Procmod_compiler.external_context

(********************************)
(* Module context accessors     *)
(********************************)

let procctxt_from_compiled_prolog compiled_prolog =
  Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog

let nsenv_from_compiled_prolog compiled_prolog =
  Procmod_compiler.nsenv_from_compiled_prolog compiled_prolog

let procctxt_from_prepared_prolog prepared_prolog =
  Procmod_compiler.processing_context_from_compiled_prolog prepared_prolog

let nsenv_from_prepared_prolog prepared_prolog =
  Procmod_compiler.nsenv_from_compiled_prolog prepared_prolog

(*************)
(* AST Types *)
(*************)

type compiled_statement = Algebra_type.algop_expr
type compiled_prolog = Procmod_compiler.compiled_prolog
type prepared_prolog = Procmod_compiler.compiled_prolog

(********************************)
(* Document I/O functions       *)
(********************************)

let aux_load_document proc_ctxt gio =
  let aux_load_document' proc_ctxt  = 
    let (dtdopt, xml_stream)= Streaming_parse.open_xml_stream_from_io gio in
    let resolved_xml_stream = Streaming_ops.resolve_xml_stream xml_stream in
    let typed_xml_stream    = Streaming_ops.typed_of_resolved_xml_stream resolved_xml_stream in
    let nodeid_context      = Nodeid_context.default_nodeid_context () in
    let v = Physical_load.load_xml_document_from_typed_stream nodeid_context typed_xml_stream in
    (List.map (fun n -> getNode n) v)
  in
  wrap_monitor proc_ctxt (Document_Toplevel_ParsingLoading_Phase (Parse_io.name_of_input_spec gio)) aux_load_document' proc_ctxt

let poly_serialize proc_ctxt output_spec serialize_fun v =
  let gout = Parse_io.galax_output_from_output_spec output_spec in
  let fmt = Parse_io.formatter_of_galax_output gout in
  if (!Conf.verbose) then
    begin
      Format.fprintf !Conf.xml_formatter "%s" !Conf.xml_header;
      pp_print_flush !Conf.xml_formatter ()
    end;
  (* Serialization call is HERE ! *)
  serialize_fun proc_ctxt fmt v;
  Format.pp_print_newline fmt ();
  if (!Conf.verbose) then
    begin
      Format.fprintf !Conf.xml_formatter "%s" !Conf.xml_footer;
      pp_print_flush !Conf.xml_formatter ()
    end;
  (* Note that a galax_output must be closed, as below *)
  Parse_io.close_galax_output gout

let aux_serialize proc_ctxt output_spec v =
  begin
    (* Serialization should be part of the processing model and all
    this code should be moved into Procmod *)
    let serialize_datamodel v =
      poly_serialize proc_ctxt output_spec Serialization.fserialize_datamodel v
    in 
    wrap_monitor proc_ctxt Document_Serialization_Phase serialize_datamodel v
  end

let aux_serialize_typed_xml_stream proc_ctxt output_spec v =
  begin
    (* Serialization should be part of the processing model and all
    this code should be moved into Procmod *)
    let serialize_datamodel v =
      poly_serialize proc_ctxt output_spec Serialization.fserialize_typed_xml_stream v
    in 
    wrap_monitor proc_ctxt Document_Serialization_Phase serialize_datamodel v
  end

let aux_serialize_to_string proc_ctxt v =
  let buff = Buffer.create 64 in
  begin
    aux_serialize proc_ctxt (Galax_io.Buffer_Output buff) v;
    wrap_monitor proc_ctxt Document_Serialization_Phase Buffer.contents buff
  end

let default_external_context () = Procmod_compiler.default_external_context ()

let build_external_context proc_ctxt opt_ctxt_item opt_timezone var_val_list =
  Procmod_compiler.create_external_context
    proc_ctxt
    opt_ctxt_item
    opt_timezone
    var_val_list

(*********************************)
(* Monitored Galax API Functions *)
(*********************************)

(* 
   All monitored calls must be wrapped so that if the API function
   raises an error, it is caught, the monitoring is terminated and
   end_monitor_call is called.

*)
let load_document proc_ctxt gio =
  start_monitor_call proc_ctxt Prolog "load_document" ;
  try
    let doc = aux_load_document proc_ctxt gio in
    end_monitor_call proc_ctxt; 
    doc
  with
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let serialize proc_ctxt output_spec v =
  start_monitor_call proc_ctxt Serialization "serialize" ; 
  try
    aux_serialize proc_ctxt output_spec (Cursor.cursor_of_list v);
    end_monitor_call proc_ctxt;
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let serialize_to_string proc_ctxt v =
  start_monitor_call proc_ctxt Serialization "serialize_to_string" ; 
  try
    let s = aux_serialize_to_string proc_ctxt (Cursor.cursor_of_list v) in
    end_monitor_call proc_ctxt; 
    s
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let load_standard_library proc_ctxt =
  start_monitor_call proc_ctxt Prolog "load_standard_library" ; 
  try
    proc_ctxt.system <- true;
    let m = Procmod_compiler.compile_standard_library_module proc_ctxt in
    end_monitor_call proc_ctxt;
    proc_ctxt.system <- false;
    m
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let load_standard_library_no_monitor proc_ctxt =
    Procmod_compiler.compile_standard_library_module proc_ctxt

let import_prolog_only compiled_prolog ext_ctxt_item gio =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Prolog "import_prolog_only" ; 
  try
    let m = Procmod_compiler.compile_prolog ext_ctxt_item compiled_prolog gio in
    end_monitor_call proc_ctxt;
    m
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let import_library_module compiled_prolog gio =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Prolog "import_library_module" ; 
  try
    let m = Procmod_compiler.compile_library_module compiled_prolog gio in
    end_monitor_call proc_ctxt;
    m
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let import_main_module ext_ctxt_item compiled_prolog gio =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Prolog "import_main_module" ; 
  try
    let m = Procmod_compiler.compile_main_module ext_ctxt_item compiled_prolog gio in 
    end_monitor_call proc_ctxt;
    m
  with
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let eval_prolog compiled_prolog opt_ext_ctxt =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Prolog "eval_prolog" ; 
  try
    let p = Procmod_compiler.execute_compiled_prolog opt_ext_ctxt compiled_prolog in
    end_monitor_call proc_ctxt;
    p
  with
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let eval_statement compiled_prolog gio =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Statement "eval_statement" ; 
  try
    let compiled_statement = Procmod_compiler.compile_statement compiled_prolog gio in
    let v = Procmod_compiler.execute_compiled_statement compiled_prolog compiled_statement in
    (**************************************************************************)
    (* NB!! For queries with element construction, evaluation does not        *)
    (* actually occur until the sequence is materialized in list_of_sequence! *)
    (* So list_of_sequence must be included in evaluation phase.              *)
    (**************************************************************************)
    let v' =  wrap_monitor proc_ctxt (PQuery Evaluation_Phase) (fun v -> list_of_sequence (dom_value_of_physical_value v)) v in
    end_monitor_call proc_ctxt ; 
    v'
    with 
    | ex -> (end_monitor_call proc_ctxt; raise ex)

let eval_compiled_statement compiled_prolog cs =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Statement "eval_compiled_statement" ; 
  try
    let v = Procmod_compiler.execute_compiled_statement compiled_prolog cs in 
    (**************************************************************************)
    (* NB!! For queries with element construction, evaluation does not        *)
    (* actually occur until the sequence is materialized in list_of_sequence! *)
    (* So list_of_sequence must be included in evaluation phase.              *)
    (**************************************************************************)
    let v' =  wrap_monitor proc_ctxt (PQuery Evaluation_Phase) (fun v -> list_of_sequence (dom_value_of_physical_value v)) v in
    end_monitor_call proc_ctxt ; 
     v'
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)
	
let compile_serialized_optimized_main_module mod_ctxt input =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog mod_ctxt in
  start_monitor_call proc_ctxt Statement "compile_serialized_optimized_main_module" ; 
  try
    let comp_ctxt = Procmod_compiler.compile_context_from_compiled_prolog mod_ctxt in
    let parsed_plan, comp_ctxt= 
      Planio_parse_plan_xml.parse_serialized_module_plan proc_ctxt comp_ctxt input 
    in
    let m = Procmod_compiler.compile_main_module_from_optimized_logical_plan
      mod_ctxt (comp_ctxt, parsed_plan) in
    end_monitor_call proc_ctxt ; 
    m
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let compile_serialized_logical_main_module mod_ctxt input =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog mod_ctxt in
  start_monitor_call proc_ctxt Statement "compile_serialized_logical_main_module" ; 
  try
    let comp_ctxt = Procmod_compiler.compile_context_from_compiled_prolog mod_ctxt in
    let parsed_plan, comp_ctxt= 
      Planio_parse_plan_xml.parse_serialized_module_plan proc_ctxt comp_ctxt input 
    in
    let m = Procmod_compiler.compile_main_module_from_logical_plan
      mod_ctxt (comp_ctxt, parsed_plan) in
    end_monitor_call proc_ctxt ; 
    m
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)


let compile_serialized_logical_statement mod_ctxt input =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog mod_ctxt in
  start_monitor_call proc_ctxt Statement "compile_serialized_statement" ; 
  try
    let comp_ctxt = Procmod_compiler.compile_context_from_compiled_prolog mod_ctxt in
    let parsed_expr = 
      Planio_parse_plan_xml.parse_serialized_expression_plan proc_ctxt comp_ctxt input 
    in
    let s = Procmod_compiler.compile_statement_from_logical_plan
      mod_ctxt (comp_ctxt, parsed_expr) in
    end_monitor_call proc_ctxt ; 
    s
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)


(************************)
(* Streaming evaluation *)
(************************)

(* Note:
     the following are prototype operations that may return the result
     as a cursor or XML stream.
*)


let eval_statement_as_item_cursor compiled_prolog gio =
let _ = Printf.printf "************ eval statement\n" in
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Statement "eval_statement_as_item_cursor" ; 
  try
    let compiled_statement = Procmod_compiler.compile_statement compiled_prolog gio in
    let v = Procmod_compiler.execute_compiled_statement compiled_prolog compiled_statement in
    (**************************************************************************)
    (* NB!! For queries with element construction, evaluation does not        *)
    (* actually occur until the sequence is materialized in list_of_sequence! *)
    (* So list_of_sequence must be included in evaluation phase.              *)
    (**************************************************************************)
    let v' =  wrap_monitor proc_ctxt (PQuery Evaluation_Phase) (fun x -> x) (item_cursor_of_physical_value v) in
    end_monitor_call proc_ctxt ; 
    v'
    with 
    | ex -> (end_monitor_call proc_ctxt; raise ex)

let eval_compiled_statement_as_item_cursor compiled_prolog cs =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Statement "eval_compiled_statement_as_item_cursor" ; 
  try
    let v = Procmod_compiler.execute_compiled_statement compiled_prolog cs in 
    (**************************************************************************)
    (* NB!! For queries with element construction, evaluation does not        *)
    (* actually occur until the sequence is materialized in list_of_sequence! *)
    (* So list_of_sequence must be included in evaluation phase.              *)
    (**************************************************************************)
    let v' =  wrap_monitor proc_ctxt (PQuery Evaluation_Phase) (fun x -> x) (item_cursor_of_physical_value v) in
    end_monitor_call proc_ctxt ; 
     v'
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let eval_statement_as_sax compiled_prolog gio =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Statement "eval_statement_as_sax" ; 
  try
    let compiled_statement = Procmod_compiler.compile_statement compiled_prolog gio in
    let v = Procmod_compiler.execute_compiled_statement compiled_prolog compiled_statement in
    (**************************************************************************)
    (* NB!! For queries with element construction, evaluation does not        *)
    (* actually occur until the sequence is materialized in list_of_sequence! *)
    (* So list_of_sequence must be included in evaluation phase.              *)
    (**************************************************************************)
    let v' =  wrap_monitor proc_ctxt (PQuery Evaluation_Phase) (fun x -> x) (sax_value_of_physical_value v) in
    end_monitor_call proc_ctxt ; 
    v'
    with 
    | ex -> (end_monitor_call proc_ctxt; raise ex)

let eval_compiled_statement_as_sax compiled_prolog cs =
  let proc_ctxt = Procmod_compiler.processing_context_from_compiled_prolog compiled_prolog in
  start_monitor_call proc_ctxt Statement "eval_compiled_statement_as_sax" ; 
  try
    let v = Procmod_compiler.execute_compiled_statement compiled_prolog cs in 
    (**************************************************************************)
    (* NB!! For queries with element construction, evaluation does not        *)
    (* actually occur until the sequence is materialized in list_of_sequence! *)
    (* So list_of_sequence must be included in evaluation phase.              *)
    (**************************************************************************)
    let v' =  wrap_monitor proc_ctxt (PQuery Evaluation_Phase) (fun x -> x) (sax_value_of_physical_value v) in
    end_monitor_call proc_ctxt ; 
     v'
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)


let serialize_as_item_cursor proc_ctxt output_spec v =
  start_monitor_call proc_ctxt Serialization "serialize_as_item_cursor" ; 
  try
    aux_serialize proc_ctxt output_spec v;
    end_monitor_call proc_ctxt;
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

let serialize_as_sax proc_ctxt output_spec v =
  start_monitor_call proc_ctxt Serialization "serialize_as_sax" ; 
  try
    let typed_stream_value =
      v
    in
    aux_serialize_typed_xml_stream proc_ctxt output_spec typed_stream_value;
    end_monitor_call proc_ctxt;
  with 
  | ex -> (end_monitor_call proc_ctxt; raise ex)

