(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: galax.mli,v 1.27 2006/03/24 19:01:47 mff Exp $ *)

(* Module: Galax
   Description:
     This module contains the Galax Caml user API.
*)

open Dm_atomic
open Dm

open Physical_value


(********************)
(* External context *)
(********************)

type external_context

(***********)
(* Queries *)
(***********)

(*   A compiled prolog contains a processing context, extended
     with global/external variables and their values.

     A prepared prolog has all global variables (external and
     internal) bound to values.
*)

type compiled_prolog 
type compiled_statement
type prepared_prolog 

(********************************)
(* Module context accessors     *)
(********************************)

val nsenv_from_compiled_prolog              : compiled_prolog -> Namespace_context.nsenv
val procctxt_from_compiled_prolog           : compiled_prolog -> Processing_context.processing_context
val nsenv_from_prepared_prolog              : prepared_prolog -> Namespace_context.nsenv
val procctxt_from_prepared_prolog           : prepared_prolog -> Processing_context.processing_context

(********************************)
(* Document I/O functions       *)
(********************************)

val load_document       : Processing_context.processing_context -> Galax_io.input_spec -> node list
    (* [load_document io] load the XML document from the
       file/string/http uri [io] *)

val serialize           : Processing_context.processing_context -> Galax_io.output_spec -> item list -> unit
    (* [serialize gout x] serializes an XML value using to the given
        galax output *)

val serialize_to_string : Processing_context.processing_context -> item list -> string
    (* [serialize_to_string x] serializes an XML value to a string *)

(*
   External context:
   Context item (optional item)
   Timezone (optional dayTimeDuration) 
   External variables and their values
   External functions 

   Redefining any symbol raises an error
*)

val default_external_context : unit -> external_context
val build_external_context : 
    Processing_context.processing_context -> 
      (item option) -> 
	(atomicDayTimeDuration option) ->
	  (string * item list) list -> external_context

(* 
   Compile a library module or a main module : includes normalization,
   typing, annotation, and logical optimization.

   When importing a main module, must specify whether a context-item
   value will be passed as an external value -- the actual value is passed 
   to eval_prolog. 
*)

val load_standard_library : Processing_context.processing_context -> compiled_prolog 
val import_prolog_only    : 
    compiled_prolog -> 
      bool (* does prolog require external context item? *) -> 
	Galax_io.input_spec ->
	  compiled_prolog
val import_library_module : compiled_prolog -> Galax_io.input_spec -> compiled_prolog
val import_main_module    :
    bool (* does module require external context item? *) -> 
      compiled_prolog -> 
	Galax_io.input_spec -> 
	  compiled_prolog * (compiled_statement list)

(* Evaluate a compiled prolog, so that it can be used during evaluation of a statement: *)
val eval_prolog           : compiled_prolog -> external_context option -> prepared_prolog

(* Evaluation functions require a prepared prolog: *)
val eval_statement           : prepared_prolog -> Galax_io.input_spec -> item list
val eval_compiled_statement  : prepared_prolog -> compiled_statement -> item list

(* Compilation interfaces for external plans *)
val compile_serialized_logical_statement     : compiled_prolog -> Galax_io.input_spec -> compiled_statement
val compile_serialized_logical_main_module   : compiled_prolog -> Galax_io.input_spec -> compiled_prolog * (compiled_statement list)
val compile_serialized_optimized_main_module : compiled_prolog -> Galax_io.input_spec -> compiled_prolog * (compiled_statement list)


(************************)
(* Streaming evaluation *)
(************************)

(* Note:
     the following are prototype operations that may return the result
     as a cursor or XML stream.
*)


val eval_statement_as_item_cursor :
    prepared_prolog -> Galax_io.input_spec -> item Cursor.cursor
val eval_compiled_statement_as_item_cursor  :
    prepared_prolog -> compiled_statement -> item Cursor.cursor

val eval_statement_as_sax :
    prepared_prolog -> Galax_io.input_spec -> Streaming_types.typed_xml_stream
val eval_compiled_statement_as_sax  :
    prepared_prolog -> compiled_statement -> Streaming_types.typed_xml_stream

val serialize_as_item_cursor :
    Processing_context.processing_context -> Galax_io.output_spec -> item Cursor.cursor -> unit
val serialize_as_sax :
    Processing_context.processing_context -> Galax_io.output_spec -> Streaming_types.typed_xml_stream -> unit

