(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: default_lexer.mll,v 1.30 2006/01/23 19:22:45 simeon Exp $ *)

(* Module: Default_lexer
   Description:
     This module contains the part of the XQuery lexer corresponding
     to the DEFAULT lexical state.
*)

{

open Lexing

open Finfo
open Error

open Parse_xquery   (* tokens definition *)
open Lexing_util     (* lexical state handling *)

}

(*****************************)
(* Lexer regular expressions *)
(*****************************)

(* Whitespace *)

let space = ['\032' '\009' '\013' '\010']
let whitespace = space*
let whitespaceplus = space+

(* XML character sets *)

(* Note:
     Some of these Char codes have been taken from the PXP ISO
     parser. I am not sure that they are doing the right thing and
     should be double checked for conformance.
   - Jerome
 *)

let base_char    = ['A'-'Z' 'a'-'z']
let ideographics = ['\192'-'\214' '\216'-'\246' '\248'-'\255']
let extender     = '\183'
let digit        = ['0'-'9']
let ascii_digit = ['0'-'9']
let ascii_hexdigit = ['0'-'9' 'A'-'F' 'a'-'f']

let char_ref_number = ascii_digit+
let char_ref_hex_number = ascii_hexdigit+

(* XML names *)

let letter   = base_char (* | ideographics *)
let nmstart  = letter | '_'
let nmchar   = letter | extender | digit | '.' | '-' | '_'
 
let ncname   = nmstart nmchar*

let qname    = ncname (':' ncname)?

(* Numerics *)

let sign            = '-' | '+'
let exponent        = 'e' | 'E'

let integer_numeric = digit+
let decimal_numeric = (('.' digit+) | (digit+ '.' (digit+)?))
let double_numeric  = (('.' digit+) | (digit+ ('.' (digit+)?)?)) exponent sign? digit+

(****************)
(* Lexing rules *)
(****************)

rule token lh = parse

(********************)
(* Numeric literals *)
(********************)

(* Double *)
  | double_numeric
      { let v = float_of_string (lexeme lexbuf) in
        pop_state lh; push_operator lh; (DOUBLE v) }

(* Decimals *)
  | decimal_numeric
      { let v = Decimal._decimal_of_string (lexeme lexbuf) in
        pop_state lh; push_operator lh; (DECIMAL v) }

(* Integers *)
  | integer_numeric
      { let v = Decimal._integer_of_string (lexeme lexbuf) in
        pop_state lh; push_operator lh; (INT v) }

(* Arithmetic operators *)
  | "-"
      { MINUS }
  | "+"
      { PLUS }

(* Whitespace *)
  | space                  (* for all others, skip blanks *)
      { token lh lexbuf }

(* Comments *)
  | ":)"
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Unmatched end of comment"))) }
  | "(:"
      { set_depth lh 1; xquery_comment lh lexbuf; token lh lexbuf }
  | "{--" 
      { raise (Query (Lexing(lexing_locinfo lexbuf, "ATTENTION: The syntax of comments in XQuery has now changed from {-- --} to (: :)"))) }

  | "*"
      { pop_state lh; push_operator lh; STAR }
  | "*" whitespace ":" whitespace (ncname as n)
      { pop_state lh; push_operator lh; STARNCNAME n }
  | (ncname as n) whitespace ":" "*"
      { pop_state lh; push_operator lh; NCNAMESTAR n }

(* Axis names *)
  | (ncname as n) whitespace "::"
      { AXIS (get_axis n) }

(* Declaration keywords *)

  | "declare" whitespaceplus "default" whitespace "collation"
      { pop_state lh; push_namespacedecl lh; DECLAREDEFAULTCOLLATION }

  | "declare" whitespaceplus "namespace"
      { pop_state lh; push_namespacedecl lh; DECLARENAMESPACE }

  | "declare" whitespaceplus "base-uri"
      { pop_state lh; push_namespacedecl lh; DECLAREBASEURI }

  | "module" whitespaceplus "namespace"
      { pop_state lh; push_namespacedecl lh; MODULENAMESPACE }

  | "declare" whitespaceplus "base-uri"
      { pop_state lh; push_namespacedecl lh; DECLAREDEFAULTCOLLATION }

  | "declare" whitespace "default" whitespaceplus "element"
      { pop_state lh; push_namespacekeyword lh; DECLAREDEFAULTELEMENT }

  | "declare" whitespace "default" whitespaceplus "function"
      { pop_state lh; push_namespacekeyword lh; DECLAREDEFAULTFUNCTION }

  | "import" whitespaceplus "schema"
      { pop_state lh; push_namespacekeyword lh; IMPORTSCHEMA }

  | "import" whitespaceplus "service"
      { pop_state lh; push_namespacekeyword lh; IMPORTSERVICE }

  | "import" whitespaceplus "module"
      { pop_state lh; push_namespacekeyword lh; IMPORTMODULE }

  | "xquery" whitespaceplus "version" whitespace '"'
      { reset_string lh; version_string lh lexbuf }

  | "xquery" whitespaceplus "version" whitespace '''
      { reset_string lh; version_string2 lh lexbuf }

  | "declare" whitespaceplus "value" whitespace "index"
      { DECLAREVALUEINDEX }

  | "declare" whitespaceplus "name" whitespace "index"
      { DECLARENAMEINDEX }

  | "declare" whitespaceplus "boundary-space"
      { pop_state lh; push_operator lh; DECLAREBOUNDARYSPACE }

  | "declare" whitespaceplus "construction"
      { pop_state lh; push_operator lh; DECLARECONSTRUCTION }

  | "declare" whitespaceplus "ordering"
      { pop_state lh; push_operator lh; DECLAREORDERING }

  | "declare" whitespaceplus "default" whitespaceplus "order"
      { pop_state lh; push_operator lh; DECLAREDEFAULTORDER }

  | "declare" whitespace "copy-namespaces"
      { pop_state lh; push_copynamespaces lh; DECLARECOPYNAMESPACES }

  | "declare" whitespaceplus "schema" whitespace "{"
      { push_schema_declaration lh; DECLARESCHEMALCURLY }

  | "declare" whitespaceplus "attribute" whitespaceplus (qname as q)
      { let sn = Namespace_names.uqname_element_of_string q in
        push_type_declaration lh;
        DECLAREATTRIBUTE (sn) }

  | "declare" whitespaceplus "element" whitespaceplus (qname as q)
      { let sn = Namespace_names.uqname_element_of_string q in
        push_type_declaration lh;
        DECLAREELEMENT (sn) }

  | "declare" whitespaceplus "simple" whitespaceplus "type" whitespaceplus (qname as q)
      { let sn = Namespace_names.uqname_element_of_string q in
        push_type_declaration lh;
        DECLARESIMPLETYPE (sn) }

  | "declare" whitespaceplus "complex" whitespaceplus "type" whitespaceplus (qname as q)
      { let sn = Namespace_names.uqname_element_of_string q in
        push_type_declaration lh;
        DECLARECOMPLEXTYPE (sn) }

  | "declare" whitespaceplus "group" whitespaceplus (qname as q)
      { let sn = Namespace_names.uqname_element_of_string q in
        push_type_declaration lh;
        DECLAREGROUP (sn) }

  | "declare" whitespaceplus "attrGroup" whitespaceplus (qname as q)
      { let sn = Namespace_names.uqname_element_of_string q in
        push_type_declaration lh;
        DECLAREATTRGROUP (sn) }

  | "declare" whitespaceplus "function"
      { DECLAREFUNCTION }

  | "external"
      { EXTERNAL }

  | "declare" whitespaceplus "variable" whitespaceplus "$"
      { pop_state lh; push_varname lh; DECLAREVARIABLE }

  | "as"
      { pop_state lh; push_itemtype lh; AS }

(* Keywords followed by variables *)

  | "some" whitespaceplus "$"
      { pop_state lh; push_varname lh; SOMEDOLLAR }

  | "every" whitespaceplus "$"
      { pop_state lh; push_varname lh; EVERYDOLLAR }

  | "for" whitespaceplus "$"
      { pop_state lh; push_varname lh; FORDOLLAR }

  | "let" whitespaceplus "$"
      { pop_state lh; push_varname lh; LETDOLLAR }

  | "validate" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh ; VALIDATELCURLY }

  | "validate" whitespaceplus "lax" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; VALIDATELAXLCURLY }

  | "validate" whitespaceplus "strict" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; VALIDATESTRICTLCURLY }

(* Update tokens *)
  | "delete" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; DELETELCURLY }
  | "detach" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; DETACHLCURLY }
  | "insert" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; INSERTLCURLY }
  | "rename" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; RENAMELCURLY }
  | "replace" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; REPLACELCURLY }
  | "replace" whitespaceplus "value" whitespaceplus "of" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; REPLACEVALUEOFLCURLY }
  
  | "snap" whitespace "delete" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPDELETELCURLY }
  | "snap" whitespace "detach" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPDETACHLCURLY }
  | "snap" whitespace "insert" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPINSERTLCURLY }
  | "snap" whitespace "rename" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPRENAMELCURLY }
  | "snap" whitespace "replace" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPREPLACELCURLY }
  | "snap" whitespace "replace" whitespaceplus "value" whitespaceplus "of" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPREPLACEVALUEOFLCURLY }

  | "copy" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; COPYLCURLY }
  | "snap" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPLCURLY }
  | "snap" whitespace "nondeterministic" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPLCURLYNONDETERMINISTIC } 
  | "snapo" whitespaceplus "{"
      { pop_state lh; push_operator lh; push_default lh ; SNAPLCURLYORDERED } 
(* Slash and SlashSlash *)
  | "/"
      { SLASH }
  | "//"
      { SLASHSLASH }

(* XML tags *)
  | "<?" (ncname as n) whitespace (* Processing instruction *)
      { pop_state lh; push_operator lh; push_processing_instruction lh; OPENINGPI n }
  | "<!--"                  (* Opening comment *)
      { pop_state lh; push_operator lh; push_comment lh; LOPENINGCOMMENT }

  | "<"                     (* Opening XML tag *)
      { pop_state lh; push_operator lh; push_opening_tag lh; LOPENINGTAG }

(* Choice *)
  | "|"
      { BAR }

(* Parenthesis *)
  | "("
      { LPAR }
  | ")" 
      { pop_state lh; push_operator lh; RPAR }
  | "[" 
      { LBRACK }
  | "]" 
      { pop_state lh; push_operator lh; RBRACK }
  | "{"
      { push_default lh;
	LCURLY }
  | "}"
      { try
	  pop_state lh; RCURLY
        with
          Stack.Empty ->
	    raise (Query (Lexing(lexing_locinfo lexbuf, "Unmatched end of enclosed expression"))) }

(* Misc symbols *)
  | "@"
      { ATSIGN }
  | ","
      { COMMA }
  | ";"
      { SEMICOLON }
  | "."
      { pop_state lh; push_operator lh; DOT }
  | ".."
      { pop_state lh; push_operator lh; DOTDOT }

(* Variables *)
  | "$"
      { pop_state lh; push_varname lh; DOLLAR }

(* Function calls *)
  | (qname as q) whitespace "(:"
      { begin
	  set_depth lh 1; xquery_comment lh lexbuf;
          pop_state lh; push_operator lh; 
          match process_qname_string q with
	  | NCNAME_KIND s ->
	      NCNAME s
	  | QNAME_KIND qname ->
	      QNAME qname
        end }

(* From the XQuery Spec - Jerome

A.3 Reserved Function Names

The following names are not recognized as function names in an
unprefixed form because expression syntax takes precedence.

    * attribute
    * comment
    * document-node
    * element
    * item
    * if
    * item
    * node
    * processing-instruction
    * schema-attribute
    * schema-element
    * text
    * typeswitch
    * empty-sequence

*)

  | (qname as q) whitespace "("
      { let actual_qname = Namespace_names.uqname_function_of_string q in
        match actual_qname with
	| (Namespace_names.NSDefaultFunctionPrefix, "attribute") -> ATTRIBUTELPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "comment") -> COMMENTLPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "document-node") -> DOCUMENTNODELPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "element") -> ELEMENTLPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "item") -> ITEMLPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "if") -> IFLPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "node") -> NODELPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "processing-instruction") -> PROCESSINGINSTRUCTIONLPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "schema-attribute") -> SCHEMAATTRIBUTELPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "schema-element") -> SCHEMAELEMENTLPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "text") -> TEXTLPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "typeswitch") -> TYPESWITCHLPAR
	| (Namespace_names.NSDefaultFunctionPrefix, "empty-sequence") -> EMPTYSEQUENCELPAR
	| _ -> FUNCTIONNAMELPAR actual_qname }

(* Constructors *)

  | "attribute" whitespaceplus (qname as q) whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; ATTRIBUTEQNAMECURLY (Namespace_names.uqname_element_of_string q) }

  | "namespace" whitespaceplus (ncname as n) whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; NAMESPACENCNAMECURLY n }

  | "element" whitespaceplus (qname as q) whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; ELEMENTQNAMECURLY (Namespace_names.uqname_element_of_string q) }

  | "document" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; DOCUMENTCURLY }

  | "text" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; TEXTCURLY }

  | "processing-instruction" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; PICURLY }

  | "processing-instruction" whitespaceplus (ncname as n) whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; PINCNAMECURLY n }

  | "comment" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; COMMENTCURLY }

  | "element" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; ELEMENTCURLY }

  | "attribute" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh; ATTRIBUTECURLY }

  | "ordered" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh ; ORDEREDCURLY }

  | "unordered" whitespace "{"
      { pop_state lh; push_operator lh; push_default lh ; UNORDEREDCURLY }

  | "(#" whitespace (qname as q)
      { reset_string lh; pop_state lh; push_operator lh; push_pragma lh; OPENINGPRAGMA (Namespace_names.uqname_element_of_string q) }

(* XML Names *)
  | qname
      { let var_s = (lexeme lexbuf) in
        pop_state lh; push_operator lh; 
        match process_qname_string var_s with
	| NCNAME_KIND s ->
	    NCNAME s
	| QNAME_KIND qname ->
	    QNAME qname }

(*******************)
(* String literals *)
(*******************)

  | "\""
      { reset_string lh; string lh lexbuf }
  | "'"
      { reset_string lh; string_single lh lexbuf }

  | "at" whitespace "\""
      { pop_state lh; push_default lh; reset_string lh; at_string lh lexbuf }
  | "at" whitespace "'"
      { pop_state lh; push_default lh; reset_string lh; at_string_single lh lexbuf }

(***************)
(* End of file *)
(***************)

  | eof
      { EOF }

(********************)
(* Unknown caracter *)
(********************)

(* Note:
     If the lexer arrives here, this is an error.
   - Jerome
*)

  | _
      { raise (Query (Lexing(lexing_locinfo lexbuf, ("Illegal character '" ^ (lexeme lexbuf) ^ "' in expression")))) }

and xquery_comment lh = parse
  | "(:"
      { set_depth lh (succ (get_depth lh)); xquery_comment lh lexbuf }
  | ":)"
      { set_depth lh (pred (get_depth lh)); if get_depth lh > 0 then xquery_comment lh lexbuf }
  | eof
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Comment not terminated"))) }
  | [^ '\n']
      { xquery_comment lh lexbuf }
  | "\n"
      { xquery_comment lh lexbuf }

(****************)
(* String lexer *)
(****************)

and version_string lh = parse
  | "\"\"" { add_char_to_string lh '"'; version_string lh lexbuf }                       (* Escaped quote *)
  | '"'    { let s = get_string lh in pop_state lh; push_operator lh; XQUERYVERSION s }  (* End of string *)
  | eof    { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _      { add_char_to_string lh (lexeme_char lexbuf 0); version_string lh lexbuf }

and version_string2 lh = parse
  | "''" { add_char_to_string lh '\''; version_string2 lh lexbuf }                       (* Escaped quote *)
  | '''    { let s = get_string lh in pop_state lh; push_operator lh; XQUERYVERSION s }  (* End of string *)
  | eof    { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _      { add_char_to_string lh (lexeme_char lexbuf 0); version_string2 lh lexbuf }

and string lh = parse
  | "\"\"" { add_char_to_string lh '"'; string lh lexbuf }                         (* Escaped quote *)
  | "&#" (char_ref_number as charref) ";"
      { let charref_int = int_of_string charref in
        let s = Encoding.character (Encoding.get_internal_encoding ()) charref_int in
        add_string_to_string lh s; string lh lexbuf }
  | "&#x" (char_ref_hex_number as hexcharref) ";"
      { let hexcharref_int = int_of_string ("0x" ^ hexcharref) in
	let s = Encoding.character (Encoding.get_internal_encoding ()) hexcharref_int in
        add_string_to_string lh s; string lh lexbuf }
  | "&lt;"
      { add_char_to_string lh '<'; string lh lexbuf }
  | "&gt;"
      { add_char_to_string lh '>'; string lh lexbuf }
  | "&amp;"
      { add_char_to_string lh '&'; string lh lexbuf }
  | "&quot;"
      { add_char_to_string lh '"'; string lh lexbuf }
  | "&apos;"
      { add_char_to_string lh '\''; string lh lexbuf }
  | "\013\n" { add_char_to_string lh '\n'; string lh lexbuf }
  | '"'    { let s = get_string lh in pop_state lh; push_operator lh; STRING s }  (* End of string *)
  | eof    { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _      { add_char_to_string lh (lexeme_char lexbuf 0); string lh lexbuf }

and string_single lh = parse
  | "''"   { add_char_to_string lh '\''; string_single lh lexbuf }                 (* Escaped ' *)
  | "&#" (char_ref_number as charref) ";"
      { let charref_int = int_of_string charref in
        let s = Encoding.character (Encoding.get_internal_encoding ()) charref_int in
        add_string_to_string lh s; string_single lh lexbuf }
  | "&#x" (char_ref_hex_number as hexcharref) ";"
      { let hexcharref_int = int_of_string ("0x" ^ hexcharref) in
	let s = Encoding.character (Encoding.get_internal_encoding ()) hexcharref_int in
        add_string_to_string lh s; string_single lh lexbuf }
  | "&lt;"
      { add_char_to_string lh '<'; string_single lh lexbuf }
  | "&gt;"
      { add_char_to_string lh '>'; string_single lh lexbuf }
  | "&amp;"
      { add_char_to_string lh '&'; string_single lh lexbuf }
  | "&quot;"
      { add_char_to_string lh '"'; string_single lh lexbuf }
  | "&apos;"
      { add_char_to_string lh '\''; string_single lh lexbuf }
  | "\013\n" { add_char_to_string lh '\n'; string_single lh lexbuf }
  | '\''   { let s = get_string lh in pop_state lh; push_operator lh; STRING s }  (* End of string *)
  | eof    { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _      { add_char_to_string lh (lexeme_char lexbuf 0); string_single lh lexbuf }

and at_string lh = parse
  | "\"\"" { add_char_to_string lh '"'; at_string lh lexbuf }                      (* Escaped quote *)
  | '"'    { let s = get_string lh in ATSTRING s }                               (* End of string *)
  | eof    { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _      { add_char_to_string lh (lexeme_char lexbuf 0); at_string lh lexbuf }

and at_string_single lh = parse
  | "''"   { add_char_to_string lh '\''; at_string_single lh lexbuf }              (* Escaped ' *)
  | '\''   { let s = get_string lh in ATSTRING s }                               (* End of string *)
  | eof    { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _      { add_char_to_string lh (lexeme_char lexbuf 0); at_string_single lh lexbuf }

