(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lexing_util.ml,v 1.10 2006/02/08 19:00:48 pmichiel Exp $ *)

(* Module: Lexing_util
   Description:
     This module contains some basic functions used during
     lexing. Notably this support lexical state handling during XML
     and XQuery parsing.
*)

open Error

open Xquery_common_ast
open Xquery_ast
open Xquery_core_ast


(****************************)
(* Lexing global parameters *)
(****************************)

(* Character encoding *)

type encoding_kind =
  | Utf8
  | Iso88591


(* Language currently parsed *)

type language_kind =
  | XPath_2_0
  | XQuery_1_0


(* Lexical state *)

type lex_state =
      (* XQuery states *)
  | DefaultState
  | OperatorState
  | NamespaceDeclState
  | NamespaceKeywordState
  | CopyNamespacesState
  | XMLSpaceDeclState
  | ItemTypeState
  | KindTestState
  | KindTestForPIState
  | SchemaContextStepState
  | VarNameState
  | PragmaState
      (* XML & XQuery states *)
  | StartTagState
  | ElementContentState
  | EndTagState
  | XMLCommentState
  | PIState
  | CDATASectionState
  | QuoteAttributeContentState      (* Note: There are three kinds of attribute value lexing. - Jerome *)
  | AposAttributeContentState
  | EntityIncludedInLiteralState
      (* Galax extensions states *)
  | SchemaDeclarationState
  | TypeDeclarationState
  | XTypeState


(**************************)
(* Lexical state handling *)
(**************************)

type lexing_handler =
    { current_lex_state     : lex_state Stack.t;
      current_parsing_kind  : language_kind;
      mutable depth         : int;
      mutable string_buffer : string;
      mutable string_end    : int }

let initial_string_buffer () = String.create 256

let new_lexing_handler language_kind =
  { current_lex_state    = Stack.create ();
    current_parsing_kind = language_kind;
    depth                = 0;
    string_buffer        = initial_string_buffer ();
    string_end           = 0 }

let get_current_lex_state lh =
  Stack.top lh.current_lex_state


(****************************)
(* String and text handling *)
(****************************)

(* Manipulation of the string buffer *)

let reset_string lh =
  lh.string_buffer <- initial_string_buffer ();
  lh.string_end <- 0

let add_char_to_string lh ch =
  let len = String.length (lh.string_buffer) in
  if lh.string_end >= len
  then
    begin
      let newBuffer = String.create (len * 2) in
      String.blit (lh.string_buffer) 0 newBuffer 0 len;
      lh.string_buffer <- newBuffer
    end;
  String.unsafe_set lh.string_buffer lh.string_end ch;
  lh.string_end <- succ lh.string_end

let add_string_to_string lh str =
  String.iter (add_char_to_string lh) str

let get_string lh =
  let s = String.sub lh.string_buffer 0 lh.string_end in
  lh.string_buffer <- initial_string_buffer ();
  s


(****************************)
(* Push a new lexical state *)
(****************************)

(* XQuery states *)

let push_default lh =
  reset_string lh;
  Stack.push DefaultState lh.current_lex_state

let push_operator lh =
  reset_string lh;
  Stack.push OperatorState lh.current_lex_state

let push_namespacedecl lh =
  Stack.push NamespaceDeclState lh.current_lex_state

let push_namespacekeyword lh =
  Stack.push NamespaceKeywordState lh.current_lex_state

let push_copynamespaces lh =
  Stack.push CopyNamespacesState lh.current_lex_state

let push_xmlspacedecl lh =
  Stack.push XMLSpaceDeclState lh.current_lex_state

let push_itemtype lh =
  Stack.push ItemTypeState lh.current_lex_state

let push_kindtest lh =
  Stack.push KindTestState lh.current_lex_state

let push_kindtestforpi lh =
  Stack.push KindTestForPIState lh.current_lex_state

let push_schemacontextstep lh =
  Stack.push SchemaContextStepState lh.current_lex_state

let push_varname lh =
  Stack.push VarNameState lh.current_lex_state

let push_pragma lh =
  reset_string lh;
  Stack.push PragmaState lh.current_lex_state

(* XML & XQuery states *)

let push_opening_tag lh =
  Stack.push StartTagState lh.current_lex_state

let push_text lh =
  reset_string lh;
  Stack.push ElementContentState lh.current_lex_state

let push_closing_tag lh =
  Stack.push EndTagState lh.current_lex_state

let push_comment lh =
  reset_string lh;
  Stack.push XMLCommentState lh.current_lex_state

let push_processing_instruction lh =
  reset_string lh;
  Stack.push PIState lh.current_lex_state

let push_cdata lh =
  Stack.push CDATASectionState lh.current_lex_state

let push_attribute_text_double lh =
  reset_string lh;
  Stack.push QuoteAttributeContentState lh.current_lex_state

let push_attribute_text_single lh =
  reset_string lh;
  Stack.push AposAttributeContentState lh.current_lex_state

let push_entity_included lh =
  reset_string lh;
  Stack.push ElementContentState lh.current_lex_state

let push_entity_included_in_literal lh =
  reset_string lh;
  Stack.push EntityIncludedInLiteralState lh.current_lex_state

(* Galax extensions states *)

let push_schema_declaration lh =
  Stack.push SchemaDeclarationState lh.current_lex_state

let push_type_declaration lh =
  Stack.push TypeDeclarationState lh.current_lex_state

let push_xtype lh =
  Stack.push XTypeState lh.current_lex_state

(****************************)
(* Poping the lexical state *)
(****************************)

let pop_state lh =
  reset_string lh;
  ignore(Stack.pop lh.current_lex_state)

let pop_state_keep_buffer lh =
  ignore(Stack.pop lh.current_lex_state)


(* Access the comments depth *)

let get_depth lh   = lh.depth
let set_depth lh i = lh.depth <- i


(****************)
(* Parsing kind *)
(****************)

let get_parsing_kind lh =
  lh.current_parsing_kind

type attribute_quote_kind =
  | SingleQuoteKind
  | DoubleQuoteKind
  | EntityIncludedInLiteralKind

let attribute_quote_kind lh =
  match get_current_lex_state lh with
  | AposAttributeContentState -> SingleQuoteKind
  | QuoteAttributeContentState -> DoubleQuoteKind
  | EntityIncludedInLiteralState -> EntityIncludedInLiteralKind
  | _ ->
      raise (Query (Internal_Error "Requiring quote kind outside of an attribute"))

let init_xquery_lexing () =
  let lh = new_lexing_handler XQuery_1_0 in
  Stack.push DefaultState lh.current_lex_state;
  lh

let init_type_lexing () =
  let lh = new_lexing_handler XQuery_1_0 in
  Stack.push XTypeState lh.current_lex_state;
  lh

(********************)
(* Token processing *)
(********************)

(* Extracts an axis name *)

let resolve_axis =
  function
    | "attribute" -> Attribute
    | "child" -> Child
    | "descendant" -> Descendant
    | "descendant-or-self" -> Descendant_or_self
    | "parent" -> Parent
    | "self" -> Self
    | "ancestor" -> Ancestor
    | "ancestor-or-self" -> Ancestor_or_self
    | "following-sibling" -> Following_sibling
    | "preceding-sibling" -> Preceding_sibling
    (* full axis feature -- Philippe *)
    | "following" -> Following
    | "preceding" -> Preceding
    | s -> raise (Query (Parsing (Finfo.parsing_locinfo (),"Unkown XQuery axis " ^ s)))

(* The following axis are not in XQuery - Jerome
    | "namespace" -> Namespace
*)

let get_axis axis_name =
  resolve_axis axis_name

(* Extracts escaped name *)

let get_escaped_name var_s =
  let len = (String.length var_s) - 1 in
  (String.sub var_s 1 len)

type qname_kind =
  | NCNAME_KIND of string
  | QNAME_KIND of Namespace_names.uqname

let process_qname_string s =
  try
    let i = String.index s ':' in
    if i = 0
    then
      NCNAME_KIND (get_escaped_name s)
    else
      QNAME_KIND (Namespace_names.uqname_element_of_string s)
  with
  | _ ->
      NCNAME_KIND s

(* Check the validity of a PI target name *)

let get_target_pi pi_target =
  if (String.lowercase pi_target = "xml")
  then
    raise (Query (Parsing (Finfo.parsing_locinfo (),"Process instruction target cannot match ['X' 'x']['M' 'm']['L' 'l']")))
  else
    pi_target

(* Extracts the character encoding *)

let get_xml_encoding_quotes var_s =
  let mid = String.index var_s '"' in
  let sta = mid+1 in
  let len = (String.length var_s) - sta - 1 in
  String.sub var_s sta len

let get_xml_encoding_single_quotes var_s =
  let mid = String.index var_s '\'' in
  let sta = mid+1 in
  let len = (String.length var_s) - sta - 1 in
  String.sub var_s sta len

(* Extracts character and entity references *)

let get_char_ref       l =
  let len = String.length l in
  String.sub l 2 (len-3)

let get_hexchar_ref    l =
  let len = String.length l in
  String.sub l 3 (len-4)

let get_entity_ref l =
  let len = String.length l in
  String.sub l 1 (len-2)

