(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: namespacedecl_lexer.mll,v 1.4 2006/02/23 20:08:02 simeon Exp $ *)

(* Module: Namespacedecl_lexer
   Description:
     This module contains the part of the XQuery lexer corresponding
     to the NAMESPACEDECL lexical state.
*)

{

open Lexing

open Finfo
open Error
open Parse_xquery   (* tokens definition *)
open Lexing_util     (* lexical state handling *)

}

(*****************************)
(* Lexer regular expressions *)
(*****************************)

(* Whitespace *)

let space = ['\032' '\009' '\013' '\010']
let whitespace = space*
let whitespaceplus = space+

(* XML character sets *)

(* Note:
     Some of these Char codes have been taken from the PXP ISO
     parser. I am not sure that they are doing the right thing and
     should be double checked for conformance.
   - Jerome
 *)

let base_char    = ['A'-'Z' 'a'-'z']
let ideographics = ['\192'-'\214' '\216'-'\246' '\248'-'\255']
let extender     = '\183'
let digit        = ['0'-'9']

(* XML names *)

let letter   = base_char (* | ideographics *)
let nmstart  = letter | '_'
let nmchar   = letter | extender | digit | '.' | '-' | '_'
 
let ncname   = nmstart nmchar*

let ascii_digit = ['0'-'9']
let ascii_hexdigit = ['0'-'9' 'A'-'F' 'a'-'f']
let char_ref_number = ascii_digit+
let char_ref_hex_number = ascii_hexdigit+

(****************)
(* Lexing rules *)
(****************)

rule token lh = parse

(* Whitespace *)
  | space                  (* for all others, skip blanks *)
      { token lh lexbuf }

(* Comments *)
  | ":)"
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Unmatched end of comment"))) }
  | "(:" 
      { set_depth lh 1; xquery_comment lh lexbuf; token lh lexbuf }
  | "{--" 
      { raise (Query (Lexing(lexing_locinfo lexbuf, "ATTENTION: The syntax of comments in XQuery has now changed from {-- --} to (: :)"))) }

(* Equals and EqualsEquals *)
  | "="
      { EQUALS }

(* XML Names *)
  | ncname
      { NCNAME (lexeme lexbuf) }

(*******************)
(* String literals *)
(*******************)

  | "\""
      { reset_string lh; string lh lexbuf }
  | "'"
      { reset_string lh; string_single lh lexbuf }

(***************)
(* End of file *)
(***************)

  | eof
      { EOF }

(********************)
(* Unknown caracter *)
(********************)

(* Note:
     If the lexer arrives here, this is an error.
   - Jerome
*)

  | _
      { raise (Query (Lexing(lexing_locinfo lexbuf, ("Illegal character '" ^ (lexeme lexbuf) ^ "' in expression")))) }

and xquery_comment lh = parse
  | "(:"
      { set_depth lh (succ (get_depth lh)); xquery_comment lh lexbuf }
  | ":)"
      { set_depth lh (pred (get_depth lh)); if get_depth lh > 0 then xquery_comment lh lexbuf }
  | eof
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Comment not terminated"))) }
  | [^ '\n']
      { xquery_comment lh lexbuf }
  | "\n"
      { xquery_comment lh lexbuf }

(****************)
(* String lexer *)
(****************)

and string lh = parse
  | "\"\"" { add_char_to_string lh '"'; string lh lexbuf }    (* Escaped quote *)
  | "\013\n" { add_char_to_string lh '\n'; string lh lexbuf }
  | '"'     { let s = get_string lh in pop_state lh; push_default lh; STRING s }                  (* End of string *)
  | eof     { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _       { add_char_to_string lh (lexeme_char lexbuf 0); string lh lexbuf }

and string_single lh = parse
  | "''"  { add_char_to_string lh '\''; string_single lh lexbuf }    (* Escaped ' *)
  | "\013\n" { add_char_to_string lh '\n'; string_single lh lexbuf }
  | '\''  { let s = get_string lh in pop_state lh; push_default lh; STRING s }                    (* End of string *)
  | eof   { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _     { add_char_to_string lh (lexeme_char lexbuf 0); string_single lh lexbuf }

