(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: operator_lexer.mll,v 1.16 2006/05/16 23:33:59 simeon Exp $ *)

(* Module: Operator_lexer
   Description:
     This module contains the part of the XQuery lexer corresponding
     to the OPERATOR lexical state.
*)

{

open Lexing

open Finfo
open Error
open Parse_xquery   (* tokens definition *)
open Lexing_util     (* lexical state handling *)

}

(*****************************)
(* Lexer regular expressions *)
(*****************************)

(* Whitespace *)

let space = ['\032' '\009' '\013' '\010']
let whitespace = space*
let whitespaceplus = space+

(* XML character sets *)

(* Note:
     Some of these Char codes have been taken from the PXP ISO
     parser. I am not sure that they are doing the right thing and
     should be double checked for conformance.
   - Jerome
 *)

let base_char    = ['A'-'Z' 'a'-'z']
let ideographics = ['\192'-'\214' '\216'-'\246' '\248'-'\255']
let extender     = '\183'
let digit        = ['0'-'9']
let ascii_digit = ['0'-'9']
let ascii_hexdigit = ['0'-'9' 'A'-'F' 'a'-'f']

let char_ref_number = ascii_digit+
let char_ref_hex_number = ascii_hexdigit+


(* XML names *)

let letter   = base_char (* | ideographics *)
let nmstart  = letter | '_'
let nmchar   = letter | extender | digit | '.' | '-' | '_'
 
let ncname   = nmstart nmchar*

let qname    = ncname (':' ncname)?

(* Numerics *)

let sign            = '-' | '+'
let exponent        = 'e' | 'E'

let integer_numeric = digit+
let decimal_numeric = (('.' digit+) | (digit+ '.' (digit+)?))
let double_numeric  = (('.' digit+) | (digit+ ('.' (digit+)?)?)) exponent sign? digit+

(****************)
(* Lexing rules *)
(****************)

rule token lh = parse

(* Arithmetic operators *)
  | "-"
      { pop_state lh; push_default lh; MINUS }
  | "+"
      { pop_state lh; push_default lh; PLUS }
  | "*"
      { pop_state lh; push_default lh; MULT }

(* Whitespace *)
  | space                  (* for all others, skip blanks *)
      { token lh lexbuf }

(* Comments *)
  | ":)"
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Unmatched end of comment"))) }
  | "(:" 
      { set_depth lh 1; xquery_comment lh lexbuf; token lh lexbuf }
  | "{--" 
      { raise (Query (Lexing(lexing_locinfo lexbuf, "ATTENTION: The syntax of comments in XQuery has now changed from {-- --} to (: :)"))) }

(* Reserved words *)
  | "encoding"
      { ENCODING }
  | "or"
      { pop_state lh; push_default lh; OR }
  | "and"
      { pop_state lh; push_default lh; AND }
  | "div"
      { pop_state lh; push_default lh; DIV }
  | "idiv"
      { pop_state lh; push_default lh; IDIV }
  | "mod"
      { pop_state lh; push_default lh; MOD }
  | "in"
      { pop_state lh; push_default lh; IN }
  | "satisfies"
      { pop_state lh; push_default lh; SATISFIES }
  | "return"
      { pop_state lh; push_default lh; RETURN }
  | "then"
      { pop_state lh; push_default lh; THEN }
  | "else"
      { pop_state lh; push_default lh; ELSE }
  | "to"
      { pop_state lh; push_default lh; TO }
  | "where"
      { pop_state lh; push_default lh; WHERE }
  | "intersect"
      { pop_state lh; push_default lh; INTERSECT }
  | "union"
      { pop_state lh; push_default lh; UNION }
  | "except"
      { pop_state lh; push_default lh; EXCEPT }
  | "<<"
      { pop_state lh; push_default lh; PRECEDES }
  | ">>"
      { pop_state lh; push_default lh; FOLLOWS }
  | "case"
      { pop_state lh; push_itemtype lh; CASE }
  | "ascending"
      { ASCENDING }
  | "descending"
      { DESCENDING }
  | "empty" whitespaceplus "greatest"
      { EMPTYGREATEST }
  | "empty" whitespaceplus "least"
      { EMPTYLEAST }
  | "default"
      { DEFAULT }

(* Composed reserved words *)
  | "instance" whitespaceplus "of"
      { pop_state lh; push_itemtype lh; INSTANCEOF }
  | "assert" whitespaceplus "as"
      { pop_state lh; push_itemtype lh; ASSERTAS }
  | "cast" whitespaceplus "as"
      { pop_state lh; push_itemtype lh; CASTAS }
  | "castable" whitespaceplus "as"
      { pop_state lh; push_itemtype lh; CASTABLEAS }
  | "treat" whitespaceplus "as"
      { pop_state lh; push_itemtype lh; TREATAS }

  (* Updates *)
  | "as" whitespaceplus "first"
      { ASFIRST }
  | "as" whitespaceplus "last"
      { ASLAST }
  | "into" whitespace "{"
      { push_default lh; INTOLCURLY }
  | "before" whitespace "{"
      { push_default lh; BEFORELCURLY }
  | "after" whitespace "{"
      { push_default lh; AFTERLCURLY }
  | "with" whitespace "{"
      { push_default lh; WITHLCURLY }
  | "to" whitespace "{"
      { push_default lh; TOLCURLY }

(* Declaration keywords *)

  | "preserve"
      { PRESERVE }

  | "strip"
      { STRIP }

  | "ordered"
      { ORDERED }

  | "unordered"
      { UNORDERED }

  | "external"
      { EXTERNAL }

  | "as"
      { pop_state lh; push_itemtype lh; AS }

  | "at"
      { pop_state lh; push_default lh; AT }

(* Keywords followed by variables *)

  | "some" whitespaceplus "$"
      { pop_state lh; push_varname lh; SOMEDOLLAR }

  | "every" whitespaceplus "$"
      { pop_state lh; push_varname lh; EVERYDOLLAR }

  | "for" whitespaceplus "$"
      { pop_state lh; push_varname lh; FORDOLLAR }

  | "let" whitespaceplus "$"
      { pop_state lh; push_varname lh; LETDOLLAR }

(* Keywords followed by a parenthesis *)

  | "stable" whitespace "order" whitespaceplus "by"
      { pop_state lh; push_default lh; STABLEORDERBY }

  | "order" whitespaceplus "by"
      { pop_state lh; push_default lh; ORDERBY }

(* Type thingies *)
  | ":"
      { COLON }

(* Slash and SlashSlash *)
  | "/"
      { pop_state lh; push_default lh; SLASH }
  | "//"
      { pop_state lh; push_default lh; SLASHSLASH }

(* Equals and EqualsEquals *)
  | "eq"
      { pop_state lh; push_default lh; EQ }
  | "="
      { pop_state lh; push_default lh; EQUALS }

(* NotEquals and NotEqualsEquals *)
  | "ne"
      { pop_state lh; push_default lh; NE }
  | "!="
      { pop_state lh; push_default lh; NOTEQUALS }

(* Comparators *)
  | "<"
      { pop_state lh; push_default lh; LT }
  | ">"
      { pop_state lh; push_default lh; GT }

  | "lt" 
      { pop_state lh; push_default lh; LTOP }
  | "gt"
      { pop_state lh; push_default lh; GTOP }

  | "le"
      { pop_state lh; push_default lh; LTE }
  | "ge"
      { pop_state lh; push_default lh; GTE }

  | "<="
      { pop_state lh; push_default lh; LTEQUALS }
  | ">="
      { pop_state lh; push_default lh; GTEQUALS }

  | "is"
      { pop_state lh; push_default lh; IS }

(* Assignment *)
  | ":="
      { pop_state lh; push_default lh; COLONEQUALS }

(* Choice *)
  | "|"
      { pop_state lh; push_default lh; BAR }

(* Parenthesis *)
  | "("
      { pop_state lh; push_default lh; LPAR }
  | ")"
      { RPAR }
  | "[" 
      { pop_state lh; push_default lh; LBRACK }
  | "]" 
      { RBRACK }
  | "{"
      { push_default lh;
	LCURLY }
  | "}"
      { try
	  pop_state lh; RCURLY
        with
          Stack.Empty ->
	    raise (Query (Lexing(lexing_locinfo lexbuf, "Unmatched end of enclosed expression"))) }

(* Misc symbols *)
  | ","
      { pop_state lh; push_default lh; COMMA }
  | ";"
      { pop_state lh; push_default lh; SEMICOLON }
  | "."
      { DOT }
  | ".."
      { DOTDOT }

(* Variables *)
  | "$"
      { pop_state lh; push_varname lh; DOLLAR }

(* XML Names *)
  | qname
      { let var_s = (lexeme lexbuf) in
        match process_qname_string var_s with
	| NCNAME_KIND s ->
	    NCNAME s
	| QNAME_KIND qname ->
	    QNAME qname }

(********************)
(* Numeric literals *)
(********************)

(* Double *)
  | double_numeric
      { let v = float_of_string (lexeme lexbuf) in
        (DOUBLE v) }

(* Decimals *)
  | decimal_numeric
      { let v = Decimal._decimal_of_string (lexeme lexbuf) in
        (DECIMAL v) }

(* Integers *)
  | integer_numeric
      { let v = Decimal._integer_of_string (lexeme lexbuf) in
        (INT v) }

(*******************)
(* String literals *)
(*******************)

  | "\""
      { reset_string lh; string lh lexbuf }
  | "'"
      { reset_string lh; string_single lh lexbuf }

(***************)
(* End of file *)
(***************)

  | eof
      { EOF }

(********************)
(* Unknown caracter *)
(********************)

(* Note:
     If the lexer arrives here, this is an error.
   - Jerome
*)

  | _
      { raise (Query (Lexing(lexing_locinfo lexbuf, ("Illegal character '" ^ (lexeme lexbuf) ^ "' in expression")))) }

and xquery_comment lh = parse
  | "(:"
      { set_depth lh (succ (get_depth lh)); xquery_comment lh lexbuf }
  | ":)"
      { set_depth lh (pred (get_depth lh)); if get_depth lh > 0 then xquery_comment lh lexbuf }
  | eof
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Comment not terminated"))) }
  | [^ '\n']
      { xquery_comment lh lexbuf }
  | "\n"
      { xquery_comment lh lexbuf }

(****************)
(* String lexer *)
(****************)

and string lh = parse
  | "\"\""  { add_char_to_string lh '"'; string lh lexbuf }          (* Escaped quote *)
  | "&#" (char_ref_number as charref) ";"
      { let charref_int = int_of_string charref in
        let s = Encoding.character (Encoding.get_internal_encoding ()) charref_int in
        add_string_to_string lh s; string lh lexbuf }
  | "&#x" (char_ref_hex_number as hexcharref) ";"
      { let hexcharref_int = int_of_string ("0x" ^ hexcharref) in
	let s = Encoding.character (Encoding.get_internal_encoding ()) hexcharref_int in
        add_string_to_string lh s; string lh lexbuf }
  | "&lt;"
      { add_char_to_string lh '<'; string lh lexbuf }
  | "&gt;"
      { add_char_to_string lh '>'; string lh lexbuf }
  | "&amp;"
      { add_char_to_string lh '&'; string lh lexbuf }
  | "&quot;"
      { add_char_to_string lh '"'; string lh lexbuf }
  | "&apos;"
      { add_char_to_string lh '\''; string lh lexbuf }
  | "\013\n" { add_char_to_string lh '\n'; string lh lexbuf }
  | '"'     { STRING (get_string lh) }                  (* End of string *)
  | eof     { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _       { add_char_to_string lh (lexeme_char lexbuf 0); string lh lexbuf }

and string_single lh = parse
  | "''"  { add_char_to_string lh '\''; string_single lh lexbuf }    (* Escaped ' *)
  | "&#" (char_ref_number as charref) ";"
      { let charref_int = int_of_string charref in
        let s = Encoding.character (Encoding.get_internal_encoding ()) charref_int in
        add_string_to_string lh s; string_single lh lexbuf }
  | "&#x" (char_ref_hex_number as hexcharref) ";"
      { let hexcharref_int = int_of_string ("0x" ^ hexcharref) in
	let s = Encoding.character (Encoding.get_internal_encoding ()) hexcharref_int in
        add_string_to_string lh s; string_single lh lexbuf }
  | "&lt;"
      { add_char_to_string lh '<'; string_single lh lexbuf }
  | "&gt;"
      { add_char_to_string lh '>'; string_single lh lexbuf }
  | "&amp;"
      { add_char_to_string lh '&'; string_single lh lexbuf }
  | "&quot;"
      { add_char_to_string lh '"'; string_single lh lexbuf }
  | "&apos;"
      { add_char_to_string lh '\''; string_single lh lexbuf }
  | "\013\n" { add_char_to_string lh '\n'; string_single lh lexbuf }
  | '\''  { STRING (get_string lh) }                    (* End of string *)
  | eof   { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _     { add_char_to_string lh (lexeme_char lexbuf 0); string_single lh lexbuf }

