(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: namespace_context.ml,v 1.3 2006/03/02 17:40:52 mff Exp $ *)

(* Module: Namespace_context
   Description:
     This module implements support for the namespace context.
*)

open Error

open Namespace_names
open Namespace_builtin


(**********************)
(* Namespace bindings *)
(**********************)

(* Namespace binding *)

type binding = prefix * uri

(* Type for binding tables *)

type binding_table = binding list

(* Create a binding for a uri with a fresh prefix *)

let create_binding uri =
  (new_prefix "glx",uri)

(* Filters the built-in prefixes and the default function prefix *)

let cleanup_bindings bindings required_prefixes =
  let real_required_prefixes =
    List.filter (fun (pref,uri) -> not (is_xml_built_in_namespace (pref,uri))) required_prefixes
  in
  List.filter (fun (pref,uri) -> not(is_built_in_namespace (pref,uri) || pref = NSDefaultFunctionPrefix) || (List.exists (fun y -> (pref,uri) = y) real_required_prefixes)) bindings


(*************************)
(* Namespace environment *)
(*************************)

(* type for namespace environments *)

type nsenv_content =
    { nsenv_bindings : binding_table;
      nsenv_previous : nsenv }

and nsenv = nsenv_content option ref

(* creates an empty environment *)

let empty_nsenv = ref None

let default_xquery_nsenv =
  let nsenv_content =
    { nsenv_bindings = xquery_built_in_namespaces;
      nsenv_previous = empty_nsenv }
  in
  ref (Some nsenv_content)

let default_xml_nsenv =
  let nsenv_content =
    { nsenv_bindings = xml_built_in_namespaces;
      nsenv_previous = empty_nsenv }
  in
  ref (Some nsenv_content)

(* adds a new prefix,uri binding in a namespace environment *)

let add_ns pref uri =
  (* Note:
       In fact, adding the "" namespace corresponds to removing the
       corresponding prefix/uri binding!  See the namespace rec for
       more on this.
   - Jerome *)
  begin
    match (pref,uri) with
      (* Hack for the 'empty' prefix which is used in the XML
	 Schema to types mapping *)
    | (NSPrefix "empty", NSUri "") ->
	()
    | (NSPrefix pre, NSUri "") ->
	raise (Query (Namespace_Error ("Namespace name cannot be empty when using the namespace prefix "^pre^" [See Namespace REC Section 2. Declaring Namespaces]")))
    | (NSWildcardPrefix, _) ->
	raise (Query (Namespace_Internal ("Cannot redefine the wildcard prefix")))
    | _ ->
	()
  end

let add_all_ns nsenv binding_table =
  (* Note: Important! if there is no new bindings,
     just point to the previous namespace environment.
     - Jerome *)
  if binding_table = []
  then
    nsenv
  else
    begin
      (* Checking some namespace consistency *)
      List.iter (fun (x,y) -> add_ns x y) binding_table;
      (* Creates a new namespace environment, pointing to the previous context *)
      let new_nsenv_content =
	{ nsenv_bindings = binding_table;
	  nsenv_previous = nsenv }
      in
      ref (Some new_nsenv_content)
    end

(* gets the namespace of a given prefix *)

let rec get_ns_of_prefix nsr pref =
  match !nsr with
  | None -> 
      begin
	match pref with
	| NSDefaultElementPrefix ->
	    raise (Query (Namespace_Internal ("Default element namespace not found.")))
	| NSDefaultFunctionPrefix ->
	    raise (Query (Namespace_Internal ("Default function namespace not found.")))
	| NSWildcardPrefix ->
	    raise (Query (Namespace_Internal ("Wildcard namespace not found.")))
        | NSPrefix nc ->
	    raise (Query (Namespace_Internal ("Namespace prefix "
					  ^ nc
					  ^ " not found.")))
      end
  | Some nsenv_content ->
      begin
	try
	  List.assoc pref nsenv_content.nsenv_bindings
	with
	| Not_found ->
	    get_ns_of_prefix nsenv_content.nsenv_previous pref
      end

let rec get_prefix_of_uri nsenv uri =
  match !nsenv with
  | None -> 
      begin
	match uri with
	| NSUri uri ->
	    raise (Query (Namespace_Internal ("Default prefix in scope for uri: " ^ uri)))
	| NSWildcardUri ->
	    raise (Query (Namespace_Internal ("Wildcard uri never has a prefix.")))
      end
  | Some nsenv_content ->
      begin
	try
	  fst(List.find
		(fun (y,x) -> x = uri && (y != NSDefaultFunctionPrefix)) 
		nsenv_content.nsenv_bindings)
	with
	| Not_found ->
	    get_prefix_of_uri nsenv_content.nsenv_previous uri
      end

let make_binding nsenv (orig_prefix,uri,ncname) =
  (* Original prefix is ignored for now - Jerome *)
  try
    let pref = get_prefix_of_uri nsenv uri in
    ((pref,ncname), None, (pref,uri))
  with
  | _ ->
      let (pref,uri) = create_binding uri in
      ((pref,ncname),Some (pref,uri), (pref,uri))

let make_attribute_binding nsenv (orig_prefix,uri,ncname) =
  (* Original prefix is ignored for now - Jerome *)
  if uri = NSUri ""
  then
    ((NSDefaultElementPrefix,ncname),None,(NSDefaultElementPrefix,uri))
  else
    begin
      try
	let pref = get_prefix_of_uri nsenv uri in
	((pref,ncname), None, (pref,uri))
      with
      | _ ->
	  let (pref,uri) = create_binding uri in
	  ((pref,ncname),Some (pref,uri), (pref,uri))
    end

(* Returns the new bindings for the current context *)

let active_bindings nsenv =
  match !nsenv with
  | None -> []
  | Some nsenv_content ->
      nsenv_content.nsenv_bindings

let rec delta_bindings nsenv1 nsenv2 =
  if nsenv1 == nsenv2
  then []
  else
    match !nsenv1 with
    | None -> []
    | Some nsenv_content ->
	nsenv_content.nsenv_bindings
	@ (delta_bindings (nsenv_content.nsenv_previous) nsenv2)

(* Returns all of the bindings *)

let rec flatten_bindings nsenv =
  match !nsenv with
  | None -> []
  | Some nsenv_content ->
      nsenv_content.nsenv_bindings @ (flatten_bindings nsenv_content.nsenv_previous)

(* Are the nsenvs the identical ?*)

let same_nsenv nsenv1 nsenv2 = (nsenv1 == nsenv2)

