(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies, AT&T Labs and IBM.       *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: namespace_resolve.ml,v 1.3 2006/05/08 12:09:39 simeon Exp $ *)

(* Module: Namespace_resolve
   Description:
     This module implements namespace resolution.
*)


open Namespace_names
open Namespace_context
open Namespace_symbols

open Error


(**********************************************)
(* Resolving QNames to Expanded QNames        *)
(* Expanded QNames are represented by symbols *)
(**********************************************)

(* Note:
     Given namespace environment an unresolved symbol, returns the
     resolved symbol for the resolved qname
   - J&B
*)

let resolve_element_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSWildcardPrefix in
      (NSWildcardPrefix, uri, localname)
  | (NSDefaultElementPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSDefaultElementPrefix in
      (NSDefaultElementPrefix, uri, localname)
  | (NSPrefix nc, localname) ->
      let uri = get_ns_of_prefix nsr (NSPrefix nc) in
      (NSPrefix nc, uri, localname)
  | (NSDefaultFunctionPrefix, localname) ->
      raise (Query (Namespace_Error "Trying to resolve an element name as a function name"))

let resolve_function_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      raise (Query (Namespace_Error ("Resolution of wildcard namespace not supported!")))
  | (NSDefaultElementPrefix, localname) ->
      raise (Query (Namespace_Error "Trying to resolve a function name as an element name"))
  | (NSPrefix nc, localname) ->
      let uri = get_ns_of_prefix nsr (NSPrefix nc) in
      (NSPrefix nc, uri, localname)
  | (NSDefaultFunctionPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSDefaultFunctionPrefix in
      (NSDefaultFunctionPrefix, uri, localname)

let resolve_variable_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      raise (Query (Namespace_Error ("Resolution of wildcard namespace not supported!")))
  | (NSDefaultElementPrefix, localname) ->
      (* Note:
	   Variables with no prefix are always in no namespace. See
	   XQuery spec [Section 3.1.2 Variable References].
         - Jerome *)
      (NSDefaultElementPrefix,NSUri "", localname)
  | (NSPrefix nc, localname) ->
      let uri = get_ns_of_prefix nsr (NSPrefix nc) in
      (NSPrefix nc, uri, localname)
  | (NSDefaultFunctionPrefix, localname) ->
      raise (Query (Namespace_Error "Trying to resolve a variable name as a function name"))

let resolve_type_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSWildcardPrefix in
      (NSWildcardPrefix, uri, localname)
  | (NSDefaultElementPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSDefaultElementPrefix in
      (NSDefaultElementPrefix, uri, localname)
  | (NSPrefix nc, localname) ->
      let uri = get_ns_of_prefix nsr (NSPrefix nc) in
      (NSPrefix nc, uri, localname)
  | (NSDefaultFunctionPrefix, localname) ->
      raise (Query (Namespace_Error "Trying to resolve an element name as a function name"))

let resolve_group_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSWildcardPrefix in
      (NSWildcardPrefix, uri, localname)
  | (NSDefaultElementPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSDefaultElementPrefix in
      (NSDefaultElementPrefix, uri, localname)
  | (NSPrefix nc, localname) ->
      let uri = get_ns_of_prefix nsr (NSPrefix nc) in
      (NSPrefix nc, uri, localname)
  | (NSDefaultFunctionPrefix, localname) ->
      raise (Query (Namespace_Error "Trying to resolve a group name as a function name"))

let resolve_attrGroup_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSWildcardPrefix in
      (NSWildcardPrefix, uri, localname)
  | (NSDefaultElementPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSDefaultElementPrefix in
      (NSDefaultElementPrefix, uri, localname)
  | (NSPrefix nc, localname) ->
      let uri = get_ns_of_prefix nsr (NSPrefix nc) in
      (NSPrefix nc, uri, localname)
  | (NSDefaultFunctionPrefix, localname) ->
      raise (Query (Namespace_Error "Trying to resolve an attrGroup name as a function name"))

let resolve_attribute_qname nsr uqname =
  match uqname with 
  | (NSWildcardPrefix, localname) ->
      let uri = get_ns_of_prefix nsr NSWildcardPrefix in
      (NSWildcardPrefix, uri, localname)
  | (NSDefaultElementPrefix, localname) ->
      (NSDefaultElementPrefix, NSUri "", localname)
  | (NSPrefix prefix, localname) -> 
      let uri = get_ns_of_prefix nsr (NSPrefix prefix) in 
      (NSPrefix prefix,uri,localname)
  | (NSDefaultFunctionPrefix, _) ->
      raise (Query (Namespace_Error "Trying to resolve an attribute name as a function name"))

let resolve_element_qname_to_symbol nsr ename = 
  relem_symbol (resolve_element_qname nsr ename)

let resolve_attribute_qname_to_symbol nsr aname = 
  rattr_symbol (resolve_attribute_qname nsr aname)

let resolve_type_qname_to_symbol nsr tname = 
  rtype_symbol (resolve_type_qname nsr tname)

let resolve_anon_qname_to_symbol nsr tname = 
  anon_symbol (resolve_element_qname nsr tname)

