(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: norm_context.ml,v 1.18 2006/03/02 17:40:52 mff Exp $ *)

(* Module Norm_context
   Description:
     This module implements the part of the context which is required
     during the normalization phase.
*)

open Error

open Namespace_names
open Namespace_util
open Namespace_context

open Xquery_common_ast
open Xquery_core_ast

open Processing_context


(* Type for the normalization context *)

type norm_context = 
    { norm_processing_context : processing_context;
      norm_nsenv    	      : nsenv;
      norm_cxschema   	      : Xquery_type_core_ast.cxschema ;
      norm_fun_sigs 	      : (cfname * cfunction_signature) RQNameIntHashtbl.t }

(* Builds a new normalization context *)

let build_norm_context proc_ctxt nsenv cxschema siglist =
  let norm_context =
    { norm_processing_context = proc_ctxt;
      norm_nsenv    		 = nsenv;
      norm_cxschema              = cxschema;
      norm_fun_sigs 		 = RQNameIntHashtbl.create 167 }
  in
  let _ =
    List.iter
      (fun ((cfname, arity), (tfname, cfunc_signature)) ->
	let arity = List.length (fst cfunc_signature) in
	RQNameIntHashtbl.add norm_context.norm_fun_sigs (cfname,arity) (tfname,cfunc_signature)) siglist
  in
  norm_context

(* Default normalization context *)

let default_norm_context proc_ctxt =
  let default_nsenv = default_xquery_nsenv in
  let default_schema = Schema_builtin.built_in_cxschema in
  let default_signatures = [] in
  build_norm_context
    proc_ctxt
    default_nsenv
    default_schema
    default_signatures
  

(* Replace the previous processing context *)

let replace_processing_context_in_norm_context proc_ctxt norm_ctxt =
  { norm_processing_context = proc_ctxt;
    norm_nsenv    	    = norm_ctxt.norm_nsenv;
    norm_cxschema           = norm_ctxt.norm_cxschema;
    norm_fun_sigs 	    = norm_ctxt.norm_fun_sigs }

(* Replace the namespace environment  *)

let replace_namespace_env_in_norm_context nsenv norm_ctxt =
  { norm_processing_context = norm_ctxt.norm_processing_context;
    norm_nsenv    	    = nsenv;
    norm_cxschema           = norm_ctxt.norm_cxschema;
    norm_fun_sigs 	    = norm_ctxt.norm_fun_sigs }

(* We need to keep the hashtable of function signatures in tact. *)

let copy_norm_context_with_sigs ctxt nsenv cxschema = 
  let ctxt = 
  { norm_processing_context = ctxt.norm_processing_context;
    norm_nsenv    	    = nsenv;
    norm_cxschema           = Xquery_type_core_ast_util.merge_cxschema ctxt.norm_cxschema cxschema;
    norm_fun_sigs 	    = RQNameIntHashtbl.copy ctxt.norm_fun_sigs }
in
(*  Print_type_core.print_cxschema Format.std_formatter ctxt.norm_cxschema; *)
  ctxt

let copy_norm_context ctxt =
  { norm_processing_context = ctxt.norm_processing_context;
    norm_nsenv    	    = ctxt.norm_nsenv;
    norm_cxschema 	    = ctxt.norm_cxschema;
    norm_fun_sigs 	    = RQNameIntHashtbl.copy ctxt.norm_fun_sigs }

let add_sigs_to_norm_context context siglist =
  let norm_context = copy_norm_context context in 
  let _ =
    List.iter 
      (fun ((cfname, arity), (tfname, cfunc_signature)) ->
	RQNameIntHashtbl.add norm_context.norm_fun_sigs (cfname,arity) (tfname, cfunc_signature)) siglist
  in
  norm_context

(* Extract components from normalization the context *)

let processing_context_from_norm_context norm_ctxt = norm_ctxt.norm_processing_context
let nsenv_from_norm_context norm_ctxt  		   = norm_ctxt.norm_nsenv
let cxschema_from_norm_context norm_ctxt           = norm_ctxt.norm_cxschema

let one_sig_from_norm_context context (cfname,arity) =
  let l = RQNameIntHashtbl.find_all context.norm_fun_sigs (cfname,arity) in
  match l with
  | [] ->
      raise (Query(Undefined(
		   " Signature of function " ^ (prefixed_string_of_rqname cfname) ^
		   " with arity " ^ (string_of_int arity) ^
		   " undefined")))
  | [(f,s)] -> s
  | _ ->
      raise (Query(Symbol_Already_Defined (" Function " , (prefixed_string_of_rqname cfname) ^
				   " with arity " ^ (string_of_int arity) ^
				   " and signature (" ^ (string_of_int(List.length l)))))

