(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: norm_top.ml,v 1.90 2006/05/16 23:33:59 simeon Exp $ *)

(* Module: Norm_top
   Description:
     This module implements the XQuery normalization phase.
*)

open Format

open Error

open Namespace_names
open Namespace_builtin
open Namespace_symbols
open Namespace_resolve

open Datatypes
open Datatypes_util

open Xquery_common_ast
open Xquery_common_ast_util

open Xquery_ast
open Xquery_ast_util

open Xquery_type_ast
open Xquery_type_ast_util

open Xquery_core_ast
open Xquery_core_ast_annotation
open Xquery_core_ast_util

open Norm_context
open Norm_util
open Norm_expr


(****************************************************)
(* functions to map each components in query_module *)
(****************************************************)

(* Global variable declaration *)

let check_dup_vars ht cvname =
  if Namespace_util.RQNameHashtbl.mem ht cvname
  then
    raise (Query (Malformed_Core_Expr "[err:XQST0049] Two global variables with the same name"))
  else
    Namespace_util.RQNameHashtbl.add ht cvname ()

let normalize_var ht norm_context v =
  let nsenv = nsenv_from_norm_context norm_context in 
  let var_desc =
    match v.pvar_decl_desc with
    | (vname,dt,Some e) ->
	let cvname = resolve_variable_qname nsenv vname in
	check_dup_vars ht cvname;
	let ce = normalize_expr norm_context e in
	let cdt = normalize_optional_sequencetype norm_context dt in
	(cvname, cdt, Some ce)
    | (vname,dt,None) ->
	let cvname = resolve_variable_qname nsenv vname in
	check_dup_vars ht cvname;
	let cdt = normalize_optional_sequencetype norm_context dt in
	(cvname, cdt, None)
  in
  fmkcvar_decl var_desc v.pvar_decl_loc

let normalize_variables norm_context vs =
  let ht = Namespace_util.RQNameHashtbl.create 17 in
  List.map (normalize_var ht norm_context) vs

(* Key definition *)

let normalize_index norm_context v =
  let nsenv = nsenv_from_norm_context norm_context in 
  let index_def =
    match v.pindex_def_desc with
    | ValueIndex (kn,e1,e2) ->
	let ce1 = normalize_expr norm_context e1 in
	let ce2 = normalize_expr norm_context e2 in
	CValueIndex (kn, ce1, ce2)
    | NameIndex ename ->
	let cename = resolve_element_qname nsenv ename in
	CNameIndex cename
  in
  fmkcindex_def index_def v.pindex_def_loc

(* Function Body *)

let normalize_body norm_context b output_dt =
  match b with
  | EFunctionBltIn ->
      CEFunctionBltIn
  | EFunctionUser expr ->
      let ce = (normalize_expr norm_context expr) in
      let fi = (expr.pexpr_loc) in
      let (v1, var1) = gen_new_cvar (Some expr) fi in
      (* Apply rules for convert output value of function to simple
      value, if necessary *)
      let ce' = convert_function_input_output norm_context ((v1, var1), (ce, output_dt)) var1 in 
      (* Type match checks that result of evaluating function body
      matches declared return type *)
      (CEFunctionUser ce')

(* Function Definition *)

let check_proper_sigs system ht (cfname,arity) =
  let (_,uri,_) = cfname in
  if not(system) then
    if (uri = xml_uri) || (uri = xmlns_uri) || (uri = xs_uri) || (uri = xsi_uri) || (uri = fn_uri)
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0045] Function declared in a reserved URI"));
  if Namespace_util.RQNameIntHashtbl.mem ht (cfname,arity)
  then
    raise (Query (Malformed_Core_Expr "[err:XQST0034] Two function declarations with the same name and same arity"))
  else
    Namespace_util.RQNameIntHashtbl.add ht (cfname,arity) ()

let normalize_function_signature ht norm_context f = 
  let nsenv = nsenv_from_norm_context norm_context in
  let procctxt = processing_context_from_norm_context norm_context in
  let system = procctxt.Processing_context.system in
  let (fname, vars, (input_types, return_type), _) = f.pfunction_def_desc in
  let rfname = resolve_function_qname nsenv fname in
  let csign =
    (List.map
       (fun dt ->
	 (normalize_optional_sequencetype_strong norm_context dt))
       input_types,
     (normalize_optional_sequencetype_strong norm_context return_type)) in
  let arity = List.length (fst csign) in
  check_proper_sigs system ht (rfname,arity);
  ((rfname, arity), (rfname, csign))

let normalize_function_signatures norm_context fs =
  let ht = Namespace_util.RQNameIntHashtbl.create 17 in
  List.map (normalize_function_signature ht norm_context) fs

let check_dup_params ht cvname =
  if Namespace_util.RQNameHashtbl.mem ht cvname
  then
    raise (Query (Malformed_Core_Expr "[err:XQST0039] Two parameters with the same name in function declaration"))
  else
    Namespace_util.RQNameHashtbl.add ht cvname ()

let check_distinct_parameters cvnames =
  let ht = Namespace_util.RQNameHashtbl.create 17 in
  List.iter (check_dup_params ht) cvnames

let normalize_function_body norm_context f = 
  let nsenv = nsenv_from_norm_context norm_context in
  let (fname, vnames, sign, body) = f.pfunction_def_desc in
  let rfname = resolve_function_qname nsenv fname in
  let arity = List.length (fst sign) in
  let (input_types, output_type) =
    one_sig_from_norm_context norm_context (rfname, arity)
  in
  let cvnames = List.map (fun vname -> resolve_variable_qname nsenv vname) vnames in
  check_distinct_parameters cvnames;
  let cbody = normalize_body norm_context body output_type in
  fmkcfunction_def ((rfname,arity), cvnames, (input_types, output_type), cbody) f.pfunction_loc 


(**************)
(* Statements *)
(**************)

(* Top-level snap normalization: 
   Ensure every query is wrapped by a top-level snap 
*)
let normalize_toplevel_snap ce = 
  match ce.pcexpr_desc with
    | CESnap _ -> ce
	(* Default to unordered, unless there is a snap *)
    | _ -> fmkcexpr (CESnap (Snap_Unordered_Deterministic, ce)) None ce.pcexpr_loc
	
let normalize_statement norm_context statement =
  normalize_toplevel_snap 
    (normalize_expr norm_context statement)



let normalize_statement_list norm_context ss = 
  List.map (normalize_statement norm_context) ss


(**************************)
(* Performs module import *)
(**************************)

let rec apply_module_imports_aux proc_ctxt (p,decls) c =
  match c.pcontext_decl_desc with 
    (* Don't do anything for most declarations *)
  | EBaseURIDecl _
  | ENamespaceDecl _
  | EDefaultElementNamespaceDecl _
  | EDefaultFunctionNamespaceDecl _
  | EXmlSpaceDecl _
  | EDefaultCollationDecl _
  | EConstructionDecl _
  | EOrderingDecl _
  | EDefaultEmptyOrderDecl _
  | ECopyNamespacesDecl _
  | ESchemaDecl _ ->
      (p, c :: decls)
  | EImportServiceDecl (ncn, wsdl_uri, loc_hint_option) ->
      let wsdl_loc = Schema_util.resolve_location_hint proc_ctxt loc_hint_option wsdl_uri in 
      (* Note: The namespace URI of the service MUST be equal to the
         targetNamespace in the WSDL definition file.  This should be
         checked in Wsdl_load.xml_to_wsdl_ast. - Jerome & Mary *)

      (* 1. Create a namespace declaration for the module target namespace *)
      let new_ns_decl =
	fmkcontext_decl (ENamespaceDecl (ncn,(NSUri wsdl_uri))) c.pcontext_decl_loc
      in
      (* 2. Open & parse the file at the location hint. *)
      let (_, wsdl_stream) =
	Streaming_parse.open_xml_stream_from_io (Galax_io.Http_Input wsdl_loc)
      in
      (* 3. Load the WSDL AST from the WSDL file. *)
      let wsdl_ast = Wsdl_load.xml_to_wsdl_ast proc_ctxt wsdl_uri wsdl_stream  in
      (* 4. Create the WSDL import stub as an XQuery module. *)
      let wsdl_mod = Wsdl_import.wsdl_ast_to_xquery_ast_import ncn wsdl_ast None None in
      (* 5. Merges the imported module with the calling module. *)
      (* Note:
	   This also checks that the target uri is the one declared in
	   the module.
	 - Jerome
       *)
      (merge_library_module_in_prolog (NSUri wsdl_uri) p wsdl_mod, new_ns_decl :: decls)

  | EImportModuleDecl (ncn, mod_uri, loc_hint_option) ->
      (* 1. Create a namespace declaration for the module target namespace *)
      let mod_loc = Schema_util.resolve_location_hint proc_ctxt loc_hint_option mod_uri in 
      let new_ns_decl =
	fmkcontext_decl (ENamespaceDecl (ncn,(NSUri mod_uri))) c.pcontext_decl_loc
      in
      (* 2. Parse the library module *)
      let imported_mod =
	let parse_ctxt = Parse_context.build_xquery_parse_context proc_ctxt in
	let (_,m) =
	  Parse_top.parse_library_module_from_io parse_ctxt
	    (Galax_io.Http_Input mod_loc)
	in
	m
      in
      (* 3. Merges the imported module with the calling module. *)
      (* Note:
	   This also checks that the target uri is the one declared in
	   the module.
	 - Jerome
       *)
      (merge_library_module_in_prolog (NSUri mod_uri) p imported_mod, new_ns_decl :: decls)

let apply_module_imports proc_ctxt p =
  let cdl = p.pprolog_contexts in
  let (imported_mod, remaining_cdl) =
    List.fold_left (apply_module_imports_aux proc_ctxt) (empty_prolog,[]) cdl
  in
  merge_prologs_with_decls remaining_cdl imported_mod p

(*******************************)
(* Mapping context of a module *)
(*******************************)

(* Note:
     normalize_context takes a context and an expression map the expression
     and return the updated context
   - Jerome *)

(* HACK! -- See comment to Vladimir *)
let rec filter1_ns_decls ns_decls =
  match ns_decls with
  | [] -> []
  | (pref,uri) :: ns_decls -> (NSPrefix pref,uri) :: (filter1_ns_decls ns_decls)

let rec filter2_ns_decls ns_decls =
  match ns_decls with
  | [] -> []
  | (NSPrefix pref,uri) :: ns_decls -> (pref,uri) :: (filter2_ns_decls ns_decls)
  | _ :: ns_decls -> (filter2_ns_decls ns_decls)

let normalize_context_decl proc_ctxt (ns_decls, schemas) c =
  match c.pcontext_decl_desc with
  | EBaseURIDecl base_uri_string ->
      let base_uri = AnyURI._uri_of_string base_uri_string in
      begin
	Processing_context.set_base_uri proc_ctxt (Some base_uri);
	(ns_decls, schemas)
      end
  | ENamespaceDecl (ncn, uri) ->
      if (uri = (NSUri ("http://www.w3.org/XML/1998/namespace"))) || (ncn = "xml") || (ncn = "xmlns")
      then raise (Query (Namespace_Error "[err:XQST0070] Namespace declaration redefines the xml or xmlns namespace"));
      let new_binding = (NSPrefix ncn,uri) in
      (ns_decls@[new_binding], schemas)
  | EDefaultElementNamespaceDecl uri ->
      let new_binding = (NSDefaultElementPrefix,uri) in
      (ns_decls@[new_binding], schemas)
  | EDefaultFunctionNamespaceDecl uri ->
      let new_binding = (NSDefaultFunctionPrefix,uri) in
      (ns_decls@[new_binding], schemas)
  | ESchemaDecl (prefix_option, target_uri, loc_hint_option) ->
      let actual_uri = Schema_util.resolve_location_hint proc_ctxt loc_hint_option target_uri in
      let (_, xml_schema_stream) =
	Streaming_parse.open_xml_stream_from_io (Galax_io.Http_Input actual_uri)
      in
      let resolved_xml_stream =
	Streaming_ops.resolve_xml_stream xml_schema_stream
      in
      let xschema = Schema_import.import_schema_document proc_ctxt prefix_option resolved_xml_stream in
      let new_schemas = schemas@[xschema] in
      begin
	match prefix_option with
	| None ->
	    (ns_decls, new_schemas)
	| Some (NSPrefix ncn) -> 
	    let new_binding = ((NSPrefix ncn),NSUri target_uri) in
	    (ns_decls@[new_binding], new_schemas)
	| Some NSDefaultElementPrefix ->
	    let new_binding = (NSDefaultElementPrefix,NSUri target_uri) in
	    (ns_decls@[new_binding], new_schemas)
	| Some _ ->
	    raise (Query (Namespace_Error "Bogus namespace prefix in schema import AST"))
      end
  | EImportServiceDecl (ncn, wsdl_uri, wsdl_file) ->
      raise (Query (Prototype "Nested service imports are not supported yet"))
  | EImportModuleDecl (ncn, mod_uri, mod_file) ->
      raise (Query (Prototype "Nested module imports are not supported yet"))

    (* New declarations are ignored for now - Jerome *)
  | EXmlSpaceDecl strip_or_preserve ->
      Processing_context.set_boundary_space_kind proc_ctxt strip_or_preserve;
      (ns_decls, schemas)
  | EDefaultCollationDecl uri ->
      Processing_context.set_default_collation proc_ctxt uri;
      (ns_decls, schemas)
  | EConstructionDecl strip_or_preserve ->
      Processing_context.set_construction_kind proc_ctxt strip_or_preserve;
      (ns_decls, schemas)
  | EOrderingDecl ordered_or_unordered ->
      Processing_context.set_ordering_kind proc_ctxt ordered_or_unordered;
      (ns_decls, schemas)
  | EDefaultEmptyOrderDecl emptysortkind ->
      Processing_context.set_default_order_kind proc_ctxt emptysortkind;
      (ns_decls, schemas)
  | ECopyNamespacesDecl (preserve_or_no_preserve,inherit_or_no_inherit) ->
      Processing_context.set_ns_preserve_kind proc_ctxt preserve_or_no_preserve;
      Processing_context.set_ns_inherit_kind proc_ctxt inherit_or_no_inherit;
      (ns_decls, schemas)

(* Checking constraints on the prolog setters *)

let multiple_boundary_space cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with EXmlSpaceDecl _ -> true | _ -> false) cdl) > 1

let multiple_base_uri cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with EBaseURIDecl _ -> true | _ -> false) cdl) > 1

let multiple_default_element_namespace cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with EDefaultElementNamespaceDecl _ -> true | _ -> false) cdl) > 1

let multiple_default_function_namespace cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with EDefaultFunctionNamespaceDecl _ -> true | _ -> false) cdl) > 1

let multiple_construction_decl cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with EConstructionDecl _ -> true | _ -> false) cdl) > 1

let multiple_ordering_decl cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with EOrderingDecl _ -> true | _ -> false) cdl) > 1

let multiple_empty_order_decl cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with EDefaultEmptyOrderDecl _ -> true | _ -> false) cdl) > 1

let multiple_default_collation_decl cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with EDefaultCollationDecl _ -> true | _ -> false) cdl) > 1

let multiple_copy_namespaces_decl cdl =
  List.length (List.find_all (fun x -> match x.pcontext_decl_desc with ECopyNamespacesDecl _ -> true | _ -> false) cdl) > 1

let check_prolog_consistent cdl =
  begin
(*    if incorrect_schema_imports cdl
    then
    else
      ; *)
    if multiple_boundary_space cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0068] Multiple boundary-space declarations"));
    if multiple_base_uri cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0032] Multiple base-uri declarations"));
    if multiple_default_element_namespace cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0066] Multiple default element namespace declarations"));
    if multiple_default_function_namespace cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0066] Multiple default function namespace declarations"));
    if multiple_construction_decl cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0067] Multiple construction declarations"));
    if multiple_ordering_decl cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0065] Multiple ordering declarations"));
    if multiple_empty_order_decl cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0069] Multiple empty order declarations"));
    if multiple_default_collation_decl cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0068] Multiple default collation declarations"));
    if multiple_copy_namespaces_decl cdl
    then
      raise (Query (Malformed_Core_Expr "[err:XQST0055] Multiple copy namespaces declarations"));
  end

let normalize_context_with_env proc_ctxt cdl =
  check_prolog_consistent cdl;
  List.fold_left (normalize_context_decl proc_ctxt) ([], []) cdl


(****************************************************************)
(* Toplevel functions to map the Full XQuery to the core XQuery *)
(****************************************************************)

(* Prolog *)

let normalize_prolog norm_context prolog1 =
  let nsenv      	= nsenv_from_norm_context norm_context in
  let proc_ctxt  	= processing_context_from_norm_context norm_context in
  begin
    (* 1. Apply module imports to that prolog *)
    let prolog2 = apply_module_imports proc_ctxt prolog1 in
    (* 2. Normalize the context *)
    let { pprolog_xschemas = inlined_xschemas;
	  pprolog_contexts = ns;
	  pprolog_functions = fs;
	  pprolog_vars = vs;
	  pprolog_indices = ks } = prolog2
    in
    let (ns_decls, imported_xschemas) = normalize_context_with_env proc_ctxt ns in
    (* Any redeclarations of namespaces that are inconsistent should raise an error *)
    let nsenv2 = Namespace_context.add_all_ns nsenv ns_decls in

    (* 3. Processing ISSDS *)
    (* Note:
         The namespace environment is not quite correct here!
         We will need to revisit this.
       - Jerome *)
    let imported_cxschemas = List.map (Schema_norm.normalize Namespace_context.default_xml_nsenv) imported_xschemas in
    let inlined_cxschemas = List.map (Schema_norm.normalize nsenv2) inlined_xschemas in
    let all_cxschemas = imported_cxschemas @ inlined_cxschemas in
    let complete_cxschema = Xquery_type_core_ast_util.merge_cxschemas all_cxschemas in
if !Conf.debugging then Print_type_core.print_cxschema Format.std_formatter complete_cxschema;
    (* 4. Builds the new normalization context *)
    let norm_context' = copy_norm_context_with_sigs norm_context nsenv2 complete_cxschema in
    let sigs' = normalize_function_signatures norm_context' fs in
    let norm_context'' = add_sigs_to_norm_context norm_context' sigs' in

    (* 5. map all function bodies: *)
    let functions = List.map (normalize_function_body norm_context'') fs in
    let vars = normalize_variables norm_context'' vs in
    let indices = List.map (normalize_index norm_context'') ks in
    let cprolog =
      { pcprolog_functions = functions;
	pcprolog_vars = vars;
	pcprolog_indices = indices }
    in
    (norm_context'', cprolog)
  end

(* Library module *)

let normalize_library_module norm_context q =
  (* 1. Get the original query prolog *)
  let prolog1 = q.plibrary_module_prolog in
  (* 2. Get the module target namespace *)
  let (pref,uri) = q.plibrary_module_decl in
  (* 3. Extend the prolog with the corresponding namespace declaration *)
  let new_decl = fmkcontext_decl (ENamespaceDecl (pref,NSUri uri)) Finfo.bogus in
  let patched_prolog =
    { pprolog_xschemas 	= prolog1.pprolog_xschemas;
      pprolog_contexts 	= new_decl :: prolog1.pprolog_contexts;
      pprolog_functions = prolog1.pprolog_functions;
      pprolog_vars  	= prolog1.pprolog_vars;
      pprolog_indices  	= prolog1.pprolog_indices }
  in
  (* 4. Normalize the prolog *)
  let (norm_context',cprolog) = normalize_prolog norm_context patched_prolog in
  (* 5. Build the resulting core module *)
  let cmodule =
    { pcmodule_prolog = cprolog;
      pcmodule_statements = [] }
  in
  (norm_context', cmodule)

(* Main module *)

let normalize_main_module norm_context q = 
  (* 1. Get the original query prolog *)
  let prolog1 = q.pmain_module_prolog in
  (* 2. Normalize the prolog *)
  let (norm_context',cprolog) = normalize_prolog norm_context prolog1 in
(* Print_top.printf_cxschema "Input Schema :" (cxschema_from_norm_context norm_context') ;  *)
  (* 3. Normalize the statements *)
  let es = q.pmain_module_statements in
  let ces = normalize_statement_list norm_context' es in
  (* 4. Build the main module *)  
  let cmodule =
    { pcmodule_prolog = cprolog;
      pcmodule_statements = ces }
  in
  (norm_context', cmodule)

(* Module *)

let normalize_module norm_context qm =
  match qm with
  | ELibraryModule q ->
      normalize_library_module norm_context q
  | EMainModule q ->
      normalize_main_module norm_context q

(* No normalization -- just return dummy core expressions *)
let nonorm_statement norm_context statement =
  fmkcexpr CEEmpty None Finfo.bogus

let nonorm_prolog norm_context prolog =
  let cprolog =
    { pcprolog_functions = [];
      pcprolog_vars = [];
      pcprolog_indices = [] }
  in
  (norm_context, cprolog)

let nonorm_library_module norm_context q =
  (* 1. Get the original query prolog *)
  let prolog = q.plibrary_module_prolog in
  let (_, cprolog) = nonorm_prolog norm_context prolog in
  let cmodule =
    { pcmodule_prolog = cprolog;
      pcmodule_statements = [] }
  in
  (norm_context, cmodule)

let nonorm_main_module norm_context q = 
  (* 1. Get the original query prolog *)
  let prolog = q.pmain_module_prolog in
  let (_, cprolog) = nonorm_prolog norm_context prolog in
  let cmodule =
    { pcmodule_prolog = cprolog; 
      pcmodule_statements = [] }
  in
  (norm_context, cmodule)

