(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)


(* $Id: optimization_judge.ml,v 1.16 2006/05/11 10:52:49 pmichiel Exp $ *)

(* Module: Optimization judgements. Currently these all have to deal
   with side-effects. *)

open Xquery_algebra_ast
open Xquery_algebra_ast_util
open Optimization_util
open Namespace_builtin

type ('a,'b) optimization_judge_env =  ('a,'b) Compile_context.compile_context

(* Atomic test *)
let rec op_contains_side_effects ((visited_hash, comp_ctxt) as env) op = 
  match op.palgop_expr_name with      
    | AOEDelete
    | AOEDetach
    | AOEInsert _ 
    | AOERename
    | AOEReplace _ -> true
	(* We know that overloaded functions are side-effect free,
	   also because of phase ordering (they are listed in
	   normalization not factorization) we do not include them in
	   the table. This is hackish but, works for now - Chris *)
    | AOECallOverloaded (((prefix,uri,local), _), _) -> false
    | AOECallBuiltIn    ((((prefix,uri,local) as cfname),_),_,_) ->
	not (Factorize_side_effect.global_is_side_effect_free_function cfname)
    | AOECallUserDefined  (fname, _,_) -> 
	if Hashtbl.mem visited_hash fname
	then false
	else begin
	  Hashtbl.add visited_hash fname ();
	  let (params,expr_ref,opt) = Compile_context.get_function_from_compile_context comp_ctxt fname in 
	    has_side_effect_internal env !expr_ref
	end
    | _ -> false

(* This judgment determines if the result has side-effects, we need to
   process the user-defined functions to mark them as well. This has a
   placeholder for the envirnonment for now. *)
and has_side_effect_internal ((visited_hash, comp_ctxt) as env) algop = 
  let has_side_effect_subexpr env sexpr = 
    match sexpr with 
      | NoSub -> false
      | OneSub op -> has_side_effect_internal env op
      | TwoSub (op1,op2) ->
	  (has_side_effect_internal env op1) ||
	  (has_side_effect_internal env op2) 
      | ManySub ops ->
	  Array.fold_left 
	    (fun b op -> b || has_side_effect_internal env op) false ops
  in
  let is_update = op_contains_side_effects env algop in
    is_update || 
    (has_side_effect_subexpr env algop.psub_expression) ||
    (has_side_effect_subexpr env algop.pdep_sub_expression) 

let has_side_effect comp_ctxt algop =
  let visited_hash = Hashtbl.create 1 in 
    has_side_effect_internal (visited_hash, comp_ctxt) algop

(* Just walk for a snap, then see if there are side-effects inside *)
(* NOTE: This is not *quite* right because we regard trace as a
   side-effect function..  Perhaps we should pass in the env of
   functions with side-effects.
*)
let rec has_non_trivial_snap comp_ctxt algop = 
  let has_non_trivial_snap_subexpr comp_ctxt sexpr = 
    match sexpr with 
      | NoSub -> false
      | OneSub op -> has_non_trivial_snap comp_ctxt op
      | TwoSub (op1,op2) ->
	  (has_non_trivial_snap comp_ctxt op1) ||
	  (has_non_trivial_snap comp_ctxt op2) 
      | ManySub ops ->
	  Array.fold_left 
	    (fun b op -> b || has_non_trivial_snap comp_ctxt op) false ops
  in
    match algop.palgop_expr_name with
      | AOESnap sm -> 
	  has_side_effect comp_ctxt algop
      | _ -> 
	  (has_non_trivial_snap_subexpr comp_ctxt algop.psub_expression) ||
	  (has_non_trivial_snap_subexpr comp_ctxt algop.pdep_sub_expression) 
 
(*** HACK CONDITION
 * Several of the optimization rules use the singleton 
 * judgement. Unitl it is actually available, we will use
 * the name of the field to determine the cardinality
 **)

let is_dot_field f =
  let (prefix, uri, name) = f in
  let len = String.index name '_' in
  let substr = String.sub name 0 len in
    (uri = Namespace_builtin.glx_uri) && (substr = "dot")

(*** HACK CONDITION
 * Is this a glx:sequence field?
 *)
let is_seq_field f =
  let (prefix, uri, name) = f in
  let len = String.index name '_' in
  let substr = String.sub name 0 len in
    (uri = Namespace_builtin.glx_uri) && (substr = "sequence")



(***********************************)
(* is the algebraic op a singleton *)
(***********************************)
let rec is_singleton_tuple comp_ctxt algop = 
  match algop.palgop_expr_name with
      (* Input tuple *)
    | AOEInputTuple -> true

    (**************************)
    (* Basic Tuple operations *)
    (**************************)
    | AOECreateTuple _ -> true
    | AOEAccessTuple _ -> false (* Returns items *)
    | AOEConcatTuples  -> true                    

    (**************)
    (* Tuple Maps *)
    (**************)
    (* | AOEMapFromItem   (* Need a singleton judgment about the item. which we don't have *) *)
    | AOEMap                                (* Tuple iteration *)
    | AOEMapConcat                          (* Map with built-in tuple concatenation *)
    | AOEOuterMapConcat _           (* Same with null value on the given name *)
      -> (* For these maps need indep and dep are singleton *)
	let indep = access_onesub algop.psub_expression in 
	let dep   = access_onesub algop.pdep_sub_expression in 
	  (is_singleton_tuple comp_ctxt indep) && 
	  (is_singleton_tuple comp_ctxt dep)

    | AOESelect _ -> 
	let indep = access_onesub algop.psub_expression in 
	  is_singleton_tuple comp_ctxt indep
	
    (* Singleton Maps *)
    | AOENullMap _ 
    | AOEMapIndex _ 
    | AOEMapIndexStep _ -> 
	let indep = access_onesub algop.psub_expression in 
	  is_singleton_tuple comp_ctxt indep

    (* Map Concats *)

    (******************************)
    (* Products, Joins Selections *)
    (******************************)
    | AOEProduct           
    | AOEJoin   _          
    | AOELeftOuterJoin _ ->
	let i1, i2 = access_twosub algop.psub_expression in 
	(is_singleton_tuple comp_ctxt i1) && (is_singleton_tuple comp_ctxt i2)

    (************************************)
    (* Grouping and Ordering operations *)
    (************************************)
    | AOEGroupBy _ -> (* If indep is singleton *)
	let indep = access_onesub algop.psub_expression in 
	  is_singleton_tuple comp_ctxt indep

    | AOEOrderBy _  -> (* if indep is *)
    	let indep = access_onesub algop.psub_expression in 
	  is_singleton_tuple comp_ctxt indep
    | AOEProject p ->
	let indep = access_onesub algop.psub_expression in 
	  is_singleton_tuple comp_ctxt indep

    | _ -> false

(******************************************)
(* Function may not actually be recursive *)
(******************************************)
let is_recursive_function ctxt check_name =
  let visited_functions = Namespace_util.RQNameIntHashtbl.create 1 in 
  let rec check_for_call op = 
    match op.palgop_expr_name with
      | AOECallUserDefined (fname,_,_) ->	  
	  (Namespace_names.rqname_int_equal check_name fname) ||
	  (check_function_body fname)    ||
	  (check_subexpr op.psub_expression) ||
	  (check_subexpr op.pdep_sub_expression)

      | _ -> 
	  (check_subexpr op.psub_expression) ||
	  (check_subexpr op.pdep_sub_expression)

  and check_subexpr sexpr = 
    match sexpr with
      | NoSub     -> false
      | OneSub op -> check_for_call op
      | TwoSub (op1,op2) -> 
	  (check_for_call op1) ||
	  (check_for_call op2) 
      | ManySub a ->
	  Array.fold_left 
	    (fun b op -> b || (check_for_call op))
	    false a

  and check_function_body fname = 
    if Namespace_util.RQNameIntHashtbl.mem visited_functions fname
    then false
    else begin
      Namespace_util.RQNameIntHashtbl.add visited_functions fname ();
      let (params, expr_ref, opt) = Compile_context.get_function_from_compile_context ctxt fname in 
	check_for_call !expr_ref
    end
  in
    check_function_body check_name

(* Checks if a operator tree maps one field to another *)
(* 
  if true -> Some(source_field, dest_field)
  else    -> None
 *)
let is_sequence_to_field_mapping op =
  let fail = None in
  match op.palgop_expr_name with
  | AOEMapFromItem crname1 ->
      begin
	let map_dep = access_onesub op.pdep_sub_expression in
	let map_indep = access_onesub op.psub_expression in
	match map_dep.palgop_expr_name with
	| AOECreateTuple a 
	  when Array.length a = 1 ->
	    begin
	      let (asequencetype,output) = a.(0) in
	      let sub = (access_manysub map_dep.psub_expression).(0) in
	      match sub.palgop_expr_name with
	      | AOEVar crname2 
		when crname1 = crname2 ->
		  begin
		    match map_indep.palgop_expr_name with
		    | AOEAccessTuple seq ->
			Some (seq, output)
		    | _ -> fail
		  end
	      | _ -> fail
	    end
	|  _ -> fail
      end
  | _ -> fail


let is_map_to_item_step op =
  match op.palgop_expr_name with
  | AOEMapToItem ->
      begin
	let map_dep = access_onesub op.pdep_sub_expression in
	let map_indep = access_onesub op.psub_expression in
	match map_dep.palgop_expr_name with
	| AOEAccessTuple step_output ->
	    begin
	      match map_indep.palgop_expr_name with
	      | AOETupleTreePattern (out, p) ->
		  if is_input_tuple (access_onesub map_indep.psub_expression)
		  then  Some (step_output, map_indep)
		  else  None
	      | _ -> None
	    end
	| _ -> None
      end
  | _ -> None

(* Checks if an operator tree maps a TupleTreePattern to a 
   sequence in a tuple field *)
(* if true -> Some (seq_field, TTP_operator)
   else    -> None
 *)
let is_step_to_sequence_mapping op =
  let fail = None in
  match op.palgop_expr_name with
  | AOEMapConcat ->
      begin
	let map_dep = access_onesub op.pdep_sub_expression in
	let map_indep = access_onesub op.psub_expression in
	match map_dep.palgop_expr_name with
	| AOECreateTuple a 
	  when Array.length a = 1 ->
	    begin
	      let (asequencetype,seq) = a.(0) in
	      let sub = (access_manysub map_dep.psub_expression).(0) in
	      match is_map_to_item_step sub with
	      | Some (step_output, tp_op) -> 
		  Some (seq, step_output, tp_op, map_indep)
	      | None -> None
	    end
	| _ -> fail
      end
  | _ -> fail


(* DOES ORDER MATTER ? *)

type distinct_order_matters =
  | Manipulating_Ord
  | Mainpulating_Distinct
  | Manipulating_Ord_Distinct
  | Observing
  | Neutral

(* FUNCTIONS *)
(* o  All user defined functions are considered to be Observing              *)
(* o All functions not mentioned below that are in FN FS or GLX namespaces   *)
(*   are considered to be neutral                                            *)
(* o Below should be all the functions that take node sequences              *)
(*   i.e., node()+ or node()*                                                *)
(*   as an argument, because those are only ones that can observe/manipulate *)
(*   order and duplicates.                                                   *)
let distinct_order_handle_fun fname =
  if 
    fname = fn_reverse || fname = fn_unordered || fname = fs_docorder
  then Manipulating_Ord

  else if 
    fname = fs_distinct
  then Mainpulating_Distinct

  else if 
     fname = fn_boolean 
  || fname = fn_empty 
  || fname = fn_distinct_values
  || fname = fn_zero_or_one
  || fname = fn_exactly_one
  || fname = fn_one_or_more
  || fname = fn_deep_equal
  || fname = fs_distinct_docorder_or_atomic_sequence
  || fname = fs_distinct_docorder
  || fname = glx_deep_distinct
  then Manipulating_Ord_Distinct

  else if 
     fname = fn_data
  || fname = fn_index_of
  || fname = fn_subsequence
  || fname = fn_count
  || fname = fn_position
  || fname = fn_last
  || fname = fs_first
  || fname = fs_last_fn
  || fname = glx_get_order
  || fname = fs_convert_simple_operand
(* to cover all of min/max/sum/avg, we consider the function 
   fs:convert-simple-operand() and fn:data() above to be 
   observing order/dupl. -- Philippe
 *)
  then Observing

  else if
    fname = fn_trace
  || fname = fn_insert_before
  || fname = fn_remove
  || fname = op_union
  || fname = op_intersect
  || fname = op_except
  || fname = glx_get_docid
  then Neutral

  (* NOTE: we could play safe here and consider everything else to observing -- Philippe *)
  else Neutral 



(* OPERATORS *)
let distinct_order_handle op =

  match op with
  (* Depends on function *) 
  | AOECallBuiltIn ((name,arity), optintypes, opttypes) ->
      distinct_order_handle_fun name
  (* Manipulating order *)
  | AOEGroupBy _ | AOEOrderBy _ 
    -> Manipulating_Ord

  (* Manipulating duplicates *)
  | AOEDistinct _ |AOEPrune _ 
    -> Mainpulating_Distinct

  (* Manipulating order and duplicates *)
  | AOETreeJoin _ | AOETupleTreePattern _ 
    -> Manipulating_Ord_Distinct

  (* observing order/duplicates *)
  | AOETypeswitch _ | AOEDocument | AOEPI _ | AOEPIComputed 
  | AOEComment _ | AOECommentComputed | AOEText _ | AOETextComputed
  | AOEElem _ | AOEAnyElem _ | AOEAttr _ | AOEAnyAttr _ | AOEValidate _
  | AOECast _ |  AOECastable _ | AOEConvertSimple _ | AOEPromoteNumeric _
  | AOEUnsafePromoteNumeric _ | AOEMapIndex _ | AOEMapIndexStep _ 
  | AOECopy | AOEDelete | AOEDetach | AOEInsert _ | AOERename
  | AOEReplace _ | AOESnap _ 
  | AOESome _  (* Type subexpr - O is safe *)
  | AOEEvery _ (* Type subexpr - O is safe *)
  | AOECallUserDefined _ 
  | AOECallOverloaded _ 
    -> Observing

  (* neutral *)
  | AOEIf | AOELet _ | AOEVar _ | AOEScalar _ | AOESeq | AOEEmpty 
  | AOEError | AOETreat _ | AOEInputTuple |  AOEKeyRef _ | AOEExecute 
  | AOECreateTuple _ | AOEAccessTuple _ | AOEConcatTuples
  | AOEMapFromItem _ | AOEMapToItem  | AOEMap 
  | AOENullMap _ | AOEMapConcat | AOEOuterMapConcat _ 
  | AOEMaterializeTable | AOEProduct | AOESelect _ 
  | AOEJoin _ | AOELeftOuterJoin _  |  AOEParse _  | AOEProject _
    -> Neutral

let rec get_anc_op_list op_node op =
  if op_node == op then
    [], true
  else
    let dep_subexprs = subexpr_to_list op_node.pdep_sub_expression in
    let indep_subexprs = subexpr_to_list op_node.psub_expression in
    let subexprs = Array.of_list (dep_subexprs @ indep_subexprs) in
    let list = ref [] in 
    let found = ref false in
    let pos = ref 0 in
    while not(!found) && !pos < (Array.length subexprs) do
      let op_node' = subexprs.(!pos) in
      let (list', found') = get_anc_op_list op_node' op in
      list := list'; found := found'; incr(pos)
    done;
    if !found then
      op_node::!list , true
    else
      [], false

(* All of this can be done with annotations on the syntax tree
 * but, I'll wait until the dust settles on the reorganising
 * everything before switching to that approach -- Philippe
 *)
let ord_dup_matters root op =
  let not_is_order_dupfree_op (o,d) op =
    match distinct_order_handle op.palgop_expr_name with
    | Manipulating_Ord -> (false, d)
    | Mainpulating_Distinct -> (d, false)
    | Manipulating_Ord_Distinct -> (false, false)
    | Observing -> (true, true)
    | Neutral -> (o,d)
  in
  let ancs, _ = get_anc_op_list root op in
  let o,d =  List.fold_left not_is_order_dupfree_op (true, true) ancs in
  o || d


let is_update_op op =
  match op.palgop_expr_name with
  (*| AOEGroupBy _ | AOEOrderBy _ *) (* why?? -- Philippe *)
  | AOECopy  | AOEDelete 
  | AOEDetach | AOEInsert _
  | AOERename | AOEReplace _ -> true
  | _ -> false

let rec contains_update op =
  if is_update_op op then true
  else
    let fold_fun b x = b || contains_update x in
    let dep =
      match op.pdep_sub_expression with
      | NoSub -> false
      | OneSub op' -> contains_update op'
      | TwoSub (op1,op2) -> (contains_update op1 || contains_update op2)
      | ManySub ops -> Array.fold_left fold_fun false ops
    in
    let indep =
      match op.psub_expression with
      | NoSub -> false
      | OneSub op' ->  contains_update op'
      | TwoSub (op1,op2) -> (contains_update op1 || contains_update op2)
      | ManySub ops -> Array.fold_left fold_fun false ops
    in
    indep || dep
