(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)


(* Module: Optimization Rewrite SBDO
   Description:
     This module implements the SBDO optimization, described in

       Optimizing Sorting and Duplicate Elimination 
       in XQuery Path Expressions
       
       Mary Fernndez, Jan Hidders, Philippe Michiels, 
       Jrme Simon, Roel Vercammen
       
       DEXA 2005
       http://www.adrem.ua.ac.be/bibrem/pubs/fernandez.1.pdf
*)

open Error
open Namespace_builtin

open Xquery_algebra_ast
open Xquery_algebra_ast_util
open Xquery_algebra_ast_annotation_util

open Compile_annotate
open Compile_context

open Processing_context

open Optimization_util
open Optimization_sbdo_automaton

(*
 *  WARNING: Before being able to apply the SBDO optimization
 *           to a TTP, we must be sure that the input op the
 *           operator itself is always sorted by document 
 *           order and free from duplicates.
 *           -- Ph
 *)

let set_sbdo pattern index action_needed = 
  match action_needed with 
  | Nothing          -> 
      let _ = if !Conf.debugging then Printf.printf "Action is NOTHING\n" in
      pattern.(index).requires_sbdo <- (false,false)
  | Docorder         -> 
      let _ = if !Conf.debugging then Printf.printf "Action is ORD\n" in
      pattern.(index).requires_sbdo <- (true ,false)
  | Distinct         -> 
      let _ = if !Conf.debugging then Printf.printf "Action is NODUP\n" in
      pattern.(index).requires_sbdo <- (false,true )
  | DistinctDocorder -> 
      let _ = if !Conf.debugging then Printf.printf "Action is ORD/NODUP\n" in
      pattern.(index).requires_sbdo <- (true ,true )

let is_last pattern index =
  match pattern.(index).child_twig with
  | Some _ -> false
  | None -> true

(*
 * Processes an entire tree pattern using a given automaton (sbdo_kind)
 *)
let rec process_tree_pattern sbdo_kind pattern index state =
  let twig_node = pattern.(index) in
  let _ =
    (* (i) recursively process child steps *)
    match twig_node.child_twig with
    | Some (axis, i) ->
	begin
	  let state', act = 
	    do_transition sbdo_kind state (axis_adapter axis) (is_last pattern i)
	  in
	  let _ = set_sbdo pattern i act in
	  let _ = if !Conf.debugging then Printf.printf "State = %s\n%!" (print_state state') in
	  process_tree_pattern sbdo_kind pattern i state'
	end
    | None -> ()
  in
  let start_automaton_for_pred (axis, i) =
    let state  = init_state sbdo_kind in
    let state', act = do_transition sbdo_kind state (axis_adapter axis) (is_last pattern i) in
    let _ = set_sbdo pattern i act in
    let _ = if !Conf.debugging then Printf.printf "State = %s\n%!" (print_state state') in
    process_tree_pattern sbdo_kind pattern i state'
  in
  let _ =
    (* (ii) start new automaton for predicates *)
    List.map start_automaton_for_pred twig_node.pred_twigs
  in
  ()

let unsup = Query(Internal_Error("This SBDO mode is not supported any more."))

(*
 * Top level rewrite rule for the SBDO optim 
 *)
let sbdo_rewrite ctxt op =
  let proc_ctxt = processing_context_from_compile_context ctxt in
  match op.palgop_expr_name with
  | AOETupleTreePattern (i, p) ->
      begin
	match proc_ctxt.sbdo_kind with
	| SBDO_Remove   ->
	    let _ = Array.map (fun x -> x.requires_sbdo <- (false, false)) p in
	    op, false
	| SBDO_Preserve -> op, false
	| SBDO_Tidy | SBDO_DupTidy | SBDO_Sloppy -> 
	    let _ = 
	      process_tree_pattern proc_ctxt.sbdo_kind p 0 (init_state proc_ctxt.sbdo_kind)
	    in op, false
	| SBDO_AdHoc -> raise unsup
      end
  | _ -> op, false


