(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

open Xquery_algebra_ast
open Xquery_algebra_ast_util
open Xquery_algebra_ast_annotation_util

open Logical_algebra_types
type ('a,'b) logical_op = ('a, 'b) Xquery_algebra_ast.aalgop_expr 
open Error

type sub_expr_kind = 
  | Dependent 
  | Independent

type ('a, 'b) child_desc = ('a, 'b) aalgop_expr * sub_expr_kind * int


let get_cardinality se =
  match se with
    | NoSub     -> 0
    | OneSub _  -> 1
    | TwoSub _  -> 2
    | ManySub a -> Array.length a

let check_cardinality se index =
  if ((get_cardinality se) <= index) then
    begin
      let card = get_cardinality se in
	raise (Query (Malformed_Algebra_Expr("Attempt to access out of bounds index in subexpr: " ^ (string_of_int index) ^ " : " ^ (string_of_int card))))
    end

let get_sub_expr op dep_indep =
  match dep_indep with
    | Dependent   -> op.pdep_sub_expression 
    | Independent -> op.psub_expression
	

let assign_sub_expr op dep_indep se = 
  match dep_indep with
    | Dependent   -> op.pdep_sub_expression <- se
    | Independent -> op.psub_expression <-se

let replace_subexpr se index new_op =
  check_cardinality se index;
  match se with
    | NoSub -> (* Check cardinality should have caught this *)
	raise (Query (Malformed_Algebra_Expr("check_cardinality should have caught this. Attempt to assign subexpr to a NoSub")))
    | OneSub s -> OneSub new_op
    | TwoSub (s0, s1) ->
	if (index = 0) then
	  TwoSub (new_op, s1)
	else
	  TwoSub (s0, new_op)
    | ManySub a -> 
	a.(index) <- new_op;
	ManySub a

let get_op_from_sub_expr op dep_indep index = 
  let se = get_sub_expr op dep_indep in
    check_cardinality se index;
    match se with
      | NoSub ->
   	  raise (Query (Malformed_Algebra_Expr("check_cardinality should have caught this."
					^ " Attempt to retrieve subexpr to a NoSub")))
      | OneSub s -> s
      | TwoSub (s0,s1) ->
	  if (index = 0) then
	    s0
	  else
	    s1
      | ManySub a -> a.(index)

let replace_op parent dep_indep index new_op =
  let se = get_sub_expr parent dep_indep in
    assign_sub_expr parent dep_indep (replace_subexpr se index new_op)

(* Exposed: Update the parent of this op 
   by using the given child description *)

let update_parent cd op =
  match cd with
      None -> (* we are the root *) ()
    | Some (parent,dep_indep,index) ->
	replace_op parent dep_indep index op


let intersect_list l1 l2 = 
  List.filter (fun x -> List.mem x l1) l2

(* l1 - l2 *)
let difference_list l1 l2 =
  List.filter (fun x -> not (List.mem x l2)) l1

let make_child_desc parent dep_indep index = 
  parent,dep_indep, index
(*******************************************)
(*** Below Here should probably be moved ***)
(*******************************************)

(* should these type of utils be moved into xquery_algebra_ast_util.ml? *)
let is_let op =
  match op.palgop_expr_name with
    | AOELet _ -> true
    | _ -> false

let is_access_tuple op =
  match op.palgop_expr_name with
    | AOEAccessTuple _ -> true
    | _ -> false

let is_empty_tuple op =
  match op.palgop_expr_name with
    | AOECreateTuple names ->
	names = [||]
    | _ -> false

let is_any_join op =
  match op.palgop_expr_name with
    | AOELeftOuterJoin _
    | AOEJoin _ -> true
    | _ -> false

let is_outer_join op = 
  match op.palgop_expr_name with
    | AOELeftOuterJoin _ -> true
    | _ -> false


let is_regular_join op =
  match op.palgop_expr_name with
    | AOEJoin _ -> true
    | _ -> false


let is_input_tuple op =
  match op.palgop_expr_name with
    | AOEInputTuple  -> true	
    | _ -> false


let is_a_map_concat op =
  match op.palgop_expr_name with
    | AOEMapConcat -> 
	true
    | _ -> false

let is_any_map_concat op =
  match op.palgop_expr_name with
    | AOEOuterMapConcat _
    | AOEMapConcat -> 
	true
    | _ -> false

let is_an_outer_map op =
  match op.palgop_expr_name with
    | AOEOuterMapConcat _ -> 
	true
    | _ -> false

let is_a_sep_sequence op = 
  match op.palgop_expr_name with
    | AOEMapIndexStep _ -> true
    | _ -> false

let is_a_map_index op = 
  match op.palgop_expr_name with
    | AOEMapIndex _ -> true
    | _ -> false

let is_a_map op = 
  match op.palgop_expr_name with
    | AOEMapConcat 
	(* Tuple operations *)
    | AOEMapFromItem _
    | AOEMapToItem                        (* Tuple to item iteration *)
    | AOEMap                                 (* Tuple iteration *)
    | AOEMapIndex _ ->  true
    | _ -> 
	false

(* is a dependent map *)
let is_a_dep_map op = 
  match op.palgop_expr_name with
    | AOEMapConcat 
	(* Tuple operations *)
    | AOEMapFromItem _
    | AOEMapToItem                        (* Tuple to item iteration *)
    | AOEMap                                 (* Tuple iteration *)
      ->  true
    | _ -> 
	false


let is_a_non_concat_map op = 
  (is_a_map op) && (not (is_a_map_concat op))

let is_a_map_from_item op =
  match op.palgop_expr_name with
    | AOEMapFromItem _ -> true
    | _ -> false

let is_a_map_to_item op =
  match op.palgop_expr_name with
    | AOEMapToItem -> true
    | _ -> false

let is_a_tuple_input_map op =
  (is_a_dep_map op) && (not (is_a_map_from_item op))

let is_group op = 
  match op.palgop_expr_name with 
    | AOEGroupBy _ -> true
    | _ -> false

let is_select op =
   match op.palgop_expr_name with
     | AOESelect _ -> true
     | _ -> false

let is_product op =
  match op.palgop_expr_name with
    | AOEProduct -> true
    | _ -> false

let is_a_select op =
  match op.palgop_expr_name with
    | AOESelect _ -> true
    | _ -> false

let is_var op = 
  match op.palgop_expr_name with
    | AOEVar _ -> true
    | _ -> false

let is_some op =
  match op.palgop_expr_name with
    | AOESome _ -> true
    | _ -> false

let is_null_map op =
  match op.palgop_expr_name with
    | AOENullMap _ -> true
    | _ -> false
    
(* Added to remove materialize *)
(* in all cases except updates *)

let is_a_materialize op =
  match op.palgop_expr_name with
    | AOEMaterializeTable -> 
	true
    | _ -> false

(*************)
(* Functions *)
(*************)

(* get the function name or None if it 
   isn't a function *)
let get_function_name op =
    match op.palgop_expr_name with
      | AOECallBuiltIn ((cfname,_),_,_)
      | AOECallOverloaded ((cfname,_), _) 
      | AOECallUserDefined ((cfname,_),_,_) ->
	  Some cfname
      | _ -> None

let match_name op name =
  match get_function_name op with
      None -> false
    | Some n ->
	Namespace_names.rqname_equal n name

let is_boolean op        = match_name op Namespace_builtin.fn_boolean
let is_distinct_value op = match_name op Namespace_builtin.fn_distinct_values
let is_fn_data op        = match_name op Namespace_builtin.fn_data

let is_equal op = 
  let overloaded_equality = 
    [Namespace_builtin.op_equal;
     Namespace_builtin.op_equal_left_empty;
     Namespace_builtin.op_equal_right_empty] 
  in
  let builtin_equality = 
    [Namespace_builtin.op_anyURI_equal ;
     Namespace_builtin.op_boolean_equal;
     Namespace_builtin.op_decimal_equal;
     Namespace_builtin.op_double_equal;
     Namespace_builtin.op_float_equal;] in

  let names = overloaded_equality @ builtin_equality in
    List.exists (match_name op) names

let is_gt op = 
  let overloaded_gt = 
    [Namespace_builtin.op_gt;
     Namespace_builtin.op_gt_left_empty;
     Namespace_builtin.op_gt_right_empty] 
  in
  let builtin_gt = 
    [Namespace_builtin.op_integer_gt;
     Namespace_builtin.op_date_gt;
     Namespace_builtin.op_time_gt;
     Namespace_builtin.op_dateTime_gt;
     Namespace_builtin.op_yearMonthDuration_gt;
     Namespace_builtin.op_dayTimeDuration_gt;
     Namespace_builtin.op_string_gt;
     Namespace_builtin.op_boolean_gt;
     Namespace_builtin.op_decimal_gt;
     Namespace_builtin.op_double_gt;
     Namespace_builtin.op_float_gt;] in

  let names = overloaded_gt @ builtin_gt in
    List.exists (match_name op) names


let is_ge op = 
  let overloaded_ge = 
    [Namespace_builtin.op_ge;
     Namespace_builtin.op_ge_left_empty;
     Namespace_builtin.op_ge_right_empty] 
  in
  let builtin_ge = 
    [Namespace_builtin.op_integer_ge;
     Namespace_builtin.op_date_ge;
     Namespace_builtin.op_time_ge;
     Namespace_builtin.op_dateTime_ge;
     Namespace_builtin.op_yearMonthDuration_ge;
     Namespace_builtin.op_dayTimeDuration_ge;
     Namespace_builtin.op_string_ge;
     Namespace_builtin.op_boolean_ge;
     Namespace_builtin.op_decimal_ge;
     Namespace_builtin.op_double_ge;
     Namespace_builtin.op_float_ge;] in

  let names = overloaded_ge @ builtin_ge in
    List.exists (match_name op) names


let is_lt op = 
  let overloaded_lt = 
    [Namespace_builtin.op_lt;
     Namespace_builtin.op_lt_left_empty;
     Namespace_builtin.op_lt_right_empty] 
  in
  let builtin_lt = 
    [Namespace_builtin.op_integer_lt;
     Namespace_builtin.op_date_lt;
     Namespace_builtin.op_time_lt;
     Namespace_builtin.op_dateTime_lt;
     Namespace_builtin.op_yearMonthDuration_lt;
     Namespace_builtin.op_dayTimeDuration_lt;
     Namespace_builtin.op_string_lt;
     Namespace_builtin.op_boolean_lt;
     Namespace_builtin.op_decimal_lt;
     Namespace_builtin.op_double_lt;
     Namespace_builtin.op_float_lt;] in

  let names = overloaded_lt @ builtin_lt in
    List.exists (match_name op) names


let is_le op = 
  let overloaded_le = 
    [Namespace_builtin.op_le;
     Namespace_builtin.op_le_left_empty;
     Namespace_builtin.op_le_right_empty] 
  in
  let builtin_le = 
    [Namespace_builtin.op_integer_le;
     Namespace_builtin.op_date_le;
     Namespace_builtin.op_time_le;
     Namespace_builtin.op_dateTime_le;
     Namespace_builtin.op_yearMonthDuration_le;
     Namespace_builtin.op_dayTimeDuration_le;
     Namespace_builtin.op_string_le;
     Namespace_builtin.op_boolean_le;
     Namespace_builtin.op_decimal_le;
     Namespace_builtin.op_double_le;
     Namespace_builtin.op_float_le;] in

  let names = overloaded_le @ builtin_le in
    List.exists (match_name op) names

let is_fs_untyped_to_any op =
  match_name op Namespace_builtin.fs_untyped_to_any

let get_let_name op = 
  match op.palgop_expr_name  with
    | AOELet (odt, vn) -> vn
    | _ -> raise (Query (Malformed_Algebra_Expr ("Attempted to get name of op we thought was a let but is not")))

let get_outer_map_name op =
  match op.palgop_expr_name with 
      AOEOuterMapConcat vn ->
	vn
    | _ -> raise (Query (Malformed_Algebra_Expr ("Attempted to get name of op we thought was an outer map but is not")))

let get_map_from_item_name op =
  match op.palgop_expr_name with 
    | AOEMapFromItem vn -> vn
    | _ -> raise (Query (Malformed_Algebra_Expr ("Attempted to get name of op we thought was a mapfromitem but is not")))


let get_null_map_name op =
  match op.palgop_expr_name with
    | AOENullMap v -> v
    | _ -> raise (Query (Malformed_Algebra_Expr ("Attempted to get name of op we thought was a null map but is not")))

let get_sep_sequence_name op =
  match op.palgop_expr_name with
    | AOEMapIndexStep v -> v
    | _ -> raise (Query (Malformed_Algebra_Expr ("Attempted to get name of op we thought was a sep sequence but is not")))

let get_outer_join_name op =
  match op.palgop_expr_name with
    | AOELeftOuterJoin (vn,pd) -> vn
    | _ -> raise (Query (Malformed_Algebra_Expr ("Attempted to get name of op we thought was an outer join but is not")))

(* t1 is indep of t2 if t1 accesses 
   no fields returned by t2.
   returned(t2) intersect accessed(t1) = empty_set
   --- 
   (returned (t2) - returned (t2.indep)) intersect accessed (t1) 
*)
let subexpr_get_returned se =
  List.concat 
    (match se with
	NoSub -> []
      | OneSub s0 -> (algop_get_returned_fields s0) :: []
      | TwoSub (s0,s1) ->
	  (algop_get_returned_fields s0) :: 
	    (algop_get_returned_fields s1) :: []
      | ManySub sa ->
	  List.map algop_get_returned_fields (Array.to_list sa))

let string_of_tuple_fields tf =
  List.fold_left (fun out cur ->
		    out ^ "," ^ (Namespace_names.prefixed_string_of_rqname cur)) "" tf

let tuple_independent t1 t2 = 
  let af = algop_get_accessed_fields t1 in
  let rf = algop_get_returned_fields t2 in
  let t2_indep = subexpr_get_returned t2.psub_expression in 
  let t2_intro = difference_list rf t2_indep in      
  let int_map = intersect_list af t2_intro in
  let result =     [] = int_map in
    result

