(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: planio_parse_plan_xml.ml,v 1.43 2006/05/11 10:52:49 pmichiel Exp $ *)

(* Module Algebra_parse_plan_xml
   Description:
   This module parses the algebraic plans which are serialized by algebra_print_xml.ml
   so that they may be run again after modification

   NOTE: This code is not currently hooked into the runtime.
*)

open Error

open Occurrence

open Namespace_names
open Namespace_context
open Namespace_symbols

open Datatypes

open Dm_atomic
open Dm
open Dm_util

open Small_stream_ast
open Streaming_types

open Print_common

open Xquery_common_ast
open Xquery_core_ast
open Xquery_algebra_ast 
open Xquery_algebra_ast_util 

open Planio_print_common
open Planio_print_util

open Compile_annotate

let nsenv = default_xquery_nsenv
let string_id x = x
let int_id x = int_of_string x

(* Apply a function multiple times to the stream *)
let get_multiple cardinality fn st =
  let rec get_multiple_helper cur accum =
    if (cur <= 0) then
      accum
    else
      get_multiple_helper (cur-1) ((fn st) :: accum)
  in
    List.rev (get_multiple_helper cardinality [])

(* Helper functions for stream parsing *)
let rec start_element_get_name_no_consume st =
  let v = Stream.peek st in    
  match v with
  | Some event ->
      begin
	match event.rse_desc with 
	  ((RSAX_startElement (e, attrs, has, baseuri, env))) ->
	    e, attrs
    (* Comments and characters are ignored *)
	| (RSAX_comment _) 
	| (RSAX_characters _) ->
	    Stream.junk st;
	    start_element_get_name_no_consume st
	| (RSAX_endElement ) ->
	    raise (Query (Algebra_Parsing_Error ("Expecting Element start but found end element"))	)
	| _ -> 
	    raise (Query (Algebra_Parsing_Error ("Expecting Element start but did not find it")))
      end
  | _ -> 
      raise (Query (Algebra_Parsing_Error ("Expecting Element start but did not find it")))

(* Start element and extract its name from the stream, consume it *)
let start_element_get_name st = 
  let res = start_element_get_name_no_consume st in
    Stream.junk st;
    res

(* Start element, with expectected element name *)
let start_element elem_name st = 
  let name, attrs = 
    try
      start_element_get_name st     
    with Query (Algebra_Parsing_Error p) ->
      raise (Query (Algebra_Parsing_Error ("Named: " ^ (string_of_relem_symbol elem_name) ^ "|" ^ p ))	)
  in
    if not (relem_equal name elem_name) then      
      raise (Query (Algebra_Parsing_Error ("Expecting " ^ (string_of_relem_symbol elem_name) ^ " did not find it" ^
	    " found: " ^ (string_of_relem_symbol name))))
    else
      attrs

let rec is_element_end st = 
  let v = Stream.peek st in
    match v with
    | Some event ->
	begin
	  match event.rse_desc with 
	  | RSAX_endElement -> true
	  | (RSAX_comment _) 
	  | (RSAX_characters _) ->
	      Stream.junk st;
	      is_element_end st
	  | _ -> false
	end
    | _ -> false

(* Assert there is an endtag and consume it *)
let rec consume_end_element (st:resolved_sax_event Stream.t) =
  let v = Stream.peek st in
    match v with
    | Some event ->
	begin
	  match event.rse_desc with 
	  | (RSAX_endElement) -> Stream.junk st
	  | (RSAX_comment _) 
	  | (RSAX_characters _) ->
	      Stream.junk st;
	      consume_end_element st 
	  | _ -> raise (Query (Algebra_Parsing_Error ("Expecting end tag - did not find it")))
	end
    | _ -> raise (Query (Algebra_Parsing_Error ("Expecting end tag - did not find it")))


(* Util to check if an optional element is present *)
let rec check_opt_element elem st =
  let relem = relem_symbol elem in
  let rec check_opt_element_helper () =
  let v = Stream.peek st in
    match v with
    | Some event ->
	begin
	  match event.rse_desc with 
	  | (RSAX_comment _) 
	  | (RSAX_characters _) ->
	      Stream.junk st;
	      check_opt_element elem st
	  | ((RSAX_startElement (e, attrs, has, baseuri, env))) ->	
	      e = relem
	  | _ -> false
	end
      | _ -> false
  in
    check_opt_element_helper ()
(* 
   Consumes the start and end tags of elem_name
To help make side-effects occur in as few places as possible *)
let element_parser elem_name fn error_msg st  =
  try
    let attrs = start_element (relem_symbol elem_name) st in
    let return_value = fn attrs st in
    let _ = consume_end_element st in
    return_value
  with (Query (Algebra_Parsing_Error q)) ->
    raise (Query (Algebra_Parsing_Error (q ^ " | " ^ error_msg)))

(* Attribute Parsing functions *)
(* coerce_fun: string -> 'a, empty_fn: unit -> 'a *)
let rec parse_attr_helper coerce_fn attr_name empty_fn attr_list =
  match attr_list with
      [] -> empty_fn ()
    | (aname, value) :: rest when attr_name = aname -> 
	coerce_fn value
    | _ :: rest -> parse_attr_helper coerce_fn attr_name empty_fn rest 

(* coerce_fun: string -> 'a *)
let parse_get_attr_from_attr_list attr_list coerce_fn a_name =
  let error_fn  () = 
     raise (Query (Algebra_Parsing_Error ("Looking for " ^ (prefixed_string_of_rqname a_name))))
  in
  let attr_name = rattr_symbol a_name in
  parse_attr_helper coerce_fn attr_name error_fn attr_list
 

let parse_opt_get_attr_from_attr_list attr_list coerce_fn a_name =
  let attr_name = rattr_symbol a_name in
  parse_attr_helper (fun x -> Some (coerce_fn x)) attr_name (fun () -> None) attr_list



(********************* END PARSE HELPER SECTION *****************)


(* Type Symbols and Namespace parsing section *)
let occurrence_sym = rattr_symbol occurrence_attr_name 
let parse_occurrence_option attrs =
  match attrs with 
    | [] -> None
    |[(n,str)] when n = occurrence_sym ->        
       begin
	 match str with
	   | "None" -> None
	   | "*" -> Some (UP_INT 0, UNBOUNDED) 
	   | "+" -> Some (UP_INT 1, UNBOUNDED)
	   | "?" -> Some (UP_INT 0, UP_INT 1)
	   | _   -> raise (Query (Algebra_Parsing_Error ("Unknown Occurence Marker")))
       end
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown Occurence Marker")))


(* Relem symbol *)
let relem_symbol_constructor attrs st =
  let prefix = parse_get_attr_from_attr_list attrs xml_prefix_of_string prefix_attr_name in
  let (_,uri,ncname)   = parse_get_attr_from_attr_list attrs parse_curly_uri_string rqname_attr_name in
    relem_symbol (prefix, uri, ncname)

let relem_symbol_element_parser st =
  element_parser relem_elem_name relem_symbol_constructor "relem_symbol" st 

let rattr_symbol_constructor attrs st =
  let prefix = parse_get_attr_from_attr_list attrs xml_prefix_of_string prefix_attr_name in
  let (_,uri,ncname)   = parse_get_attr_from_attr_list attrs parse_curly_uri_string rqname_attr_name in
    rattr_symbol (prefix, uri, ncname)

let rattr_symbol_element_parser st =
  element_parser rattr_elem_name rattr_symbol_constructor "rattr_symbol" st

let rtype_symbol_constructor attrs st = 
  let prefix = parse_get_attr_from_attr_list attrs xml_prefix_of_string prefix_attr_name in
  let (_,uri,ncname)   = parse_get_attr_from_attr_list attrs parse_curly_uri_string rqname_attr_name in
    rtype_symbol (prefix, uri, ncname)

let rtype_symbol_element_parser st =
  element_parser rtype_elem_name rtype_symbol_constructor "rtype_symbol" st

(* AIT_xXxX Parsing Section *)
(* Default parsing function for those that do not need paramaters *)
let no_param_constructor name ret_type emsg st = 
  element_parser name (fun attrs st -> ret_type) emsg st 

let ait_attributetype_constructor attrs st =
  if (is_element_end st)
  then
    AAttributeTest None
  else
    let rname = rattr_symbol_element_parser st in
    let rtype = rtype_symbol_element_parser st in
    AAttributeTest (Some (rname, Some rtype))

let ait_attributetype_element_parser st =
  element_parser ait_attributetype_elem_name ait_attributetype_constructor "ait_attributetype" st

(* AIT_element_type *)
let ait_elementtype_constructor attrs st =
  if (is_element_end st)
  then
    AElementTest None
  else
    let rname = relem_symbol_element_parser st in
    let rtype = rtype_symbol_element_parser st in
    AElementTest (Some (rname, Some rtype))

let ait_elementtype_element_parser st =
  element_parser ait_elementtype_elem_name ait_elementtype_constructor "ait_elementtype" st

(* Schema attribute *)
let ait_schemaattribute_constructor attr st =
  ASchemaAttributeTest (rattr_symbol_element_parser st)

let ait_schemaattribute_element_parser st =
  element_parser ait_schemaattribute_elem_name ait_schemaattribute_constructor "ait_schemaattribute" st

(* Schema Elementtype *)
let ait_schemaelementtype_constructor attr st =
  ASchemaElementTest (relem_symbol_element_parser st)

let ait_schemaelementtype_element_parser st =
  element_parser ait_schemaelementtype_elem_name ait_schemaelementtype_constructor "ait_schemaelementtype" st

(* typeref *)
let ait_typeref_constructor attrs st =
  AITTypeRef (rtype_symbol_element_parser st)

let ait_typeref_element_parser st =
  element_parser ait_typeref_elem_name ait_typeref_constructor "ait_typeref" st

(* Processing Instruction *)
let ait_processing_intruction_constructor attrs st =
  let pi = parse_opt_get_attr_from_attr_list attrs string_id pi_arg_attr_name in
  APIKind pi

let ait_processing_instruction_element_parser st =
  element_parser ait_processing_instruction_elem_name ait_processing_intruction_constructor "ait_processing_instruction" st

(* Atomic *)
let ait_atomic_constructor attrs st =
  AITAtomic (rtype_symbol_element_parser st)

let ait_atomic_element_parser st =
  element_parser ait_atomic_elem_name ait_atomic_constructor "ait_atomic_element" st 

(* ElementKind *)
let ait_element_kind_parser st =
  let t = ait_schemaelementtype_element_parser st in
  AElementKind t

(* AttributeKind *)
let ait_attribute_kind_parser st =
  let t = ait_schemaattribute_element_parser st in
  AAttributeKind t

(* Document *)  
let ait_document_constructor attrs st =
  if (is_element_end st) then
    ADocumentKind None
  else
    let rname = ait_elementtype_element_parser st in
    ADocumentKind (Some rname)
  
let ait_document_element_parser st = 
  element_parser ait_document_elem_name ait_document_constructor "ait_document" st

(* without params *)
let ait_node_element_parser st = no_param_constructor ait_node_elem_name AAnyKind "ait_node" st
let ait_item_element_parser st = no_param_constructor ait_item_elem_name AITItem "ait_item" st
let ait_numeric_element_parser st = no_param_constructor ait_numeric_elem_name AITNumeric "ait_numeric" st
let ait_text_element_parser st = no_param_constructor ait_text_elem_name ATextKind "ait_text" st
let ait_comment_element_parser st = no_param_constructor ait_comment_elem_name ACommentKind "ait_comment" st
let ait_empty_element_parser st = no_param_constructor ait_empty_elem_name AITEmpty "ait_empty" st

let akind_parsers =
  List.map (fun (x,y) -> ((alg_elem x), y))
    [ (ait_attributetype_elem_name,ait_attribute_kind_parser);
      (ait_elementtype_elem_name,ait_element_kind_parser); 
      (ait_node_elem_name,ait_node_element_parser);
      (ait_text_elem_name,ait_text_element_parser);
      (ait_comment_elem_name,ait_comment_element_parser);
      (ait_processing_instruction_elem_name,ait_processing_instruction_element_parser);
      (ait_processing_instruction_elem_name,ait_processing_instruction_element_parser);
      (ait_document_elem_name,ait_document_element_parser)
   ]

let parse_akind_test_element st = 
  let name, attrs = start_element_get_name_no_consume st in
  let cur_parser = 
    try
      (List.assoc name akind_parsers)
    with Not_found ->
      raise (Query (Algebra_Parsing_Error ("Invalid subelement in parse_akind_test_element: *" ^ (string_of_relem_symbol name) ^"*")))
  in
  let kind_test = (cur_parser st)
  in 
    kind_test 


let ait_kind_test_parser st =
  let _ = start_element (relem_symbol ait_kindtest_elem_name) st in
  let t = parse_akind_test_element st in
  consume_end_element st;
  AITKindTest t

(* NOTE: All of these parser functions need to consume their first 
   element token before proceeding.  We fixed ait_kind_test_parser,
   but none of the others.  Look for all calls to start_element_get_name_no_consume to 
   see where tokens are not being consumed. 
*)
let aitem_parsers =
  List.map (fun (x,y) -> ((alg_elem x), y))
    [ (ait_kindtest_elem_name,ait_kind_test_parser);
      (ait_typeref_elem_name,ait_typeref_element_parser);
      (ait_item_elem_name,ait_item_element_parser);
      (ait_numeric_elem_name,ait_numeric_element_parser); 
      (ait_empty_elem_name,ait_empty_element_parser);
      (ait_atomic_elem_name,ait_atomic_element_parser) ]
 
let parse_aitemtype_element st = 
  let name, attrs = start_element_get_name_no_consume st in
  let cur_parser = 
    try
      List.assoc name aitem_parsers 
    with Not_found ->
      raise (Query (Algebra_Parsing_Error ("Invalid subelement in parse_aitemtype_element " ^ (string_of_relem_symbol name) )))
  in
  cur_parser st

(* Parse a csequence element *)

let asequencetype_constructor attrs st =
  let occur = parse_occurrence_option attrs in
  let item_portion = parse_aitemtype_element st in
  { pasequencetype_desc = (item_portion, occur);
    pasequencetype_loc = Finfo.bogus }

let asequencetype_element_parser st = 
  element_parser asequencetype_elem_name asequencetype_constructor "asequencetype" st

(* If an optional sequence type is present grab it else return None *)
let parse_optasequencetype_element st = 
  if (check_opt_element asequencetype_elem_name st) then
    Some (asequencetype_element_parser st)
  else
    None

(* Functions for parsing the Datamodel signatures *)
let parse_input_sig (st: resolved_sax_event Stream.t) = 
  let attrs = start_element (alg_elem input_datamodel_elem_name) st in
  let index = parse_get_attr_from_attr_list attrs int_of_string index_attr_name in
  let datamodel = parse_get_attr_from_attr_list attrs physical_type_of_string datamodel_attr_name in
  let _ = consume_end_element st in
    datamodel, index

(* Datamodel signature *)
let make_input_signature in_sigs str =
  let types, index = List.split in_sigs in
  let assert_cardinality c l = 
    if (List.length l) != c then
      raise (Query (Algebra_Parsing_Error ("Input Cardinality Mismatch")))
    else
      ()
  in
    
  (* signatures should be in order since they are elements *)
  match str with
    | "NoInput" -> 
	NoInput
    | "OneInput" -> 
	OneInput (List.nth types 0)
    | "TwoInput" -> 
	TwoInput ((List.nth types 0), (List.nth types 1))
    | "ManyInput" ->  (* Many means it is replicated *)
	ManyInput (Array.of_list types)
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown Input Cardinality for signature")))
      
(* DM signature *)
let parse_output_sig st = 
  let attrs = start_element (alg_elem output_datamodel_elem_name) st in
  let _ = consume_end_element st in
    parse_get_attr_from_attr_list attrs physical_type_of_string datamodel_attr_name


let parse_element_of_expr_eval_sig st =
  let attrs = start_element (alg_elem datamodel_signature) st in
  let cardinality = parse_get_attr_from_attr_list attrs get_cardinality arg_count_attr_name in 
    (* This implicitly checks the cardinality of the inputs *)
  let input_sigs = get_multiple cardinality parse_input_sig st in 
  let output_sig = parse_output_sig st in
  let _ = consume_end_element st in

  let inputs = parse_get_attr_from_attr_list attrs (make_input_signature input_sigs) arg_count_attr_name in
    input, output_sig 


(* Function Type Signature *)
let input_signature_constructor attrs st =
  let fn_index = parse_get_attr_from_attr_list attrs int_of_string index_attr_name in
    parse_optasequencetype_element st    

let input_signature_element_parser st =
  element_parser input_type_elem_name input_signature_constructor "input_signature" st

(* Should assert that attrs = []? *)
let output_signature_constructor attrs st = 
  asequencetype_element_parser st 
  

let output_signature_element_parser st =
  element_parser output_type_elem_name output_signature_constructor "output_signature" st

(* Function Singature parser *)
let parse_function_signature_constructor attrs st =
  let arity = parse_get_attr_from_attr_list attrs int_of_string arity_attr_name in
  let fname  = parse_get_attr_from_attr_list attrs parse_curly_uri_string fn_name_attr_name in
  let input_sig = Array.of_list (get_multiple arity input_signature_element_parser st) in
  let output_sig = output_signature_element_parser st in
    fname,arity, input_sig, output_sig

let function_signature_element_parser st =
  element_parser function_signature_elem_name parse_function_signature_constructor "function_signature" st

(* Tuple Name parsing *)

let tuple_slot_constructor attrs st =
  let tindex = parse_get_attr_from_attr_list attrs int_of_string index_attr_name in
  let tname  = parse_get_attr_from_attr_list attrs parse_curly_uri_string var_attr_name in
  let odt    = parse_optasequencetype_element st in
    (tindex, (odt,tname))

let tuple_slot_element_parser st =
  element_parser tuple_slot_elem_name tuple_slot_constructor "tuple_slot" st 

(* index attribute kept around just for debugging since the elements are ordered*)
let make_tuple_names arity st =
  let index_type_name_tuple = get_multiple arity tuple_slot_element_parser st in
  let _, type_names = List.split index_type_name_tuple in
    (Array.of_list type_names)


(* Node Test Parsing *)
  
let name_test_constructor attrs st = 
  let qn = parse_get_attr_from_attr_list attrs parse_curly_uri_string name_test_attr_name in
    APNameTest (relem_symbol qn)

let name_test_element_parser st =
  element_parser name_test_elem_name name_test_constructor "name_test" st

let pi_kind_test_constructor attrs st = 
  if attrs = [] then
    None
  else
    Some (parse_get_attr_from_attr_list attrs string_id pi_kind_attr_name)

let pi_kind_element_parser st =
  element_parser pi_kind_test_elem_name pi_kind_test_constructor "pi_kind" st

let kind_test_constructor attrs st = 
  let node_kind = parse_akind_test_element st in
  APNodeKindTest node_kind

let kind_test_element_parser st =
  element_parser kind_test_elem_name kind_test_constructor "kind_test" st

let node_test_constructor attrs st =
  (* Can be one of two elements name test or kind test *)
  let name, _ = start_element_get_name_no_consume st in
    if (name = (alg_elem name_test_elem_name)) then
      name_test_element_parser st
    else if (name = (alg_elem kind_test_elem_name)) then
      kind_test_element_parser st
    else
      raise (Query (Algebra_Parsing_Error("Expecting a node test sub-element but did not find a known one")))


let node_test_element_parser st = 
  element_parser node_test_elem_name node_test_constructor "node_test" st

let opt_node_test_element_parser st =
  if check_opt_element node_test_elem_name st then
    Some (element_parser node_test_elem_name node_test_constructor "node_test" st)
  else
    None


let twig_index_ctor attrs st =
  let ix = parse_get_attr_from_attr_list attrs int_of_string index_attr_name in
  let ax = parse_get_attr_from_attr_list attrs axis_of_string axis_attr_name in
  ax, ix


let child_twig_parser st = 
  if check_opt_element child_twig_name st then
    Some (element_parser child_twig_name twig_index_ctor "child_twig" st)
  else
    None

let rec pred_twig_parser st =
   if check_opt_element pred_twig_name st then
    let x = element_parser pred_twig_name twig_index_ctor "pred_node" st in
    x :: (pred_twig_parser st)
  else 
    [ ]

(* check_opt_element *)
let twig_node_constructor attrs st =
  let nt = opt_node_test_element_parser st in
  let child_index = child_twig_parser st in
  let pred_index_list = pred_twig_parser st in
  let out_field = 
    parse_opt_get_attr_from_attr_list attrs parse_curly_uri_string out_field_attr_name in
  let restore =  parse_get_attr_from_attr_list attrs bool_of_string restore_attr_name in
  {
    node_test = nt;
    out = out_field;
    restore_out = restore;
    child_twig = child_index;
    pred_twigs = pred_index_list;
    requires_sbdo = (true, true); (* fixme, include in xml export *)
  }

let rec tree_pattern_parser st =
  if check_opt_element twig_node_name st then
    let x = element_parser twig_node_name twig_node_constructor "twig_node" st in
    x :: (tree_pattern_parser st)
  else 
    [ ]

(* Typeswitch parsing *)

let typeswitch_case_constructor attrs st =
  let variable = parse_opt_get_attr_from_attr_list attrs parse_curly_uri_string var_attr_name in
  let vtype    = asequencetype_element_parser st in
  let apattern = { papattern_desc = ACase vtype;
		   papattern_loc  = Finfo.bogus} in

    apattern, variable

let typeswitch_default_constructor attrs st =
  let variable = parse_opt_get_attr_from_attr_list attrs parse_curly_uri_string var_attr_name in
  let apattern = { papattern_desc = ADefault;
		   papattern_loc  = Finfo.bogus} in
    apattern, variable

let typeswitch_case_element_parser st = 
  element_parser ts_case_elem_name typeswitch_case_constructor "typeswitch_case" st
    
let typeswitch_default_element_parser st =
  element_parser ts_default_elem_name typeswitch_default_constructor "typeswitch_default" st


let parse_typeswitch_branches arity st = 
  (* We know that the last branch is always default *)
  let case_list = get_multiple (arity-1) typeswitch_case_element_parser st in
  let default = typeswitch_default_element_parser st in
    Array.of_list ( case_list @ default :: [] )


(* Projection Parsing *)
let project_name_constructor attrs st =
  parse_get_attr_from_attr_list attrs parse_curly_uri_string project_name_attr_name

let project_name_element_parser st = 
  element_parser project_elem_name project_name_constructor "project_name_parser" st

let rec project_constructor attrs st =
  if check_opt_element project_name_elem_name st then
    let x =  project_name_element_parser st in
    x :: (project_constructor attrs st)
  else 
    []

(* OrderBy parsing *)
let sort_spec_constructor attrs st =
  let sk  = parse_get_attr_from_attr_list attrs sortkind_of_string sort_kind_attr_name in
  let esk = parse_get_attr_from_attr_list attrs emptysortkind_of_string empty_sort_kind_attr_name in
    (sk, esk)

let sort_spec_element_parser st =
  element_parser sort_spec_elem_name sort_spec_constructor "sort spec" st

(**********************)
(* Grouping Operation *)
(**********************)
let variable_name_constructor attrs st =
  parse_get_attr_from_attr_list attrs parse_curly_uri_string vname_attr_name

let named_variable_name_element_parser name st =
  element_parser name variable_name_constructor "named variable name" st

let variable_name_element_parser st =
  named_variable_name_element_parser vname_elem_name st 

let variable_list_constructor attrs st =
  let arity = parse_get_attr_from_attr_list attrs int_of_string arity_attr_name in 
    get_multiple arity variable_name_element_parser st

let variable_list_element_parser name st =
  element_parser name variable_list_constructor "variable list" st

let group_desc_constructor attrs st =
  let gbn = variable_list_element_parser gbn_elem_name st in
  let ind = variable_list_element_parser induced_elem_name st in
  let mbv = variable_list_element_parser mbv_elem_name st in
  let agg = named_variable_name_element_parser agg_elem_name st      in
  let odt = parse_optasequencetype_element st in
    mk_group_desc gbn ind mbv (odt,agg)

let group_desc_element_parser st =
  element_parser group_desc_elem_name group_desc_constructor "group description" st

let stable_of_string str = 
  match str with
    | "Stable" -> Stable
    | "Unstable" -> NonStable
    | _ -> raise (Query (Malformed_Expr ("Unknown stable type string'"^str^"'")))
(**************)
(* Predicates *)
(**************)
let rec simple_conjunct_constructor attrs st =  
  let s = parse_get_attr_from_attr_list attrs int_of_string simple_conjunct_start_attr_name in 
  let e = parse_get_attr_from_attr_list attrs int_of_string simple_conjunct_end_attr_name   in 
  let _ = consume_end_element st in
    SimpleConjunct(s,e)

and complex_conjunct_constructor attrs st = 
  let l = element_of_pred_desc st in
  let r = element_of_pred_desc st in
  let _ = consume_end_element st in
    ComplexConjunct (l,r)

and disjunct_constructor attrs st = 
  let l = element_of_pred_desc st in
  let r = element_of_pred_desc st in
  let _ = consume_end_element st in
    Disjunct (l,r)

and element_of_pred_desc st =
  let name, attrs = start_element_get_name st in
    if relem_equal name (alg_elem simple_conjunct_elem_name) then
      simple_conjunct_constructor attrs st
    else if relem_equal name (alg_elem complex_conjunct_elem_name) then
      complex_conjunct_constructor attrs st
    else if relem_equal name (alg_elem disjunct_elem_name) then
      disjunct_constructor attrs st
    else raise (Query (Algebra_Parsing_Error ("Attempting to parse a predicate description but found " ^  
				   (string_of_relem_symbol name))))
(***********)
(* Updates *)
(***********)
let insert_location_of_string str =
  match str with 
    | "LastInto" -> AOUAsLastInto
    | "FirstInto" -> AOUAsFirstInto 
    | "Into" -> AOUInto 
    | "After" -> AOUAfter 
    | "Before" -> AOUBefore
    | _ -> raise (Query (Malformed_Expr ("Unknown Location String '"^str^"'")))

let snap_modifier_of_string sm = 
  match sm with
    | "ordered deterministic" -> Snap_Ordered_Deterministic
    | "unordered deterministic" -> Snap_Unordered_Deterministic
    | "nondeterministic"      -> Snap_Nondeterministic  
    | _ -> raise (Query (Malformed_Expr ("Unknown Snap Modifier String '"^sm^"'")))


let value_flag_of_string str =
  match str with 
    |  "Normal"  -> Normal_Replace
    | "Value of" -> Value_Of_Replace 
    | _ -> raise (Query (Malformed_Expr ("Value of flag '"^str^"'")))

	(* Name parsing *)


let parse_algop_name comp_ctxt st =
  let element_name, attrs = start_element_get_name st in
  let parse_attrs x = parse_get_attr_from_attr_list attrs x in
  let parse_opt_attrs x = parse_opt_get_attr_from_attr_list attrs x in
    match (get_moniker_of_serialized_moniker element_name) with 

      | AOEIf_n ->
	  AOEIf

      | AOELet_n ->
	  let odt = parse_optasequencetype_element st in
	  let vn = parse_attrs parse_curly_uri_string vname_attr_name in
	    AOELet ( odt, vn )
	    
      | AOEInputTuple_n -> AOEInputTuple

      | AOETypeswitch_n ->
	  let arity = parse_attrs int_of_string ts_num_branches_attr_name in
	  let pat_array = parse_typeswitch_branches arity st in
	    AOETypeswitch( pat_array ) 

      | AOEVar_n ->
	  let vname = parse_attrs parse_curly_uri_string vname_attr_name in 	  
	    AOEVar(vname)

      | AOEScalar_n ->
	  let scale_value = parse_attrs (fun x -> x) value_attr_name in
	  let atomic_type = parse_attrs atomic_type_of_string atomic_type_attr_name in 
	  let literal = literal_of_string atomic_type scale_value in
	  AOEScalar literal

      | AOESeq_n ->
	  AOESeq 
	    
      | AOEEmpty_n ->
	  AOEEmpty	    
	    
      | AOEDocument_n ->
	  AOEDocument
	    
      | AOEPI_n  ->
	  let tg = parse_attrs string_id target_attr_name in
	  let ct = parse_attrs string_id content_attr_name in
	    AOEPI(tg, ct)

      | AOEPIComputed_n  ->
	  AOEPIComputed
	    
      | AOEComment_n ->
	  let cmt = parse_attrs string_id value_attr_name in
	    AOEComment( cmt )
	      
      | AOECommentComputed_n ->
	  AOECommentComputed

      | AOEText_n ->
	  let v = parse_attrs string_id value_attr_name in
	  AOEText v

      | AOETextComputed_n ->
	  AOETextComputed
	    
    (* NB!  We are not printing out the namespace environment associated with an element constructor, 
       so when we parse the element here, we end up with an incomplete namespace environment! *)
    (* This will break round-tripping *)
      | AOEElem_n ->
	  let e_name = relem_symbol_element_parser st in
	  let nsenv = empty_nsenv in
	    AOEElem (e_name, nsenv)
	      
      | AOEAnyElem_n ->
	  let nsenv = empty_nsenv in
	    AOEAnyElem nsenv
	      
      | AOEAttr_n -> 
	  let a_name = rattr_symbol_element_parser st in
	    AOEAttr a_name
	      
      | AOEAnyAttr_n ->
	  let nsenv = empty_nsenv in
	    AOEAnyAttr nsenv
    	      
      | AOEError_n -> 
	  AOEError
	    
      | AOETreat_n -> 
	  let cst = asequencetype_element_parser st in
	    AOETreat( cst )
	      
      | AOEValidate_n ->
	  AOEValidate Strict
	    
      | AOECast_n  ->
	  let cst = asequencetype_element_parser st in
	  let nsenv = empty_nsenv in
	    AOECast (nsenv, cst)
	      
      | AOECastable_n ->
	  let cst = asequencetype_element_parser st in
	  let nsenv = empty_nsenv in
	    AOECastable (nsenv, cst)

      | AOETreeJoin_n->
	  let axis_name = parse_attrs axis_of_string axis_attr_name in
	  let nt = node_test_element_parser st in
	    AOETreeJoin(axis_name, nt)

      | AOETupleTreePattern_n ->
	  let input_field = parse_attrs parse_curly_uri_string in_field_attr_name in
	  let pattern = Array.of_list (tree_pattern_parser st) in
	  AOETupleTreePattern (input_field, pattern)

      | AOESome_n ->
	  let odt = parse_optasequencetype_element st in
	  let vn = parse_attrs parse_curly_uri_string vname_attr_name in
	    AOESome ( odt, vn )
	      
      | AOEEvery_n ->
	  let odt = parse_optasequencetype_element st in
	  let vn = parse_attrs parse_curly_uri_string vname_attr_name in
	    AOEEvery ( odt, vn )
	      
      (* Function Calls *)
      | AOECallBuiltIn_n -> 
	  (* There are attributes, but they are name and arity which are replicated 
	     inside the signature *)
	  let fname,arity,input_types, output_type = function_signature_element_parser st in
	    AOECallBuiltIn((fname,arity), input_types, output_type)
	      

      | AOECallOverloaded_n ->
	  let fn = parse_attrs parse_curly_uri_string fn_name_attr_name in
	  let arity = parse_attrs int_of_string arity_attr_name in
	    (* Is that going to work here ?? PLEASE CHECK - Jerome *)
	  let norm_ctxt = Compile_context.norm_context_from_compile_context comp_ctxt in
	  let sigs      = Norm_overloaded.table_for_overloaded_function norm_ctxt (fn,arity) in
	  let asigs     = Compile_util.compile_overloaded_table_sigs comp_ctxt sigs in
	    
	    AOECallOverloaded((fn,arity),asigs)
	      
      | AOECallUserDefined_n -> 
	  let fname,arity,input_types, output_type = function_signature_element_parser st in
	    AOECallUserDefined( (fname,arity), input_types, output_type)

      | AOEConvertSimple_n -> 
	  let atomic_type = parse_attrs atomic_type_of_string atomic_type_attr_name in  
	  AOEConvertSimple atomic_type

      | AOEPromoteNumeric_n  -> 
	  let atomic_type = parse_attrs atomic_type_of_string atomic_type_attr_name in 
	  AOEPromoteNumeric atomic_type

      | AOEUnsafePromoteNumeric_n -> 
	  let atomic_type = parse_attrs atomic_type_of_string atomic_type_attr_name in 
	  AOEUnsafePromoteNumeric atomic_type

      | AOEKeyRef_n ->
	  let fname = parse_attrs string_id kname_attr_name in
	    AOEKeyRef fname
	      
      | AOEExecute_n ->
(*
	  let host = parse_attrs string_id host_attr_name in
	  let port = parse_attrs int_id port_attr_name in
*)
	    AOEExecute 
	      
      (* Tuples *)
      | AOECreateTuple_n -> 
	  let arity = parse_attrs int_of_string arity_attr_name in
	  let tnames =  make_tuple_names arity st in
	    AOECreateTuple ( tnames ) 
	      
      | AOEAccessTuple_n ->
	  let tname = parse_attrs parse_curly_uri_string tuple_name_attr_name in
	    AOEAccessTuple tname

      | AOEProject_n ->
	  let tnames =  
	    element_parser project_elem_name project_constructor "project_parser" st
	  in
	  AOEProject (Array.of_list tnames)

      | AOEProduct_n -> AOEProduct
      | AOESelect_n -> 
	  let pd = element_of_pred_desc st in
	    AOESelect pd

      | AOEJoin_n -> 	  
	  let pd = element_of_pred_desc st in
	    AOEJoin pd

      | AOELeftOuterJoin_n -> 
	  let null_index = parse_attrs parse_curly_uri_string null_index_attr_name in
	  let pd         = element_of_pred_desc st in
	    AOELeftOuterJoin (null_index, pd)

      | AOEGroupBy_n ->
	  let arity = parse_attrs int_of_string arity_attr_name in
	  let gd_list =  get_multiple arity group_desc_element_parser  st in
	    AOEGroupBy gd_list

      | AOEMap_n    -> AOEMap
      | AOEMaterializeTable_n -> AOEMaterializeTable

      | AOEMapConcat_n -> AOEMapConcat    

      | AOEOuterMapConcat_n -> 
	  let null_name = parse_attrs parse_curly_uri_string null_index_attr_name in
	    AOEOuterMapConcat null_name

      | AOEMapFromItem_n -> 
	  let vn = parse_attrs parse_curly_uri_string item_tuple_attr_name in
	    AOEMapFromItem vn

      | AOENullMap_n  ->
	  let null_name = parse_attrs parse_curly_uri_string null_index_attr_name in
	    AOENullMap null_name

      | AOEMapIndex_n  ->
	  let index_name = parse_attrs parse_curly_uri_string tuple_name_attr_name in
	    AOEMapIndex index_name 

      | AOEMapIndexStep_n  ->
	  let index_name = parse_attrs parse_curly_uri_string tuple_name_attr_name in
	    AOEMapIndexStep index_name 
	      
      | AOEMapToItem_n ->
	  AOEMapToItem	  

      | AOEConcatTuples_n ->
	  AOEConcatTuples

      | AOEDelete_n -> AOEDelete
      | AOEDetach_n -> AOEDetach
      | AOEInsert_n ->
	  let il = parse_attrs insert_location_of_string insert_location_attr_name in
	    AOEInsert il
      | AOERename_n -> AOERename
      | AOEReplace_n ->
	  let val_flag = parse_attrs value_flag_of_string value_of_flag_attr_name in
	    AOEReplace val_flag
      | AOESnap_n ->
	  let attr_modifier = parse_attrs snap_modifier_of_string snap_modifier_attr_name in
	    AOESnap attr_modifier 
      | AOESequencing_n -> 
	  let sm = parse_attrs snap_modifier_of_string snap_modifier_attr_name in 
	    AOESnap sm

      | AOEOrderBy_n ->
	  let n_attrs = parse_attrs int_of_string n_sort_criteria_attr_name in
	  let stable  = parse_attrs stable_of_string stable_attr_name in
	  let sk_esk  = get_multiple n_attrs sort_spec_element_parser st in
	  let norm_ctxt = Compile_context.norm_context_from_compile_context comp_ctxt in
	  let sigs    = Norm_overloaded.table_for_op_gt norm_ctxt in
	  let asigs   = Compile_util.compile_overloaded_table_sigs comp_ctxt sigs in
	  AOEOrderBy (stable, sk_esk, asigs)

      | AOEPrune_n ->
          let prunefield = parse_attrs parse_curly_uri_string input_prune_field_attr_name in
          let axis_name = parse_attrs axis_of_string axis_attr_name in
            AOEPrune(prunefield,axis_name)

      | AOEDistinct_n ->
          let distinctfield = parse_attrs parse_curly_uri_string input_distinct_field_attr_name in
            AOEDistinct(distinctfield)

      | AOEParse_n ->
	  let doc_name = parse_attrs string_id docname_attr_name in
	    AOEParse doc_name


(* Do not consume end element *)

let parse_subexpression parse_algop sub_type arity st =
  match sub_type, arity with
    | "NoSub", 0   -> NoSub
    | "OneSub", 1  -> OneSub  (parse_algop st)
    | "TwoSub", 2  -> 
	let a1 = (parse_algop st) in 
	let a2 = (parse_algop st) in 
	  TwoSub (a1, a2)
    | "ManySub", x -> 
	let sub_list = ref [] in
	  for i = 1 to x do
	    sub_list := (parse_algop st :: !sub_list)
	  done;
	  ManySub (Array.of_list (List.rev !sub_list))
    | _ -> raise (Query (Algebra_Parsing_Error ("Number of Subexpressions and type do not match")))


(* Below here is for the actual structural parsing *)
let parse_subexpression_constructor parse_algop attrs st =
  let arity = parse_get_attr_from_attr_list attrs int_of_string sub_expression_arity_attr_name in
  let sub_type = parse_get_attr_from_attr_list attrs string_id sub_expression_type_attr_name in
    parse_subexpression parse_algop sub_type arity st 

let parse_dep_subexpression_element_parser parse_algop st =
  if (check_opt_element dep_elem_name st) then
    element_parser dep_elem_name (parse_subexpression_constructor parse_algop) "dep_subexpression" st
  else
    NoSub

(* Independent Expressions *)
let parse_indep_subexpression_element_parser parse_algop st =
  if (check_opt_element indep_elem_name st) then
    element_parser indep_elem_name (parse_subexpression_constructor parse_algop) "indep_subexpression" st
  else
    NoSub

let rec algop_of_undecorated_element comp_ctxt st =
  let name  = parse_algop_name comp_ctxt st in    
  let dep   = parse_dep_subexpression_element_parser (algop_of_undecorated_element comp_ctxt) st in
  let indep = parse_indep_subexpression_element_parser (algop_of_undecorated_element comp_ctxt) st in
  let _ = consume_end_element st in
    logical_aalgop_mkop name indep dep None None Finfo.bogus

(* Module handlers *)

(*************)
(* Functions *)
(*************)

let fn_body_constructor parse_algop attrs st =
  let aname, terms = List.split attrs in
  let body = ref (parse_algop st) in
  let terms' = Array.of_list (List.map parse_curly_uri_string terms) in
    (* function body *)
    terms', body

let fn_body_element_parser parse_algop st =
  element_parser function_body_elem_name (fn_body_constructor parse_algop) "fn_body" st

let function_decl_constructor parse_algop attrs st =
  let name  = parse_get_attr_from_attr_list attrs parse_curly_uri_string fn_name_attr_name in
  let fname,arity,input_types, output_type = function_signature_element_parser st in
  let input_types' = List.fold_left (fun cur_types opt ->
				       match opt with 
					   None -> raise (Query (Algebra_Parsing_Error ("User function contains None for input type")))
					 | Some s -> cur_types @ [s]) [] (Array.to_list input_types) in
  let args,body_ref  = fn_body_element_parser parse_algop st in
  let fn_body = args, body_ref, (Some output_type) in

    (* build return structure *)
  let decl_desc = (name, arity), (input_types',output_type), fn_body in
    fmkalgop_function_decl decl_desc Finfo.bogus

let function_decl_element_parser parse_algop st =
  element_parser function_decl_elem_name (function_decl_constructor parse_algop) "function_decls" st

let module_function_constructor parse_algop attrs st = 
  let n_functions = parse_get_attr_from_attr_list attrs int_of_string fn_count_attr_name in
    get_multiple n_functions (function_decl_element_parser parse_algop) st

let module_function_element_parser parse_algop st =
  element_parser function_tag_elem_name (module_function_constructor parse_algop) "module functions" st

(***********************)
(* prolog_algop parser *)
(***********************)
let parse_prolog_algop_name attrs st =  
  let parse_attrs x = parse_get_attr_from_attr_list attrs x in
  let moniker_str   = parse_attrs (fun x -> x) prolog_attr_name in
  match (get_prolog_algop_moniker moniker_str) with 
    | AOEVarDeclExternal_n -> 
	let odt = parse_optasequencetype_element st in
	let vn  = parse_attrs parse_curly_uri_string variable_name_attr_name in
	  AOEVarDeclExternal (odt, vn)
    | AOEVarDecl_n ->
	let odt = parse_optasequencetype_element st in
	let vn  = parse_attrs parse_curly_uri_string variable_name_attr_name in
	  AOEVarDecl (odt, vn)
    | AOEValueIndexDecl_n ->
	let kn = parse_attrs (fun x -> x) kname_attr_name in
	  AOEValueIndexDecl kn
    | AOENameIndexDecl_n ->
	let e_name = relem_symbol_element_parser st in
	AOENameIndexDecl e_name

(* Do not consume. *)
let prolog_constructor parse_algop attrs st =
  let algop_name = parse_prolog_algop_name attrs st in
  let dep = parse_dep_subexpression_element_parser parse_algop st in
  let indep = parse_indep_subexpression_element_parser parse_algop st in
    logical_aalgop_decl_mkop algop_name indep dep Finfo.bogus

let prolog_element_parser parse_algop st =
  element_parser prolog_elem_name (prolog_constructor parse_algop) "prolog" st

(********)
(* Vars *)
(********)

let module_var_constructor parse_algop attrs st = 
  let n_vars = parse_get_attr_from_attr_list attrs int_of_string prolog_var_count_attr_name in
    get_multiple n_vars (prolog_element_parser parse_algop) st 

let module_var_element_parser parse_algop st = 
  element_parser prolog_vars_elem_name (module_var_constructor parse_algop) "var declaration" st

(********)
(* Keys *)
(********)
let module_index_definitions_constructor parse_algop attrs st = 
  let count = parse_get_attr_from_attr_list attrs int_of_string prolog_index_count_attr_name in
    get_multiple count (prolog_element_parser parse_algop) st 
    

let module_index_definitions_element_parser parse_algop st =
  element_parser prolog_indices_elem_name (module_index_definitions_constructor parse_algop) "index definitions" st

(**********)
(* Prolog *)
(**********)
let prolog_constructor parse_algop attrs st =
  let fn_decls  = module_function_element_parser parse_algop st in
  let var_decls = module_var_element_parser parse_algop st in
  let index_defs  = module_index_definitions_element_parser parse_algop st in
    fmkalgop_prolog fn_decls var_decls index_defs 

let prolog_element_parser parse_algop st =
  element_parser algop_prolog_decl_elem_name (prolog_constructor parse_algop) "whole prolog" st

(**************)
(* Statements *)
(**************)

let pass_element_parser elem_name parse_algop emsg st =
  element_parser elem_name (fun attrs st -> parse_algop st) emsg st 

let expression_element_parser parse_algop st = 
  pass_element_parser expression_elem_name parse_algop "expression parser" st 


let module_statements_constructor parse_algop attrs st =
  let n_statements = parse_get_attr_from_attr_list attrs int_of_string number_of_statements_attr_name in
    get_multiple n_statements (expression_element_parser parse_algop) st
  
let module_statements_element_parser parse_algop st =
  element_parser statement_declaration_elem_name (module_statements_constructor parse_algop) "statements" st

(***********)
(* Modules *)
(***********)

let parse_algebra_module_constructor parse_algop attrs st =
  let prolog = prolog_element_parser parse_algop st in
  let stmts  = module_statements_element_parser parse_algop st in
    fmkalgop_xmodule prolog stmts

let algebra_module_element_parser parse_algop st = 
  element_parser algop_module_decl_elem_name (parse_algebra_module_constructor parse_algop) "algebra module" st

(******************************************************************)
(***************** BELOW HERE ARE EXPOSED FUNCTIONS ***************)
(******************************************************************)

let parse_module_logical_plan_from_xml_stream comp_ctxt st =
  algebra_module_element_parser (algop_of_undecorated_element comp_ctxt) st


let parse_expression_logical_plan_from_xml_stream comp_ctxt st =
  expression_element_parser (algop_of_undecorated_element comp_ctxt) st


(******************************************************************)
(***************** BELOW HERE ARE TEST FUNCTIONS ******************)
(******************************************************************)


(* Strip off the document event and then parse it *)
let plan_of_serialized_stream parser comp_ctxt the_stream =
  let st = Cursor.stream_of_cursor
      (Streaming_ops.resolve_xml_stream 
	 the_stream) in
  
  match Stream.peek st with
  | Some event ->
      begin
	match event.rse_desc with 
	  (RSAX_startDocument _ ) ->
	    Stream.junk st;
	    parser comp_ctxt st
	| _ -> 
	    raise (Query (Algebra_Parsing_Error ("No document start for serialized plan")))
      end
  | _ -> 
      raise (Query (Algebra_Parsing_Error ("No document start for serialized plan")))

let parse_serialized_module_plan proc_ctxt (comp_ctxt:Logical_algebra_types.logical_alg_compile_context) input =
  let nsenv                = add_all_ns (Processing_context.get_external_nsenv proc_ctxt) algebra_bindings in
  let _                    = Processing_context.set_namespace_env proc_ctxt nsenv in
  let (dtdopt, the_stream) = Streaming_parse.open_xml_stream_from_io input in   
  let parsed_module        = plan_of_serialized_stream parse_module_logical_plan_from_xml_stream comp_ctxt the_stream in
  let _                    = annotate_algebraic_module parsed_module in
  let comp_ctxt            = Compile_context.update_compile_context_from_module comp_ctxt parsed_module in 
    (parsed_module,comp_ctxt)

let parse_serialized_expression_plan proc_ctxt (comp_ctxt:Logical_algebra_types.logical_alg_compile_context) input =
  let nsenv                = add_all_ns (Processing_context.get_external_nsenv proc_ctxt) algebra_bindings in
  let _                    = Processing_context.set_namespace_env proc_ctxt nsenv in
  let (dtdopt, the_stream) = Streaming_parse.open_xml_stream_from_io input in   
  let parsed_expr          = plan_of_serialized_stream parse_expression_logical_plan_from_xml_stream comp_ctxt the_stream in
  let _                    = annotate_algebraic_expression parsed_expr in
  (parsed_expr)

