(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies and AT&T Labs.            *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* Module: Algebgra_print_common
   Description:

   This module contains common data for the parser and the printer

   NOTE: The parser is not currently hooked into the system
   NOTE: This code does not currently support the serialization of updates

*)
open Error

open Dm_util
open Small_stream_ast
open Streaming_types
open Namespace_names

open Print_common

open Xquery_common_ast
open Xquery_core_ast
open Xquery_algebra_ast 

exception AlgebraParseError of string
(* This is what we would like but, it causes 
   serialization to act strangely *)

let algebra_prefix = NSPrefix "alg"
let algebra_uri = NSUri "http://www.galaxquery.org/algebra/2006"

(*let algebra_prefix = Namespace_builtin.glx_prefix
let algebra_uri    = Namespace_builtin.glx_uri*)
let algebra_bindings = [(algebra_prefix, algebra_uri)]

let alg_elem x      = Namespace_symbols.relem_symbol x
let alg_attr x      = Namespace_symbols.rattr_symbol x
 
(* These should probably be merged into a util area *)
(*
let elem_name name =  (algebra_prefix, (algebra_uri, name))
let attr_name name =  (algebra_prefix, (algebra_uri, name))
*)

let elem_name name =  (algebra_prefix, algebra_uri, name)
let attr_name name =  (algebra_prefix, algebra_uri, name)


let string_of_rattr_symbol rsya = Namespace_symbols.rattr_prefix_string rsya 
let string_of_relem_symbol rsye = Namespace_symbols.relem_prefix_string rsye
let string_of_rtype_symbol rsyg = Namespace_symbols.rtype_prefix_string rsyg



(* constants for element and attribute names *)
let function_tag_elem_name = elem_name "Functions"
let function_body_elem_name = elem_name "function_body"

let input_datamodel_elem_name = elem_name "inputdatamodel"
let output_datamodel_elem_name = elem_name "outputdatamodel"
let datamodel_signature = elem_name "DatamodelSignature"


let dep_elem_name   = elem_name "dep"
let indep_elem_name = elem_name "indep"

let sub_expression_arity_attr_name = attr_name "subarity"
let sub_expression_type_attr_name  = attr_name "subtype"

let input_type_elem_name = elem_name "inputtype"
let output_type_elem_name = elem_name "outtype"
let function_signature_elem_name = elem_name "functionsignature"

let out_type_attr_name = attr_name "outtype"
let arg_count_attr_name = attr_name "argcount"
let index_attr_name = attr_name "index"
let datamodel_attr_name = attr_name "datamodel"

let input_prune_field_attr_name = attr_name "prunefield"
let input_distinct_field_attr_name = attr_name "distinctfield"

(* Put in a sample format of the prolog *)

let no_input   = "NoInput"
let one_input  = "OneInput"
let two_input  = "TwoInput"
let many_input = "ManyInput"

(* Axis Related *)
let node_test_elem_name = elem_name "nodetest"

let name_test_attr_name = attr_name "nametest"
let name_test_elem_name = elem_name "nametest"

let kind_test_elem_name = elem_name "kindtest"
let kind_test_attr_name = attr_name "kind"

let pi_kind_test_elem_name = elem_name "pikindtest"
let pi_kind_attr_name = attr_name "piarg"
let axis_attr_name = attr_name "axis"

(* Typeswitch Element *)
let ts_case_elem_name = elem_name "typeswitchcase"
let ts_default_elem_name = elem_name "typeswitchdefault"
let ts_num_branches_attr_name = attr_name "numbranches"
(* Tuple element *)
let tuple_slot_elem_name = elem_name "tuple"

let var_attr_name = attr_name "variable"
let type_attr_name = attr_name "type"

let vname_attr_name = attr_name "vname"
let stablekind_attr_name = attr_name "stablekind"
let sortkind_attr_name = attr_name "sortkind"
let emptysortkind_attr_name = attr_name "empty_sortkind"
let optdatatype_attr_name = attr_name "optdatatype"
let pos_attr_name = attr_name "pos"
let usage_attr_name = attr_name "usage"
let value_attr_name = attr_name "value"
let atomic_type_attr_name = attr_name "atomictype"
let target_attr_name = attr_name "target"
let content_attr_name = attr_name "content"
let attr_name_attr_name = attr_name "attr_name"
let elem_name_attr_name = attr_name "elem_name"
let fn_name_attr_name = attr_name "fname"
let arity_attr_name = attr_name "arity"
let kname_attr_name = attr_name "kn"
let host_attr_name = attr_name "host"
let port_attr_name = attr_name "port"
let tuple_name_attr_name = attr_name "tuple"
let occurrence_attr_name = attr_name "occurrence"


let prefix_attr_name = attr_name "prefix"
let rqname_attr_name = attr_name "rqname"

let relem_elem_name = elem_name "relem"
let rattr_elem_name = elem_name "rattr"
let rtype_elem_name = elem_name "rtype"

let n_sort_criteria_attr_name = attr_name "NumberSortCriteria"
let stable_attr_name = attr_name "stable"
let sort_spec_elem_name = elem_name "sortspec"
let empty_sort_kind_attr_name = attr_name "emptysortkind"
let sort_kind_attr_name = attr_name "sortkind"

let item_tuple_attr_name = attr_name "itemtuplevar"

let project_elem_name = elem_name "project"
let project_name_elem_name = elem_name "projectfield"
let project_name_attr_name = attr_name "name"
let value_of_flag_attr_name = attr_name "valueofflag"
let insert_location_attr_name = attr_name "insertlocation"
let snap_modifier_attr_name   = attr_name "snapmodifier"
let docname_attr_name   = attr_name "docname"

let asequencetype_elem_name = elem_name "asequencetype"
(* ait element names *)
let ait_attributetype_elem_name = elem_name "aitelementname"
let ait_elementtype_elem_name = elem_name "aitelementtype"
let ait_schemaattribute_elem_name = elem_name "aitschemaattribute"
let ait_schemaelementtype_elem_name = elem_name "aitschemaelementtype"
let ait_kindtest_elem_name = elem_name "aitkindtest"
let ait_typeref_elem_name = elem_name "aittyperef"
let ait_node_elem_name = elem_name "aitnode"
let ait_item_elem_name = elem_name "aititem"
let ait_numeric_elem_name = elem_name "aitnumeric"
let ait_text_elem_name = elem_name "aittext"
let ait_comment_elem_name = elem_name "aitcomment"
let ait_processing_instruction_elem_name = elem_name "aitpi"
let ait_empty_elem_name = elem_name "aitempty"
let ait_anyatomic_elem_name = elem_name "aitanyatomic"
let ait_atomic_elem_name = elem_name "aitatomic"
let ait_document_elem_name = elem_name "aitdoc"

(* ait attribute names *)
let schema_attr_name_attr_name = attr_name "schemaattr"
let schema_elem_name_attr_name = attr_name "schemaelem"
let type_ref_attr_name = attr_name "typeref"
let pi_arg_attr_name = attr_name "piarg"
let atomic_type_attr_name = attr_name "atomictype"

let function_decl_elem_name = elem_name "functiondecl"
let fn_count_attr_name = attr_name "countoffunctions"

let variable_declarations_elem_name = elem_name "variabledeclarations"
let variable_decl_elem_name         = elem_name "variabledecl"
let variable_count_attr_name        = attr_name "numberofvariables"
let variable_name_attr_name         = attr_name "variablename"

let null_index_attr_name = attr_name "NullIndex"
let input_dot_attr_name  = attr_name "InputDot"
let output_dot_attr_name = attr_name "OutputDot"

let prolog_elem_name  = elem_name "prolog_element"
let index_definitions_elem_name = elem_name "indexdefinitions"
let index_def_elem_name = elem_name "indexdef"
let op_1_elem_name    = elem_name "op1"
let op_2_elem_name    = elem_name "op2"
let kname_attr_name   = attr_name "indexname"
let indices_count_attr_name = attr_name "nindices"

let statement_declaration_elem_name = elem_name "Statements"
let expression_elem_name = elem_name "Expression"
let number_of_statements_attr_name = attr_name "nstatements"

let algop_module_decl_elem_name = elem_name "ModuleDeclaration"
(**********************)
(* Compile Annotation *)
(**********************)

let compile_annotation_elem_name  = elem_name "CompileAannotationName"
let accessed_fields_elem_name     = elem_name "AccessedFields"
let returned_fields_elem_name     = elem_name "ReturnedFields"
let free_variables_elem_name      = elem_name "FreeVariables"
    (* Attributes *)
let num_accessed_fields_attr_name = attr_name "NAccessedFields"
let num_returned_fields_attr_name = attr_name "NReturnedFields"
let num_free_variables_attr_name  = attr_name "NFreeVariables"
(**************)
(* Overloaded *)
(**************)

let overloaded_call_elem_name    = elem_name "OverloadedCallTable"

(**********************)
(* Grouping operation *)
(**********************)
let vname_elem_name          = elem_name "variablename"
(* group_desc *)
let gbn_elem_name            = elem_name "GroupByName"
let induced_elem_name        = elem_name "InducedGroups"
let mbv_elem_name            = elem_name "MustBeValid"
let agg_elem_name            = elem_name "AggregateName"
let group_desc_elem_name     = elem_name "GroupDesc"
let distinct_value_attr_name = attr_name "distinct"

(**************)
(* Predicates *)
(**************)
let simple_conjunct_elem_name       = elem_name "SimpleConjunctElem"
let complex_conjunct_elem_name      = elem_name "ComplexConjunctElem"
let disjunct_elem_name              = elem_name "DisjunctElem"
let simple_conjunct_start_attr_name = attr_name "SimpleConjunctStart"
let simple_conjunct_end_attr_name   = attr_name "SimpleConjunctEnd"

(**********************)
(* TupleTreePatterns  *)
(**********************)
let ttp_name            = elem_name "TupleTreePattern"
let twig_node_name      = elem_name "TwigNode"
let in_field_attr_name = attr_name  "in_field"
let restore_attr_name   = attr_name "restore"
let out_field_attr_name = attr_name "out_field"
let child_twig_name     = elem_name "ChildTwig"
let pred_twig_name      = elem_name "PredTwig"

(**********)
(* Prolog *)
(**********)
let prolog_var_decl_external_name = "AOEVarDeclExternal"
let prolog_var_decl_name          = "AOEVarDecl"
let prolog_value_index_decl_name  = "AOEValueIndexDecl"
let prolog_name_index_decl_name   = "AOENameIndexDecl"

let prolog_var_count_attr_name    = attr_name "count"
let prolog_index_count_attr_name  = attr_name "count"
let prolog_vars_elem_name         = elem_name "var_decls"
let prolog_indices_elem_name      = elem_name "index_defs"
let prolog_attr_name              = attr_name "prolog_op"

let algop_prolog_decl_elem_name   = elem_name "Prolog"
type prolog_algop_moniker = 
    AOEVarDeclExternal_n 
  | AOEVarDecl_n
  | AOEValueIndexDecl_n
  | AOENameIndexDecl_n

let get_prolog_algop_moniker str = 
  match str with
    | "AOEVarDeclExternal" -> AOEVarDeclExternal_n
    | "AOEVarDecl" -> AOEVarDecl_n
    | "AOEValueIndexDecl" -> AOEValueIndexDecl_n
    | "AOENameIndexDecl" -> AOENameIndexDecl_n
    | _ -> raise (Query (Malformed_Algebra_Expr ("Unknown prolog algop moniker: " ^str)))

let attribute_of_prolog_algop_name n =
  let name = 
    match n with
    | AOEVarDeclExternal (ocdt, vn) ->
	prolog_var_decl_external_name
    | AOEVarDecl( ocdt, vn ) ->
	prolog_var_decl_name
    | AOEValueIndexDecl kn -> 
	prolog_value_index_decl_name
    | AOENameIndexDecl kn -> 
	prolog_name_index_decl_name
  in
  prolog_attr_name, name 

type algop_moniker = 
  | AOELet_n
  | AOEIf_n
  | AOETypeswitch_n
  | AOEVar_n
  | AOEScalar_n
  | AOESeq_n
  | AOEEmpty_n
  | AOEDocument_n
  | AOEPI_n
  | AOEPIComputed_n
  | AOEComment_n
  | AOECommentComputed_n
  | AOEText_n
  | AOETextComputed_n
  | AOEElem_n
  | AOEAnyElem_n
  | AOEAttr_n
  | AOEAnyAttr_n
  | AOEError_n
  | AOETreat_n
  | AOEValidate_n
  | AOECast_n
  | AOECastable_n
  | AOESome_n 
  | AOEEvery_n

  | AOETreeJoin_n
  | AOETupleTreePattern_n
  | AOEProject_n
  | AOEPrune_n
  | AOEDistinct_n

  (* Function Calls *)
  | AOECallBuiltIn_n
  | AOECallOverloaded_n 
  | AOECallUserDefined_n
  | AOEConvertSimple_n
  | AOEPromoteNumeric_n
  | AOEUnsafePromoteNumeric_n

  | AOEKeyRef_n
  | AOEExecute_n

  (* Tuples *)
  | AOECreateTuple_n          (* Creates a new tuple *)
  | AOEAccessTuple_n          (* Accesses a new tuple *)
  | AOEOrderBy_n
  | AOEMaterializeTable_n
  | AOEProduct_n
  | AOESelect_n
  | AOEJoin_n
  | AOELeftOuterJoin_n
  | AOEConcatTuples_n         (* Tuple concatenation *)
  | AOEMapFromItem_n          (* Item to tuple iteration *)
  | AOEMapToItem_n            (* Tuple to item iteration *)
  | AOEMap_n                  (* Tuple iteration *)
  | AOEMapIndex_n
  | AOEMapIndexStep_n
  | AOEMapConcat_n
  | AOEOuterMapConcat_n
  | AOEInputTuple_n
  | AOEGroupBy_n
  | AOENullMap_n

  (* Updates *)
  | AOESnap_n
  | AOEDelete_n
  | AOEDetach_n
  | AOEInsert_n
  | AOERename_n
  | AOEReplace_n
  | AOESequencing_n 

  (* Added because it didn't exist already *)
  | AOEParse_n

(* 
   This is here so we don't have to expose elem_name and attr_name constructors
   to force all the names to be registered between the parser and printer *)

(* This is here so we don't have to expose elem_name and attr_name
   constructors to force all the names to be registered between the
   parser and printer *)

let make_input_attr_name index = attr_name  ("input" ^ (string_of_int index))
    
let moniker_list = 
  List.map (fun (x,y) ->
	      (alg_elem (elem_name x)), y)
  [  ("AOELet"), AOELet_n;
     ("AOEIf") , AOEIf_n;
     ("AOETypeSwitch"), AOETypeswitch_n;
     ("AOEVar")   , AOEVar_n;
     ("AOEScalar"), AOEScalar_n;
     ("AOESeq")   , AOESeq_n;
     ("AOEEmpty") , AOEEmpty_n;
     ("AOEDocument"), AOEDocument_n;
     ("AOEPI"), AOEPI_n;
     ("AOEPIComputed"), AOEPIComputed_n;
     ("AOEComment"), AOEComment_n;
     ("AOECommentComputed"), AOECommentComputed_n;
     ("AOEText"), AOEText_n;
     ("AOETextComputed"), AOETextComputed_n;
     ("AOEElem"), AOEElem_n;
     ("AOEAnyElem"), AOEAnyElem_n;
     ("AOEAttr"), AOEAttr_n;
     ("AOEAnyAttr"), AOEAnyAttr_n;
     ("AOEError") , AOEError_n;
     ("AOETreat") , AOETreat_n;
     ("AOEValidate"), AOEValidate_n;
     ("AOECast")  , AOECast_n;
     ("AOECastable"), AOECastable_n;
     ("AOESome"), AOESome_n ;
     ("AOEEvery"), AOEEvery_n;

     ("AOETreeJoin"), AOETreeJoin_n;
     ("AOETupleTreePattern"), AOETupleTreePattern_n;
     ("AOEPrune"), AOEPrune_n;
     ("AOEDistinct"), AOEDistinct_n;
	
     (* Function Calls *)
     ("AOECallBuiltIn"), AOECallBuiltIn_n;
     ("AOECallOverloaded"), AOECallOverloaded_n ;
     ("AOECallUserDefined"), AOECallUserDefined_n;
     ("AOEConvertSimple"), AOEConvertSimple_n;
     ("AOEPromoteNumeric"), AOEPromoteNumeric_n;
     ("AOEUnsafePromoteNumeric"), AOEUnsafePromoteNumeric_n;
     ("AOEKeyRef")  , AOEKeyRef_n;
     ("AOEExecute")  , AOEExecute_n;

    (* Tuples *)
     ("AOECreateTuple"), AOECreateTuple_n;         (* Creates a new tuple *)
     ("AOEAccessTuple"), AOEAccessTuple_n;            (* Accesses a new tuple *)
     ("AOEProject"),  AOEProject_n;
     ("AOEOrderBy"),  AOEOrderBy_n;
     ("AOEMaterializeTable"), AOEMaterializeTable_n;
     ("AOEProduct"), AOEProduct_n;
     ("AOESelect"), AOESelect_n;     
     ("AOEJoin", AOEJoin_n);
     ("AOELeftOuterJoin", AOELeftOuterJoin_n);
     ("AOEConcatTuples", AOEConcatTuples_n);
     ("AOEMapFromItem", AOEMapFromItem_n);
     ("AOEMapToItem",  AOEMapToItem_n );
     ("AOEMap" , AOEMap_n);
     ("AOEMapIndex", AOEMapIndex_n);
     ("AOEMapIndexStep", AOEMapIndexStep_n);
     ("AOEMapConcat", AOEMapConcat_n);
     ("AOEOuterMapConcat", AOEOuterMapConcat_n);
     ("AOEInputTuple", AOEInputTuple_n);
     ("AOEGroupBy", AOEGroupBy_n);
     ("AOENullMap", AOENullMap_n);

     (* Updates *)     
     ("AOESnap"), AOESnap_n;
     ("AOEDelete"), AOEDelete_n;
     ("AOEDetach"), AOEDetach_n;
     ("AOEInsert"),AOEInsert_n;
     ("AOERename"), AOERename_n;
     ("AOEReplace"), AOEReplace_n;
     ("AOESequencing"), AOESequencing_n;

     (* Added because it didn't exist for some reason - Doug *)
     ("AOEParse"), AOEParse_n;
]                  (* Iteration over a sequence of tuples *)
  ;;
let moniker_hash = Hashtbl.create 167 ;;
let _ = 
  try
    List.iter (fun (x,y) -> Hashtbl.add moniker_hash x y) moniker_list
  with
  | e ->
      begin
	eprintf_error "  " e;
	Format.fprintf (!Conf.glx_err_formatter) "@."
      end

let get_moniker_of_serialized_moniker alg_str = 
  try
    Hashtbl.find moniker_hash alg_str
  with Not_found ->
    raise (AlgebraParseError ("* Unknown Moniker Error: "
			      ^ (Namespace_symbols.relem_string alg_str)))

let __debug_verify_moniker = true
let verify_moniker m =
  let m' = alg_elem (elem_name m) in
    begin
      if (not (Hashtbl.mem moniker_hash m')) then	  
	  raise (Query (Malformed_Algebra_Expr ("Moniker not present: " ^ m)))
    end
    

let elem_create m = 
  if __debug_verify_moniker then
    verify_moniker m
  ;
  elem_name m 


let get_serializable_moniker_of_algop_name alg_name = 
  match alg_name with
  | AOELet _  -> elem_create "AOELet"
  | AOEIf     -> elem_create "AOEIf"
  | AOETypeswitch _ -> elem_create "AOETypeSwitch"
  | AOEVar _   -> elem_create "AOEVar"
  | AOEScalar _ -> elem_create "AOEScalar"
  | AOESeq    -> elem_create "AOESeq"
  | AOEEmpty  -> elem_create "AOEEmpty"
  | AOEDocument -> elem_create "AOEDocument"
  | AOEPI _ -> elem_create "AOEPI"
  | AOEPIComputed -> elem_create "AOEPIComputed"
  | AOEComment _ -> elem_create "AOEComment"
  | AOECommentComputed -> elem_create "AOECommentComputed"
  | AOEText _ -> elem_create "AOEText"
  | AOETextComputed -> elem_create "AOETextComputed"
  | AOEElem _ -> elem_create "AOEElem"
  | AOEAnyElem _ -> elem_create "AOEAnyElem"
  | AOEAttr _ -> elem_create "AOEAttr"
  | AOEAnyAttr _ -> elem_create "AOEAnyAttr"
  | AOEError  -> elem_create "AOEError"
  | AOETreat _  -> elem_create "AOETreat"
  | AOEValidate _ -> elem_create "AOEValidate"
  | AOECast  _  -> elem_create "AOECast"
  | AOECastable _ -> elem_create "AOECastable"
  | AOESome _ -> elem_create "AOESome" 
  | AOEEvery _ -> elem_create "AOEEvery"

	(* Function Calls *)
  | AOECallBuiltIn _ -> elem_create "AOECallBuiltIn"
  | AOECallOverloaded _ -> elem_create "AOECallOverloaded" 
  | AOECallUserDefined _ -> elem_create "AOECallUserDefined"
  | AOEUnsafePromoteNumeric _ -> elem_create "AOEUnsafePromoteNumeric"
  | AOEPromoteNumeric _ -> elem_create "AOEPromoteNumeric"
  | AOEConvertSimple _ -> elem_create "AOEConvertSimple"

  | AOEKeyRef _   -> elem_create "AOEKeyRef"
  | AOEExecute   -> elem_create "AOEExecute"

	(* Tuples *)
  | AOEMapConcat       -> elem_create "AOEMapConcat"
  | AOEOuterMapConcat _ -> elem_create "AOEOuterMapConcat"
  | AOECreateTuple _ -> elem_create "AOECreateTuple"    (* Creates a new tuple *)
  | AOEAccessTuple _ -> elem_create "AOEAccessTuple"    (* Accesses a new tuple *)
  | AOEMap -> elem_create "AOEMap"
  | AOEMapToItem -> elem_create "AOEMapToItem"
  | AOEMapFromItem _ -> elem_create "AOEMapFromItem"
  | AOEMapIndex _ -> elem_create "AOEMapIndex"
  | AOEMapIndexStep _ -> elem_create "AOEMapIndexStep"
  | AOEInputTuple -> elem_create "AOEInputTuple"
  | AOEConcatTuples -> elem_create "AOEConcatTuples"
  | AOEProduct -> elem_create "AOEProduct"
  | AOESelect _ -> elem_create "AOESelect"
  | AOEJoin _ -> elem_create "AOEJoin"
  | AOEOrderBy _ -> elem_create "AOEOrderBy"
  | AOECopy    -> elem_create "AOECopy"
  | AOEDelete    -> elem_create "AOEDelete"
  | AOEDetach    -> elem_create "AOEDetach"
  | AOEInsert _ -> elem_create "AOEInsert"
  | AOERename -> elem_create "AOERename"
  | AOEReplace _ -> elem_create "AOEReplace"
  | AOESnap _ -> elem_create "AOESnap"
  | AOEMaterializeTable -> elem_create "AOEMaterializeTable"
  | AOEGroupBy _  -> elem_create "AOEGroupBy"
  | AOELeftOuterJoin _ -> elem_create "AOELeftOuterJoin"
  | AOENullMap _ -> elem_create "AOENullMap"
  | AOETreeJoin _ -> elem_create "AOETreeJoin"
  | AOETupleTreePattern _ -> elem_create "AOETupleTreePattern"
  | AOEParse _ -> elem_create "AOEParse"
  | AOEPrune _ -> elem_create "AOEPrune"
  | AOEDistinct _ -> elem_create "AOEDistinct"
  | AOEProject _ -> elem_create "AOEProject"
