(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies and AT&T Labs.            *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* Module: Planio_print_utils
   Description:
   This module contains some util functions for algebra printers and parsers
   
*)

open Error
open Occurrence

open Namespace_names
open Namespace_symbols

open Xquery_algebra_ast
open Xquery_physical_type_ast
open Xquery_common_ast
open Xquery_core_ast (* CITxxx *)

open Streaming_types
open Small_stream_ast

open Namespace_names
open Datatypes

open Print_common
open Planio_print_common

(* This seems a bit hackish but, since we are in a well known attribute we are ok *)
let xml_string_of_prefix pref = 
  match pref with
    | NSDefaultElementPrefix  -> "NSDefaultElementPrefix"
    | NSDefaultFunctionPrefix -> "NSDefaultFunctionPrefix"
    | NSWildcardPrefix        -> "NSWildcardPrefix"
    | NSPrefix(pref_name)     -> pref_name (* This is the hackish portion since it won't give warning...*)

let xml_prefix_of_string str  = 
  match str with
    | "NSDefaultElementPrefix"  -> NSDefaultElementPrefix
    | "NSDefaultFunctionPrefix" -> NSDefaultFunctionPrefix
    | "NSWildcardPrefix"        -> NSWildcardPrefix
    | x                         -> NSPrefix x 
      
let string_of_optint oi = 
  match oi with
      None -> "None"
    | Some i' -> string_of_int i'

let optint_of_string str =
  match str with
      "None" -> None
    | x -> Some (int_of_string str)

(* SortKind *)
let string_of_sortkind sk = 
  match sk with
    | Ascending -> "Ascending"
    | Descending -> "Descending"

let sortkind_of_string str = 
  match str with
    | "Ascending" -> Ascending
    | "Descending" -> Descending
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown SortKind: " ^ str)))

(* Stablekind *)
let string_of_stablekind stable =
  match stable with
    | Stable -> "Stable"
    | NonStable -> "NonStable"

let stablekind_of_string str =
  match str with 
    | "Stable" -> Stable
    | "NonStable" -> NonStable
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown StableKind: " ^ str)))

(* Empty Sort Kind *)
let string_of_emptysortkind esk = 
  match esk with
    | EmptyGreatest -> "EmptyGreatest"
    | EmptyLeast -> "EmptyLeast"

let emptysortkind_of_string str = 
  match str with
    | "EmptyGreatest" -> EmptyGreatest 
    | "EmptyLeast" -> EmptyLeast 
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown EmptySortKind: " ^ str)))

let axis_of_string str =
  match str with
  | "ancestor" -> Ancestor
  | "ancestor-or-self" -> Ancestor_or_self 
  | "attribute" -> Attribute
  | "child" -> Child
  | "descendant" -> Descendant     
  | "descendant-or-self" -> Descendant_or_self
  | "following-sibling" -> Following_sibling
  | "preceding-sibling" -> Preceding_sibling
  | "parent" -> Parent
  | "self" -> Self
  | _ -> raise (Query (Algebra_Parsing_Error ("Unknown axis expression in algebra parsing")))


(*
let string_of_physical_type pt = 
  match pt with    
    | PT_xml_value      -> "Xml Value"
    | PT_sax            -> "Sax"   (* Stream of SAX events *)
    | PT_item_cursor    -> "Item Cursor"
    | PT_item_list      -> "Item List"
    | PT_tuple_cursor   -> "Tuple Cursor"
    | PT_physical_value -> "Physical Value" 

(* Should be moved to util  *)
let physical_type_of_string str = 
  match str with 
    | "Xml Value"      -> PT_xml_value 
    | "Sax"            -> PT_sax   (* Stream of SAX events *)
    | "Item Cursor"    -> PT_item_cursor
    | "Item List"      -> PT_item_list      
    | "Tuple Cursor"   -> PT_tuple_cursor
    | "Physical Value" -> PT_physical_value  
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown Physical Type " ^ str)))
*)

let string_of_physical_dom_type mt =
  match mt with 
  | PT_CursorSeq -> "item_cursor"
  | PT_ListSeq -> "item_list"

let string_of_physical_sax_type st =
  match st with 
  | PT_Stream -> "sax_stream"
  | PT_Discarded -> "sax_discarded"

let string_of_physical_xml_type pt =
  match pt with 
  | PT_Dom mt -> string_of_physical_dom_type mt
  | PT_Sax st -> string_of_physical_sax_type st

let physical_type_of_xml_type_string xtstr = 
  match xtstr with 
  | "item_cursor" -> (PT_Dom PT_CursorSeq)
  | "item_list" -> (PT_Dom PT_ListSeq)
  | "sax_stream" -> (PT_Sax PT_Stream)
  | "sax_discarded" -> (PT_Sax PT_Discarded)
  | _ -> raise (Query (Algebra_Parsing_Error ("Unknown Physical Type " ^ xtstr)))

let physical_type_of_tuple_type_string tupstr = 
  List.map (fun vxt -> 
    let (v, xtstr) = Gmisc.split_left_on_char vxt ':' 
    in (Namespace_names.parse_curly_uri_string v, physical_type_of_xml_type_string xtstr) 
    ) (Gmisc.split_on_char tupstr ';') 

let string_of_physical_tuple_type tups = 
  String.concat ";" (List.map (fun (v, xt) -> "("^(Namespace_names.curly_uri_string_of_rqname v)^":"^(string_of_physical_xml_type xt)^")") tups)

let string_of_physical_type pt =
  match pt with 
  | PT_XML xt -> "xml("^(string_of_physical_xml_type xt)^")"
  | PT_Table tt -> "tup("^(string_of_physical_tuple_type tt)^")"

(* Should be moved to util  *)
let physical_type_of_string str = 
  match String.sub str 0 4 with 
    | "xml("  -> PT_XML(physical_type_of_xml_type_string (String.sub str 4 ((String.length str) - 5)))
    | "tup("  -> PT_Table(physical_type_of_tuple_type_string (String.sub str 4 ((String.length str) - 5)))
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown Physical Type " ^ str)))

let get_cardinality str =
  match str with 
    | "NoInput" -> 0
    | "OneInput" -> 1
    | "TwoInput" -> 2
    | "ManyInput" -> 1 (* Many means it is replicated *)
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown Input Cardinality")))

let get_subexpr_arity sub =
  match sub with
    | NoSub -> 0
    | OneSub _ -> 1
    | TwoSub _ -> 2
    | ManySub x -> Array.length x (* must serialize along with the type (i.e. Many) *)

let get_subexpr_type sub =
  match sub with
      NoSub -> "NoSub"
    | OneSub _ -> "OneSub"
    | TwoSub _ -> "TwoSub"
    | ManySub x -> "ManySub"
	  

let atomic_type_of_string str =
  let rqname = parse_curly_uri_string str in 
  let type_sym = Namespace_symbols.rtype_symbol rqname in
  try
    Datatypes_util.lookup_bltin_type type_sym
  with
    | _ -> raise (Query (Algebra_Parsing_Error ("Unknown Atomic type: " ^ str)))
