(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies and AT&T Labs.            *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* Module: Algebgra_print_xml
   Description:
    This module prints serializes undecorated algebraic to XML

*)

(* Algbera printing module *)
open Error
open Occurrence

open Namespace_names
open Namespace_symbols

open Dm_util

open Xquery_common_ast

open Xquery_algebra_ast
open Xquery_algebra_ast_util
open Xquery_algebra_ast_annotation_util

open Small_stream_ast
open Streaming_types
open Print_common

open Planio_print_common
open Planio_print_util
open Planio_print_xml_util

    
(* We need both of these because there 
   are things that have type ManySub but have 2 elements (for example the serialization of some builtins)
*)

(* 
   <nodetest>
     <nametest name='..'/>
   | <kindtest kind='..'>
       ?<processinginstruction kind=>
     </kindtest>
   </nodetest> *)


let attr_of_stable stable =
  stable_attr_name, (
  match stable with
      Stable ->  "Stable"
    | NonStable -> "Unstable"
  )
let attr_of_sortkind sk =
  let sk_str  = string_of_sortkind sk in
    sort_kind_attr_name, sk_str 

let attr_of_emptysortkind esk =
  let esk_str = string_of_emptysortkind esk in
    empty_sort_kind_attr_name, esk_str 
    
let element_of_sort_kinds (sk,esk) =
  let sk_attr  = attr_of_sortkind sk in
  let esk_attr = attr_of_emptysortkind esk in
    construct_element sort_spec_elem_name (sk_attr :: esk_attr :: []) []


let make_node_test_element nt =
  let child_type =
   match nt with
    | APNameTest qn ->
	let attr = [name_test_attr_name, curly_uri_string_of_rqname (relem_name qn)] in
	construct_element  name_test_elem_name attr []
    | APNodeKindTest ckt ->
	construct_element kind_test_elem_name [] [element_of_kind_test ckt]
  in
    construct_element node_test_elem_name [] [child_type]

let elem_of_project_name names =
  let values = 
    Array.map 
      (fun p -> 
	construct_element project_name_elem_name 
                          [project_name_attr_name, (curly_uri_string_of_rqname p)] []) 
      names in
    construct_element project_elem_name [] (Array.to_list values)

let attr_of_insert_location loc = 
  let loc_str =
    match loc with 
      | AOUAsLastInto -> "LastInto"
      | AOUAsFirstInto -> "FirstInto"
      | AOUInto -> "Into"
      | AOUAfter -> "After"
      | AOUBefore -> "Before"
  in
    insert_location_attr_name, loc_str

let attr_of_snap_modifier sm = 
  let sm_str = 
    match sm with
      | Snap_Ordered_Deterministic   -> "ordered deterministic"
      | Snap_Unordered_Deterministic -> "unordered deterministic"
      | Snap_Nondeterministic   -> "nondeterministic"
  in
    snap_modifier_attr_name, sm_str

let attr_of_value_of_flag avf =
  let avf_str = 
    match avf with
      | Normal_Replace -> "Normal"
      | Value_Of_Replace -> "Value of"
  in
    value_of_flag_attr_name, avf_str

let make_typeswitch_element (p,ovn) = 
    match p.papattern_desc with
      | ACase m ->
	  let ctype = element_of_asequencetype m in 
	  let attrs = 
	    match ovn with 
		Some s -> [ var_attr_name, (curly_uri_string_of_rqname s)]
	      | None -> []
	  in
	    construct_element ts_case_elem_name attrs [ctype] 
      | ADefault ->
	  let attrs = 
	    match ovn with 
		Some s -> [ var_attr_name, 
			   (curly_uri_string_of_rqname s)]
	      | None -> []
	  in
	  construct_element ts_default_elem_name attrs [] 


let make_tuple_slots tindex (odt,tname) =
  let index = index_attr_name, (string_of_int tindex) in
  let name  = var_attr_name, (curly_uri_string_of_rqname tname) in 
  let attrs = index :: name :: [] in
  let dt    = element_of_optasequencetype odt in 
    construct_element tuple_slot_elem_name attrs dt
	

(* <datamodel_signature arg_count_attr_name='no_input|one_input|two_input|many_input'>
      <input_datamodel index_attr_name='<index>' datamodel_attr_name='<physical_type_str>'/>
       ...
      <output_datamodel datamodel_attr_name='<physical_type_str>'/>
   </datamodel_signature>
*)
let element_of_expr_eval_sig expr_sig_opt =
  let generate_outdm d = out_type_attr_name, (string_of_physical_type d) in
  let generate_input (d:input_signature) = 
    match d with
	NoInput -> (arg_count_attr_name, no_input),[]
      | OneInput s -> (arg_count_attr_name, one_input),[(string_of_physical_type s)]
      | TwoInput (s0,s1) -> (arg_count_attr_name, two_input),[(string_of_physical_type s0); (string_of_physical_type s1)]
      | ManyInput sargs -> (arg_count_attr_name, many_input), (Array.to_list(Array.map string_of_physical_type sargs))
  in

  let generate_input_sig d =
    let attr, in_types = generate_input d in
    let elems,_  = List.fold_left 
		     (fun (constructed, index) cur_term ->
			let attr0 = index_attr_name, (string_of_int index) in
			let attr1 = datamodel_attr_name, cur_term in
			let new_elem = construct_element input_datamodel_elem_name (attr0 :: attr1 :: []) [] in
			  (new_elem :: constructed), (index + 1)) ([], 0) in_types in
      attr,(List.rev elems)
  in

  let generate_output_dm d =
    let alg = datamodel_attr_name, (string_of_physical_type d) in
    construct_element output_datamodel_elem_name [alg] [] 
  in

    
    match expr_sig_opt with
      | None -> construct_element datamodel_signature [] []
      | Some (physop, input_dm, outputdm) ->
	  let attr,elems = generate_input_sig input_dm in 
	  let out = generate_output_dm outputdm in
	    construct_element datamodel_signature [attr]  (elems @ out :: [])

let make_function_signature_element fname intypes otype =
  let make_input_sig index t = 
    let in_elem = element_of_optasequencetype t in
    let i = index_attr_name, (string_of_int index) in
    let attrs = i :: [] in
      construct_element input_type_elem_name attrs in_elem
  in
  let make_output_sig t =
    let out_elem = element_of_asequencetype t in
    let attrs = [] in
      construct_element output_type_elem_name attrs [out_elem] 
  in

  let in_types = Array.to_list (Array.mapi make_input_sig intypes) in
  let out_types = make_output_sig otype in
  let name = fn_name_attr_name, (curly_uri_string_of_rqname fname) in
  let arity = arity_attr_name, (string_of_int (List.length in_types)) in
  let attrs = name :: arity :: [] in

    construct_element function_signature_elem_name attrs (in_types @ out_types :: [] )

(*******************************)
(* Compile annotation printing *)
(*******************************)
let text_wrapper_construct name value =
    construct_element name [] [RSText (value)]

let create_accessed_fields_element af = 
  List.map (fun x -> text_wrapper_construct accessed_fields_elem_name (curly_uri_string_of_rqname x)) af 

let create_returned_fields_element rf =
  List.map (fun x -> text_wrapper_construct returned_fields_elem_name (curly_uri_string_of_rqname x)) rf

let create_free_variables_element fv =
  List.map (fun x -> text_wrapper_construct free_variables_elem_name (curly_uri_string_of_rqname x)) fv

let element_of_compile_annotation algop = 
  let accessed_fields = algop_get_accessed_fields algop in
  let returned_fields = algop_get_returned_fields algop in
  let free_variables  = algop_get_free_variables algop  in
  (* Attributes *)
  let n_accessed_fields = construct_attribute num_accessed_fields_attr_name
			    (string_of_int (List.length accessed_fields)) in

  let n_returned_fields = construct_attribute num_returned_fields_attr_name
			    (string_of_int (List.length returned_fields)) in

  let n_free_variables  = construct_attribute num_free_variables_attr_name 
			    (string_of_int (List.length free_variables)) in
    
  let attrs = n_accessed_fields :: n_returned_fields :: n_free_variables :: [] in

  (* element children *)
  let e_accessed_fields = create_accessed_fields_element accessed_fields in
  let e_returned_fields = create_returned_fields_element returned_fields in
  let e_free_variables  = create_free_variables_element  free_variables  in

  let elems = e_accessed_fields @ e_returned_fields @ e_free_variables in
    
    construct_element compile_annotation_elem_name attrs elems

(*************************)
(* Overloaded Call Table *)
(*************************)
let make_overloaded_call_table_element table = 
  let n_entries = List.length table in
  let entries     = List.map (fun (name, (in_sig, out_sig)) ->
				let wrapped = List.map (fun x -> Some x) in_sig in
				make_function_signature_element
				name (Array.of_list wrapped) out_sig) table in

  let arity     = arity_attr_name, (string_of_int n_entries) in
    construct_element overloaded_call_elem_name [arity] entries 

(*******************)
(* Group Operation *)
(*******************)
let element_of_named_variable_name ename vn =
  let name = vname_attr_name, (curly_uri_string_of_rqname vn) in
    construct_element ename [name] []

let element_of_variable_name vn =
  element_of_named_variable_name vname_elem_name vn

let element_of_variable_list evl_name vlist =
  let arity = arity_attr_name, (string_of_int (List.length vlist)) in
  let elems = List.map element_of_variable_name vlist in
    construct_element evl_name [arity] elems

let element_of_group_desc gd =
  let gbn_elem = element_of_variable_list gbn_elem_name (get_group_names gd)     in
  let ind_elem = element_of_variable_list induced_elem_name (get_group_names gd) in
  let mbv_elem = element_of_variable_list mbv_elem_name (get_valid_names gd)     in
  let agg_elem = element_of_named_variable_name agg_elem_name (get_aggregate_name gd) in
  let agg_type = element_of_optasequencetype (get_aggregate_type gd) in 
  let elems    = gbn_elem :: ind_elem :: mbv_elem :: agg_elem :: agg_type in
    construct_element group_desc_elem_name [] elems

(*********************)
(* TupleTreePatterns *)
(*********************)
let get_child_twig_elems twig_node =
  let create_elem_of_axis_index_pair name (axis, index) =
    let ax = axis_attr_name, (string_of_axis axis) in
    let ix = index_attr_name, (string_of_int index) in
    construct_element name [ax; ix] []
  in
  (match twig_node.child_twig with
  | Some (axis, index) ->
      [create_elem_of_axis_index_pair child_twig_name (axis,index)]
  | None -> [] ) @
  List.map (create_elem_of_axis_index_pair pred_twig_name) twig_node.pred_twigs

let element_of_twig_node twig_node =
  let attrs = [(restore_attr_name, string_of_bool twig_node.restore_out)] in
  let elems = get_child_twig_elems twig_node in
  let attrs', elems' =
    match twig_node.out, twig_node.node_test with
    | Some f, Some nt ->
	let nte = make_node_test_element nt in
	(out_field_attr_name, curly_uri_string_of_rqname f) :: attrs, nte :: elems
    | None, Some nt -> 
	let nte = make_node_test_element nt in
	attrs, nte :: elems
    | Some f, None ->
	(out_field_attr_name, curly_uri_string_of_rqname f) :: attrs, elems
    | _ -> attrs, elems
  in
  construct_element twig_node_name attrs' elems'

(*let elements_of_ttp i tp =
  let twig_list = Array.to_list tp in
  let twig_node_elems = List.map element_of_twig_node twig_list in
  let input_attr = (in_field_attr_name, curly_uri_string_of_rqname i) in 
  construct_element ttp_name [input_attr] twig_node_elems
*)

(**************)
(* Predicates *)
(**************)
let rec element_of_predicate_desc pd =
  match pd with
    | SimpleConjunct (s,e) ->
	let start_attr = simple_conjunct_start_attr_name, (string_of_int s) in
	let end_attr   = simple_conjunct_end_attr_name  , (string_of_int e) in
	  construct_element simple_conjunct_elem_name (start_attr :: end_attr :: []) []
    | ComplexConjunct (l,r) ->
	let l_elem = element_of_predicate_desc l in
	let r_elem = element_of_predicate_desc r in
	  construct_element complex_conjunct_elem_name [] (l_elem :: r_elem :: [])
    | Disjunct (l,r) ->
	let l_elem = element_of_predicate_desc l in
	let r_elem = element_of_predicate_desc r in
	  construct_element disjunct_elem_name [] (l_elem :: r_elem :: [])

(* Take an algebraic operation,
   return a pair of the string name and the information 
   contained in the name of the algebraic operation *)
let element_of_algop_name algop = 
  let attrs, elems =
    match algop.palgop_expr_name with
    | AOELet (ocdt, vn) -> 
	let dt = element_of_optasequencetype ocdt in
	let vname = vname_attr_name, (curly_uri_string_of_rqname vn) in 
	let attrs = vname :: [] in
	attrs, dt

    | AOEIf ->
	[], []

    | AOETypeswitch(pat_name_array) ->
	let switch_list = Array.to_list pat_name_array in
	let elems = List.map make_typeswitch_element switch_list in
	let attrs = [ts_num_branches_attr_name, (string_of_int (List.length elems))] in
	attrs, elems 

    | AOEVar(vn) ->
	let vname = vname_attr_name, (curly_uri_string_of_rqname vn) in 	  
	let attrs = [vname] in
	attrs, []

    | AOEScalar(v) ->	
	let v' = value_attr_name, (string_of_literal v) in
	let t  = atomic_type_attr_name, (curly_uri_string_of_rqname (Namespace_symbols.rtype_name (Datatypes_util.symbol_of_primitive_type (atomic_type_of_literal v)))) in 
(*
      Question: How should we print/parse symbols in builtin namespaces?

 let t  = atomic_type_attr_name, (prefixed_string_of_rqname (Namespace_symbols.rtype_name (Datatypes_util.symbol_of_primitive_type (atomic_type_of_literal v)))) in
*)
	let attrs = v' :: t :: [] in
	attrs, []

    | AOESeq ->
	[], []

    | AOEEmpty ->
	[], []

    | AOEDocument ->
	[], []

    | AOEPI(target, content) ->
	let tg = target_attr_name, target in
	let ct = content_attr_name, content in
	let attrs = tg :: ct :: [] in
	attrs, []

    | AOEPIComputed ->
	[], []

    | AOEComment comment ->
	let cmt = value_attr_name, comment in
	let attrs = cmt :: [] in
	attrs, []

    | AOECommentComputed ->
	[], []

    | AOEText dt ->
	let t = value_attr_name, dt in
	let attrs = t :: [] in
	attrs, []

    | AOETextComputed ->
	[], []

    (* NB!  We are not printing out the namespace environment
       associated with an element constructor, so when we parse the
       element here, we end up with an incomplete namespace
       environment! *)

    (* This will break round-tripping *)
    | AOEElem(rname,nsenv) ->
	let elems = element_of_relem_symbol rname in
	[], elems :: []

    | AOEAnyElem nsenv ->
	[], []

    | AOEAttr(rname) -> 
	let elems = element_of_rattr_symbol rname in
	[], elems :: []

    | AOEAnyAttr nsenv ->
	[], []

    | AOEError -> 
	[], []

    | AOETreat(cst) ->
	let cst' = element_of_asequencetype cst in
	[], [cst']

    | AOEValidate _ ->
	[], []

    | AOECast(nsenv, cst) ->
	let cst' = element_of_asequencetype cst in
	[], [cst']

    | AOECastable(nsenv, cst) ->
	let cst' = element_of_asequencetype cst in
	[], [cst']

    | AOESome (ocdt, vn) ->
	let dt = element_of_optasequencetype ocdt in
	let vname = vname_attr_name, (curly_uri_string_of_rqname vn) in 
	let attrs = vname :: [] in
	attrs, dt

    | AOEEvery(ocdt, vn) ->
	let dt = element_of_optasequencetype ocdt in
	let vname = vname_attr_name, (curly_uri_string_of_rqname vn) in 
	let attrs = vname :: [] in
	attrs,dt

	  (* Function Calls *)
    | AOECallBuiltIn((fname,arity), input_types, output_type) -> 
	let function_signature_element =
	  make_function_signature_element fname input_types output_type
	in
	(* This information is replicated in the signature but, is
	   used to identify the function as well *)
	let name = fn_name_attr_name, (curly_uri_string_of_rqname fname) in
	let arity = arity_attr_name, (string_of_int arity) in
	let attrs = name :: arity :: [] in
	attrs, [function_signature_element]

    | AOECallOverloaded((fname, arity),table) ->
	let name = fn_name_attr_name, (curly_uri_string_of_rqname fname) in
	let arity = arity_attr_name, (string_of_int arity) in
	(* We are not currently serializing the call table, though we will
	   when static typing comes back *)
	(*	let table = make_overloaded_call_table_element table in *)
	let attrs = name :: arity :: [] in
	attrs, []

    | AOECallUserDefined((fname,arity), input_types, output_type) -> 
	let name = fn_name_attr_name, (curly_uri_string_of_rqname fname) in
	let arity' = arity_attr_name, (string_of_int arity) in
	let function_signature_element = make_function_signature_element fname input_types output_type in
	let attrs = name :: arity' :: [] in 
	attrs,[function_signature_element]

    | AOEConvertSimple at ->  
	let t  = atomic_type_attr_name, (curly_uri_string_of_rqname (Namespace_symbols.rtype_name (Datatypes_util.symbol_of_primitive_type at))) in 
	[t], []

    | AOEPromoteNumeric at  -> 
	let t  = atomic_type_attr_name, (curly_uri_string_of_rqname (Namespace_symbols.rtype_name (Datatypes_util.symbol_of_primitive_type at))) in 
	[t], []

    | AOEUnsafePromoteNumeric at  -> 
	let t  = atomic_type_attr_name, (curly_uri_string_of_rqname (Namespace_symbols.rtype_name (Datatypes_util.symbol_of_primitive_type at))) in 
	[t], []

    | AOEKeyRef(kname) ->
	let kn = kname_attr_name, kname in
	let attrs = kn :: [] in
	attrs, []

    | AOEExecute ->
(*
	let hn = host_attr_name, host in
	let pn = port_attr_name, (string_of_int port) in
	let attrs = hn :: pn :: [] in
	attrs, []
*)
	[], []

    | AOEInputTuple ->
	[],[]

	  (* Tuples *)
    | AOECreateTuple(tnames) -> 	
	let arity' = arity_attr_name, (string_of_int (Array.length tnames)) in
	let elems = Array.to_list (Array.mapi make_tuple_slots tnames) in
	[arity'], elems

    | AOEAccessTuple(tname) ->
	let tname = tuple_name_attr_name, (curly_uri_string_of_rqname tname) in
	let attrs = tname :: [] in 
	attrs, []

    | AOEProject (tnames) ->
	[], [elem_of_project_name tnames]

    | AOEConcatTuples ->
	[], []

    | AOEMapConcat ->
	[],[] 

    | AOEOuterMapConcat vn ->  
	let vname = null_index_attr_name, (curly_uri_string_of_rqname vn) in
	(vname :: []), [] 

    | AOEMap ->
	[], []

    | AOEMapFromItem vn ->
	let vname = item_tuple_attr_name, (curly_uri_string_of_rqname vn) in
	(vname :: []), []

    | AOEMapToItem ->
	[], []

    | AOENullMap null_index ->
	let tname = null_index_attr_name, (curly_uri_string_of_rqname null_index) in
	let attrs = tname :: [] in 
	attrs, []

    | AOEMapIndex crname ->
	let tname = tuple_name_attr_name, (curly_uri_string_of_rqname crname) in
	let attrs = tname :: [] in 
	attrs, []

    | AOEMapIndexStep crname ->
	let tname = tuple_name_attr_name, (curly_uri_string_of_rqname crname) in
	let attrs = tname :: [] in 	    
	attrs, []

    | AOEProduct ->
	[], []

    | AOESelect pd ->
	let pd_elem = element_of_predicate_desc pd in
	[], (pd_elem :: [])

    | AOELeftOuterJoin (vn, pd) ->
	let vname   = null_index_attr_name, (curly_uri_string_of_rqname vn) in
	let pd_elem = element_of_predicate_desc pd in
	(vname :: []), (pd_elem :: [])

    | AOEJoin pd  -> 
	let pd_elem = element_of_predicate_desc pd in
	[], (pd_elem :: [])

    | AOEGroupBy gd_list ->
	let arity = arity_attr_name, (string_of_int (List.length gd_list)) in
	let attrs = arity :: [] in
	let gds   = List.map element_of_group_desc gd_list in 
	attrs, gds

    | AOEOrderBy (stable, sort_kinds, osig) ->
	let n_attrs = n_sort_criteria_attr_name, (string_of_int (List.length sort_kinds)) in
	let stable = attr_of_stable stable in
	let sorts = List.map element_of_sort_kinds sort_kinds in
	n_attrs :: stable :: [], sorts

    | AOEMaterializeTable -> 
	[], []

	  (* Updates *)
    | AOECopy -> 
	[], []

    | AOEDelete -> 
	[], []

    | AOEDetach -> 
	[], []

    | AOEInsert il -> 
	let il_attr = attr_of_insert_location il in
	il_attr :: [], []

    | AOERename ->
	[], []

    | AOEReplace val_flag -> 
	let val_flag = attr_of_value_of_flag val_flag in
	val_flag :: [], []

    | AOESnap sm ->
	let sm_attr = attr_of_snap_modifier sm in 
	sm_attr :: [], []

    | AOETreeJoin (axis, anode_test) ->
	let ax = axis_attr_name, (string_of_axis axis) in
	let nt = make_node_test_element anode_test in
	let attrs = ax :: [] in
	attrs, [nt]

    | AOETupleTreePattern (input_dot, pattern) ->
	let twig_list = Array.to_list pattern in
	let twig_node_elems = List.map element_of_twig_node twig_list in
	let input_attr = (in_field_attr_name, curly_uri_string_of_rqname input_dot) in 
	[input_attr], twig_node_elems

    | AOEPrune (name, axis) -> 
        let ax = axis_attr_name, (string_of_axis axis) in
	let input_vname = input_prune_field_attr_name, (curly_uri_string_of_rqname name) in
	[ax; input_vname],[]

    | AOEDistinct name ->
	let input_vname = input_distinct_field_attr_name, (curly_uri_string_of_rqname name) in
	[input_vname],[]

    | AOEParse docname ->
        (docname_attr_name, docname) :: [], []

  in
  (get_serializable_moniker_of_algop_name algop.palgop_expr_name), attrs, elems


(* Return list of elements corresponding to sub-expressions *)
let rec element_of_subexpr element_of_algop se =
  match se with
      NoSub -> None
    | OneSub s1 -> Some [element_of_algop s1] 
    | TwoSub(s1,s2) -> Some [element_of_algop s1; element_of_algop s2]
    | ManySub s_array ->
	Some (Array.to_list (Array.map element_of_algop s_array))

(* Wrappers for the two kinds of subexpressions *)

and subexpr subexpr_elem_name element_of_algop sub =
  let arity = sub_expression_arity_attr_name, (string_of_int (get_subexpr_arity sub)) in
  let sub_type = sub_expression_type_attr_name, (get_subexpr_type sub) in
  let attrs = arity :: sub_type :: [] in
  let subexprs = element_of_subexpr element_of_algop sub in
  (* If the operator has no subexpressions, we don't print them out *)
  match subexprs with 
  | None -> []
  | Some subexprs ->
      [ construct_element subexpr_elem_name attrs subexprs ]

and dep_sub element_of_algop sub =
  subexpr dep_elem_name element_of_algop sub 

and indep_sub element_of_algop sub = 
  subexpr indep_elem_name element_of_algop sub 

(* Pre code selection *)
and undecorated_element_of_algop algop =
  let name, attributes, (sub_elem: rsexpr list) = element_of_algop_name algop in
  let dependent_exprs     = dep_sub undecorated_element_of_algop algop.pdep_sub_expression in
  let independent_exprs   = indep_sub undecorated_element_of_algop algop.psub_expression in
  (* let compile_annotations = element_of_compile_annotation algop in  *)

  let child_elements      =  sub_elem @ dependent_exprs @ independent_exprs in
    construct_element name attributes child_elements
      
(* Need to also have an identifier placed in.. or an annotation for code pieces *)
and decorated_element_of_algop algop =
  let name, attributes, sub_elem = element_of_algop_name algop in
  let dependent_exprs = dep_sub decorated_element_of_algop algop.pdep_sub_expression in
  let independent_exprs = indep_sub decorated_element_of_algop algop.psub_expression in
  let data_model_signature = element_of_expr_eval_sig algop.palgop_expr_eval_sig in
  let signature_checks = check_signatures construct_comment algop.palgop_expr_eval_sig algop.psub_expression in
  let child_elements = signature_checks @ sub_elem @ data_model_signature :: dependent_exprs @ independent_exprs in
    construct_element name attributes child_elements

(* Statement handler *)
let handle_statement element_of_algop s =
  construct_element expression_elem_name [] [(element_of_algop s)]
	  
let handle_statements element_of_algop stmts = 
  let n_statements = number_of_statements_attr_name, (string_of_int (List.length stmts)) in
  let all_statements = List.map (handle_statement element_of_algop) stmts in

    construct_element statement_declaration_elem_name (n_statements :: []) all_statements 


let element_of_fn_body element_of_algop (var_names, fn_body, _ )(*type info reported elsewhere *) =
  let attrs = Array.to_list (Array.mapi
		(fun index term -> (make_input_attr_name index), (curly_uri_string_of_rqname term))
		var_names) in
  let fn_b = element_of_algop !fn_body in

    construct_element function_body_elem_name attrs [fn_b]

let handle_fn element_of_algop fn = 
  let (name,arity), signature, body = fn.palgop_function_decl_desc in

    (* Attributes *)
  let arity_attr = arity_attr_name, (string_of_int arity) in
  let name_attr = fn_name_attr_name, (curly_uri_string_of_rqname name) in

    (* Elements *)
  let input_sig, output_sig = signature in
    (* this is nasty code *)
  let input_sig' = Array.of_list (List.map (fun x -> Some x) input_sig) in
  let signature_element = make_function_signature_element name input_sig' output_sig in
  let body_element = element_of_fn_body element_of_algop body in

  let children = signature_element :: body_element :: [] in    
  let attrs = arity_attr :: name_attr :: [] in

    construct_element function_decl_elem_name attrs children

let handle_fns element_of_algop fdecl_list =
  let all_fns = List.map (handle_fn element_of_algop) fdecl_list in
  let attrs = [fn_count_attr_name, (string_of_int (List.length all_fns))] in
    construct_element function_tag_elem_name attrs all_fns

(********************)
(* Prolog Functions *)
(********************)
let name_of_prolog_op op =
  let attrs, sub_elems = 
    match op.alg_decl_name with
      | AOEVarDeclExternal (ocdt, vn) ->	  
	  let e = element_of_optasequencetype ocdt in
	  let a_vn = variable_name_attr_name, (curly_uri_string_of_rqname vn) in
	    a_vn :: [], e

      | AOEVarDecl( ocdt, vn ) ->
	  let e = element_of_optasequencetype ocdt in
	  let a_vn = variable_name_attr_name, (curly_uri_string_of_rqname vn) in
	    a_vn :: [], e

      | AOEValueIndexDecl kn ->
	  let attrs = (kname_attr_name, kn) :: [] in
	    attrs, []

      | AOENameIndexDecl rname ->
	  let elems = element_of_relem_symbol rname in
	  let attrs = [] in
	  attrs, elems :: []

  in
  let attrs = (attribute_of_prolog_algop_name op.alg_decl_name) :: attrs in 
    attrs,sub_elems

let handle_prolog_op element_of_algop prolog_op =
  let indep_exprs = indep_sub element_of_algop prolog_op.alg_decl_indep in
  let dep_exprs   = dep_sub   element_of_algop prolog_op.alg_decl_dep   in
  let attributes, sub_elem = name_of_prolog_op prolog_op in
  let child_elements = sub_elem @ dep_exprs @ indep_exprs in

    construct_element prolog_elem_name attributes child_elements

let handle_vars element_of_algop var_list =
  let all_vars = List.map (handle_prolog_op element_of_algop) var_list in
  let attrs = [prolog_var_count_attr_name, (string_of_int (List.length all_vars))] in
    construct_element prolog_vars_elem_name attrs all_vars
 
let handle_indices element_of_algop index_list =  
  let all_indices = List.map (handle_prolog_op element_of_algop) index_list in
  let attrs =
    [prolog_index_count_attr_name, (string_of_int (List.length all_indices))]
  in
  construct_element prolog_indices_elem_name attrs all_indices

let algop_prolog_to_xml element_of_algop prolog =
  let fns     = handle_fns  element_of_algop prolog.palgop_prolog_functions in
  let vars    = handle_vars element_of_algop prolog.palgop_prolog_vars in
  let indices = handle_indices element_of_algop prolog.palgop_prolog_indices in
  let prolog  = fns :: vars :: indices :: [] in
  construct_element algop_prolog_decl_elem_name [] prolog

(* Only doing expressions *)
let algop_module_to_xml element_of_algop am = 
  let prolog     = algop_prolog_to_xml element_of_algop am.palgop_module_prolog in
  let statements = handle_statements element_of_algop am.palgop_module_statements in
  let module_xml = prolog :: statements :: [] in
    construct_element algop_module_decl_elem_name [] module_xml 
 

    


(* Logical handlers *)
let xml_stream_of_logical_plan_expr nsenv stmt =
  let nsenv' = Namespace_context.add_all_ns nsenv [(algebra_prefix, algebra_uri)]  in 
  Small_stream_context.resolved_xml_stream_of_sexpr 
    (Small_stream_context.sexpr_of_rsexpr nsenv'
       (handle_statement undecorated_element_of_algop stmt))


let xml_stream_of_logical_plan_prolog nsenv xmod = 
  let nsenv' = Namespace_context.add_all_ns nsenv [(algebra_prefix, algebra_uri)]  in 
  Small_stream_context.resolved_xml_stream_of_sexpr 
    (Small_stream_context.sexpr_of_rsexpr nsenv'
       (algop_prolog_to_xml undecorated_element_of_algop xmod))

let xml_stream_of_logical_plan_module nsenv xmod = 
  let nsenv' = Namespace_context.add_all_ns nsenv [(algebra_prefix, algebra_uri)]  in 
  Small_stream_context.resolved_xml_stream_of_sexpr 
    (Small_stream_context.sexpr_of_rsexpr nsenv'
       (algop_module_to_xml undecorated_element_of_algop xmod))

let xml_stream_of_physical_plan_expr nsenv stmt = 
  let nsenv' = Namespace_context.add_all_ns nsenv [(algebra_prefix, algebra_uri)]  in 
  Small_stream_context.resolved_xml_stream_of_sexpr 
    (Small_stream_context.sexpr_of_rsexpr nsenv'
       (handle_statement decorated_element_of_algop stmt))

let xml_stream_of_physical_plan_module nsenv xmod = 
  let nsenv' = Namespace_context.add_all_ns nsenv [(algebra_prefix, algebra_uri)]  in 
  Small_stream_context.resolved_xml_stream_of_sexpr 
    (Small_stream_context.sexpr_of_rsexpr nsenv'
       (algop_module_to_xml decorated_element_of_algop xmod))

(************************************ TEST HANDLING CODE *************************)
(* Top-level helper modules *)


let serialize_algebra_statement_helper element_of_algop proc_ctxt output stmt = 
  let gout  = Parse_io.galax_output_from_output_spec output in
    (* is this correct? *)
  let nsenv = Processing_context.get_external_nsenv proc_ctxt in
  let nsenv' = Namespace_context.add_all_ns nsenv [(algebra_prefix, algebra_uri)]  in 
    try
      begin	
	let whole_plan = handle_statement element_of_algop stmt in
	let algop_module_doc = Small_stream_context.sexpr_of_rsexpr nsenv' (RSDocument( Dm_atomic_util.default_no_uri_dm, [whole_plan])) in 
	let rxmlstr = Small_stream_context.resolved_xml_stream_of_sexpr algop_module_doc in	  
	let fmt = Parse_io.formatter_of_galax_output gout in
	  (Serialization.fserialize_resolved_xml_stream proc_ctxt fmt rxmlstr;
	   Format.pp_print_newline fmt ();
	 Parse_io.close_galax_output gout)
      end
    with
      | exn -> (Parse_io.close_galax_output gout; raise exn)

(* These are for debugging with ocaml-galax*)
let serialize_algebra_module_helper element_of_algop proc_ctxt output am = 
  (* Note that a galax_output must be closed, as below *)
  (* is this correct? *)
  let nsenv = Processing_context.get_external_nsenv proc_ctxt in
  let nsenv' = Namespace_context.add_all_ns nsenv [(algebra_prefix, algebra_uri)]  in 
  let gout = Parse_io.galax_output_from_output_spec output in
    try
      begin

	let rxmlstr = Small_stream_context.resolved_xml_stream_of_sexpr 
			(Small_stream_context.sexpr_of_rsexpr nsenv'
			   (algop_module_to_xml element_of_algop am)) in

	let fmt = Parse_io.formatter_of_galax_output gout in
	  (Serialization.fserialize_resolved_xml_stream proc_ctxt fmt rxmlstr;
	   Format.pp_print_newline fmt ();
	   Parse_io.close_galax_output gout)
      end
    with
      | exn -> (Parse_io.close_galax_output gout; raise exn)

let init_proc_ctxt proc_ctxt = 
  Processing_context.set_serialization_kind
    proc_ctxt 
    Processing_context.Serialize_As_Well_Formed

(* Exposed modules *)
(* Logical versions *)
let serialize_logical_algebra_statement proc_ctxt output stmt =
  init_proc_ctxt proc_ctxt;
  serialize_algebra_statement_helper undecorated_element_of_algop proc_ctxt output stmt


let serialize_logical_algebra_module proc_ctxt output am =
  init_proc_ctxt proc_ctxt;
  serialize_algebra_module_helper undecorated_element_of_algop proc_ctxt output am

(* Decorated or after code selection versions *)
let serialize_physical_algebra_statement proc_ctxt output stmt =
  init_proc_ctxt proc_ctxt;
  serialize_algebra_statement_helper decorated_element_of_algop proc_ctxt output stmt

let serialize_physical_algebra_module proc_ctxt output am =
  init_proc_ctxt proc_ctxt;
  serialize_algebra_module_helper decorated_element_of_algop proc_ctxt output am

