(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies and AT&T Labs.            *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* Module: Algebgra_print_utils
   Description:
   This module contains some util functions for algebra printers and parsers
   
*)

(* Algbera printing module *)
open Error
open Occurrence

open Namespace_names
open Namespace_symbols

open Xquery_algebra_ast
open Xquery_common_ast
open Xquery_core_ast (* AITxxx *)

open Streaming_types
open Small_stream_ast

open Namespace_names
open Datatypes

open Print_common
open Planio_print_common
open Planio_print_util

let construct_element name attrs sub_elements =
  RSElem(name, [], attrs, Dm_atomic_util.default_no_uri_dm, sub_elements)

let construct_attribute attr_name str = (attr_name, str)

(* This code is to check when materializations occur in the decorated plan and
   generate a comment appropriately *)
let construct_comment msg x = [RSComment(msg ^ x)];;


let string_of_occurrence occ =
  match occ with
  | None ->
      "None"
  | Some (l,h) ->
      begin
	match (l,h) with
	| (UP_INT 0, UNBOUNDED) ->
	    "*"
	| (UP_INT 1, UNBOUNDED) ->
	    "+"
	| (UP_INT 0, UP_INT 1) ->
	    "?"
	| _ ->
	    raise (Query (Malformed_Type("Cannot have complex minOccurs and maxOccurs in a Datatype")))
      end

let attribute_of_occurrence occ = occurrence_attr_name, (string_of_occurrence occ)



(* Symbol (reelem/rattr) parsing functions *)
let element_of_rtype_symbol rtype = 
  let prefix = prefix_attr_name, (xml_string_of_prefix (rtype_prefix rtype)) in
  let name   = rqname_attr_name, (curly_uri_string_of_rqname (rtype_name rtype)) in
  let attrs = prefix :: name :: [] in
    
    construct_element rtype_elem_name attrs [] 

let element_of_relem_symbol rname = 
  let prefix = prefix_attr_name, (xml_string_of_prefix (relem_prefix rname)) in
  let name   = rqname_attr_name, (curly_uri_string_of_rqname (relem_name rname)) in
  let attrs = prefix :: name :: [] in
    
    construct_element relem_elem_name attrs [] 

let element_of_rattr_symbol rname = 
  let prefix = prefix_attr_name, (xml_string_of_prefix (rattr_prefix rname)) in
  let name   = rqname_attr_name, (curly_uri_string_of_rqname (rattr_name rname)) in
  let attrs  = prefix :: name :: [] in 
    
    construct_element rattr_elem_name attrs [] 

let element_of_element_test cet =
  match cet with
  | ASchemaElementTest nam ->
      let child1 = element_of_relem_symbol nam in
      let children = child1 :: [] in
      construct_element ait_schemaelementtype_elem_name [] children
  | AElementTest None ->
      let children = [] in
      construct_element ait_elementtype_elem_name [] children
  | AElementTest (Some (nam,None)) ->
      let child1 = element_of_relem_symbol nam in
      let children = child1 :: [] in
      construct_element ait_elementtype_elem_name [] children
  | AElementTest (Some (nam, Some typ)) ->
      let child1 = element_of_relem_symbol nam in
      let child2 = element_of_rtype_symbol typ in
      let children = child1 :: child2 :: [] in
      construct_element ait_elementtype_elem_name [] children

let element_of_attribute_test cet =
  match cet with
  | ASchemaAttributeTest nam ->
      let child1 = element_of_rattr_symbol nam in
      let children = child1 :: [] in
      construct_element ait_schemaattribute_elem_name [] children
  | AAttributeTest None ->
      let children = [] in
      construct_element ait_attributetype_elem_name [] children
  | AAttributeTest (Some (nam,None)) ->
      let child1 = element_of_rattr_symbol nam in
      let children = child1 :: [] in
      construct_element ait_attributetype_elem_name [] children
  | AAttributeTest (Some (nam, Some typ)) ->
      let child1 = element_of_rattr_symbol nam in
      let child2 = element_of_rtype_symbol typ in
      let children = child1 :: child2 :: [] in
      construct_element ait_attributetype_elem_name [] children

let element_of_kind_test ckt =
  match ckt with
  | AAttributeKind cet ->
      element_of_attribute_test cet
  | AElementKind cet ->
      element_of_element_test cet
  | AAnyKind ->
      construct_element ait_node_elem_name [] []
  | ATextKind ->
      construct_element ait_text_elem_name [] []
  | ACommentKind ->
      construct_element ait_comment_elem_name [] []
  | APIKind None ->
      construct_element ait_processing_instruction_elem_name [] []
  | APIKind (Some s) ->
      let attrs = [pi_arg_attr_name, s] in
      construct_element ait_processing_instruction_elem_name attrs []
  | ADocumentKind None ->
      construct_element ait_document_elem_name [] [] 
  | ADocumentKind (Some cet) ->
      let elem = element_of_element_test cet in
	construct_element ait_document_elem_name [] (elem :: [])


let element_of_aitemtype cdtk =
  match cdtk with
  | AITKindTest ckt ->
      let elem = element_of_kind_test ckt in
      construct_element ait_kindtest_elem_name [] (elem :: [])
  | AITTypeRef type_symbol ->
      let elem = element_of_rtype_symbol type_symbol in
      construct_element ait_typeref_elem_name [] (elem :: [])
  | AITItem ->
      construct_element ait_item_elem_name [] [] 
  | AITNumeric ->
      construct_element ait_numeric_elem_name [] [] 
  | AITEmpty ->
      construct_element ait_empty_elem_name [] [] 
  | AITAtomic type_symbol ->
      let elem = element_of_rtype_symbol type_symbol in
      construct_element ait_atomic_elem_name [] (elem :: [])

let element_of_asequencetype cdt =
  let cdtk,occ = cdt.pasequencetype_desc in
  let attrs = [attribute_of_occurrence occ] in
  let elems = [element_of_aitemtype cdtk] in
    construct_element  asequencetype_elem_name attrs elems

let element_of_optasequencetype ocdt = 
  match ocdt with 
      None -> []
    | Some cdt -> [element_of_asequencetype cdt]

