(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: processing_context.ml,v 1.44 2006/03/03 18:10:26 mff Exp $ *)

(* Module: Processing_context
   Description:
     This module contains basic configuration parameters for Galax,
     notably set up of the processing phases.
*)

open Error

open Monitoring_context

(***********)
(* Options *)
(***********)

type sbdo_kind =
  | SBDO_Remove
  | SBDO_Preserve
  | SBDO_AdHoc
  | SBDO_Tidy
  | SBDO_DupTidy
  | SBDO_Sloppy

type typing_kind =
  | Typing_None
  | Typing_Weak
  | Typing_Strong

type serialization_kind =
  | Serialize_As_Standard
  | Serialize_As_Well_Formed
  | Serialize_As_XQuery
  | Serialize_As_Canonical

type projection_kind =
  | Projection_None
  | Projection_Standard
  | Projection_Optimized

type treejoin_logical_kind =
  | Default       (* no rewritings for bulk *)
  | TreeJoin      (* apply bulk TreeJoin rewrites 
                      NOTES 
			- Use this for streaming *)
  | Twig
  
type treejoin_physical_kind =
  | NestedLoop
  | SCJoin
  | IndexSortJoin
  | TwigJoin
  | Streaming


(**********************)
(* Processing context *)
(**********************)

type processing_context =
    { (* Processing phases *)
      mutable normalization : bool; (* true when normalization is on [default: on] *)
      mutable normalization_ident : bool; (* true when normalization is identity [default: off] *)
      mutable typing        : bool; (* true when static typing is on [default: off] *)
      mutable rewriting     : bool; (* true when rewriting of the normalized core on [default: on] *)
      mutable factorization : bool; (* true when factorization is on [default: on] *)
      mutable optimization  : bool; (* true when algebraic optimization is on [default: on] *)
      mutable evaluation    : bool; (* true when evaluation is on [default: on] *)

      (* Data model options *)
      mutable xml_whitespace       : bool;
      mutable xml_pis_and_comments : bool;

      (* Behavior options *)
      mutable sbdo_kind        	  : sbdo_kind;         (* SBDO optim kind *)
      mutable typing_kind 	  : typing_kind;       (* Static analysis kind *)
      mutable serialization_kind  : serialization_kind;(* Serialization kind *)
      mutable boundary_space_kind : Xquery_common_ast.strip_or_preserve;
      mutable construction_kind   : Xquery_common_ast.strip_or_preserve;
      mutable ordering_kind       : Xquery_common_ast.ordered_or_unordered;
      mutable default_order_kind  : Xquery_common_ast.emptysortkind;
      mutable ns_preserve_kind    : Xquery_common_ast.preserve_or_no_preserve;
      mutable ns_inherit_kind     : Xquery_common_ast.inherit_or_no_inherit;

      (* Optimization options *)
      mutable inline_functions 	 : bool;              (* Function inline *)
      mutable inline_variables 	 : bool;              (* Variable inline *)
      mutable projection_kind    : projection_kind;   (* Document projection *) 
      mutable treejoin_phys      : treejoin_physical_kind;
      mutable treejoin_log       : treejoin_logical_kind;
      mutable pruning            : bool;
      mutable streaming          : bool;
      mutable system             : bool;

      monitor_context            : monitor_context;   (* Function monitoring context *)

      (* External namespace bindings *)
      mutable external_nsenv     : Namespace_context.nsenv;

      (* Location hints *)
      mutable external_schema_locs : (string * string) list;

      (* Base URI *)
      mutable base_uri           : AnyURI._uri option;

      (* Default collation *)
      mutable default_collation  : AnyURI._uri }

(* Default processing context *)

let default_processing_context () = 
  {
   (* Processing phases *)
   normalization 	     = true;  (* true when normalization is on [default: on] *)
   normalization_ident 	     = false; (* true when normalization is identity [default: off] *)
   typing        	     = true;  (* true when static typing is on [default: on] *)
   rewriting   	 	     = true;  (* true when rewriting of the normalized core on [default: on] *)
   factorization             = false; (* true when factorization is on [default: on] *)
   optimization              = true;  (* true when algebraic optimization is on [default: on] *)
   evaluation 	 	     = true;  (* true when evaluation is on [default: on] *)

   (* Data model options *)
   xml_whitespace  = true; (* true when preserving boundary whitespace [default: on] *)
   xml_pis_and_comments = true; (* true when preserving pis and comments [default: on ] *)

   (* Other options *)
   inline_functions         	     = false;
   inline_variables         	     = false;
   sbdo_kind          	     = SBDO_Preserve; (* What kind of SBDO optimization? [default: duptidy] *)
   typing_kind        	     = Typing_Weak;   (* What kind of static analysis?   [default: weak] *)
   serialization_kind 	     = Serialize_As_Standard;
   boundary_space_kind 	     = Xquery_common_ast.Strip;
   construction_kind   	     = Xquery_common_ast.Preserve;
   ordering_kind             = Xquery_common_ast.Ordered;
   default_order_kind        = Xquery_common_ast.EmptyGreatest;
   ns_preserve_kind          = Xquery_common_ast.NSPreserve;
   ns_inherit_kind           = Xquery_common_ast.NSInherit;
   projection_kind    	     = Projection_None ;

   treejoin_log              = Default;  
   treejoin_phys             = NestedLoop;
   pruning                   = false;
   streaming                 = false;
   system                    = false;

   (* API Function Monitoring Context *)
   monitor_context           = default_monitor_context() ;

   (* External namespace bindings *)
   external_nsenv            = Namespace_context.default_xquery_nsenv;

   (* Location hints *)
   external_schema_locs      = [];

   (* Base URI *)
   base_uri                  = AnyURI.default_base_uri();

   (* Default collation *)
   default_collation         = AnyURI.default_collation_uri() }

(* Returns the external namespace environment *)

let get_external_nsenv proc_ctxt = proc_ctxt.external_nsenv

(* Returns the base URI *)

let get_base_uri proc_ctxt = proc_ctxt.base_uri
let get_default_collation proc_ctxt = proc_ctxt.default_collation

(* Set specific parameters in the processing context *)

let set_normalization_phase  proc_ctxt b = proc_ctxt.normalization <- b
let set_normalization_ident  proc_ctxt b = proc_ctxt.normalization_ident <- b
let set_typing_phase         proc_ctxt b = proc_ctxt.typing <- b
let set_rewriting_phase      proc_ctxt b = proc_ctxt.rewriting <- b
let set_factorization_phase  proc_ctxt b = proc_ctxt.factorization <- b
let set_optimization_phase   proc_ctxt b = proc_ctxt.optimization <- b
let set_evaluation_phase     proc_ctxt b = proc_ctxt.evaluation <- b

let set_xml_whitespace       proc_ctxt b = proc_ctxt.xml_whitespace <- b
let set_xml_pis_and_comments proc_ctxt b = proc_ctxt.xml_pis_and_comments <- b

let set_inline_variables     proc_ctxt b = proc_ctxt.inline_variables <- b
let set_inline_functions     proc_ctxt b = proc_ctxt.inline_functions <- b
let set_sbdo_kind            proc_ctxt sbdo_kind             = proc_ctxt.sbdo_kind <- sbdo_kind
let set_typing_kind          proc_ctxt typing_kind           = proc_ctxt.typing_kind <- typing_kind
let set_serialization_kind   proc_ctxt serialization_kind    = proc_ctxt.serialization_kind <- serialization_kind
let set_boundary_space_kind  proc_ctxt boundary_space_kind = proc_ctxt.boundary_space_kind <- boundary_space_kind
let set_construction_kind    proc_ctxt construction_kind = proc_ctxt.construction_kind <- construction_kind
let set_ordering_kind        proc_ctxt ordering_kind = proc_ctxt.ordering_kind <- ordering_kind
let set_default_order_kind   proc_ctxt defaultorder = proc_ctxt.default_order_kind <- defaultorder
let set_ns_preserve_kind     proc_ctxt nspreserve_kind = proc_ctxt.ns_preserve_kind <- nspreserve_kind
let set_ns_inherit_kind      proc_ctxt nsinherit_kind = proc_ctxt.ns_inherit_kind <- nsinherit_kind

let set_projection_kind      proc_ctxt projection_kind       = proc_ctxt.projection_kind <- projection_kind

let set_treejoin_log         proc_ctxt treejoin_log        = proc_ctxt.treejoin_log  <- treejoin_log
let set_treejoin_phys        proc_ctxt treejoin_phys       = proc_ctxt.treejoin_phys <- treejoin_phys
let set_pruning              proc_ctxt pruning             = proc_ctxt.pruning <- pruning
let set_streaming            proc_ctxt streaming           = proc_ctxt.streaming <- streaming

let set_namespace_env        proc_ctxt nsenv =
  proc_ctxt.external_nsenv <- nsenv

let set_schema_locations     proc_ctxt sl =
  proc_ctxt.external_schema_locs <- sl

let set_base_uri             proc_ctxt base_uri =
  proc_ctxt.base_uri       <- base_uri

(* Only default collation is known for now - Jerome *)
let known_collation uri =
  uri = Conf.collns

let set_default_collation    proc_ctxt uri =
  if known_collation uri
  then
    proc_ctxt.default_collation       <- (AnyURI._uri_of_string uri)
  else
    raise
      (Query(Wrong_Args("[err:XQST0038] Unknown collation:" ^ uri)))



(***************************)
(* Check the configuration *)
(***************************)

let check_config proc_ctxt =
  begin
    if proc_ctxt.treejoin_phys = Streaming && not(proc_ctxt.treejoin_log = TreeJoin) then
      raise (Query(Wrong_Args("XPath streaming requires logical bulk treejoins (use -treejoin-log treejoin)")));
    if proc_ctxt.typing && (not(proc_ctxt.normalization) && not(proc_ctxt.normalization_ident)) then
      raise (Query(Wrong_Args("Static typing requires normalization")));
    if proc_ctxt.rewriting && (not(proc_ctxt.normalization) && not(proc_ctxt.normalization_ident)) then
      raise (Query(Wrong_Args("Optimization requires normalization")));
    if proc_ctxt.evaluation && (not(proc_ctxt.normalization) && not(proc_ctxt.normalization_ident)) then
      raise (Query(Wrong_Args("Dynamic evaluation requires normalization")));
    if !Conf.print_type && not(proc_ctxt.typing) then
      raise (Query(Wrong_Args("Cannot print type without static typing on")));
    if !Conf.print_core_expr && (not(proc_ctxt.normalization)) then
      raise (Query(Wrong_Args("Cannot print normalized expression without normalization on")));
    if !Conf.print_optimized_expr && not(proc_ctxt.rewriting) then
      raise (Query(Wrong_Args("Cannot print optimized expression without optimization on")));
    if !Conf.print_xml && not(proc_ctxt.evaluation) then
      raise (Query(Wrong_Args("Cannot print xml result without dynamic evaluation on")));
(*  Removing this temporarily -- note that SBDO should work with weak typing also !! -- Philippe  
    if (proc_ctxt.sbdo_kind = SBDO_Tidy || proc_ctxt.sbdo_kind = SBDO_DupTidy || proc_ctxt.sbdo_kind = SBDO_Sloppy) &&
       not(proc_ctxt.typing_kind = Typing_Strong) 
    then 
      raise (Query(Wrong_Args("Automata SBDO requires strong typing"))); *)
  end

let set_proc pc dpc =
  pc.normalization <- dpc.normalization;
  pc.normalization_ident <- dpc.normalization_ident;
  pc.typing <- dpc.typing;
  pc.rewriting <- dpc.rewriting;
  pc.factorization <- dpc.factorization;
  pc.optimization <- dpc.optimization;
  pc.evaluation <- dpc.evaluation;
  pc.xml_whitespace <- dpc.xml_whitespace;
  pc.xml_pis_and_comments <- dpc.xml_pis_and_comments;
  pc.inline_functions <- dpc.inline_functions;
  pc.sbdo_kind <- dpc.sbdo_kind;
  pc.typing_kind <- dpc.typing_kind;
  pc.serialization_kind <- dpc.serialization_kind;
  pc.boundary_space_kind <- dpc.boundary_space_kind;
  pc.construction_kind <- dpc.construction_kind;
  pc.ordering_kind <- dpc.ordering_kind;
  pc.default_order_kind <- dpc.default_order_kind;
  pc.ns_preserve_kind <- dpc.ns_preserve_kind;
  pc.ns_inherit_kind <- dpc.ns_inherit_kind;
  pc.projection_kind <- dpc.projection_kind;
  pc.treejoin_phys <- dpc.treejoin_phys;
  pc.treejoin_log <- dpc.treejoin_log;
  pc.pruning <- dpc.pruning;
  pc.streaming <- dpc.streaming;
  pc.external_schema_locs <- dpc.external_schema_locs;
  pc.base_uri <- dpc.base_uri

let reset_proc pc =
  let dpc = default_processing_context() in
  set_proc pc dpc

