(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: processing_context.mli,v 1.31 2006/02/23 20:08:02 simeon Exp $ *)

(* Module: Processing_context
   Description:

     The processing context contains the state that is shared across
     multiple Galax API calls and between the processing-model phases
     of the Galax engine.

     The processing context contains : 

     1. basic configuration parameters for Galax, notably set up of
        the processing phases.
     2. monitoring data, which includes the call stack of all Galax
        API calls
     3. cached documents, i.e., all documents loaded by Galax API
        calls
*)

open Monitoring_context

(***********)
(* Options *)
(***********)

type sbdo_kind =
  | SBDO_Remove
  | SBDO_Preserve
  | SBDO_AdHoc
  | SBDO_Tidy
  | SBDO_DupTidy
  | SBDO_Sloppy

type typing_kind =
  | Typing_None
  | Typing_Weak
  | Typing_Strong

type serialization_kind =
  | Serialize_As_Standard
  | Serialize_As_Well_Formed
  | Serialize_As_XQuery
  | Serialize_As_Canonical

type projection_kind =
  | Projection_None
  | Projection_Standard
  | Projection_Optimized

type treejoin_logical_kind =
  | Default       (* no rewritings for bulk *)
  | TreeJoin      (* apply bulk TreeJoin rewrites 
                      NOTES 
		        - these include TupleTreeJoin rewrites 
			- Use this for streaming *)
  | Twig
  
type treejoin_physical_kind =
  | NestedLoop
  | SCJoin
  | IndexSortJoin
  | TwigJoin
  | Streaming


(**********************)
(* Processing context *)
(**********************)

type processing_context =
    { (* Processing phases *)
      mutable normalization : bool; (* true when normalization is on [default: on] *)
      mutable normalization_ident : bool; (* true when normalization is identity [default: off] *)
      mutable typing        : bool; (* true when static typing is on [default: off] *)
      mutable rewriting     : bool; (* true when rewriting of the normalized core on [default: on] *)
      mutable factorization : bool; (* true when factorization is on [default: on] *)
      mutable optimization  : bool; (* true when algebraic optimization is on [default: on] *)
      mutable evaluation    : bool; (* true when evaluation is on [default: on] *)

      (* Data model options *)
      mutable xml_whitespace       : bool;
      mutable xml_pis_and_comments : bool;

      (* Behavior options *)
      mutable sbdo_kind        	  : sbdo_kind;         (* SBDO optim kind *)
      mutable typing_kind 	  : typing_kind;       (* Static analysis kind *)
      mutable serialization_kind  : serialization_kind;(* Serialization kind *)
      mutable boundary_space_kind : Xquery_common_ast.strip_or_preserve;
      mutable construction_kind   : Xquery_common_ast.strip_or_preserve;
      mutable ordering_kind       : Xquery_common_ast.ordered_or_unordered;
      mutable default_order_kind  : Xquery_common_ast.emptysortkind;
      mutable ns_preserve_kind    : Xquery_common_ast.preserve_or_no_preserve;
      mutable ns_inherit_kind     : Xquery_common_ast.inherit_or_no_inherit;

      (* Optimization options *)
      mutable inline_functions   : bool;              (* Function inline *)
      mutable inline_variables   : bool;              (* Variable inline *)
      mutable projection_kind    : projection_kind;   (* Document projection *) 
      mutable treejoin_phys      : treejoin_physical_kind;
      mutable treejoin_log       : treejoin_logical_kind;
      mutable pruning            : bool;
      mutable streaming          : bool;
      mutable system             : bool;

      monitor_context            : monitor_context;   (* Function monitoring context *)

      (* External namespace bindings *)
      mutable external_nsenv      : Namespace_context.nsenv;

      (* Location hints *)
      mutable external_schema_locs : (string * string) list;

      (* Base URI *)
      mutable base_uri           : AnyURI._uri option;
      (* Default collation *)
      mutable default_collation  : AnyURI._uri }

(* Returns the default processing context *)

val default_processing_context : unit -> processing_context

(* Returns the external namespace environment *)

val get_external_nsenv       : processing_context -> Namespace_context.nsenv

(* Returns the base URI *)

val get_base_uri             : processing_context -> AnyURI._uri option
val get_default_collation    : processing_context -> AnyURI._uri

(* Set specific parameters in the processing context *)

val set_normalization_phase  : processing_context -> bool -> unit
val set_normalization_ident  : processing_context -> bool -> unit
val set_typing_phase         : processing_context -> bool -> unit
val set_rewriting_phase      : processing_context -> bool -> unit
val set_factorization_phase  : processing_context -> bool -> unit
val set_optimization_phase   : processing_context -> bool -> unit
val set_evaluation_phase     : processing_context -> bool -> unit

val set_xml_whitespace       : processing_context -> bool -> unit
val set_xml_pis_and_comments : processing_context -> bool -> unit

val set_inline_functions     : processing_context -> bool -> unit
val set_inline_variables     : processing_context -> bool -> unit
val set_sbdo_kind            : processing_context -> sbdo_kind -> unit
val set_typing_kind          : processing_context -> typing_kind -> unit
val set_serialization_kind   : processing_context -> serialization_kind -> unit
val set_boundary_space_kind  : processing_context -> Xquery_common_ast.strip_or_preserve -> unit
val set_construction_kind    : processing_context -> Xquery_common_ast.strip_or_preserve -> unit
val set_ordering_kind        : processing_context -> Xquery_common_ast.ordered_or_unordered -> unit
val set_default_order_kind   : processing_context -> Xquery_common_ast.emptysortkind -> unit
val set_ns_preserve_kind     : processing_context -> Xquery_common_ast.preserve_or_no_preserve -> unit
val set_ns_inherit_kind      : processing_context -> Xquery_common_ast.inherit_or_no_inherit -> unit
val set_projection_kind      : processing_context -> projection_kind -> unit

val set_treejoin_log         : processing_context -> treejoin_logical_kind  -> unit
val set_treejoin_phys        : processing_context -> treejoin_physical_kind -> unit
val set_pruning              : processing_context -> bool -> unit
val set_streaming            : processing_context -> bool -> unit

val set_namespace_env        : processing_context -> Namespace_context.nsenv -> unit
val set_schema_locations     : processing_context -> (string * string) list -> unit

val set_base_uri             : processing_context -> AnyURI._uri option -> unit
val set_default_collation    : processing_context -> string -> unit


(***************************)
(* Check the configuration *)
(***************************)

val check_config : processing_context -> unit


val set_proc : processing_context -> processing_context -> unit
val reset_proc : processing_context -> unit

