(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: procmod_compiler.ml,v 1.29 2006/02/23 20:08:02 simeon Exp $ *)

(* Module: Procmod_compiler
   Description:
     This module contains top-level operations over the processing
     model. This notably supports various levels of compilations for
     queries.
*)

open Error

open Xquery_common_ast
open Xquery_ast
open Xquery_core_ast
open Xquery_algebra_ast
open Logical_algebra_types
open Algebra_type

open Print_top

open Dm_atomic
open Physical_value

open Processing_context
open Parse_context
open Norm_context
open Typing_context
open Compile_context
open Code_selection_context
open Execution_context
open Monitoring_context

open Monitor
open Procmod_types
open Procmod_phases


(*******************************************)
(* Type for intermediate compilation steps *)
(*******************************************)

type ginput =
  | IOStatement of Galax_io.input_spec
  | IOProlog of Galax_io.input_spec
  | IOLibraryModule of Galax_io.input_spec
  | IOMainModule of Galax_io.input_spec

type ast =
  | ASTStatement of statement
  | ASTProlog of prolog
  | ASTLibraryModule of library_module
  | ASTMainModule of main_module

type core_ast =
  | CoreASTStatement of acstatement
  | CoreASTProlog of acprolog
  | CoreASTLibraryModule of acxmodule
  | CoreASTMainModule of acxmodule

type logical_plan =
  | LogicalPlanStatement of logical_algop_expr
  | LogicalPlanProlog of logical_algop_prolog
  | LogicalPlanLibraryModule of logical_algop_xmodule
  | LogicalPlanMainModule of logical_algop_xmodule

type physical_plan =
  | PhysicalPlanStatement of algop_expr
  | PhysicalPlanProlog of algop_prolog
  | PhysicalPlanLibraryModule of algop_xmodule
  | PhysicalPlanMainModule of algop_xmodule


(*********************************)
(* Types used during compilation *)
(*********************************)

(* External context *)

type external_context =
    { external_context_item : item option;
      external_timezone     : atomicDayTimeDuration option;
      external_variables    : (cvname * item list) list }

(* Compiled statements *)

type logical_plan_statement = logical_algop_expr
type compiled_statement = algop_expr

(* Compiled prolog *)

type logical_plan_prolog = logical_algop_prolog
type compiled_prolog =
    { compiled_prolog_external_context       : external_context option;
      compiled_prolog_plan  	             : algop_prolog;
      compiled_prolog_code_selection_context : code_selection_context;
      compiled_prolog_algebra_context        : algebra_context }

(* In a prepared prolog, all global variables' values are bound: *)
type prepared_prolog = compiled_prolog

(* Compiled module *)

type logical_plan_main_module = logical_algop_xmodule
type compiled_library_module = compiled_prolog
type compiled_main_module = compiled_prolog * compiled_statement list

(*******************************)
(* The default compiled prolog *)
(*******************************)

let default_compiled_prolog proc_ctxt =
  { compiled_prolog_external_context       = None;
    compiled_prolog_plan                   = Xquery_algebra_ast_util.empty_prolog_plan;
    compiled_prolog_code_selection_context = default_code_selection_context proc_ctxt;
    compiled_prolog_algebra_context        = default_algebra_context () }


(****************************)
(* Module context accessors *)
(****************************)

let processing_context_from_compiled_prolog m =
  processing_context_from_code_selection_context
    m.compiled_prolog_code_selection_context

let nsenv_from_compiled_prolog m =
  nsenv_from_norm_context
    (norm_context_from_stat_context
       (static_context_from_code_selection_context
	  m.compiled_prolog_code_selection_context))

let norm_context_from_compiled_prolog m =
  (norm_context_from_stat_context
     (static_context_from_code_selection_context
	m.compiled_prolog_code_selection_context))

let static_context_from_compiled_prolog m =
  static_context_from_code_selection_context
    m.compiled_prolog_code_selection_context

let compile_context_from_compiled_prolog m =
  compile_context_from_code_selection_context
    m.compiled_prolog_code_selection_context

let code_selection_context_from_compiled_prolog m =
  m.compiled_prolog_code_selection_context

let algebra_context_from_compiled_prolog m =
  m.compiled_prolog_algebra_context



(********************)
(* External context *)
(********************)


let create_external_context proc_ctxt opt_ctxt_item opt_timezone var_val_list = 
  let nsenv = get_external_nsenv proc_ctxt in
  let interpret_variable (vname_string,vl) =
    let vname = Namespace_names.uqname_element_of_string vname_string in
    let cvname = Namespace_resolve.resolve_variable_qname nsenv vname in
    (cvname,vl)
  in
  let interpreted_variables = List.map interpret_variable var_val_list in
  { external_context_item = opt_ctxt_item;
    external_timezone     = opt_timezone;
    external_variables    = interpreted_variables }

let default_external_context () =
  { external_context_item = None;
    external_timezone     = None;
    external_variables    = [] }
  

(*
   If a context item is provided in the external context, then 
   we add an external declaration for $fs:dot. 
*)

let fix_prolog_with_external_context_item ext_ctxt_item prolog_ast = 
  if (ext_ctxt_item) then
    { pprolog_xschemas = prolog_ast.pprolog_xschemas;
      pprolog_contexts = prolog_ast.pprolog_contexts;
      pprolog_functions = prolog_ast.pprolog_functions;
      pprolog_vars  	= (Xquery_ast_util.fmkvar_decl ((Namespace_builtin.fs_prefix,"dot"), Some (Xquery_ast_util.fmksequencetype (ITItem, None) Finfo.bogus), None) Finfo.bogus) :: prolog_ast.pprolog_vars; 
      pprolog_indices  	= prolog_ast.pprolog_indices;
    } 
  else prolog_ast

let fix_main_module_with_external_context_item ext_ctxt_item main_module_ast =
  if (ext_ctxt_item) then
    let prolog_ast = main_module_ast.pmain_module_prolog in
    let prolog =
      { pprolog_xschemas = prolog_ast.pprolog_xschemas;
	pprolog_contexts = prolog_ast.pprolog_contexts;
	pprolog_functions = prolog_ast.pprolog_functions;
	pprolog_vars  	= (Xquery_ast_util.fmkvar_decl ((Namespace_builtin.fs_prefix,"dot"), Some (Xquery_ast_util.fmksequencetype (ITItem, None) Finfo.bogus), None) Finfo.bogus) :: prolog_ast.pprolog_vars; 
	pprolog_indices	= prolog_ast.pprolog_indices }
    in
    { pmain_module_prolog     = prolog;
      pmain_module_statements = main_module_ast.pmain_module_statements }
  else main_module_ast
  

(* set_external_context extends the algebra context with bindings of
   variables to values.  We currently do not have an interface for
   adding external bindings of variables to types. *)

let set_external_context opt_ext_ctxt compiled_mod =
  let code_ctxt = compiled_mod.compiled_prolog_code_selection_context in
  let alg_ctxt = compiled_mod.compiled_prolog_algebra_context in 
  match opt_ext_ctxt with
  | None -> compiled_mod
  | Some ext_ctxt -> 
      begin
	let code_ctxt =
	  match ext_ctxt.external_context_item with
	  | None -> code_ctxt
	  | Some v -> 
	      (add_external_variable_value code_ctxt Xquery_common_ast.fs_dot [v])
	in
	let code_ctxt =
	  List.fold_left
	    (fun code_ctxt (vn, v) ->
	(* item_list *)
	      add_external_variable_value code_ctxt vn v)
	    code_ctxt ext_ctxt.external_variables
	in
	let new_alg_ctxt =
	  match ext_ctxt.external_timezone with
	  | None -> alg_ctxt
	  | Some tz -> set_timezone alg_ctxt tz
	in
	{ compiled_prolog_external_context       = opt_ext_ctxt;
	  compiled_prolog_plan                   = compiled_mod.compiled_prolog_plan;
	  compiled_prolog_code_selection_context = compiled_mod.compiled_prolog_code_selection_context;
	  compiled_prolog_algebra_context        = new_alg_ctxt }
      end

(*************************)
(* Statement compilation *)
(*************************)

(* Note:
     Compilation occurs within the context of a given compiled module.
   - Jerome
*)

(* Parsing *)

let apply_parsing_phase procmod_ctxt comp_mod ginput =
  (* 1. Obtain context information from the compiled module *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 2. Parse *)
  match ginput with
  | IOStatement gio ->
      let parse_ctxt = build_xquery_parse_context proc_ctxt in
      let statement = procmod_ctxt.parsing_phase.parse_statement parse_ctxt gio in
      ASTStatement statement
  | IOProlog gio ->
      let parse_ctxt = build_xquery_parse_context proc_ctxt in
      let (new_parse_ctxt,prolog) = procmod_ctxt.parsing_phase.parse_prolog parse_ctxt gio in
      ASTProlog prolog
  | IOLibraryModule gio ->
      let parse_ctxt = build_xquery_parse_context proc_ctxt in
      let (new_parse_ctxt,library_module) = procmod_ctxt.parsing_phase.parse_library_module parse_ctxt gio in
      ASTLibraryModule library_module
  | IOMainModule gio ->
      let parse_ctxt = build_xquery_parse_context proc_ctxt in
      let (new_parse_ctxt,main_module) = procmod_ctxt.parsing_phase.parse_main_module parse_ctxt gio in
      ASTMainModule main_module


(* Normalization *)

let apply_normalize_phase procmod_ctxt comp_mod ast =
  (* 1. Obtain context information from the compiled module *)
  let norm_ctxt = norm_context_from_compiled_prolog comp_mod in
  (* 3. Normalization *)
  match ast with
  | ASTStatement statement ->
      let cstatement =
	procmod_ctxt.normalization_phase.normalize_statement norm_ctxt statement
      in
      (norm_ctxt, CoreASTStatement cstatement)
  | ASTProlog prolog ->
      let (norm_ctxt',cprolog) =
	procmod_ctxt.normalization_phase.normalize_prolog norm_ctxt prolog
      in
      (norm_ctxt',CoreASTProlog cprolog)
  | ASTLibraryModule library_module ->
      let (norm_ctxt',cxmodule) =
	procmod_ctxt.normalization_phase.normalize_library_module norm_ctxt library_module
      in
      (norm_ctxt',CoreASTLibraryModule cxmodule)
  | ASTMainModule main_module ->
      let (norm_ctxt',cxmodule) =
	procmod_ctxt.normalization_phase.normalize_main_module norm_ctxt main_module
      in
      (norm_ctxt',CoreASTMainModule cxmodule)

let apply_rewriting_phase procmod_ctxt comp_mod new_norm_ctxt core_ast =
  (* 1. Obtain context information from the compiled module *)
  let orig_stat_ctxt = static_context_from_compiled_prolog comp_mod in

  (* 2. Set up the new static context *)
  let new_stat_ctxt =
    replace_norm_context_in_static_context new_norm_ctxt orig_stat_ctxt
  in
  match core_ast with
  | CoreASTStatement acstatement ->
      let rtcstatement =
	procmod_ctxt.rewriting_phase.rewriting_statement new_stat_ctxt acstatement
      in
      (new_stat_ctxt,CoreASTStatement rtcstatement)
  | CoreASTProlog acprolog ->
      let (new_stat_ctxt',rtcprolog) =
	procmod_ctxt.rewriting_phase.rewriting_prolog new_stat_ctxt acprolog
      in
      (new_stat_ctxt',CoreASTProlog rtcprolog)
  | CoreASTLibraryModule acxmodule ->
      let (new_stat_ctxt',rtcxmodule) =
	procmod_ctxt.rewriting_phase.rewriting_library_module new_stat_ctxt acxmodule
      in
      (new_stat_ctxt',CoreASTLibraryModule rtcxmodule)
  | CoreASTMainModule acxmodule ->
      let (new_stat_ctxt',rtcxmodule) =
	procmod_ctxt.rewriting_phase.rewriting_main_module new_stat_ctxt acxmodule
      in
      (new_stat_ctxt',CoreASTMainModule rtcxmodule)

let apply_factorization_phase procmod_ctxt comp_mod new_stat_ctxt core_ast =
  match core_ast with
  | CoreASTStatement acstatement ->
      let rtcstatement =
	procmod_ctxt.factorization_phase.factorize_statement new_stat_ctxt acstatement
      in
      (new_stat_ctxt,CoreASTStatement rtcstatement)
  | CoreASTProlog acprolog ->
      let (new_stat_ctxt',rtcprolog) =
	procmod_ctxt.factorization_phase.factorize_prolog new_stat_ctxt acprolog
      in
      (new_stat_ctxt',CoreASTProlog rtcprolog)
  | CoreASTLibraryModule acxmodule ->
      let (new_stat_ctxt',rtcxmodule) =
	procmod_ctxt.factorization_phase.factorize_library_module new_stat_ctxt acxmodule
      in
      (new_stat_ctxt',CoreASTLibraryModule rtcxmodule)
  | CoreASTMainModule acxmodule ->
      let (new_stat_ctxt',rtcxmodule) =
	procmod_ctxt.factorization_phase.factorize_main_module new_stat_ctxt acxmodule
      in
      (new_stat_ctxt',CoreASTMainModule rtcxmodule)


let apply_compile_phase procmod_ctxt comp_mod new_rewriting_ctxt rewritten_core_ast =
  (* 1. Obtain context information from the compiled module *)
  let orig_comp_ctxt = compile_context_from_compiled_prolog comp_mod in
  (* 2. Set up the new compilation context. *)
  let new_comp_ctxt =
    replace_static_context_in_compile_context new_rewriting_ctxt orig_comp_ctxt
  in
  match rewritten_core_ast with
  | CoreASTStatement rtcstatement ->
      let statement_plan =
	procmod_ctxt.compile_phase.compile_statement new_comp_ctxt rtcstatement
      in
      (new_comp_ctxt,LogicalPlanStatement statement_plan)
  | CoreASTProlog rtcprolog ->
      let (new_comp_ctxt',prolog_plan) =
	procmod_ctxt.compile_phase.compile_prolog new_comp_ctxt rtcprolog
      in
      (new_comp_ctxt',LogicalPlanProlog prolog_plan)
  | CoreASTLibraryModule rtcxmodule ->
      let (new_comp_ctxt',library_module_plan) =
	procmod_ctxt.compile_phase.compile_library_module new_comp_ctxt rtcxmodule
      in
      (new_comp_ctxt',LogicalPlanLibraryModule library_module_plan)
  | CoreASTMainModule rtcxmodule ->
      let (new_comp_ctxt',main_module_plan) =
	procmod_ctxt.compile_phase.compile_main_module new_comp_ctxt rtcxmodule
      in
      (new_comp_ctxt',LogicalPlanMainModule main_module_plan)

let apply_optimization_phase procmod_ctxt comp_mod new_comp_ctxt logical_plan =
  match logical_plan with
  | LogicalPlanStatement statement_plan ->
      let statement_plan' =
	procmod_ctxt.optimization_phase.optimize_statement new_comp_ctxt statement_plan
      in	
      (new_comp_ctxt,LogicalPlanStatement statement_plan')
  | LogicalPlanProlog prolog_plan ->
      let (new_comp_ctxt',prolog_plan') =
	procmod_ctxt.optimization_phase.optimize_prolog new_comp_ctxt prolog_plan
      in
      (new_comp_ctxt',LogicalPlanProlog prolog_plan')
  | LogicalPlanLibraryModule library_module_plan ->
      let (new_comp_ctxt',library_module_plan') =
	procmod_ctxt.optimization_phase.optimize_library_module new_comp_ctxt library_module_plan
      in
      (new_comp_ctxt',LogicalPlanLibraryModule library_module_plan')
  | LogicalPlanMainModule main_module_plan ->
      let (new_comp_ctxt',main_module_plan') =
	procmod_ctxt.optimization_phase.optimize_main_module new_comp_ctxt main_module_plan
      in
      (new_comp_ctxt',LogicalPlanMainModule main_module_plan')

let apply_code_selection_phase procmod_ctxt comp_mod new_comp_ctxt logical_plan =
  (* 1. Obtain context information from the compiled module *)
  let orig_cs_ctxt   = code_selection_context_from_compiled_prolog comp_mod in
    (* 2. Set up the new code selection context. *)
  let new_cs_ctxt =
    replace_compile_context_in_code_selection_context  (Cs_annotate.annotate_context new_comp_ctxt)  orig_cs_ctxt
  in
    match logical_plan with
      | LogicalPlanStatement statement_plan ->
	  let statement_plan' =
	    procmod_ctxt.selection_phase.selection_statement new_cs_ctxt statement_plan
	  in
	    (new_cs_ctxt,PhysicalPlanStatement statement_plan')
      | LogicalPlanProlog prolog_plan ->
	  let (new_cs_ctxt',prolog_plan') =
	    procmod_ctxt.selection_phase.selection_prolog new_cs_ctxt prolog_plan
	  in
	    (new_cs_ctxt',PhysicalPlanProlog prolog_plan')
      | LogicalPlanLibraryModule library_module_plan ->      
	  let (new_cs_ctxt',library_module_plan') =
	    procmod_ctxt.selection_phase.selection_library_module new_cs_ctxt library_module_plan
	  in
	    (new_cs_ctxt',PhysicalPlanLibraryModule library_module_plan')
      | LogicalPlanMainModule main_module_plan ->
	  let (new_cs_ctxt',main_module_plan') =
	    procmod_ctxt.selection_phase.selection_main_module new_cs_ctxt main_module_plan
	  in
	    (new_cs_ctxt',PhysicalPlanMainModule main_module_plan')


(*************************************)
(* Compositions of compilation steps *)
(*************************************)

(* Compile a physical plan from an input *)

let compile_physical_plan_from_input comp_mod ginput =
  (* 1. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 2. Create a processing phase handler *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 3. Parsing *)
  let ast =
    apply_parsing_phase procmod_ctxt comp_mod ginput
  in

  (* 4. Normalization *)
  let (new_norm_ctxt,core_ast) =
    apply_normalize_phase procmod_ctxt comp_mod ast
  in

  (* 6. Rewriting *)
  let (new_rewriting_ctxt,rewritten_core_ast) =
    apply_rewriting_phase procmod_ctxt comp_mod new_norm_ctxt core_ast
  in

  (* 7. Factorization *)
  let (new_factorized_ctxt,factorized_core_ast) =
    apply_factorization_phase procmod_ctxt comp_mod new_rewriting_ctxt rewritten_core_ast
  in

  (* 8. Compilation *)
  let (new_comp_ctxt,logical_plan) =
    apply_compile_phase procmod_ctxt comp_mod new_factorized_ctxt factorized_core_ast
  in

  (* 9. Optimization *)
  let (new_comp_ctxt,optimized_logical_plan) =
    apply_optimization_phase procmod_ctxt comp_mod new_comp_ctxt logical_plan
  in

  (* 10. Code selection *)
  apply_code_selection_phase procmod_ctxt comp_mod new_comp_ctxt optimized_logical_plan


(* Compile a physical plan from an input *)

let compile_typed_core_ast_from_input comp_mod ginput =
  (* 1. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 2. Create a processing phase handler *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 3. Parsing *)
  let ast =
    apply_parsing_phase procmod_ctxt comp_mod ginput
  in

  (* 4. Normalization *)
  let (new_norm_ctxt,core_ast) =
    apply_normalize_phase procmod_ctxt comp_mod ast
  in

  (* 6. Rewriting *)
  let (new_rewriting_ctxt,rewritten_core_ast) =
    apply_rewriting_phase procmod_ctxt comp_mod new_norm_ctxt core_ast
  in

  (* 7. Factorization *)
  apply_factorization_phase procmod_ctxt comp_mod new_rewriting_ctxt rewritten_core_ast


(* Compile a physical plan from an AST *)

let compile_physical_plan_from_ast comp_mod ast =
  (* 1. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 2. Create a processing phase handler *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 3. Normalization *)
  let (new_norm_ctxt,core_ast) =
    apply_normalize_phase procmod_ctxt comp_mod ast
  in

  (* 5. Rewriting *)
  let (new_rewriting_ctxt,rewritten_core_ast) =
    apply_rewriting_phase procmod_ctxt comp_mod new_norm_ctxt core_ast
  in

  (* 6. Factorization *)
  let (new_factorized_ctxt,factorized_core_ast) =
    apply_factorization_phase procmod_ctxt comp_mod new_rewriting_ctxt rewritten_core_ast
  in

  (* 7. Compilation *)
  let (new_comp_ctxt,logical_plan) =
    apply_compile_phase procmod_ctxt comp_mod new_factorized_ctxt factorized_core_ast
  in

  (* 8. Optimization *)
  let (new_comp_ctxt,optimized_logical_plan) =
    apply_optimization_phase procmod_ctxt comp_mod new_comp_ctxt logical_plan
  in

  (* 9. Code selection *)  
  apply_code_selection_phase procmod_ctxt comp_mod new_comp_ctxt optimized_logical_plan
    

(* Compile a typed core from an AST *)

let compile_typed_core_ast_from_ast comp_mod ast =
  (* 1. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 2. Create a processing phase handler *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 3. Normalization *)
  let (new_norm_ctxt,core_ast) =
    apply_normalize_phase procmod_ctxt comp_mod ast
  in

  (* 5. Rewriting *)
  let (new_rewriting_ctxt,rewritten_core_ast) =
    apply_rewriting_phase procmod_ctxt comp_mod new_norm_ctxt core_ast
  in

  (* 6. Factorization *)
  apply_factorization_phase procmod_ctxt comp_mod new_rewriting_ctxt rewritten_core_ast


(* Compile a physical plan from a logical plan *)

let compile_physical_plan_from_logical_plan comp_mod (new_comp_ctxt,logical_plan) =
  (* 1. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 2. Create a processing phase handler *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 3. Optimization *)
  let (new_comp_ctxt,optimized_logical_plan) =
    apply_optimization_phase procmod_ctxt comp_mod new_comp_ctxt logical_plan
  in

  (* 4. Code selection *)
  apply_code_selection_phase procmod_ctxt comp_mod new_comp_ctxt optimized_logical_plan


(* Compile a physical plan from a logical plan *)

let compile_physical_plan_from_optimized_logical_plan comp_mod (new_comp_ctxt,optimized_logical_plan) =
  (* 1. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 2. Create a processing phase handler *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 3. Code selection *)
  apply_code_selection_phase procmod_ctxt comp_mod new_comp_ctxt optimized_logical_plan


(*************************)
(* Statement compilation *)
(*************************)

let compile_statement_ast comp_mod statement =
  let new_comp_ctxt, physical_plan =
    compile_physical_plan_from_ast comp_mod (ASTStatement statement) 
  in
  match physical_plan with
  | PhysicalPlanStatement physical_statement_plan ->
      physical_statement_plan
  | _ ->
      raise (Query (Compilation "Statement compilation failed"))

let compile_statement comp_mod gio =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 1. Set up the processing context *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 2. Parse the statement *)
  let parse_ctxt = build_xquery_parse_context proc_ctxt in
  let statement = procmod_ctxt.parsing_phase.parse_statement parse_ctxt gio in

  (* 3. Compile the statement *)
  compile_statement_ast comp_mod statement

let compile_statement_typed_core_ast_from_ast comp_mod statement =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  let new_comp_ctxt, core_ast =
    compile_typed_core_ast_from_ast comp_mod (ASTStatement statement) 
  in
  match core_ast with
  | CoreASTStatement core_statement ->
      core_statement
  | _ ->
      raise (Query (Compilation "Statement compilation failed"))

let compile_statement_from_logical_plan comp_mod (new_comp_ctxt,statement_plan) =
  let new_comp_ctxt, physical_plan =
    compile_physical_plan_from_logical_plan comp_mod (new_comp_ctxt,(LogicalPlanStatement statement_plan))
  in
  match physical_plan with
  | PhysicalPlanStatement physical_statement_plan ->
      physical_statement_plan
  | _ ->
      raise (Query (Compilation "Statement compilation failed"))

let compile_statement_from_optimized_logical_plan comp_mod (new_comp_ctxt,optimized_statement_plan) =
  let new_comp_ctxt, physical_plan =
    compile_physical_plan_from_optimized_logical_plan comp_mod (new_comp_ctxt,(LogicalPlanStatement optimized_statement_plan))
  in
  match physical_plan with
  | PhysicalPlanStatement physical_statement_plan ->
      physical_statement_plan
  | _ ->
      raise (Query (Compilation "Statement compilation failed"))

(**********************)
(* Prolog compilation *)
(**********************)

(* Note:
     This function compiles an XQuery prolog in the context of a given
     compiled module.
   - Jerome
*)

let compile_prolog_ast ext_ctxt_item comp_mod prolog =
  (* 0. Fix-up the prolog with external context item *)
  let prolog = fix_prolog_with_external_context_item ext_ctxt_item prolog in

  (* 1. Obtain context information from the compiled module *)
  let orig_alg_ctxt  = algebra_context_from_compiled_prolog comp_mod in

  (* 2. Compile the prolog *)
  let (new_cs_ctxt',physical_plan) =
    compile_physical_plan_from_ast comp_mod (ASTProlog prolog)
  in

  (* 3. Extract the result *)
  let physical_prolog_plan =
    match physical_plan with
    | PhysicalPlanProlog physical_prolog_plan -> physical_prolog_plan
    | _ -> raise (Query (Compilation "Prolog compilation failed"))
  in

  (* 4. Merge new compiled prolog with the existing one. *)
  let new_prolog_plan =
    Xquery_algebra_ast_util.merge_alg_prologs
      comp_mod.compiled_prolog_plan
      physical_prolog_plan
  in

  (* 5. Create the final compiled module *)
  { compiled_prolog_external_context       = comp_mod.compiled_prolog_external_context;
    compiled_prolog_plan                   = new_prolog_plan;
    compiled_prolog_code_selection_context = new_cs_ctxt';
    compiled_prolog_algebra_context        = orig_alg_ctxt }

let compile_prolog_typed_core_ast_from_ast ext_ctxt_item comp_mod prolog =
  (* 0. Fix-up the prolog with external context item *)
  let prolog = fix_prolog_with_external_context_item ext_ctxt_item prolog in

  (* 1. Compile the prolog *)
  let (new_stat_ctxt,core_ast) =
    compile_typed_core_ast_from_ast comp_mod (ASTProlog prolog)
  in

  (* 2. Extract the result *)
  let core_prolog_ast =
    match core_ast with
    | CoreASTProlog acap ->
	acap
    | _ ->
	raise (Query (Compilation "Prolog compilation failed"))
  in
  (new_stat_ctxt,core_prolog_ast)


let compile_prolog ext_ctxt_item comp_mod gio =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 1. Set up the processing context *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 2. Parse the library module *)
  let parse_ctxt = build_xquery_parse_context proc_ctxt in
  let (new_parse_ctxt,prolog) = procmod_ctxt.parsing_phase.parse_prolog parse_ctxt gio in

  (* 3. Compile the prolog *)
  compile_prolog_ast ext_ctxt_item comp_mod prolog

let compile_prolog_from_logical_plan comp_mod (new_comp_ctxt,prolog_plan) =
  (* 1. Obtain context information from the compiled module *)
  let orig_alg_ctxt  = algebra_context_from_compiled_prolog comp_mod in

  let new_cs_ctxt, physical_plan =
    compile_physical_plan_from_logical_plan comp_mod (new_comp_ctxt,(LogicalPlanProlog prolog_plan))
  in
  let physical_prolog_plan =
    match physical_plan with
      | PhysicalPlanProlog physical_prolog_plan ->
	  physical_prolog_plan
      | _ ->
	  raise (Query (Compilation "Prolog compilation failed"))
  in
  (* 4. Merge new compiled prolog with the existing one. *)

  let new_prolog_plan =
    Xquery_algebra_ast_util.merge_alg_prologs
      comp_mod.compiled_prolog_plan
      physical_prolog_plan
  in

  (* 5. Create the final compiled module *)
  { compiled_prolog_external_context       = comp_mod.compiled_prolog_external_context;
    compiled_prolog_plan                   = new_prolog_plan;
    compiled_prolog_code_selection_context = new_cs_ctxt;
    compiled_prolog_algebra_context        = orig_alg_ctxt }

let compile_prolog_from_optimized_logical_plan comp_mod (new_comp_ctxt,optimized_prolog_plan) =
  (* 1. Obtain context information from the compiled module *)
  let orig_alg_ctxt  = algebra_context_from_compiled_prolog comp_mod in

  let new_cs_ctxt, physical_plan =
    compile_physical_plan_from_optimized_logical_plan comp_mod (new_comp_ctxt,(LogicalPlanProlog optimized_prolog_plan))
  in
  let physical_prolog_plan =
    match physical_plan with
      | PhysicalPlanProlog physical_prolog_plan ->
	  physical_prolog_plan
      | _ ->
	  raise (Query (Compilation "Prolog compilation failed"))
  in
  (* 4. Merge new compiled prolog with the existing one. *)

  let new_prolog_plan =
    Xquery_algebra_ast_util.merge_alg_prologs
      comp_mod.compiled_prolog_plan
      physical_prolog_plan
  in

  (* 5. Create the final compiled module *)
  { compiled_prolog_external_context       = comp_mod.compiled_prolog_external_context;
    compiled_prolog_plan                   = new_prolog_plan;
    compiled_prolog_code_selection_context = new_cs_ctxt;
    compiled_prolog_algebra_context        = orig_alg_ctxt }



(**********************)
(* Module compilation *)
(**********************)

let compile_library_module_ast comp_mod library_module =
  (* 1. Obtain context information from the compiled module *)
  let orig_alg_ctxt  = algebra_context_from_compiled_prolog comp_mod in

  (* 2. Compile the library_module *)
  let (new_cs_ctxt',physical_plan) =
    compile_physical_plan_from_ast comp_mod (ASTLibraryModule library_module)
  in

  (* 3. Extract the result *)
  let physical_library_module_plan =
    match physical_plan with
    | PhysicalPlanLibraryModule physical_library_module_plan ->
	physical_library_module_plan
    | _ ->
	raise (Query (Compilation "Library module compilation failed"))
  in

  (* 4. Merge new compiled prolog with the existing one. *)
  let new_prolog_plan =
    Xquery_algebra_ast_util.merge_alg_prologs
      comp_mod.compiled_prolog_plan
      physical_library_module_plan.palgop_module_prolog
  in

  (* 5. Create the final compiled module *)
  { compiled_prolog_external_context       = comp_mod.compiled_prolog_external_context;
    compiled_prolog_plan                   = new_prolog_plan;
    compiled_prolog_code_selection_context = new_cs_ctxt';
    compiled_prolog_algebra_context        = orig_alg_ctxt }

let compile_library_module comp_mod gio =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 1. Set up the processing context *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 2. Parse the library module *)
  let parse_ctxt = build_xquery_parse_context proc_ctxt in
  let (new_parse_ctxt,library_module) = procmod_ctxt.parsing_phase.parse_library_module parse_ctxt gio in

  (* 3. Compile the library_module *)
  compile_library_module_ast comp_mod library_module

let compile_main_module_typed_core_ast comp_mod gio =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 1. Compile the main module *)
  let (new_stat_ctxt, core_ast) =
    compile_typed_core_ast_from_input comp_mod (IOMainModule gio)
  in

  (* 2. Extract the result *)
  let core_ast_main_module =
    match core_ast with
    | CoreASTMainModule core_ast_main_module ->
	core_ast_main_module
    | _ ->
	raise (Query (Compilation "Main module compilation failed"))
  in
  (new_stat_ctxt,core_ast_main_module)

let compile_main_module ext_ctxt_item comp_mod gio =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 1. Set up the processing context *)
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in

  (* 2. Parse the main module *)
  let parse_ctxt = build_xquery_parse_context proc_ctxt in
  let (new_parse_ctxt,main_module) = procmod_ctxt.parsing_phase.parse_main_module parse_ctxt gio in

  (* 3. Fix-up the prolog with external context item *)
  let main_module = fix_main_module_with_external_context_item ext_ctxt_item main_module in

  (* 4. Obtain context information from the compiled module *)
  let orig_alg_ctxt  = algebra_context_from_compiled_prolog comp_mod in

  (* 5. Compile the whole module *)
  let (new_cs_ctxt',physical_plan) =
    compile_physical_plan_from_ast comp_mod (ASTMainModule main_module)
  in

  (* 6. Extract the result *)
  let physical_main_module_plan =
    match physical_plan with
    | PhysicalPlanMainModule physical_main_module_plan -> physical_main_module_plan
    | _ -> raise (Query (Compilation "Prolog compilation failed"))
  in

  (* 7. Separate the prolog from the statements *)
  let (physical_prolog_plan,comp_statements) = Xquery_algebra_ast_util.split_main_module_plan physical_main_module_plan in

  (* 8. Merge new compiled prolog with the existing one. *)
  let new_prolog_plan =
    Xquery_algebra_ast_util.merge_alg_prologs
      comp_mod.compiled_prolog_plan
      physical_prolog_plan
  in

  (* 9. Create the final compiled prolog *)
  let comp_mod' =
    { compiled_prolog_external_context    = comp_mod.compiled_prolog_external_context;
      compiled_prolog_plan                   = new_prolog_plan;
      compiled_prolog_code_selection_context = new_cs_ctxt';
      compiled_prolog_algebra_context        = orig_alg_ctxt }
  in

  (comp_mod', comp_statements)

let compile_main_module_from_logical_plan comp_mod (new_comp_ctxt,logical_plan) =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 1. Obtain context information from the compiled module *)
  let orig_alg_ctxt  = algebra_context_from_compiled_prolog comp_mod in

  (* 4. Compile the prolog *)
  let (new_cs_ctxt,comp_main_module) =
    compile_physical_plan_from_logical_plan comp_mod (new_comp_ctxt,LogicalPlanMainModule logical_plan)
  in
  let physical_plan= 
    match comp_main_module with
    | PhysicalPlanMainModule physical_main_module_plan ->
	physical_main_module_plan
    | _ ->
	raise (Query (Compilation "Main module compilation failed"))
  in
  let prolog =
    (* 5. Create the final compiled module *)
    { compiled_prolog_external_context       = comp_mod.compiled_prolog_external_context;
      compiled_prolog_plan                   = physical_plan.palgop_module_prolog;
      compiled_prolog_code_selection_context = new_cs_ctxt;
      compiled_prolog_algebra_context        = orig_alg_ctxt }
  in
  (prolog,physical_plan.palgop_module_statements)


let compile_main_module_from_optimized_logical_plan comp_mod (new_comp_ctxt,optimized_logical_plan) =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in

  (* 1. Obtain context information from the compiled module *)
  let orig_alg_ctxt  = algebra_context_from_compiled_prolog comp_mod in

  (* 4. Compile the prolog *)
  let (new_cs_ctxt,comp_main_module) =
    compile_physical_plan_from_optimized_logical_plan comp_mod (new_comp_ctxt,LogicalPlanMainModule optimized_logical_plan)
  in

  let physical_plan= 
    match comp_main_module with
    | PhysicalPlanMainModule physical_main_module_plan ->
	physical_main_module_plan
    | _ ->
	raise (Query (Compilation "Main module compilation failed"))
  in
  let prolog =
    (* 5. Create the final compiled module *)
    { compiled_prolog_external_context       = comp_mod.compiled_prolog_external_context;
      compiled_prolog_plan                   = physical_plan.palgop_module_prolog;
      compiled_prolog_code_selection_context = new_cs_ctxt;
      compiled_prolog_algebra_context        = orig_alg_ctxt }
  in
  (prolog,physical_plan.palgop_module_statements)


(***************************************)
(* Standard library module compilation *)
(***************************************)

let compile_standard_library_module proc_ctxt =
  (* Make sure the compilation is flagged as 'system' or
     fn: declarations will not work *)
  proc_ctxt.system <- true;
  (* Make sure to turn printing off for stdlib *)
  let printing = !Conf.print_global in
  let comp_mod =
    begin
      Conf.print_global := false;
	if !Conf.debugging then
	  begin
	    Format.printf "Parsing pervasives\n";
	    flush stdout
	  end;
      let compiled_prolog = (default_compiled_prolog proc_ctxt) in
      try
	let stdlib_io = (Galax_io.String_Input !Conf.pervasive_content) in
	compile_library_module compiled_prolog stdlib_io
      with
      | (Sys_error _) ->
          let stdlib_io = (Galax_io.File_Input "pervasive.xq") in
          compile_library_module compiled_prolog stdlib_io
    end
  in
  begin
    (* Turn printing back to its original setting *)
    Conf.print_global := printing;
    comp_mod
  end

let pre_compiled_standard_library_module =
  try
    try
      compile_standard_library_module (Processing_context.default_processing_context ())
    with
    | exn -> 
	raise (Query(Internal_Error("Cannot compile standard library: please contact the Galax maintainer\n\n" ^(Error.bprintf_error "" exn))))
  with
  | e ->
      begin
	eprintf_error "  " e;
	Format.fprintf (!Conf.glx_err_formatter) "@.";
	exit(1)
      end

let standard_processing_context =
  processing_context_from_compiled_prolog pre_compiled_standard_library_module
let standard_nsenv =
  nsenv_from_compiled_prolog pre_compiled_standard_library_module
let standard_norm_context =
  norm_context_from_compiled_prolog pre_compiled_standard_library_module
let standard_static_context =
  static_context_from_compiled_prolog pre_compiled_standard_library_module
let standard_compile_context =
  compile_context_from_compiled_prolog pre_compiled_standard_library_module
let standard_code_selection_context =
  code_selection_context_from_compiled_prolog pre_compiled_standard_library_module
let standard_algebra_context =
  algebra_context_from_compiled_prolog pre_compiled_standard_library_module

(************************************)
(* Perform copies for contexts that *)
(*   contain mutable structures     *)
(************************************)

let get_standard_processing_context     () = standard_processing_context
let get_standard_nsenv                  () = standard_nsenv
let get_standard_norm_context           () = standard_norm_context
let get_standard_static_context         () = standard_static_context 
let get_standard_compile_context        () = copy_compile_context standard_compile_context
let get_standard_code_selection_context () = copy_code_selection_context standard_code_selection_context
let get_standard_algebra_context        () = copy_algebra_context standard_algebra_context

(*****************************)
(* Execute a compiled module *)
(*****************************)

let execute_compiled_prolog opt_ext_ctxt comp_mod =
   (* Initialize the external context right before execution *)
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog comp_mod in
  let comp_mod = set_external_context opt_ext_ctxt comp_mod in
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in
  let alg_ctxt = comp_mod.compiled_prolog_algebra_context in
  let (alg_ctxt',res) =
    procmod_ctxt.evaluation_phase.eval_prolog
      alg_ctxt
      comp_mod.compiled_prolog_plan
  in
  { compiled_prolog_external_context       = comp_mod.compiled_prolog_external_context;
    compiled_prolog_plan = comp_mod.compiled_prolog_plan;
    compiled_prolog_code_selection_context = comp_mod.compiled_prolog_code_selection_context;
    compiled_prolog_algebra_context = alg_ctxt' }


(********************************)
(* Execute a compiled statement *)
(********************************)

let execute_compiled_statement prepared_prolog comp_statement =
  (* 0. Get processing context *)
  let proc_ctxt = processing_context_from_compiled_prolog prepared_prolog in
  let procmod_ctxt = phase_handler_of_processing_context proc_ctxt in
  let alg_ctxt = prepared_prolog.compiled_prolog_algebra_context in
  procmod_ctxt.evaluation_phase.eval_statement alg_ctxt comp_statement


