(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2004, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: procmod_compiler.mli,v 1.17 2006/02/02 21:03:17 mff Exp $ *)

(* Module: Procmod_compiler
   Description:
     This module contains top-level operations over the processing
     model. This notably supports various levels of compilations for
     queries.
*)

open Xquery_common_ast
open Xquery_ast
open Xquery_core_ast
open Xquery_algebra_ast
open Logical_algebra_types
open Algebra_type


open Dm_atomic
open Physical_value

open Processing_context
open Namespace_context
open Parse_context
open Norm_context
open Typing_context
open Compile_context
open Code_selection_context
open Execution_context
open Monitoring_context



(*******************************************)
(* Type for intermediate compilation steps *)
(*******************************************)

type ginput =
  | IOStatement of Galax_io.input_spec
  | IOProlog of Galax_io.input_spec
  | IOLibraryModule of Galax_io.input_spec
  | IOMainModule of Galax_io.input_spec

type ast =
  | ASTStatement of statement
  | ASTProlog of prolog
  | ASTLibraryModule of library_module
  | ASTMainModule of main_module

type core_ast =
  | CoreASTStatement of acstatement
  | CoreASTProlog of acprolog
  | CoreASTLibraryModule of acxmodule
  | CoreASTMainModule of acxmodule


type logical_plan =
  | LogicalPlanStatement of logical_algop_expr
  | LogicalPlanProlog of logical_algop_prolog
  | LogicalPlanLibraryModule of logical_algop_xmodule
  | LogicalPlanMainModule of logical_algop_xmodule

type physical_plan =
  | PhysicalPlanStatement of algop_expr
  | PhysicalPlanProlog of algop_prolog
  | PhysicalPlanLibraryModule of algop_xmodule
  | PhysicalPlanMainModule of algop_xmodule


(*********************************)
(* Types used during compilation *)
(*********************************)

(* External context *)

type external_context =
    { external_context_item : item option;
      external_timezone     : atomicDayTimeDuration option;
      external_variables    : (cvname * item list) list }

(* Compiled statements *)

type logical_plan_statement = logical_algop_expr
type compiled_statement = algop_expr

(* Compiled prolog *)

type logical_plan_prolog = logical_algop_prolog
type compiled_prolog =
    { compiled_prolog_external_context       : external_context option; 
      compiled_prolog_plan  	             : algop_prolog;
      compiled_prolog_code_selection_context : code_selection_context;
      compiled_prolog_algebra_context        : algebra_context }

(* Compiled module *)

type logical_plan_main_module = logical_algop_xmodule
type compiled_library_module = compiled_prolog
type compiled_main_module = compiled_prolog * compiled_statement list

(* In a prepared prolog, all global variables' values are bound: *)
type prepared_prolog = compiled_prolog


(*******************************)
(* The default compiled prolog *)
(*******************************)

(* A processing_context is always defined externally, because it
   can effect compilation *)
val default_compiled_prolog : Processing_context.processing_context -> compiled_prolog


(****************************)
(* Module context accessors *)
(****************************)

val processing_context_from_compiled_prolog : compiled_prolog -> processing_context
val nsenv_from_compiled_prolog              : compiled_prolog -> nsenv
val norm_context_from_compiled_prolog       : compiled_prolog -> norm_context
val static_context_from_compiled_prolog     : compiled_prolog -> static_context
val compile_context_from_compiled_prolog    : compiled_prolog -> logical_alg_compile_context
val code_selection_context_from_compiled_prolog : compiled_prolog -> code_selection_context
val algebra_context_from_compiled_prolog    : compiled_prolog -> algebra_context

(********************)
(* External context *)
(********************)

val create_external_context :
    processing_context ->
      (item option) ->
	(atomicDayTimeDuration option) ->
	  (string * item list) list ->
	    external_context

val default_external_context : unit -> external_context

(*************************)
(* Statement compilation *)
(*************************)

val compile_statement :
    compiled_prolog -> Galax_io.input_spec -> compiled_statement
val compile_statement_ast :
    compiled_prolog -> statement -> compiled_statement
val compile_statement_typed_core_ast_from_ast :
    compiled_prolog -> statement -> acstatement
val compile_statement_from_logical_plan :
    compiled_prolog -> (logical_alg_compile_context * logical_plan_statement) -> compiled_statement
val compile_statement_from_optimized_logical_plan :
    compiled_prolog -> (logical_alg_compile_context * logical_plan_statement) -> compiled_statement


(**********************)
(* Prolog compilation *)
(**********************)

val compile_prolog :
    bool (* external context item *) -> compiled_prolog -> Galax_io.input_spec -> compiled_prolog
val compile_prolog_typed_core_ast_from_ast :
    bool (* external context item *) -> compiled_prolog -> prolog -> (static_context * acprolog)
val compile_prolog_ast :
    bool (* external context item *) -> compiled_prolog -> prolog -> compiled_prolog
val compile_prolog_from_logical_plan :
    compiled_prolog -> (logical_alg_compile_context * logical_plan_prolog) -> compiled_prolog
val compile_prolog_from_optimized_logical_plan :
    compiled_prolog -> (logical_alg_compile_context * logical_plan_prolog) -> compiled_prolog


(**********************)
(* Module compilation *)
(**********************)

val compile_library_module :
    compiled_prolog -> Galax_io.input_spec -> compiled_library_module

val compile_main_module :
    bool (* external context item *) -> compiled_prolog -> Galax_io.input_spec -> compiled_main_module
val compile_main_module_typed_core_ast :
    compiled_prolog -> Galax_io.input_spec -> (static_context * acxmodule)
val compile_main_module_from_logical_plan :
    compiled_prolog -> (logical_alg_compile_context * logical_plan_main_module) -> compiled_main_module
val compile_main_module_from_optimized_logical_plan :
    compiled_prolog -> (logical_alg_compile_context * logical_plan_main_module) -> compiled_main_module


(***************************************)
(* Standard library module compilation *)
(***************************************)

val compile_standard_library_module : processing_context -> compiled_library_module

val pre_compiled_standard_library_module : compiled_library_module

val standard_processing_context     : processing_context
val standard_nsenv                  : nsenv
val standard_norm_context   	    : norm_context
val standard_static_context 	    : static_context
val standard_compile_context        : logical_alg_compile_context
val standard_code_selection_context : code_selection_context
val standard_algebra_context        : algebra_context

val get_standard_processing_context     : unit -> processing_context
val get_standard_nsenv                  : unit -> nsenv
val get_standard_norm_context   	: unit -> norm_context
val get_standard_static_context 	: unit -> static_context
val get_standard_compile_context        : unit -> logical_alg_compile_context
val get_standard_code_selection_context : unit -> code_selection_context
val get_standard_algebra_context        : unit -> algebra_context


(*****************************)
(* Execute a compiled prolog *)
(*****************************)

val execute_compiled_prolog :
    external_context option -> compiled_prolog -> prepared_prolog

(********************************)
(* Execute a compiled statement *)
(********************************)

val execute_compiled_statement :
    prepared_prolog -> compiled_statement -> physical_value

