(***********************************************************************)
(*                                                                     *)
(*                               Galax                                 *)
(*                            XQuery Engine                            *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: test_core.ml,v 1.19 2006/05/16 23:33:59 simeon Exp $ *)

(* Module: Test_core
   Description:
     Contains the core code of the test harness.
 *)

open Galax
open Galax_io

open Error
open Processing_context

open Datatypes
open Dm_atomic
open Dm

open Physical_value
open Physical_item

open Streaming_types
open Streaming_util
open Streaming_ops
open Streaming_constructors
open Streaming_diff


(*************)
(* Utilities *)
(*************)

(* Errors *)

let raise_test_error msg =
  raise (Query (Testing_Error msg))

let raise_harness_error msg =
  raise (Query (Internal_Error msg))

(* Context *)

(* We don't statically check any of the queries applied to the catalog itself. *)
let default_proc_ctxt =
  begin
    let pc = default_processing_context() in
    pc.serialization_kind <- Serialize_As_Standard;
    pc.typing_kind <- Typing_None; 
    pc
  end

(* By default, all test cases are weak type checked, except for the
   "StaticTyping" group, which is strong type checked. *)
let proc_ctxt =
  begin
    let pc = default_processing_context() in
    pc.serialization_kind <- Serialize_As_Standard;
    pc.typing_kind <- Typing_Weak; 
    pc
  end

let reset_proc () =
  begin
    proc_ctxt.boundary_space_kind <- Xquery_common_ast.Strip;
    proc_ctxt.base_uri <- AnyURI.default_base_uri()
  end
(*
  set_proc proc_ctxt default_proc_ctxt
*)

let defaultcp () = load_standard_library default_proc_ctxt (* No typing *)
let cp () = load_standard_library proc_ctxt

(* DM conversions *)

let get_string (il : item list) : xs_string =
  let x = List.nth il 0 in
  (getAtomicValue x)#getAtomicString()

let get_int (il : item list) : int =
  let x = List.nth il 0 in
  Decimal._int_of_integer ((getAtomicValue x)#getAtomicInteger())

let get_string_list (il : item list) : xs_string list =
  List.map (fun x -> (getAtomicValue x)#getAtomicString()) il

let get_item (il : item list) : item =
  List.nth il 0

let make_from_string (s : string) : item list =
  [(Item_Atomic (new atomicString s))]

let make_from_int (i : int) : item list =
  [(Item_Atomic (new atomicInteger (Decimal._integer_of_int i)))]

let get_item_list (il : item list) : item list = il

let eval_to_string eval query dot =
  get_string (eval query dot)

let eval_to_int eval query dot =
  get_int (eval query dot)

let eval_to_string_list eval query dot =
  get_string_list (eval query dot)

let eval_to_item eval query dot =
  get_item (eval query dot)

let eval_to_item_list eval query dot =
  get_item_list (eval query dot)

(* Streams *)

let typed_stream_of_io gio =
  let (_,s) = Streaming_parse.open_xml_stream_from_io gio in
  let rs = resolve_xml_stream s in
  typed_of_resolved_xml_stream rs

let cleaned_event event =
  match event.tse_desc with
  | TSAX_characters s ->
      let newchars =
	Whitespace.remove_newlines s
      in
      fmktse_event (TSAX_characters newchars) event.tse_loc
  | _ -> event

let cleaned_stream s =
  let next_event xml_stream () =
    try
      let next_event = Cursor.cursor_next xml_stream in
      Some (cleaned_event next_event)
    with
    | Stream.Failure ->
	None
  in
  Cursor.cursor_of_function (next_event s)

let cleaned_typed_stream_of_io gio =
  let s = typed_stream_of_io gio in
  cleaned_stream s

(* Fragments *)

let open_fragment = "<glx:result xmlns:glx=\"http://www.galaxquery.org\">"
let close_fragment = "</glx:result>"

let buffer_of_fragment file =
  let nb = Netbuffer.create 100 in
  begin
    Netbuffer.add_string nb open_fragment;
    let s = Gmisc.load_file file in
    (* let s = Whitespace.remove_trailing_spaces s in *)
    Netbuffer.add_string nb s;
    Netbuffer.add_string nb close_fragment;
    nb
  end

let typed_stream_of_fragment file =
  let buf = buffer_of_fragment file in
  let gio = Buffer_Input buf in
  typed_stream_of_io gio

let typed_stream_of_text_fragment file =
  let buffer = (buffer_of_fragment file) in
  let gio = Buffer_Input buffer in
  cleaned_typed_stream_of_io gio

let string_of_text_fragment file =
  let lh = Lexing_util.init_xquery_lexing () in
  let s = "\"" ^ (Gmisc.quote_quotes (Gmisc.load_file file)) ^ "\"" in
  let gin =
    Parse_io.galax_utf8_input_from_input_spec
      proc_ctxt
      (Galax_io.String_Input s)
  in
  let lexbuf = Parse_io.lexbuf_from_galax_input gin in
  let s1 = Parse_xquery.stringlit (Parse_util.xquery_lexfun lh) lexbuf in
  let target_encoding = `Enc_utf8 in
  let internal_encoding = `Enc_utf8 in
  Parse_io.close_galax_input gin;
  Encoding.write_data_string internal_encoding target_encoding s1



(*******************)
(* Data structures *)
(*******************)

(* Configuration *)

type unit_config =
    { unit_id       : string;
      unit_name     : string;
      testsuite_dir : string;
      catalog_dir   : string;
      query_dir     : string;
      sources       : (string,item list) Hashtbl.t;
      expected_dir  : string;
      catalog_file  : string;
      catalog_doc   : item option ref;
      unit_version  : string ref;
      temp_groups   : string list;
      test_count    : int ref }

type test_config =
    { bugs_file   : string;
      result_file : string;
      units       : (string * unit_config) list }

(* Bugs *)

type bug_location =
  | GalaxBug
  | XQTSBug

type bug =
    { bug_name : string;
      bug_id   : int;
      bug_in   : bug_location }

(* Results *)

type eval_kind =
  | Eval_KnownBug of int * bug_location
  | Eval_Succeeded of item list
  | Eval_Failed of exn

type case_kind =
  | Case_Pass
  | Case_Pass_Inspect of string
  | Case_Fail of string
  | Case_Bug of int * string

type case_result =
    { case_result_name : string;
      case_result_kind : case_kind }

(* Catalog *)

type compare_kind =
  | XML_Comparison of string
  | Fragment_Comparison of string
  | Text_Comparison of string
  | Error_Comparison of string     (* Error code *)
  | Ignore_Comparison of string
  | Inspect_Comparison of string
(* Type_Comparison *)

type test_case =
    { test_name : string;
      test_path : string;
      test_scen : string;
      test_desc : string;
      test_qfil : string;
      test_ifil : (string * string) list;
      test_qinp : (string * string) list;
      test_comp : compare_kind list }

type group_result = case_result list ref

type group_kind =
  | Subgroup of group list * group_result
  | Leafgroup of group_result

and group =
    { group_item : item;
      group_name : string;
      group_info : string;
      group_subs : group_kind }


(*********************)
(* Config processing *)
(*********************)

(* Create a new configuration *)

let make_unit_config id un td cd qd ed cad gps =
  let catdir = Filename.concat td cd in
  let querydir = Filename.concat catdir qd in
  let expecteddir = Filename.concat catdir ed in
  let catalogfile = Filename.concat catdir cad in
  { unit_id       = id;
    unit_name     = un;
    testsuite_dir = td;
    catalog_dir   = catdir;
    query_dir     = querydir;
    sources       = Hashtbl.create 59;
    expected_dir  = expecteddir;
    catalog_file  = catalogfile;
    catalog_doc   = ref None;
    unit_version  = ref "";
    temp_groups   = gps;
    test_count    = ref 0 }

let make_test_config testdir bf rf unit_configs =
  Printf.printf "Loading configuration...";flush stdout;
  let bugsfile = Filename.concat testdir bf in
  let resultfile = Filename.concat testdir rf in
  let conf =
    { bugs_file   = bugsfile;
      result_file = resultfile;
      units       = unit_configs }
  in
  conf

(* Those can be set explicitely *)

(* Configuration file *)

let ccp = import_prolog_only (defaultcp()) true (String_Input "")

let eval_config query dot = 
  let ec = build_external_context default_proc_ctxt (Some dot) None [] in
  let pp = eval_prolog ccp (Some ec) in
  eval_statement pp (String_Input query)

let query_test_dir    = "./config/galax-test/testdir/string()"
let query_bug_file    = "./config/galax-test/bugfile/string()"
let query_result_file = "./config/galax-test/resultfile/string()"
let query_units       = "./config/test-unit"

let query_id          = "./@id/string()"
let query_name        = "./@name/string()"
let query_testdir     = "./testdir/string()"
let query_catalogdir  = "./catalogdir/string()"
let query_catalog     = "./catalog/string()"
let query_querydir    = "./querydir/string()"
let query_expecteddir = "./expecteddir/string()"
let query_groups      = "./groups/test-group/@name/string()"

let id unit      = eval_to_string eval_config query_id unit
let name unit      = eval_to_string eval_config query_name unit
let testdir unit = eval_to_string eval_config query_testdir unit
let catalogdir unit = eval_to_string eval_config query_catalogdir unit
let catalog unit = eval_to_string eval_config query_catalog unit
let querydir unit = eval_to_string eval_config query_querydir unit
let expecteddir unit = eval_to_string eval_config query_expecteddir unit
let groups unit = eval_to_string_list eval_config query_groups unit

let make_config_from_unit unit =
  let id = id unit in
  let name = name unit in
  let testdir = testdir unit in
  let catalogdir = catalogdir unit in
  let catalog = catalog unit in
  let querydir = querydir unit in
  let expecteddir = expecteddir unit in
  let groups = groups unit in
  (id,make_unit_config
     id
     name
     testdir
     catalogdir
     querydir
     expecteddir
     catalog
     groups)

let make_test_config configdoc =
  let test_dir = eval_to_string eval_config query_test_dir configdoc in
  let bug_file = eval_to_string eval_config query_bug_file configdoc in
  let result_file = eval_to_string eval_config query_result_file configdoc in
  let units = eval_to_item_list eval_config query_units configdoc in
  make_test_config
    test_dir
    bug_file
    result_file
    (List.map make_config_from_unit units)

let top_make_test_config file =
  let make_config_doc config_file =
    Item_Node (List.nth (load_document proc_ctxt (File_Input config_file)) 0)
  in
  let configdoc = make_config_doc file in
  let c = make_test_config configdoc in
  begin
    Printf.printf "done.\n\n";flush stdout;
    c
  end

(**********************)
(* Catalog processing *)
(**********************)

(* The catalog *)

let catalog_prolog =
  "declare default element namespace \"http://www.w3.org/2005/02/query-test-XQTSCatalog\";"

let dcp = import_prolog_only (defaultcp()) true (String_Input catalog_prolog)

let eval_catalog query dot =
  let ec = build_external_context default_proc_ctxt (Some dot) None [] in
  let pp = eval_prolog dcp (Some ec) in
  eval_statement pp (String_Input query)

let query_version = "./test-suite/@version/string()"

let query_schemas = "./test-suite/sources/schema"
let query_schema_id = "./@uri/string()"
let query_schema_loc = "./@FileName/string()"

let set_location_hints dir catalog =
  let get_location_hint schema =
    (eval_to_string eval_catalog query_schema_id schema,
     Filename.concat dir (eval_to_string eval_catalog query_schema_loc schema))
  in
  let schemas = eval_to_item_list eval_catalog query_schemas catalog in
  let lhints = List.map get_location_hint schemas in
  proc_ctxt.external_schema_locs <- lhints

let make_catalog config () =
  Printf.printf "Loading catalog...";flush stdout;
  let ct = config.catalog_doc in
  match !ct with
  | Some _ -> ()
  | None ->
      let catalog =
	Item_Node
	  (List.nth (load_document proc_ctxt (File_Input config.catalog_file)) 0)
      in
      begin
	(* Added here to avoid re-loading it... *)
	Hashtbl.add config.sources "XQTSCatalog" [catalog];
	let version =
	  try
	    eval_to_string eval_catalog query_version catalog
	  with
	  | _ -> "" (* Not provided *)
	in
	config.unit_version := version;
	set_location_hints config.catalog_dir catalog;
	ct := Some catalog;
	Printf.printf "done.\n\n";flush stdout
      end

let get_catalog_from_config config =
  match !(config.catalog_doc) with
  | None -> raise_harness_error ("Catalog for unit [" ^ config.unit_id ^ "] not found")
  | Some doc -> doc

(* Generic catalog processing code *)

let query_initial_group = "./test-suite/test-group"
let query_group_name = "./@name/string()"
let query_group_info = "./GroupInfo/title/string()"
let query_group_subs = "./test-group"
let query_group_case = ".//test-case"
let query_group_case_here = "./test-case"

let query_case_name = "./@name/string()"
let query_case_path = "./@FilePath/string()"
let query_case_scen = "./@scenario/string()"
let query_case_desc = "./description/string()"
let query_case_qfil = "./query/@name/string()"
let query_case_inpt = "./input-file"
let input_query_case_inpt = "./input-query"
let query_case_ifil = "./string()"
let query_case_ivar = "./@variable/string()"
let input_query_case_ifil = "./@name/string()"
let input_query_case_ivar = "./@variable/string()"

let query_expected = "expected-error | output-file"

let query_case_comp = "if (fn:exists(self::expected-error)) then \"Error\" else self::output-file/@compare/string()"
let query_case_ofil = "if (fn:exists(self::expected-error)) then self::expected-error/string() else self::output-file/string()"

let query_initial_sources = "./test-suite/sources"

let build_input_file inpt =
  let ifil = eval_to_string eval_catalog query_case_ifil inpt in
  let ivar = eval_to_string eval_catalog query_case_ivar inpt in
  (ivar,ifil)

let build_input_query_file inpt =
  let ifil = eval_to_string eval_catalog input_query_case_ifil inpt in
  let ivar = eval_to_string eval_catalog input_query_case_ivar inpt in
  (ivar,ifil)

let make_comparison_kind expected =
  let comp = eval_to_string eval_catalog query_case_comp expected in
  let ofil = eval_to_string eval_catalog query_case_ofil expected in
  match comp with
  | "XML"      -> XML_Comparison ofil
  | "Fragment" -> Fragment_Comparison ofil
  | "Text"     -> Text_Comparison ofil
  | "Error"    -> Error_Comparison ofil
  | "Ignore"   -> Ignore_Comparison ofil
  | "Inspect"  -> Inspect_Comparison ofil
  | _ ->
      raise_test_error ("Unknown kind of comparison: [" ^ comp ^ "]")

let build_case case =
  let name = eval_to_string eval_catalog query_case_name case in
  let path = eval_to_string eval_catalog query_case_path case in
  let scen = eval_to_string eval_catalog query_case_scen case in
  let desc = eval_to_string eval_catalog query_case_desc case in
  let qfil = eval_to_string eval_catalog query_case_qfil case in
  let inpt = eval_to_item_list eval_catalog query_case_inpt case in
  let inpt = List.map build_input_file inpt in
  let qinp = eval_to_item_list eval_catalog input_query_case_inpt case in
  let qinp = List.map build_input_query_file qinp in
  let expected = eval_to_item_list eval_catalog query_expected case in
  let comp = List.map make_comparison_kind expected in
  { test_name = name;
    test_path = path;
    test_scen = scen;
    test_desc = desc;
    test_qfil = qfil ^ ".xq";
    test_ifil = inpt;
    test_qinp = qinp;
    test_comp = comp }

let rec build_groups group =
  let name = eval_to_string eval_catalog query_group_name group in
  let info = eval_to_string eval_catalog query_group_info group in
  let subgroups = eval_catalog query_group_subs group in
  let subs =
    List.map build_groups subgroups
  in
  { group_item = group;
    group_name = name;
    group_info = info;
    group_subs = if subgroups = [] then Leafgroup (ref []) else Subgroup (subs, ref []) }

let groups init_group =
  List.map build_groups init_group

let get_all_groups config () =
  let unit_catalog = get_catalog_from_config config in
  let init_group = eval_to_item_list eval_catalog query_initial_group unit_catalog in
  groups init_group


(*******************)
(* Bugs processing *)
(*******************)

let get_bugs_file config = config.bugs_file

let make_bugs_document bugs_file =
  Item_Node (List.nth (load_document proc_ctxt (File_Input bugs_file)) 0)

let bcp = import_prolog_only (defaultcp()) true (String_Input "")

let eval_bug query dot =
  let ec = build_external_context default_proc_ctxt (Some dot) None [] in
  let pp = eval_prolog bcp (Some ec) in
  eval_statement pp (String_Input query)

let query_bug       = "./bugs/bug[@status=\"open\"]"
let query_bug_loc   = "xs:string(./@in)"
let query_bug_id    = "xs:integer(./@id)"
let query_bug_names = "./test-case/@name/string()"


let make_bug_with_loc bug_id bug_loc testname =
  { bug_name = testname;
    bug_id = bug_id;
    bug_in = bug_loc }

let make_bug bug =
  let bug_id = eval_to_int eval_bug query_bug_id bug in
  let bug_loc =
    match eval_to_string eval_bug query_bug_loc bug with
    | "galax" -> GalaxBug
    | "xqts" -> XQTSBug
    | _ -> raise_test_error ("Unknow bug location in bugs file...")
  in
  let bug_names = eval_to_string_list eval_bug query_bug_names bug in
  List.map (fun x -> make_bug_with_loc bug_id bug_loc x) bug_names

let get_all_bugs config =
  let bugs_file = get_bugs_file config in
  let bugs_document = make_bugs_document bugs_file in
  let all_bugs = eval_to_item_list eval_bug query_bug bugs_document in
  List.concat ((List.map make_bug) all_bugs)


(*******************)
(* Compare results *)
(*******************)

let make_unexpected_fail case error =
  let name = case.test_name in
  let msg =
    match error with
    | Query err ->
	let errorname = bprintf_error_safe "" (Query err) in
	"Test case [" ^ name ^ "] raised error [" ^ errorname ^ "] :-("
    | _ ->
	"Test case [" ^ name ^ "] failed with a FATAL ERROR!! :-((("
  in
  { case_result_name = name;
    case_result_kind = Case_Fail msg }

let make_harness_fail case error =
  let name = case.test_name in
  let errorname = bprintf_error_safe "" error in
  let msg =
    "Test case or harness raised a really bad error [" ^ errorname ^ "] :( :("
  in
  { case_result_name = name;
    case_result_kind = Case_Fail msg }

let make_known_bug case error bug_loc =
  match bug_loc with
  | GalaxBug ->
      let name = case.test_name in
      let msg =
	"Known bug :-|"
      in
      { case_result_name = name;
	case_result_kind = Case_Bug (error,msg) }
  | XQTSBug ->
      let name = case.test_name in
      let msg =
	"[INSPECT] Known XQTS bug :-|"
      in
      { case_result_name = name;
	case_result_kind = Case_Pass_Inspect msg }

let make_unexpected_succeed case result =
  let name = case.test_name in
  let msg =
    "Test case [" ^ name ^ "] succeeded :-("
  in
  { case_result_name = name;
    case_result_kind = Case_Fail msg }

let make_pass case =
  let name = case.test_name in
  { case_result_name = name;
    case_result_kind = Case_Pass }

let make_pass_inspect case =
  let name = case.test_name in
  { case_result_name = name;
    case_result_kind = Case_Pass_Inspect "[INSPECT]" }

let make_differ case output_file =
  let name = case.test_name in
  let msg = "Result of case [" ^ name ^ "] and expected result in [" ^ output_file ^ "] differ :-(" in
  { case_result_name = name;
    case_result_kind = Case_Fail msg }

let compare_streams case output_file result_stream expected_stream =
  let comp =
    stream_boolean_diff result_stream expected_stream
  in
  if comp
  then
    make_pass case
  else
    make_differ case output_file

let try_serialize case result =
  match case.test_comp with
  | [Error_Comparison _] ->
      let result_stream =
	let result_stream1 =
	  Physical_export.typed_xml_stream_of_datamodel (Cursor.cursor_of_list result)
	in
	sequence_normalization result_stream1
      in
      discard_typed_xml_stream result_stream
  | _ -> ()

let compare_xml case output_file result =
  let expected_stream = typed_stream_of_io (File_Input output_file) in
  let result_stream =
    let result_stream1 =
      Physical_export.typed_xml_stream_of_datamodel (Cursor.cursor_of_list result)
    in
    sequence_normalization result_stream1
  in
  compare_streams case output_file result_stream expected_stream

let compare_many comp case output_file result =
  try
    comp case output_file result
  with
  | error ->
     make_harness_fail case error

let compare_strings case output_file expected_string result_string =
  let comp =
    (Whitespace.remove_whitespace expected_string) =
    (Whitespace.remove_whitespace result_string)
  in
  if comp
  then
    make_pass case
  else
    make_differ case output_file

let compare_fragments case output_file result =
  let result_stream =
    let result_stream1 =
      Physical_export.typed_xml_stream_of_datamodel (Cursor.cursor_of_list result)
    in
    let result_stream2 =
      glx_result_serialization result_stream1
    in
    sequence_normalization result_stream2
  in
  let expected_stream = typed_stream_of_fragment output_file in
  compare_streams case output_file result_stream expected_stream

let compare_text_fragments case output_file result =
  let result_string =
    let result_stream1 =
      Physical_export.typed_xml_stream_of_datamodel (Cursor.cursor_of_list result)
    in
    let result_stream2 =
      sequence_normalization result_stream1
    in
    Serialization.bserialize_typed_xml_stream proc_ctxt result_stream2
  in
  let expected_string = string_of_text_fragment output_file in
  compare_strings case output_file expected_string result_string

(* Old code comparing through XML diff
  let result_stream =
    let result_stream1 =
      Physical_export.typed_xml_stream_of_datamodel (Cursor.cursor_of_list result)
    in
    let result_stream2 =
      glx_result_serialization result_stream1
    in
    sequence_normalization result_stream2
  in
  let expected_stream = typed_stream_of_text_fragment output_file in
  compare_streams case output_file result_stream expected_stream
*)

let compare_error case expected_error actual_error =
  (* Don't worry about it for now *)
  make_pass case

let compare_results case output_dir result test_comp =
  match test_comp with
  | XML_Comparison ofil ->
      let output_file = Filename.concat output_dir ofil in
      begin
	match result with
	| Eval_KnownBug (bugid,bugloc) ->
	    make_known_bug case bugid bugloc
	| Eval_Succeeded result ->
	    compare_many compare_xml case output_file result
	| Eval_Failed error ->
	    make_unexpected_fail case error
      end
  | Fragment_Comparison ofil ->
      let output_file = Filename.concat output_dir ofil in
      begin
	match result with
	| Eval_KnownBug (bugid,bugloc) ->
	    make_known_bug case bugid bugloc
	| Eval_Succeeded result ->
	    compare_many compare_fragments case output_file result
	| Eval_Failed error ->
	    make_unexpected_fail case error
      end
  | Text_Comparison ofil ->
      let output_file = Filename.concat output_dir ofil in
      begin
	match result with
	| Eval_KnownBug (bugid,bugloc) ->
	    make_known_bug case bugid bugloc
	| Eval_Succeeded result ->
	    compare_many compare_text_fragments case output_file result
	| Eval_Failed error ->
	    make_unexpected_fail case error
      end
  | Error_Comparison err ->
      begin
	match result with
	| Eval_KnownBug (bugid,bugloc) ->
	    make_known_bug case bugid bugloc
	| Eval_Succeeded result ->
	    make_unexpected_succeed case result
	| Eval_Failed error ->
	    compare_many compare_error case err error
      end
  | Ignore_Comparison ofil ->
      raise_test_error "Test harness cannot perform ignore comparison yet"
  | Inspect_Comparison ofil ->
      make_pass_inspect case


(************************)
(* Test case processing *)
(************************)

let cases_of_group group =
  let group_item = group.group_item in
  let subcase = eval_catalog query_group_case group_item in
  List.map build_case subcase

let cases_of_group_here group =
  let group_item = group.group_item in
  let subcase = eval_catalog query_group_case_here group_item in
  List.map build_case subcase

let rec group_iter cond f all_groups =
  match all_groups with
  | [] -> ()
  | g :: groups' ->
      let subgroups =
	match g.group_subs with
	| Subgroup (group_list,_) -> group_list
	| Leafgroup _ -> []
      in
      if (cond g)
      then
	begin
	  (f g);
	  (group_iter (fun x -> true) f subgroups);
	  (group_iter cond f groups')
	end
      else
	begin
	  (group_iter cond f (subgroups @ groups'))
	end

let rec group_map_concat cond f all_groups =
  match all_groups with
  | [] -> []
  | g :: groups' ->
      let subgroups =
	match g.group_subs with
	| Subgroup (group_list,_) -> group_list
	| Leafgroup _ -> []
      in
      if (cond g)
      then
	begin
	  (f g) @ (group_map_concat (fun x -> true) f subgroups) @ (group_map_concat cond f groups')
	end
      else
	(group_map_concat cond f (subgroups @ groups'))

let rec group_map cond f all_groups =
  match all_groups with
  | [] -> []
  | g :: groups' ->
      let subgroups =
	match g.group_subs with
	| Subgroup (group_list,_) -> group_list
	| Leafgroup _ -> []
      in
      if (cond g)
      then
	begin
	  (f g) :: (group_map (fun x -> true) f subgroups) @ (group_map cond f groups')
	end
      else
	(group_map cond f (subgroups @ groups'))

let rec find_group_by_name_aux all_groups group_name =
  match all_groups with
  | [] -> []
  | g :: groups' ->
      let subgroups =
	match g.group_subs with
	| Subgroup (group_list,_) -> group_list
	| Leafgroup _ -> []
      in
      if g.group_name = group_name
      then
	g :: (find_group_by_name_aux (subgroups @ groups') group_name)
      else
	find_group_by_name_aux (subgroups @ groups') group_name

let find_group_by_name all_groups group_name =
  try
    let groups = find_group_by_name_aux all_groups group_name in
    List.nth groups 0
  with
  | _ ->
      raise_test_error ("Group [" ^ group_name ^ "] is unknown")

let cases_of_group_name all_groups group_name =
  let group = find_group_by_name all_groups group_name in
  cases_of_group group

let lookup_source_file config sourceid =
  let unit_catalog = get_catalog_from_config config in
  let init_sources = eval_to_item eval_catalog query_initial_sources unit_catalog in
  let query_source = "./source[@ID=\""^ sourceid ^"\"]/@FileName/string()" in
  eval_to_string eval_catalog query_source init_sources

let create_var config (var,sourceid) =
  if Hashtbl.mem config.sources sourceid
  then
    (var,Hashtbl.find config.sources sourceid)
  else
    let input_file = Filename.concat config.catalog_dir (lookup_source_file config sourceid) in
    let input_doc =
      [(Item_Node (List.nth (load_document proc_ctxt (File_Input input_file)) 0))]
    in
    Hashtbl.add config.sources sourceid input_doc;
    (var,input_doc)

let create_qinp_var case_dir config (var,sourceid) =
  if Hashtbl.mem config.sources sourceid
  then
    (var,Hashtbl.find config.sources sourceid)
  else
    let f = Filename.concat case_dir (sourceid ^ ".xq") in
    let input_doc =
      let (prolog,statements) = import_main_module false (cp ()) (File_Input f) in
      let statement = List.nth statements 0 in
      let pp = eval_prolog prolog None in
      eval_compiled_statement pp statement
    in
    Hashtbl.add config.sources sourceid input_doc;
    (var,input_doc)

let wrap_result all_bugs case eval_case query_file vars create_prolog_var prologvars =
  try
    let bug =
      List.find (fun x -> x.bug_name = case.test_name) all_bugs
    in
    Eval_KnownBug (bug.bug_id,bug.bug_in)
  with
  | Not_found ->
      begin
	try
	  let vars_qinp = List.map create_prolog_var  prologvars in
	  let result = eval_case query_file (vars @ vars_qinp) in
	  begin
	    try_serialize case result;
	    Eval_Succeeded result
	  end
	with
	| error ->
	    Eval_Failed error
      end

let eval_case query vars =
  let (prolog,statements) = import_main_module false (cp ()) (File_Input query) in
  let statement = List.nth statements 0 in
  let ec = build_external_context proc_ctxt None None vars in
  let pp = eval_prolog prolog (Some ec) in
  eval_compiled_statement pp statement

let rec merge_results case compared_results =
  match compared_results with
  | [] -> raise_harness_error ("Not expected result for test" ^ case.test_name)
  | res :: [] -> res
  | res1 :: others ->
      begin
	match res1.case_result_kind with
	| Case_Fail _ -> merge_results case others
	| _ -> res1
      end

let run_case config all_bugs case =
  reset_proc ();
  let count = config.test_count in
  incr count;
  if (!count) mod 10 = 0 then (Printf.printf ".";flush stdout);
  if (!count) mod 100 = 0 then (Printf.printf "%i\n" !count;flush stdout);
  let case_dir = Filename.concat config.query_dir case.test_path in
  let output_dir = Filename.concat config.expected_dir case.test_path in
  let query_file = Filename.concat case_dir case.test_qfil in
  let vars = List.map (create_var config) case.test_ifil in
  let result = wrap_result all_bugs case eval_case query_file vars (create_qinp_var case_dir config) case.test_qinp in 
(* (vars @ vars_qinp) in *)
  let test_comps = case.test_comp in
  let compared_results = List.map (compare_results case output_dir result) test_comps in
  merge_results case compared_results

let run_group_cases name config all_bugs cases =
  let typing_kind = proc_ctxt.typing_kind  in
  if (name = "StaticTyping") then proc_ctxt.typing_kind <- Typing_Strong;   
  let results = List.map (run_case config all_bugs) cases 
  in 
  proc_ctxt.typing_kind <- typing_kind;   
  results

let clean_groups all_groups =
  let cond g = true in
  let f g =
    match g.group_subs with
    | Subgroup (_,group_result) -> group_result := []
    | Leafgroup group_result ->
	group_result := []
  in
  group_iter cond f all_groups

let run_groups config all_bugs =
  Printf.printf "Running tests now:\n\n";flush stdout;
  let group_names = config.temp_groups in
  let all_groups = get_all_groups config () in
  let cond g = List.exists (fun x -> g.group_name = x) group_names in
  let f g =
    match g.group_subs with
    | Subgroup (_,group_result) ->
	let cases = cases_of_group_here g in
	let result = run_group_cases g.group_name config all_bugs cases in
	group_result := result
    | Leafgroup group_result ->
	let cases = cases_of_group g in
	let result = run_group_cases g.group_name config all_bugs cases in
	group_result := result
  in
  group_iter cond f all_groups;
  Printf.printf "Number of tests run: %i\n\n" (!(config.test_count));flush stdout;
  all_groups

let is_failed case =
  match case.case_result_kind with
  | Case_Fail _ -> true
  | _ -> false

let is_bug case =
  match case.case_result_kind with
  | Case_Bug _ -> true
  | _ -> false

let is_passed case =
  match case.case_result_kind with
  | Case_Pass | Case_Pass_Inspect _ -> true
  | _ -> false

let is_not_passed case =
  match case.case_result_kind with
  | Case_Pass | Case_Pass_Inspect _ -> false
  | _ -> true

let get_all_results_with_kind all_groups kind_test =
  let cond g = true in
  let f g =
    match g.group_subs with
    | Subgroup (_,group_result) -> List.filter kind_test !group_result
    | Leafgroup group_result ->
	List.filter kind_test !group_result
  in
  group_map_concat cond f all_groups

let get_all_results all_groups =
  get_all_results_with_kind all_groups (fun x -> true)

let get_all_failed_results all_groups =
  get_all_results_with_kind all_groups is_failed

let get_all_bug_results all_groups =
  get_all_results_with_kind all_groups is_bug

let get_all_passed_results all_groups =
  get_all_results_with_kind all_groups is_passed

let get_all_not_passed_results all_groups =
  get_all_results_with_kind all_groups is_not_passed

let all_group_names all_groups =
  let cond g = true in
  let f g = g.group_name in
  group_map cond f all_groups


(*********************)
(* Result processing *)
(*********************)

let eval_result query vars =
  let (prolog,statements) = import_main_module false (defaultcp()) (String_Input query) in
  let statement = List.nth statements 0 in
  let ec = build_external_context default_proc_ctxt None None vars in
  let pp = eval_prolog prolog (Some ec) in
  eval_compiled_statement pp statement

let eval_result_as_sax query vars =
  let (prolog,statements) = import_main_module false (defaultcp()) (String_Input query) in
  let statement = List.nth statements 0 in
  let ec = build_external_context default_proc_ctxt None None vars in
  let pp = eval_prolog prolog (Some ec) in
  eval_compiled_statement_as_sax pp statement

let query_make_passed = "declare variable $name external;(text { \"\n  \" },<test-case name=\"{$name}\" result=\"pass\"/>)"
let query_make_passed_inspect = "declare variable $name external;declare variable $comment external;(text { \"\n  \" },<test-case name=\"{$name}\" result=\"pass\" comment=\"{$comment}\"/>)"
let query_make_failed = "declare variable $name external;declare variable $comment external;(text { \"\n  \" },<test-case name=\"{$name}\" result=\"fail\" comment=\"{$comment}\"/>)"
let query_make_bug = "declare variable $name external;declare variable $comment external;declare variable $bug external;(text { \"\n  \" },<test-case name=\"{$name}\" result=\"fail\" comment=\"{$comment} [http://bugzilla.galaxquery.net/show_bug.cgi?id={$bug}]\"/>)"

let create_result_passed case_name =
  let vars = [("name",make_from_string case_name)] in
  eval_result query_make_passed vars

let create_result_passed_inspect case_name msg =
  let vars = [("name",make_from_string case_name);("comment",make_from_string msg)] in
  eval_result query_make_passed_inspect vars

let create_result_failed case_name msg =
  let vars = [("name",make_from_string case_name);("comment",make_from_string msg)] in
  eval_result query_make_failed vars

let create_result_bug case_name msg bug =
  let vars = [("name",make_from_string case_name);("comment",make_from_string msg);("bug",(make_from_int bug))] in
  eval_result query_make_bug vars

let create_case_result case_result =
  let case_name = case_result.case_result_name in
  match case_result.case_result_kind with
  | Case_Pass ->
      create_result_passed case_name
  | Case_Pass_Inspect msg ->
      create_result_passed_inspect case_name msg
  | Case_Fail msg ->
      create_result_failed case_name msg
  | Case_Bug (bug,msg) ->
      create_result_bug case_name msg bug

let query_result version =
  let date =
    DateTime.string_of_date
      (DateTime.date_from_dateTime (DateTime.current_dateTime()))
  in
  "declare variable $name external;declare variable $result external;(text { \"\n\n  \"},<test-run dateRun=\"" ^ date ^ "\"><test-suite version=\"" ^ version ^ "\"/></test-run>,text { \"\n  \"},$result)"

let total_results name version case_results =
  let l = List.concat (List.map create_case_result case_results) in
  let vars = [("name",make_from_string name);("result",l)] in
  eval_result (query_result version) vars

let query_final_result =
  "declare default element namespace \"http://www.w3.org/2005/02/query-test-XQTSResult\";declare variable $result external;<test-suite-result>{text { \"\n \"}}<implementation name=\"Galax\" version=\"0.5.1\" anonymous-result-column=\"false\">{text { \"\n  \"}}<organization name=\"Galax Team\" website=\"http://www.galaxquery.org\"/>{text { \"\n  \"}}<submittor name=\"Jerome Simeon\"/></implementation>{$result,text{\"\n\"}}</test-suite-result>"

let total_result res =
  let l = List.concat (List.map (fun (x,v,y) -> total_results x v y) res) in
  let vars = ["result",l] in
  eval_result_as_sax query_final_result vars

