(******************************************************************************)
(*                                                                            *)
(*                                 GALAX                                      *)
(*                              XQuery Engine                                 *)
(*                                                                            *)
(*  Copyright 2001-2003, Lucent Technologies.                                 *)
(*  Distributed only by permission.                                           *)
(*                                                                            *)
(******************************************************************************)

(* $Id: rewriting_ddo_annot.ml,v 1.14 2006/04/04 21:16:29 mff Exp $ *)

(* Module: Rewriting_ddo_annot
     Description: This module implements detection and removal of
     unnecessary distinct-docorder operation from the query plan.
*)

open Error

open Namespace_builtin
open Namespace_names

open Xquery_core_ast
open Xquery_core_ast_util
open Xquery_core_ast_annotation

open Processing_context

open Ast_walker_rewrite

open Rewriting_ddo_automaton
open Rewriting_ddo_util

open Print_top

let debug message =
  if !Conf.debugging then 
    Printf.printf "In module ANNOT_DDO: %s \n" message

(**
  * Temporary primitive typing function that is used to derive the max1 property
  * in the absence of weak or full blown static typing.
  **)
let hasMaxOne (ce:acexpr) =
    (* Since the introduction of fs:node-sequence calls in the normalization of
     * all slash expressions, we need to peek inside the argument of the call
     *)
  let is_atomic_max_one ce =
    begin
    let {
      pcexpr_desc = ce_desc; 
      pcexpr_annot = annot; 
      pcexpr_origin = eh; 
      pcexpr_loc = loc;} = ce 
    in
    match (get_xpath_annot annot) with
    | Some Xp_document_call
    | Some Xp_root_call
    | Some Xp_pos_pred_axis_step -> true
    | Some _
    | None -> match ce_desc with 
            | CEEmpty 
            | CEVar(_, _, "dot") -> true 
            (* for calls that occur outside xpath expressions *)
	    | CECall((prefix,uri,name), _, _) -> 
	       let uri_matches = (uri = fn_uri || uri = empty_uri) in
	       let name_matches = (name = "doc" || name = "document" || name= "root") in
                 (uri_matches && name_matches)
            | _ -> false
    end
  in 
  match ce.pcexpr_desc with
  | CECall (fs_node_sequence, [arg], sign) ->
    let ce = arg in
    is_atomic_max_one ce
  | _ ->
    is_atomic_max_one ce	

(**
 * This function triggers a new instance of the DDO automaton for a given
 * core expression, marking all distinct-docorder operations with a ddo
 * annotation indicating whether it should be removed or replaced.
 *
 * @param symtab A symbol table containing the different states for the
 *  visible variables of the expression
 * @parem ce The core expression to be annotated
 **)
let rec start_new_automaton 
  rewrite_ctxt 
  (sbdo: sbdo_kind)
  (symtab: st)
  (ce: acexpr) 
  : unit =
    let (_,_,_) = process_ast_node rewrite_ctxt ce symtab sbdo (init_state sbdo) Nothing true false in
    ()

(**
 * This is the core function of the SBDO optimization. It works at the level of 
 * an AST node and recursively process any child nodes.
 *
 * General operation:
 * Instead of trying to identify (parts of) path expressions from the core AST
 * directly, we define default handlers for all core expressions. Whenever we
 * do not have enough information to infer whether or not a part of the AST was
 * generated by normalizing an XPath expression, we use the XPath annotaations 
 * to obtain that information. Once we know an AST node does not belong to an
 * XPath normalization, we shift our automaton into the sink state.
 * 
 * There is one important parameter in this function: before_slash.
 * This parameter indicates whether or not transitions have already been made 
 * (i.e., we are at the right hand side of a slashexpr)
 * inside the currently processed path expressions normalization. This allows
 * us to deduce when the precedence in path expressions has been changed.
 * Basically, it comes down to this: whenever a FLWOR expression occurs in the
 * return statement of another FLWOR that is part of an XPath normalization, this 
 * means that either precedence is altered, or that it concerns the 
 * normalization of an expression falling outside the supported XPath fragment.
 * (Note that the use of predicates will also cause let expressions to occur
 * in the return statement of for loops, but we can use the XPath annotations to
 * detect these exceptions)
 *
 * MODUS OPERANDUS : see comments inside code
 *
 * @param ce The core expression to annotate
 * @param symtab The symbol table binding all variables to their respective states
 * @param sbdo The kind of automaton that has to be used
 * @param cur_state The current state
 * @param cur_action The current action that is needed wrt. the last transition made
 * @param before_slash The parameter indicating whether transitions have already been
 *   been made inside the current path expression
 * @param is_nested_step Variable indicating whether the current step expressions (if any)
 *   is the last step of a path (needed for sloppy and duptidy automaton)
 *
 * @return (state * action_needed  * bool) The triplet after processing this AST node
 *)
and process_ast_node 
  rewrite_ctxt
  (ce: acexpr) 
  (symtab: st) 
  (sbdo: sbdo_kind)
  (cur_state: state) 
  (cur_action: action_needed) 
  (before_slash: bool) 
  (is_nested_step: bool)
  : (state * action_needed  * bool) =
  let {
    pcexpr_desc = ce_desc; 
    pcexpr_annot = annot; 
    pcexpr_origin = eh; 
    pcexpr_loc = loc;} = ce 
  in
  debug (String.concat "" ["Entering AST node  with state:";
                     (print_state cur_state);
		     "; action:";
		     (print_action cur_action);
		     "; before:";
		     (string_of_bool before_slash);]);
		     
  match ce_desc with
  (*
   Calls can be parts of path expressions. For us, however, only
   distinct-docorder calls are important since they are the target
   of this optim
   *)
  | CECall (fname, args, sign) ->
    (
    (* only important if called function is ddo, or if max1 holds *)
    if ((fname = fs_distinct_docorder_or_atomic_sequence  || fname = fs_distinct_docorder) && before_slash) then
    begin
        (* ... *)
        (* continue with argument expression and  handle according to the returned state *)
        let (new_state, action, is_bs) = 
          process_ast_node rewrite_ctxt  (List.hd args) symtab sbdo cur_state cur_action before_slash is_nested_step in
	  (*Printf.printf "Marking ddo-call with  %s %s\n" (print_state new_state) (print_action action);*)
          (match action with
	    | DistinctDocorder ->
	       set_ddo_annot annot Needs_distinct_docorder
	    | Distinct ->
	       set_ddo_annot annot Needs_distinct
	    | Docorder ->
	       set_ddo_annot annot Needs_docorder
	    | Nothing ->
	       set_ddo_annot annot Needs_none);
        (new_state, action, is_bs)
    end
    else if (fname = fs_node_sequence) && before_slash then
    begin
      (* part of path expressions: we should just recursively process the argument *)
      let (new_state, new_action, bs) =
	  process_ast_node rewrite_ctxt  (List.hd args) symtab sbdo cur_state cur_action before_slash is_nested_step
      in
	  (new_state, new_action, bs)	  
    end
    else
      begin
      (* first apply new automaton to all args, then return sink state *)
      List.iter (start_new_automaton rewrite_ctxt sbdo symtab) args;
      if (hasMaxOne ce) then
        (init_state sbdo, Nothing, true)
      else
        (sink_state sbdo, DistinctDocorder, before_slash)
      end;
   )
  | CEForwardAxis (v, axis, cnode_test)
  | CEReverseAxis (v, axis, cnode_test) ->
    (* Axes perform a transition on the current state *)
    let (new_state, new_action) = do_transition sbdo cur_state (axis_adapter axis) (not is_nested_step) in
      debug (String.concat "" ["Automaton ended up in state "; (print_state new_state); " Action: "; (print_action new_action)]);
      (new_state, new_action, false)
    
  (* FLWOR expressions are important because
       (1) by registering let bindings in the symbol table we can optimize 
           expressions like let $x := e1 return $x/e2 as if this is
           the expression is e1/e2 -- a big improvement over the previous
           implementation
       (2) Since the iteration variable of for loops introduces the max1
           property we can register this iteration variable in the
           symbol table with the max1 property
       (3) They represent normalizations of parts of XPath expressions
       
   *)
  | CEFLWOR (cfl_list, cfl_where, cfl_order_by, cfl_return) ->
    begin
    
    match (get_xpath_annot annot) with
    | Some Xp_flwor ->
       debug "XPath FLWOR expression (from slash) detected"; 
       process_flwor_slash ce_desc rewrite_ctxt symtab sbdo cur_state cur_action before_slash is_nested_step Xp_flwor
    | Some Xp_pred_axis_step ->
       debug "XPath FLWOR expression (from predicate step) detected";
       process_flwor_slash ce_desc rewrite_ctxt symtab sbdo cur_state cur_action before_slash is_nested_step Xp_pred_axis_step
    | Some Xp_pos_pred_axis_step ->
       debug "XPath FLWOR expression (from positional predicate step) detected";
       process_flwor_slash ce_desc rewrite_ctxt symtab sbdo cur_state cur_action before_slash is_nested_step Xp_pos_pred_axis_step
    | _ ->
       process_flwor_general ce_desc rewrite_ctxt symtab sbdo cur_state cur_action before_slash is_nested_step

    end

    | CEVar (prefix, uri, name) ->
    (* variable references return the state associated to the vraiables' entry in 
       the symbol table *)
    (* fixme: what should action_needed be in this case?
       maybe we should also store it in the symtab??
     *)
      begin
	let state = (get_state sbdo symtab (prefix, uri, name)) in
(* print_string ("\nGet var: "^(Namespace_names.prefixed_string_of_rqname (prefix,uri,name)^" : "^(print_state state))); *)
	(state, cur_action, before_slash)
      end

  | CEIf (e1,e2,e3) ->
    (* this one is more complex -- 
       Which result state do we return? The one form e2 or the one from e3?
       
       Answer: We need to define a lattice over all possible
       states (in terms of positive properties) and return the weakest common 
       descendant state. Until we do so, we will use the following lattice:
       
       init_state (e.g. when max1 holds)
              <
       All other states
              <
	 Sink state
	
       For instance, what happens for normalized predicates is that one
       branch returns the $dot variable and the other returns the empty sequence.
       Since for both of them max1 holds, the state outside the IF is not
       influenced.
       
       *)
       
       (* it is very possible -or even most likely- that path expressions
          are continued from the same context inside predicates (or even
	  plain if-structures). This is why we continue processing them
	  with the current state. However, the conditional of an IF never
	  influences the state outside the IF-construct directly, so the 
	  resulting state gets swallowed. *)
       let (_,_,_) = process_ast_node rewrite_ctxt  e1 symtab sbdo cur_state cur_action before_slash is_nested_step in
       
       let (state_e2, action_e2, bs_e2) =
         process_ast_node rewrite_ctxt  e2 symtab sbdo cur_state cur_action before_slash is_nested_step in
       let (state_e3, action_e3, bs_e3) =
         process_ast_node rewrite_ctxt  e3 symtab sbdo cur_state cur_action before_slash is_nested_step in
           if (hasMaxOne e2 && hasMaxOne e3) then
	     (* both states have the max1 property -- nothing changes *)
             (cur_state, cur_action, before_slash)
	   else if (state_e2 = sink_state sbdo || state_e3 = sink_state sbdo) then
	     (* if at least one of the branches returns a sink state
	        the result is always a sink_state *)
             (sink_state sbdo, DistinctDocorder, before_slash)
	   else if (hasMaxOne e2) then
	     (state_e3, action_e3, bs_e3)
	   else if (hasMaxOne e3) then
	     (state_e2, action_e2, bs_e2)
	   else
             (* There is no weakest common descnendant state above the sink *)
	     (sink_state sbdo, cur_action, before_slash)
	     
  (* Typeswitches can occur in normalized path expressions so we propagate the
     current state. This should be ok since if it does not concern a path 
     expression we will be propagating the sink state *)
  | CETypeswitch (e1, clauses) ->
    let map_clause clause = 
      let (cpattern, vname, ce) = clause in 
        start_new_automaton rewrite_ctxt sbdo symtab ce in 
    List.iter map_clause clauses;
    process_ast_node rewrite_ctxt  e1 symtab sbdo cur_state cur_action before_slash is_nested_step 

  | CEEmpty ->
    (cur_state, DistinctDocorder, before_slash)
	
  | CEOverloadedCall (fname, arguments, sign) -> 
      List.iter (start_new_automaton rewrite_ctxt sbdo symtab) arguments;
      (undefined_state sbdo, DistinctDocorder, before_slash)
  
  | CEError el -> 
      List.iter (start_new_automaton rewrite_ctxt sbdo symtab) el;
      (undefined_state sbdo, DistinctDocorder, before_slash)

  | CETextComputed e 
  | CEOrdered e 
  | CEDocument e 
  | CECommentComputed e 
  | CETreat (e,_)
  | CECast (e,_,_)
  | CECastable (e,_,_)
  | CEUnordered e
  | CEValidate (_,e) ->
      start_new_automaton rewrite_ctxt sbdo symtab e; 
      (undefined_state sbdo, DistinctDocorder, before_slash)

  | CEElem (_,_,e)
  | CEAttr (_,e) -> 
      List.iter (start_new_automaton rewrite_ctxt sbdo symtab) e; 
      (undefined_state sbdo, DistinctDocorder, before_slash)

  | CEAnyElem (e1,_,e2)
  | CEAnyAttr (e1,_,e2)
  | CESeq (e1,e2)
  | CESome (_,_, e1, e2)
  | CEEvery(_,_, e1, e2) ->
     start_new_automaton rewrite_ctxt sbdo symtab e1;
     start_new_automaton rewrite_ctxt sbdo symtab e2;
     (undefined_state sbdo, DistinctDocorder, before_slash)

  | CEText _
  | CEScalar _
  | CEProtoValue _
  | CEPI (_,_)
  | CEPIComputed (_,_)
  | CEComment _
  | CEKeyRef (_,_) ->
    (undefined_state sbdo, DistinctDocorder, before_slash)

(* Added some stuff for updates. I don't have a good understanding of
   what the automaton is doing, so PLEASE check that I'm not screwing
   up everything. - Jerome 10/20/2004 *)
(* I start a new automaton for each sub-expression, and I return the
   current state, which is what happens for the empty sequence (),
   since atomic updates are always returning the empty sequence, that
   should be consistent. - Jerome *)

  | CECopy acexpr1 ->
      begin
	start_new_automaton rewrite_ctxt sbdo symtab acexpr1;
	(cur_state, DistinctDocorder, before_slash)
      end
  | CEDelete acexpr1 ->
      begin
	start_new_automaton rewrite_ctxt sbdo symtab acexpr1;
	(cur_state, DistinctDocorder, before_slash)
      end
  | CEDetach acexpr1 ->
      begin
	start_new_automaton rewrite_ctxt sbdo symtab acexpr1;
	(cur_state, DistinctDocorder, before_slash)
      end
  | CEInsert (acexpr1, acinsert_location) ->
      begin
	start_new_automaton rewrite_ctxt sbdo symtab acexpr1;
	begin
	  match acinsert_location with
	  | CUAsLastInto cexpr2
	  | CUAsFirstInto cexpr2
	  | CUInto cexpr2
	  | CUAfter cexpr2
	  | CUBefore cexpr2 ->
	      start_new_automaton rewrite_ctxt sbdo symtab cexpr2
	end;
	(cur_state, DistinctDocorder, before_slash)
      end
  | CERename (acexpr1,acexpr2) ->
      begin
	start_new_automaton rewrite_ctxt sbdo symtab acexpr1;
	start_new_automaton rewrite_ctxt sbdo symtab acexpr2;
	(cur_state, DistinctDocorder, before_slash)
      end
  | CEReplace (value_of_flag,acexpr1,acexpr2) ->
      begin
	start_new_automaton rewrite_ctxt sbdo symtab acexpr1;
	start_new_automaton rewrite_ctxt sbdo symtab acexpr2;
	(cur_state, DistinctDocorder, before_slash)
      end
  | CESnap (sm,acexpr) ->
      begin
	start_new_automaton rewrite_ctxt sbdo symtab acexpr;
	(cur_state, DistinctDocorder, before_slash)
      end
  (* Mary: I'm assuming that CEExecute behaves a lot like CESnap *)
  | CEExecute (host, port,acexpr) ->
      begin
	start_new_automaton rewrite_ctxt sbdo symtab host;
	start_new_automaton rewrite_ctxt sbdo symtab port;
	start_new_automaton rewrite_ctxt sbdo symtab acexpr;
	(cur_state, DistinctDocorder, before_slash)
      end


(** General processing rule for FLWOR expressions -- i.e., anything but path expr's**)
and process_flwor_general
            ce_flwor
            rewrite_ctxt
            (symtab: st) 
            (sbdo: sbdo_kind) 
            (cur_state: state) 
            (cur_action: action_needed) 
            (before_slash: bool) 
            (is_nested_step: bool)
   : (state * action_needed  * bool) =

   let map_cfl cfl = 
      match cfl with
        | CEFOR(_, v, _, e1') ->
	  register_var symtab v (init_state sbdo);
	  start_new_automaton rewrite_ctxt sbdo symtab e1';
        | CELET(_, v, e1') ->
	  (* FIXME: general flwors fall outside the supported frament: shouldn't we start a new automaton then? *)
	  let (new_state, new_action, bs) =
	    process_ast_node rewrite_ctxt  e1' symtab sbdo cur_state cur_action before_slash is_nested_step in
	      if (hasMaxOne e1') then
	        register_var symtab v (init_state sbdo)
	      else
	        register_var symtab v new_state
	      ;
    in
    
     match ce_flwor with
     | CEFLWOR (cfl_list, cfl_where, cfl_order_by, cfl_return) ->
      begin
        (* 1 process the list of For and Let expressions *)
        List.iter map_cfl cfl_list;
        let (ns,na,bs) = 
          process_ast_node rewrite_ctxt cfl_return symtab sbdo cur_state cur_action before_slash is_nested_step in
	  
        (* 2 Process WHERE clause *)
        (match cfl_where with
        | Some cfl_where_expr ->
           start_new_automaton rewrite_ctxt sbdo symtab cfl_where_expr;
        | None -> ());

        (* 3 process ORDER BY clause *)
        (match cfl_order_by with 
          | Some (_, order_by_list, osig) ->
              let process_order_by (order_by_ce,_,_) =
                start_new_automaton rewrite_ctxt sbdo symtab order_by_ce
              in
                List.iter process_order_by order_by_list
          | None -> ()
        );
    
      (* 4 unregister variables that were registered in step 1 *)
       let map_cfl cfl =
         match cfl with
         | CEFOR(_, v, _, e1')
         | CELET(_, v, e1') ->
             unregister_var symtab v;
       in
         List.iter map_cfl cfl_list;
        (sink_state sbdo, na, bs)
     end
     | _ -> raise (Query (Rewriting "In annot_ddo.ml, function 'process_flwor_general' is not applied to a FLWOR expression."))

(** Processing rule for FLWOR expressions, that are the result of
  * normalizing the SLASH (/) operator.
  *
  * We know that the list of Let and For expression contains
  * (a) a Let binding for $sequence
  * (b) a Let binding for $last
  * (c) a For iteration for the step expression
  **)
and process_flwor_slash
            ce_flwor
            rewrite_ctxt
            (symtab: st) 
            (sbdo: sbdo_kind) 
            (cur_state: state) 
            (cur_action: action_needed) 
            (before_slash: bool) 
            (is_nested_step: bool)
	    xpath_annot
   : (state * action_needed  * bool) =
 
   (* is this a predicate step ? *)
   let pred = (xpath_annot = Xp_pos_pred_axis_step || xpath_annot = Xp_pred_axis_step) in
   
   (* This local function defines the behavior for the for/let clauses of the FLWOR *)
    let map_cfl cfl = 
      match cfl with
        | CEFOR(_, v, _, e1) ->
	  begin
	  let (state_e1, action_e1, bs_e1) =
	    process_ast_node rewrite_ctxt  e1 symtab sbdo cur_state cur_action before_slash is_nested_step
	  in
          register_var symtab v (init_state sbdo);
	  (if (!Conf.debugging) then Printf.printf "Leaving FOR (from slash) with %s %s %b\n" (print_state state_e1) (print_action action_e1) bs_e1);
	  (state_e1, action_e1, bs_e1)
	  end
        | CELET(_, v, e1) ->
	  if (hasMaxOne(e1)) then
	  begin
	    register_var symtab v (init_state sbdo);
	    (init_state sbdo, Nothing, before_slash)
	  end
	  else
	  begin
	    let (state_e1, action_e1, bs_e1) =
	      (* if on left hand side of a slash or if this is a predicate step, then pass before_slash as true
	          THIS IS CRUCIAL for removing the fs:distinct-docorder calls in normalized filter steps *)
	      process_ast_node rewrite_ctxt e1 symtab sbdo cur_state cur_action (before_slash || pred) is_nested_step in
	    register_var symtab v state_e1;
            (state_e1, action_e1, bs_e1)
	  end
    in

    match ce_flwor with
     | CEFLWOR (cfl_list, cfl_where, _, cfl_return) ->
       if (before_slash || pred) then
       begin
         let states = List.map map_cfl cfl_list in
	 let (for_state, for_action, for_bs) = List.hd (List.rev states) in
	 let (let_state, let_action, let_bs) = List.hd states in
	   (* let_bs and for_bs indicate wether axis steps occured so far 
	      -- thus wether we are at the left hand side of a slash 
	      if we are not and ddo-calls occur in the return-clause, we cannot remove them 
	    *)
	 let (new_state, new_action, bs) =
	   process_ast_node rewrite_ctxt cfl_return symtab sbdo for_state cur_action (let_bs && for_bs) is_nested_step in
	 
	 (* unregister variables that were registered *)
         let map_cfl cfl =
           match cfl with
            | CEFOR(_, v, _, e1')
            | CELET(_, v, e1') ->
               unregister_var symtab v;
         in
         List.iter map_cfl cfl_list;
	 
	 (* Handle the where clause if present -- only happens in case of predicate steps
	    The let binding containing the axis step for the predicate step is the first
	    expression in the for/let list.
	 *)
	 (match cfl_where with 
	      | Some cfl_where_expr ->
                  start_new_automaton rewrite_ctxt sbdo symtab cfl_where_expr
	      | _ -> ());
		  
	let (s, a, bs) =
	    (* The result state from the let-binding is the state we need,
	       because the predicate in the return clause will make us end up 
	       in the sink state 
	       (let_bs && for_bs && bs) indicate whether axis steps occured in either 
	       the for/let/return clause 
	     *)
	  let state = if pred then let_state else new_state in
	  let action = 
	       if (new_state = (undefined_state SBDO_DupTidy) || new_state = (undefined_state SBDO_DupTidy))
	       then DistinctDocorder 
	       else if pred then let_action
	       else new_action  
	  in
	  let before_slash = (let_bs && for_bs && bs) in
	    (state, action,before_slash)
	in
	(if (!Conf.debugging) then Printf.printf "Leaving SLASHEXPR (pred=%b) with %s %s %b\n" pred (print_state s) (print_action a) bs);
	(s, a, bs)

       end
       else
       (* if let expressions occur after the slash (i.e., transitions have already been made) and it does not 
          concern normalized predicate steps, then this expression is not inside the supported XPath fragment. 
	  So, we return the sink state and start new automaton for the nested expressions. This corresponds
	  with the general handling of FLWORs 
       *)
         process_flwor_general ce_flwor rewrite_ctxt symtab sbdo cur_state cur_action before_slash is_nested_step
	 
     | _ -> raise (Query (Rewriting ("In annot_ddo.ml, this FLWOR is not a normalized path expression")))
	 
(**
  * This function replaces or removes calls to 
  * distinct-docorder-or-atomic-sequence
  * that are marked as such.
  **)
let rec remove_or_replace_ddo ce: acexpr =
  let {
    pcexpr_desc = ce_desc; 
    pcexpr_annot = annot; 
    pcexpr_origin = eh; 
    pcexpr_loc = loc;} = ce 
  in
  let ddo_annot = get_ddo_annot annot
  in
  let new_ce_desc =
  match ce_desc with
  | CECall (fname, args, sign) ->
    let newargs = List.map (remove_or_replace_ddo) args
    in (* can we remove or replace a ddo-call? *)
    if (rqname_equal fname fs_distinct_docorder_or_atomic_sequence)  then
    begin
        match ddo_annot with
        | Some Needs_none              -> CECall (fs_node_sequence_or_atomic_sequence, newargs, sign)
        | Some Needs_docorder          -> CECall (fs_docorder_or_atomic_sequence, newargs, sign)
        | Some Needs_distinct          -> CECall (fs_distinct_or_atomic_sequence, newargs, sign)
        | Some Needs_distinct_docorder 
        | _                            -> CECall (fs_distinct_docorder_or_atomic_sequence, newargs, sign)
    end
    else if (rqname_equal fname fs_distinct_docorder) then
    begin
        match ddo_annot with
        | Some Needs_none              -> CECall (fs_node_sequence, newargs, sign)
        | Some Needs_docorder          -> CECall (fs_docorder, newargs, sign)
        | Some Needs_distinct          -> CECall (fs_distinct, newargs, sign)
        | Some Needs_distinct_docorder 
        | _                            -> CECall (fs_distinct_docorder, newargs, sign)
    end
    else
      CECall (fname, newargs, sign)
    

  | CEForwardAxis (v,axis, cnode_test) ->
      CEForwardAxis (v,axis, cnode_test)
  | CEReverseAxis (v, axis, cnode_test) ->
      CEReverseAxis (v, axis, cnode_test)
    
  | CEVar (prefix, uri, name) ->
      CEVar (prefix, uri, name)

  | CEIf (e1,e2,e3) ->
      CEIf (remove_or_replace_ddo e1,
	    remove_or_replace_ddo e2,
	    remove_or_replace_ddo e3)

  | CEEmpty ->
      CEEmpty

  | CETypeswitch (e1, clauses) ->
    let map_clause clause = (
      let (cpattern, vname, ce) = clause
      in (cpattern, vname, remove_or_replace_ddo ce)
    )
    in
    let newclauses = List.map map_clause clauses
    in CETypeswitch (remove_or_replace_ddo e1, newclauses)

  | CEOverloadedCall (fname, args, sign) -> 
      let newargs = List.map (remove_or_replace_ddo) args
      in CEOverloadedCall (fname, newargs, sign)

  | CEUnordered e' ->
      CEUnordered (remove_or_replace_ddo e')

  | CEOrdered e' ->
      CEOrdered (remove_or_replace_ddo e')

  | CEDocument e ->
      CEDocument (remove_or_replace_ddo e)

  | CEText t ->
      CEText t

  | CEScalar bv ->
      CEScalar bv

  | CEProtoValue bv ->
      CEProtoValue bv

  | CETextComputed e1 ->
      CETextComputed (remove_or_replace_ddo e1)

  | CEPI (ncn,str) ->
      CEPI (ncn, str)

  | CEPIComputed (ce1, ce2) ->
      CEPIComputed (remove_or_replace_ddo ce1, remove_or_replace_ddo ce2)

  | CEComment s ->
      CEComment s

  | CECommentComputed e1 ->
      CECommentComputed (remove_or_replace_ddo e1)

  | CEElem (a,nsenv,elist) ->
      let el = List.map (remove_or_replace_ddo) elist
      in CEElem(a, nsenv, el)

  | CEAttr (a,elist) ->
      let el = List.map (remove_or_replace_ddo) elist
      in CEAttr(a, el)

  | CEAnyElem (e1,nsenv,e2) ->
      CEAnyElem(remove_or_replace_ddo e1, nsenv, remove_or_replace_ddo e2)

  | CEAnyAttr (e1,nsenv,e2) ->
      CEAnyAttr(remove_or_replace_ddo e1, nsenv, remove_or_replace_ddo e2)

  | CESeq (e1,e2) ->
      CESeq(remove_or_replace_ddo e1, remove_or_replace_ddo e2)

  | CEError el ->
    CEError (List.map (remove_or_replace_ddo) el)

  | CETreat (e1,m) ->
      CETreat (remove_or_replace_ddo e1, m)

  | CEValidate (vmode,e1) ->
      CEValidate (vmode, remove_or_replace_ddo e1)

  | CECast (e1,nsenv,m) ->
      CECast (remove_or_replace_ddo e1,nsenv,m)

  | CECastable (e1,nsenv, m) ->
      CECastable (remove_or_replace_ddo e1,nsenv, m)

  | CESome (odt, v, e1, e2) ->
      CESome (odt, v, remove_or_replace_ddo e1, remove_or_replace_ddo e2)

  | CEEvery(odt, v, e1, e2) ->
      CEEvery(odt, v, remove_or_replace_ddo e1, remove_or_replace_ddo e2)

  | CEKeyRef (kn, kv) ->
      CEKeyRef (kn, remove_or_replace_ddo kv)

(* Added some stuff for updates. I don't have a good understanding of
   what the automaton is doing, so PLEASE check that I'm not screwing
   up everything. - Jerome 10/20/2004 *)
  | CECopy ce1 ->
      let newce1 = remove_or_replace_ddo ce1 in
      CECopy newce1
  | CEDelete ce1 ->
      let newce1 = remove_or_replace_ddo ce1 in
      CEDelete newce1
  | CEDetach ce1 ->
      let newce1 = remove_or_replace_ddo ce1 in
      CEDetach newce1
  | CEInsert (ce1, cinsert_location) ->
      let newce1 = remove_or_replace_ddo ce1 in
      let newcinsert_location =
	match cinsert_location with
	| CUAsLastInto ce1 ->
	    (CUAsLastInto (remove_or_replace_ddo ce1))
	| CUAsFirstInto ce1 ->
	    (CUAsFirstInto (remove_or_replace_ddo ce1))
	| CUInto ce1 ->
	    (CUInto (remove_or_replace_ddo ce1))
	| CUAfter ce1 ->
	    (CUAfter (remove_or_replace_ddo ce1))
	| CUBefore ce1 ->
	    (CUBefore (remove_or_replace_ddo ce1))
      in
      CEInsert (newce1, newcinsert_location)
  | CERename (ce1, ce2) ->
      let newce1 = remove_or_replace_ddo ce1 in
      let newce2 = remove_or_replace_ddo ce2 in
      CERename (newce1,newce2)
  | CEReplace (vof, ce1, ce2) ->
      let newce1 = remove_or_replace_ddo ce1 in
      let newce2 = remove_or_replace_ddo ce2 in
      CEReplace (vof,newce1,newce2)
  | CESnap (sm,ce) ->
      let ce' = remove_or_replace_ddo ce in
      CESnap (sm,ce')

  (* Mary: I'm assuming that CEExecute behaves a lot like CESnap *)
  | CEExecute (host, port,ce) ->
      let ce' = remove_or_replace_ddo ce in
      CEExecute(host, port,ce')

  | CEFLWOR (elist, e1, ord_by, e2) ->
      (* fixme also handle order_by spec list? *)
      let new_ord_by =
       (match ord_by with 
         Some (stablekind, order_by_list, osig) ->
           let process_order_by (order_by_ce,sortkind,emptysortkind) =
             (remove_or_replace_ddo order_by_ce, sortkind, emptysortkind)
           in
	     Some (stablekind,List.map process_order_by order_by_list, osig)
	 | None -> ord_by
       ) in
      let map_cfl_expr cflexpr = (
        match cflexpr with
	| CELET (odt, v, e)      -> CELET (odt, v, remove_or_replace_ddo e)
	| CEFOR (odt, v1, v2, e) -> CEFOR (odt, v1, v2, remove_or_replace_ddo e)
      ) 
      in
      let el = List.map map_cfl_expr elist 
      in
      match e1 with
      | None   -> CEFLWOR (el, None, new_ord_by, remove_or_replace_ddo e2)
      | Some e -> CEFLWOR (el, Some (remove_or_replace_ddo e), new_ord_by, remove_or_replace_ddo e2)
  in
    {
      pcexpr_desc = new_ce_desc;
      pcexpr_annot = annot;
      pcexpr_origin = eh;
      pcexpr_loc = loc;
    }

(** Annotate a single expression. 

    The static context of the rewrite context contains the global
    variables that have the maxOne property.  The symbol table of
    variables is initialized with these global variables.

**)
let annot_ast rewrite_ctxt sbdo ce =
  debug  "Starting AST annotation\n";
  let symtab = create_symbol_table () in 
  let maxonetab = Typing_context.var_maxone_properties_from_static_context (Ast_walker_rewrite_context.get_context rewrite_ctxt) in 
  Hashtbl.iter (fun vname maxone -> register_var symtab vname (init_state sbdo)) maxonetab;
  start_new_automaton rewrite_ctxt sbdo symtab ce;
  if (!Conf.debugging) then
  begin
        let _ = Printf.printf "\n\n(** DDO Annotated Expression ***)\n\n" in 
  	let ff =  Format.formatter_of_out_channel stdout in
	print_acexpr ff ce
  end;
  remove_or_replace_ddo ce

(** Annotate/rewrite a global variable declaration. 

    If a global variable has the maxOne property, add it to the static
    context of the current rewrite context.  

    We have to use the static context, because currently, the rewrite
    context is not returned between API calls.
**)

let annot_vardecl rewrite_ctxt sbdo cvar_decl =
  match cvar_decl.pcvar_decl_desc with
  | (vname, model, Some ce) ->
   let ce' = annot_ast rewrite_ctxt sbdo ce in 
   let maxone = hasMaxOne ce in 
(* print_string ("\nGlobal var "^(Namespace_names.prefixed_string_of_rqname vname)^" : "^(string_of_bool maxone)) ;   *)
   if (maxone) then 
     begin
       let symtab = Typing_context.var_maxone_properties_from_static_context (Ast_walker_rewrite_context.get_context rewrite_ctxt) in 
       Hashtbl.add symtab vname maxone
     end;
   let cvar_decl' = fmkcvar_decl (vname, model, Some ce') cvar_decl.pcvar_decl_loc in
   (rewrite_ctxt, cvar_decl')

  | (vname, model, None) -> (rewrite_ctxt, cvar_decl)
  
