(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rewriting_rules_notyping.ml,v 1.35 2006/05/08 12:09:39 simeon Exp $ *)

(* Module: Rewriting_rules_notyping
   Description:
     This module contains the rewriting rules that are independant of
     typing information.
*)

open Error

open Namespace_names
open Namespace_builtin

open Datatypes

open Dm_atomic
open Dm

open Norm_util

open Xquery_common_ast
open Xquery_common_ast_util

open Xquery_core_ast
open Xquery_core_ast_util
open Xquery_core_ast_annotation

open Ast_walker_rewrite_context
open Ast_walker_rewrite

open Processing_context

open Rewriting_ddo_annot
open Rewriting_judgments

open Schema_builtin
open Schema_util
open Print_top

let atomicFalse = BooleanLiteral false
let atomicTrue = BooleanLiteral true

type clause_rewriter =
  Typing_context.static_context Ast_walker_rewrite_context.rewrite_context ->
  Xquery_core_ast.acfl_expr list ->
  Xquery_core_ast.acfl_expr ->
  Xquery_core_ast.acexpr ->
  Xquery_core_ast.acfl_expr list * Xquery_core_ast.acexpr * bool

(* 
  Generic rewrite rule for function calls.
*)
let call_rewrite (fname, arity, rewrite_fn) rewrite_ctxt ce =
  match ce.pcexpr_desc with
  | CECall (fname', arguments, sign) ->
      if (fname' = fname) then
	if (arity = List.length arguments) then rewrite_fn rewrite_ctxt ce arguments
	else raise (Norm_util.incorrect_arg_count fname (List.length arguments) arity)
      else raise Not_applied
  | _ -> raise Not_applied

(*********************)
(* IF SIMPLIFICATION *)
(*********************)

(*
   if cexpr1 then cexpr2 else cexpr3      if cexpr1 then cexpr2 else cexpr3  
   cexpr1 := true		          cexpr1 := false		      
   ---------------------------------      ---------------------------------  
   cexpr2			          cexpr3                             
*)
let if_rewrite (rewrite_ctxt : Typing_context.static_context rewrite_context) (ce : acexpr) =
  match ce.pcexpr_desc with
  | CEIf (cexpr1, cexpr2, cexpr3) ->
      begin
        match cexpr1.pcexpr_desc with
	| CEScalar (BooleanLiteral b)  ->
	    ((if b then cexpr2 else cexpr3), true)
	|  _ ->
	    raise Not_applied
      end
  | _ ->
      raise Not_applied

(*********************)
(* LET REWRITINGS    *)
(*********************)

(*

  Rule1: let $v := Expr1 return Expr2
         used_count $v in Expr2 => 0
         ----------------------------
         Expr2
 
  Rule2: let $v := Expr1 return Expr2
         used_count $v in Expr2 => 1
         Expr2 does not shadow/capture variables free in Expr1
         -----------------------------------------------------
         Expr2 [ Expr1 / $v ]
*)

(* Note:
     The first rule uses the non-deterministic semantics allowed by
     XQuery in the sense that the optimization occurs even though the
     expression might fail.
   - Jerome *)

(***************************************************)
(* Substitute Expr1 for Variable in Expr2          *)
(*     Expr2 [ Expr1 / Variable ]                  *)
(* returns (Expr2', Changed) where Changed is true *)
(* if Expr1 was substituted for Variable in Expr2  *)
(***************************************************)

let contains_path_expr ce =
  match (get_xpath_annot ce.pcexpr_annot) with
  | Some _ -> true
  | None -> false

(* Note:
     Variable substitution is in Ast_walker_rewrite, so it can be used
     during normalization. It is notably used now for normalization of
     the "order by" clause.
 - Jerome
 *)
    (* Aggressive optimization, which ignores whether the
       right-hand-side of the let clause may fail.  This is allowed by
       the XQuery semantics.  If a variable is never used, then the
       type assertion may also be ignored.  If it is used one or more
       times, the assertion cannot be ignored.  *)

(* Function to apply let-clause simplification *)
let let_clause_rewrite_noop rewrite_ctxt cfl_clause_list cfl_clause ce = 
  (cfl_clause_list@[cfl_clause], ce, false)

let let_clause_rewrite rewrite_ctxt (cfl_clause_list) (cfl_clause) (ce) =
  let stat_ctxt = get_context rewrite_ctxt in 
  let proc_ctxt = Typing_context.processing_context_from_stat_context stat_ctxt in
  match (cfl_clause_list, cfl_clause, ce.pcexpr_desc) with
    (* If this is the only let-clause in a FLWOR expression and the
       FLWOR includes a WHERE or ORDER-BY clause, we conservatively
       leave the let-clause in place. *)
  | ([], CELET _, CEFLWOR ([], Some _, _, _)) 
  | ([], CELET _, CEFLWOR ([], _, Some _, _)) ->
      begin 
	([cfl_clause], ce, false)
      end
  | (_, CELET (None, vname, cexpr1), CEFLWOR (other_clauses, opt_where, opt_orderby, ret_expr)) ->
      begin

	match (used_count vname ce) with
      (* Mary -- I put the can-fail test back, otherwise calls to
         fn:trace() were erroneously optimized away. We should probably
         have a side-effect test instead.  This change didn't work --
         led to infinite loop in execution plan!!! Why??? *)
	| 0 -> 
	    begin
(*
 	      if not(can_fail stat_ctxt cexpr1) then 
*)
 		match (cfl_clause_list, other_clauses, opt_where, opt_orderby) with
 		| ([], [],  None, None) -> ([], ret_expr, true)
 		| _ -> (cfl_clause_list, ce, true)
(*	      else (cfl_clause_list@[cfl_clause], ce, false)   *)
	    end
	| 1 ->
	    begin
	      if proc_ctxt.inline_variables 
                 || (contains_path_expr cexpr1) || (contains_path_expr ce)
	      then
		begin
		  if not(fs_dot == vname) then
		    let (ce', changed) = safe_substitute_var cexpr1 vname ce in 
 		    match (cfl_clause_list, ce'.pcexpr_desc) with
 		    | ([], CEFLWOR([],  None, None, ret_expr')) -> ([], ret_expr', true)
 		    | _ -> (cfl_clause_list, ce', true)
		  else (cfl_clause_list@[cfl_clause], ce, false)
		end
	      else
		(cfl_clause_list@[cfl_clause], ce, false)
	    end
	| _ -> 
	    (cfl_clause_list@[cfl_clause], ce, false)
      end
  | _ ->
      (cfl_clause_list@[cfl_clause], ce, false)

(**********************)
(* FOR SIMPLIFICATION *)
(**********************)

(* Function to apply for-clause simplification *)
let for_clause_rewrite (rewrite_ctxt : Typing_context.static_context rewrite_context) 
    (cfl_clause_list : acfl_expr list) (cfl_clause : acfl_expr) (ce : acexpr) =
  match cfl_clause with
  | CEFOR (odt, vname, Some vname', cexpr1) ->
      begin
	match used_count vname' ce with
	| 0 ->
	    (cfl_clause_list@[ CEFOR(odt, vname, None, cexpr1) ], ce, true) (* Rule 1 *)
	| _ ->
	    (cfl_clause_list@[cfl_clause], ce, false)
      end
  | _ ->
      (cfl_clause_list@[cfl_clause], ce, false)

(* Each FLWOR rewrite rules is applied to one clause at a time *)
let rec flclause_rewrite for_clause_rewriter let_clause_rewriter
    (rewrite_ctxt : Typing_context.static_context rewrite_context) 
    (cfl_clause_list : acfl_expr list) (cfl_clause : acfl_expr) rest_flwr_body =
  match (cfl_clause) with 
  | CEFOR(odt, vname, optvname, ce1) -> for_clause_rewriter rewrite_ctxt cfl_clause_list cfl_clause rest_flwr_body
  | CELET(odt, vname, ce1) -> let_clause_rewriter rewrite_ctxt cfl_clause_list cfl_clause rest_flwr_body

(* flwr_rewrite_aux recursively rewrites each LET/FOR clause in the
   FLWOR expression.  The function takes:
   (1) the list of clauses that have already been rewritten
   (2) the current clause to rewrite
   (3) the rest of the flowr expression
*)

and flwr_rewrite_aux for_clause_rewriter let_clause_rewriter
                     rewrite_ctxt (cfl_clause_list : acfl_expr list) ce =
  let ah = ce.pcexpr_annot in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with
  | CEFLWOR([], opt_where_expr, opt_orderby_clause, ret_expr) -> 
      begin
	match cfl_clause_list with
	| [] -> ([], ret_expr, false)
	| _ ->
	    let new_flwr_expr = fmkacexpr (CEFLWOR(cfl_clause_list, opt_where_expr, opt_orderby_clause, ret_expr)) ah eh loc
	    in ([], new_flwr_expr, false)
      end
  | CEFLWOR(cfl_list, opt_where_expr, opt_orderby_clause, ret_expr) ->
      let next_cfl_clause = (List.hd cfl_list) in 
      let next_cfl_clause_list = List.tl cfl_list in 
      let rest_flwr_expr = fmkacexpr (CEFLWOR(next_cfl_clause_list, opt_where_expr, opt_orderby_clause, ret_expr)) ah eh loc in 
      let (cfl_clause_list', rest_flwr_expr', changed') =
	flclause_rewrite for_clause_rewriter let_clause_rewriter rewrite_ctxt cfl_clause_list next_cfl_clause rest_flwr_expr
      in
      let (cfl_clause_list'', rest_flwr_expr'', changed'') =
	flwr_rewrite_aux for_clause_rewriter let_clause_rewriter rewrite_ctxt cfl_clause_list' rest_flwr_expr'
      in (cfl_clause_list'', rest_flwr_expr'', changed' || changed'')
  | _ ->
      (* We end up here if the entire FLWOR expression can be reduced
      to the RETURN clause *)
      (cfl_clause_list, ce, false)

and flwr_rewrite for_clause_rewriter let_clause_rewriter rewrite_ctxt (ce : acexpr) =
  match ce.pcexpr_desc with
  | CEFLWOR _ ->
      let (_, ce', changed) = flwr_rewrite_aux  for_clause_rewriter let_clause_rewriter rewrite_ctxt [] ce in
      if (changed) then (ce', changed) else raise Not_applied  (* Should we do the fix-point on clauses here ? *)
  | _ ->
      raise Not_applied

(*******************************)
(* Not operator simplification *)
(*******************************)

(*
   fn:not(Expr)           fn:not(Expr)	   
   Expr = true	          Expr = false	   
   -------------          -------------   
   false	          true            
*)
   
let not_value_rewrite rewrite_ctxt ce arguments = 
  let ah = ce.pcexpr_annot in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  let cexpr = List.hd arguments in
  match cexpr.pcexpr_desc with
  | CEScalar (BooleanLiteral b) -> 
      (fmkacexpr (CEScalar (BooleanLiteral (not (b)))) ah eh loc, true)
  | _ -> raise Not_applied

(*
   fs:item-sequence-to-node-sequence() is injected in the normalized
   expression just for the purposes of static typing.  It is always
   removed after static typing.  The function's semantics implemented
   in Streaming_constructors.element_constructor.
*)
let remove_function_call rewrite_ctxt ce arguments = 
  (List.hd arguments, true)

(*
 * When expressions of the form path//axis::nodetest occur without a tailing 
 * predicate, we can rewrite those as path/descendant::nodetest (with exception
 * of the attribute axis).
 *
 *)
 
(* helper function -- detects desc-or-self::node() *)
let get_desc_step ds =
  let match_step step =
    match step.pcexpr_desc with
    | CEForwardAxis(v,Descendant_or_self, CPNodeKindTest(CAnyKind, ckind_type)) -> true
    | _ -> false
  in
  match ds with
  | CELET(_, v, e) ->
      begin
        match e.pcexpr_desc with
        | CECall (fname, [argument], sign) ->
	    begin
              if (get_xpath_annot argument.pcexpr_annot = Some Xp_flwor) then
	        match argument.pcexpr_desc with
                | CEFLWOR (cfl_list, _, _, step) -> 
		    if (match_step step) then Some argument
		    else None
                | _ -> None
	      else None
	    end
	| CEFLWOR (_, _, _, step) -> 
	    if (match_step step) then Some e
	    else None
        | _ -> None
      end
  | _ -> None


(* the rewrite rule itself *)
let slashslash_rewrite (rewrite_ctxt : Typing_context.static_context rewrite_context) (ce : acexpr) =
  match get_xpath_annot ce.pcexpr_annot with
  | Some Xp_flwor ->
    begin
    match ce.pcexpr_desc with
    | CEFLWOR (cfl_list, w, ob, step) ->
       begin
         match step.pcexpr_desc with
         | CEForwardAxis(v,a, nt) -> 
	   begin
             let desc_step = List.hd cfl_list in
             match get_desc_step desc_step with
	     | Some expr ->
	       begin
	          match a with
		  | Child
		  | Descendant ->
                    let new_step = 
	            fmkacexpr (CEForwardAxis(v,Descendant, nt)) step.pcexpr_annot step.pcexpr_origin step.pcexpr_loc in
		      begin
		      match expr.pcexpr_desc with
		      | CEFLWOR (cfl_list, w, ob, step) ->
	                (fmkacexpr (CEFLWOR (cfl_list, w, ob, new_step)) ce.pcexpr_annot ce.pcexpr_origin ce.pcexpr_loc, true)
		      | _ -> (ce, true)
		      end
		  | Self
		  | Descendant_or_self ->
                    let new_step = 
	            fmkacexpr (CEForwardAxis(v,Descendant_or_self, nt)) step.pcexpr_annot step.pcexpr_origin step.pcexpr_loc in
		       begin
                       match expr.pcexpr_desc with
		       | CEFLWOR (cfl_list, w, ob, step) ->
	                 (fmkacexpr (CEFLWOR (cfl_list, w, ob, new_step)) ce.pcexpr_annot ce.pcexpr_origin ce.pcexpr_loc, true)
		       | _ -> (ce, true)
		       end
		  | _ ->
	            (ce, false)
	       end
	     | _ -> (ce, false)
	   end
	 | _ -> (ce, false)
       end
    | _ -> (ce, false)
    end
  | _ -> (ce, false)

(* ** LOOP FUSION ** *)
(*
 for $x in expr1
 return
   ddo(
     for $y in ddo(expr2)
     where expr3
     return expr4
   )

 (ddo is optional)
 where expr4 does not depend on $x
 and   expr3 does not depend on $x
 
     ==
 ddo(
   for $y in
     for $x in expr1
     return expr2
   where expr3
   return expr4
 )
*)

let var_name_equals ce name =
  match ce.pcexpr_desc with
  | CEVar name' -> name = name'
  | _ -> false

let loop_fusion rewrite_ctxt flwor1 =
  let fail = flwor1, false in
  match flwor1.pcexpr_desc with
  | CEFLWOR ([CEFOR(t_ce1,x,None,ce1)], w1, None, unk) ->
      begin
	let flwor2 =
	  match unk.pcexpr_desc with
	  | CECall (fname, [arg], sign) when rqname_equal fname fs_distinct_docorder  ->
	      arg
	  | _ -> unk
	in
	match flwor2.pcexpr_desc with
	| CEFLWOR ([CEFOR(t_ce2,y,None,ce2)], Some ce3, None, ce4) ->
	    begin
	      
	      if (used_count x ce3 = 0 && used_count x ce4 = 0)  || (var_name_equals ce4 y && x = y)
	      then
		let flwor1' = 
		  fmkacexpr (CEFLWOR ([CEFOR(t_ce1,x,None,ce1)], w1, None, ce2)) 
		    flwor1.pcexpr_annot flwor1.pcexpr_origin flwor1.pcexpr_loc in
		let flwor2' = 
		  fmkacexpr (CEFLWOR ([CEFOR(t_ce2,y,None,flwor1')], Some ce3, None, ce4)) 
		    flwor2.pcexpr_annot flwor2.pcexpr_origin flwor2.pcexpr_loc in
		flwor2', true
	      else 
		fail
	    end
	| _ -> fail
      end
  | _ -> fail



(* *** LOOP MERGING ***
   for $fs:dot in
     for $fs:dot in e1
     where c1
     return e2
   where c2
   return $fs:dot
   ==
   for $fs:dot in e1
   where c1 and c2
   return e2

  NOTE: c1/c2 may not use positional info and must be independent 
        This is currently checked by, restricting this particular 
        rewrite to XPath normalizations -- Philippe

        !Also, the outer loop variable must be returned!
*)
let is_step op =
  match op.pcexpr_desc with
  | CECall (fname, [arg], sign) 
    when rqname_equal fname fn_boolean  ->
      begin
	match arg.pcexpr_desc with
	| CEForwardAxis _ | CEReverseAxis _ -> true
	| _ -> false
      end
  | _ -> false

let mk_fn_false rewrite_ctxt eh fi =
  let stat_ctxt = get_context rewrite_ctxt in 
  (*let proc_ctxt = Typing_context.processing_context_from_stat_context stat_ctxt in*)
  let norm_ctxt = Typing_context.norm_context_from_stat_context stat_ctxt in
  Norm_util.build_core_false norm_ctxt eh fi

let loop_merging  rewrite_ctxt flwor1 =
  let eh = flwor1.pcexpr_origin in
  let fi = flwor1.pcexpr_loc in
  let fail = flwor1, false in
  match flwor1.pcexpr_desc with
  | CEFLWOR ([CEFOR(t_ce1,n_ce1,None,unk)], Some cond1, None, dot1) ->
      begin
	let flwor2 =
	  match unk.pcexpr_desc with
	  | CECall (fname, [arg], sign) when rqname_equal fname fs_distinct_docorder  ->
	      arg
	  | _ -> unk
	in
	match flwor2.pcexpr_desc with
	| CEFLWOR ([CEFOR(t_ce2,n_ce2,None,ce2)], Some cond2, None, ret) ->
	    begin
	      match dot1.pcexpr_desc with
	      | CEVar v1 ->
		  begin
		    if is_step cond1 && is_step cond2 
			&& n_ce1 = fs_dot && v1 = fs_dot &&  n_ce2 = fs_dot
		    then
(*		      let false_expr = fmkacexpr (CEScalar atomicFalse) ah eh fi in  *)
		      let false_expr = mk_fn_false rewrite_ctxt eh fi in
		      let where_clause =
			fmkcexpr (CEIf (cond1, cond2, false_expr)) eh fi
		      in
		      fmkacexpr (CEFLWOR ([CEFOR(t_ce1,n_ce1,None,ce2)], Some where_clause, None, ret)) 
			flwor1.pcexpr_annot flwor1.pcexpr_origin flwor1.pcexpr_loc , true
		    else
		      fail
		  end
	      | _ -> fail
	    end
	| _ -> fail
      end
  | _ -> fail

(*
   fn:data function simplification 

   fn:data(ScalarExpr)
   ------------------------
   ScalarExpr

*)
let fn_data_rewrite rewrite_ctxt ce arguments = 
  let cexpr = List.hd arguments in 
  match cexpr.pcexpr_desc with 
  | CEScalar _ -> (cexpr, true)
  | _ -> raise Not_applied


(*************************************)
(* Global sets of optimization rules *)
(*************************************)

(* Generic rewriting rules that are independent of other optimization & typing options *)
let generic_rule_set =
  [ if_rewrite;
    flwr_rewrite for_clause_rewrite let_clause_rewrite;
    call_rewrite (fn_not, 1, not_value_rewrite);
    slashslash_rewrite;
    call_rewrite (fn_data, 1, fn_data_rewrite);
    call_rewrite (fs_item_sequence_to_node_sequence, 1, remove_function_call);
(*  call_rewrite (fs_item_sequence_to_untypedAtomic, 1, remove_function_call); *)
    loop_fusion;
    loop_merging;

  ]

(***********************************************************************)

(* More optimization rules: *)

(* Function to simplify and boolean operator *)
(* Ask Mary/Jerome : t & _ | _ & t = _ ??? *)


(***************************************)
(* SQL-X specific simplification rules *)
(***************************************)

(* Note by Jerome / September 2002 *)

(* Descendant or self rewriting *)

(* Example:

   Input schema:
   
      define element users { element user_tuple* }
      define element user_tuple {
	element userid,
	element name,
	element rating?
      }

   Query:

      $users//user_tuple

        where $users is of type element users

   Should be rewritten as:

      $users/user_tuple/user_tuple

   $users/descendant-or-self::*/child::user_tuple

   $users,self::*     	    	     --> element users
   $users,child::* 	    	     --> element user_tuple
   $users,child::*,child::* 	     --> element userid
		       	 		 element name
		       	 		 element rating
   $users,child::*,child::*,child::* --> xsd:string

   $users/
     (child::user_tuple
     | (child::user_tuple/child::userid)
     | (child::user_tuple/child::name)
     | (child::user_tuple/child::rating)
     | (child::user_tuple/child::userid/text())
     | (child::user_tuple/child::name/text())
     | (child::user_tuple/child::rating/text()))/child::user_tuple

   two cases:
     element
     attribute
*)

