(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: fn_doc.ml,v 1.26 2006/02/02 21:03:17 mff Exp $ *)

(* Module: Fn_doc
   Description:
     This module implements the fn:doc() function.
*)

open Error
open Galax_io

open Datatypes

open Dm

open Monitoring_context
open Processing_context

(* Note:
     This module performs two main tasks: (1) it dispatches URL's to
     the appropriate source (HTTP, file system, back-end store, etc.),
     and (2) if necessary it 'stores' the resulting document locally
     in order to preserve the original node identity of the document.
   - Jerome
 *)


(*************************)
(* Back-end registration *)
(*************************)

type back_end_call =
    Processing_context.processing_context -> string * string option * string -> Physical_value.item list

let back_ends =
  ref []

let is_registered_back_end me =
  List.mem_assoc me !back_ends

let lookup_back_end_fun me =
  try
    List.assoc me !back_ends
  with
  | _ ->
      raise (Query (Internal_Error ("Accessing unregistered backend: " ^ me ^ ", how come?")))

let register_back_end me f =
  if is_registered_back_end me
  then
    raise (Query (Internal_Error ("Backend : " ^ me ^ " already registered")))
  else
    begin
      Galax_http.register_method me;       (* Do not forget to registed the backend as a URI method *)
      back_ends := (me,f) :: !back_ends  (* Add the function to the list of registered back-ends *)
    end

(* Alive documents *)

type alive_documents = (Datatypes.xs_string, Physical_value.item list) Hashtbl.t

let build_alive_documents_table () = Hashtbl.create 59

let alive_documents_table_mem alive_documents file =
  Hashtbl.mem alive_documents file

let alive_documents_table_get alive_documents file =
  Hashtbl.find alive_documents file
let alive_documents_table_put alive_documents file doc =
  Hashtbl.add alive_documents file doc

let merge_alive_documents ad1 ad2 =
  Gmisc.merge_hashtable ad1 ad2


(*********************)
(* Http/file sources *)
(*********************)

let lookup_document_from_io_with_index gio name_indices alive_ctxt proc_ctxt =
  let apply_load_document () =
    (* 1. Set the document id and nodeid context *)
    let uri_string =
      match gio with
      | File_Input uri_string
      | Http_Input uri_string -> uri_string
      | _ -> raise (Query (Internal_Error "Cannot call the fn:doc() function on something else than a file or URI"))
    in
    try
      Hashtbl.find alive_ctxt uri_string
    with
    | Not_found ->
	let result_node =
	  (* 1. Open a SAX cursor on the input document *)
	  let (dtd_opt, xml_stream) = Streaming_parse.open_xml_stream_from_io gio in
	  (* 2. Resolve namespaces *)
	  let resolved_xml_stream = Streaming_ops.resolve_xml_stream xml_stream in
	  (* 3. Apply type annotations *)
	  let typed_xml_stream = Streaming_ops.typed_of_resolved_xml_stream resolved_xml_stream in
	  (* If the doc id is not found, then build a new one and registers it *)
	  let docid = Galax_nodeid.new_docid () in
	  let nodeid_context = Nodeid_context.default_nodeid_context () in
	  (* 4. Load in the data model *)
	  Physical_index_load.load_xml_document_from_typed_stream_for_docid nodeid_context docid name_indices typed_xml_stream
	in
	begin
	  Hashtbl.add alive_ctxt uri_string result_node;
	  result_node
	end
  in
  Monitor.wrap_monitor proc_ctxt (Document_ParsingLoading_Phase (Parse_io.name_of_input_spec gio)) apply_load_document ()

let lookup_document_from_io gio =
  let name_indices = Physical_name_index.no_name_indices in
  lookup_document_from_io_with_index gio name_indices 

let lookup_doc_function_with_index uri name_indices =
  let http_method = Galax_http.glx_decode_url uri in
  match http_method with
  | Galax_http.File _
  | Galax_http.Http _ ->
      lookup_document_from_io_with_index (Galax_io.Http_Input uri) name_indices 
  | Galax_http.ExternalSource (me,host,port,local) ->
      let back_end_fun = lookup_back_end_fun me in
      (fun alive_ctxt -> fun proc_ctxt -> back_end_fun proc_ctxt (host,port,local))

let lookup_doc_function uri =
  let name_indices = Physical_name_index.no_name_indices in
  lookup_doc_function_with_index uri name_indices

let lookup_doc_function_no_table uri =
  let ad = build_alive_documents_table () in
  let name_indices = Physical_name_index.no_name_indices in
  lookup_doc_function_with_index uri name_indices ad

