(* Module added by Trevor, for http get *)
(* Modified by Nicola Onose, May 2003 *)
(* Modified by Jerome Simeon, July 2003 *)

(* open Neturl;; *)

let decode_url s =
   let len = String.length s in
   try
     let i = String.index s ':' in (* might raise Not_found *)
     let scheme = String.sub s 0 i in
     if i+3 >= len then raise Not_found;
     if String.get s (i+1) <> '/' then raise Not_found;
     if String.get s (i+2) <> '/' then raise Not_found;
     let colon_position =
       try Some(String.index_from s (i+3) ':')
       with Not_found -> None in
     let slash_position =
       try Some(String.index_from s (i+3) '/')
       with Not_found -> None in
     let host,port,after_port =
       match colon_position,slash_position with
         None,None ->
           String.sub s (i+3) (len-(i+3)),
           None,
           len
       | Some j,None ->
           String.sub s (i+3) (j-(i+3)),
           Some(String.sub s (j+1) (len-(j+1))),
           len
       | None,Some j ->
           String.sub s (i+3) (j-(i+3)),
           None,
           (j+1)
       | Some j,Some k ->
           String.sub s (i+3) (j-(i+3)),
           Some(String.sub s (j+1) (k-(j+1))),
           (k+1) in
     let rest =
       if after_port >= len then None else
       Some(String.sub s after_port (len-after_port)) in
     Some(scheme,host,port,rest)
   with Not_found -> None
          
module HTTP = struct
  exception Break
  let buflen = 8192

  let gen_request h p s meth extra = (* host port resource *)
    let buf = String.create buflen in
    let longbuf = Buffer.create buflen in
    let targetInetAddr =
      try
        let targetHostEntry = Unix.gethostbyname h in
        targetHostEntry.Unix.h_addr_list.(0)
      with Not_found ->
        raise (Failure
                 (Printf.sprintf
                    "Can't find the IP address of the server (%s)" h))
    in

    let sock = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
    Unix.connect sock (Unix.ADDR_INET(targetInetAddr,p));

    let request =
      match extra with 
      | Some (headers, msg) ->
	  Printf.sprintf
	  "%s %s HTTP/1.0\r\nHost: %s:%d\r\nConnection: close\r\nContent-Length: %d\r\n%s\r\n%s\r\n\r\n" 
          meth s h p (String.length msg) headers msg
      | None ->
	  Printf.sprintf
	  "%s %s HTTP/1.0\r\nHost: %s:%d\r\nConnection: close\r\n\r\n" 
            meth s h p
    in
    (*  print_string ("Request:\n"^request); flush(stdout); *)
    let _ = Unix.write sock request 0 (String.length request) in
    (*Printf.printf "Receiving:\n"; 
      let total_len = ref 0 in *)
    (try
      while true do
        (* Printf.printf "Receiving...\n"; flush(stdout); *)
        let len = Unix.read sock buf 0 buflen in
        if len <= 0 then raise Break;
        Buffer.add_substring longbuf buf 0 len;
(*
  Printf.printf "%s" (String.escaped (String.sub buf 0 len));
  flush(stdout);
  total_len := len + !total_len 
*)
      done
    with Break -> ());
    (* Printf.printf "Total bytes received: %d\n" !total_len;
       flush(stdout); *)
    Unix.close sock;
    Buffer.contents longbuf

  let get h p s = gen_request h p s "GET" None

(* We would prefer to use the standard url parsing routines, but
    (1) base/uri.mli makes the type of _uri abstract so we can't
        use the functions in tools/netstring/neturl.ml
    (2) those functions don't seem to decode ldap urls properly,
        e.g., they complain on
     
ldap://ldap.itd.umich.edu:389/ 
o=University%20of%20Michigan,c=US??sub?sn=john
*)

  (* From my generic web server *)
  type http_method =
      Get of string
    | Post of string

  type header =
      ContentLength of int
    | Header of string * string
    | Other of string

  (* For debugging *) 
  module Echo = struct            
    let line_of_method h =
      match h with
        Get s -> Printf.sprintf "GET %s HTTP/1.1\r\n" s
      | Post s ->  Printf.sprintf "POST %s HTTP/1.0\r\n" s

    let line_of_header h =
      match h with
        ContentLength s -> Printf.sprintf "Content-Length: %d\r\n" s
      | Header(x,y) -> Printf.sprintf "%s: %s\r\n" x y
      | Other s -> ">>>> "^s^"\n"

    let echo (meth,headers,payload) =
      Printf.eprintf "You said\n\n";
      output_string stderr (line_of_method meth);
      List.iter (fun x -> output_string stderr (line_of_header x)) headers;
      begin
        match payload with None -> ()
        | Some s ->
            output_string stderr "\r\n";
            output_string stderr s;
            output_string stderr "\n"
      end;
      output_string stderr "\n(that's all)\n";
      flush stderr;
  end (* module Echo *)

  let method_of_line line =
    try
      Scanf.sscanf line "GET %s HTTP/1.1\r" (fun s -> Get s)
    with _ -> try
      Scanf.sscanf line "GET %s HTTP/1.0\r" (fun s -> Get s)
    with _ -> try
      Scanf.sscanf line "POST %s HTTP/1.1\r" (fun s -> Post s)
    with _ ->
      (* galax is sending 1.0 *)
      Scanf.sscanf line "POST %s HTTP/1.0\r" (fun s -> Post s)

  let header_of_line line =
    try
      Scanf.sscanf line "Content-Length: %d\r" (fun s -> ContentLength s)
    with _ -> try 
      Scanf.sscanf line "%[^:]: %[^\r]\r" (fun x y -> Header(x,y))
    with _ ->
      Other line

  let get_payload headers inchan =
    let get_len x y =
      match x,y with
        None,ContentLength s -> Some s
      | Some s1,ContentLength s2 ->
          if s1<>s2 then raise(Failure "Two content lengths") else x
      | _,_ -> x in
    let len = List.fold_left get_len None headers in
    match len with
      None -> None
    | Some l ->
        let buf = String.create l in
        really_input inchan buf 0 l;
        Some(buf)

  let rec get_headers inchan =
    try
      let line = input_line inchan in
      if (line = "" or line = "\r") then [] else
      (header_of_line line)::(get_headers inchan)
    with End_of_file -> []

  let get_method inchan = method_of_line(input_line inchan)

  let get_http_request inchan =
    let meth = get_method inchan in
    let headers = get_headers inchan in
    let payload = get_payload headers inchan in
    (meth,headers,payload)

  let send_http_response outchan payload =
    let len = String.length payload in
    output_string outchan
      ("HTTP/1.1 200 OK\r\n"^
       "Connection: close\r\n"^
       "Content-Type: text/html\r\n"^
       "Content-Length: ");
    output_string outchan (string_of_int len);
    output_string outchan
      ("\r\n"^
       "\r\n");
    output_string outchan payload

  let ok_response line =
    try
      Scanf.sscanf line "HTTP/1.%d 200 OK\r" (fun d -> true)
    with _ -> false

  (* RPC for xquery *)
  let http_query_response my_host_port host port resource meth query =
    let inetaddr =
      try
        let hostentry = Unix.gethostbyname host in
        hostentry.Unix.h_addr_list.(0)
      with Not_found ->
        raise (Failure
                 (Printf.sprintf
                    "Can't find the IP address of the server (%s)" host))
    in
    let (inchan,outchan) = Unix.open_connection(Unix.ADDR_INET(inetaddr,port)) in
    let len = String.length query in
    (* TODO: choose a Content-Type.  text/xquery? *)
    Printf.fprintf outchan
      ("%s %s HTTP/1.1\r\nHost: %s:%d\r\nConnection: close\r\nContent-Length: %d\r\nConnected-Host: %s\r\n\r\n")
      meth resource host port len my_host_port;
    output_string outchan query;
    flush outchan; (* Necessary *)
    if (ok_response(input_line inchan)) then
      let headers = get_headers inchan in
      let payload = get_payload headers inchan in
      Unix.shutdown_connection inchan;
      close_in inchan;
(* (* This results in Sys_error("Bad file descriptor") *)
      Printexc.print close_out outchan;
 *)
      match payload with
        None -> raise(Failure("http_query_response: no response"))
      | Some res -> res
    else raise(Failure("http_query_response: error"));

end (* module HTTP *)
