(***********************************************************************)
(*                                                                     *)
(*                               Galax                                 *)
(*                            XQuery Engine                            *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: galax-run.ml,v 1.74 2006/05/08 12:09:39 simeon Exp $ *)

(* Module: Galax-run
   Description:
     This module contains the main function for the galax-run command.
 *)

open Format
open Print_top
open Error
open Dm

open Processing_context
open Monitoring_context

open Top_util
open Top_config
open Top_options


(************************)
(* Command-line options *)
(************************)

let process_args proc_ctxt =
(*
  (* functions to build options *)
  let bo f = f proc_ctxt in
  let bo_unit f = Arg.Unit (bo f) in
  let bo_string f = Arg.String (bo f) in
  let bo_set f = Arg.Set (bo f) in
  let bo_clear f = Arg.Clear (bo f) in
  (* initializing the remaining arguments *)
  let args = ref [] in
  (* parsing the command-line arguments *)
  Arg.parse [

  (* Misc options *)
  opt_version, (bo_unit   arg_version), msg_version;
  opt_verbose, (bo_string arg_verbose), msg_verbose;
  opt_debug,   (bo_set    arg_debug),   msg_debug ^ title_monitor_options;

  (* Monitoring options *)
  opt_monitor, 	      (bo_string arg_monitor), 	      msg_monitor;
  opt_monitor_mem,    (bo_string arg_monitor_mem),    msg_monitor_mem;
  opt_monitor_time,   (bo_string arg_monitor_time),   msg_monitor_time;
  opt_output_monitor, (bo_string arg_output_monitor), msg_output_monitor ^ title_character_encoding_options;

  (* Character encoding options *)
  opt_internal_encoding, (bo_string arg_internal_encoding), msg_internal_encoding;
  opt_output_encoding,   (bo_string arg_output_encoding),   msg_output_encoding ^ title_input_context_options;

  (* Input context options *)
  opt_base_uri,    (bo_string arg_base_uri),     msg_base_uri;
  opt_var,         (bo_string arg_var),          msg_var;
  opt_doc,         (bo_string arg_doc),          msg_doc;
  opt_context_item,(bo_string arg_context_item), msg_context_item;
  opt_context,     (bo_string arg_context),      msg_context ^ title_data_model_options;

  (* Data model options *)
  opt_xml_whitespace, (bo_unit arg_xml_whitespace), msg_xml_whitespace;
  opt_xml_pic,        (bo_unit arg_xml_pic),        msg_xml_pic ^ title_serialization_options;

  (* Serialization options *)
  opt_serialize, (bo_string arg_serialize), msg_serialize ^ title_behavior_options;

  (* Behavior options *)
  opt_typing,        (bo_string arg_typing),  	     msg_typing;
  opt_boundary_space,(bo_string arg_boundary_space), msg_boundary_space;
  opt_construction,  (bo_string arg_construction),   msg_construction;
  opt_ordering,      (bo_string arg_ordering),       msg_ordering;
  opt_default_order, (bo_string arg_default_order),  msg_default_order;
  opt_ns_preserve,   (bo_string arg_ns_preserve),    msg_ns_preserve;
  opt_ns_inherit,    (bo_string arg_ns_inherit),     msg_ns_inherit ^ title_processing_phases_options;

  (* Processing phases options *)
  opt_execute,          (bo_string arg_execute),          msg_execute;
  opt_normalize,     	(bo_string arg_normalize),        msg_normalize;
  opt_static_typing, 	(bo_string arg_static_typing),    msg_static_typing;
  opt_rewriting,     	(bo_string arg_rewriting),        msg_rewriting;
  opt_factorization, 	(bo_string arg_factorization),    msg_factorization;
  opt_optimization,  	(bo_string arg_optimization),     msg_optimization;
  opt_dynamic,       	(bo_string arg_dynamic),          msg_dynamic ^ title_printing_options;

  (* Printing options *)
  opt_print_global,      (bo_string arg_print_global),     msg_print_global;

  opt_print_expr,        (bo_string arg_print_expr),       msg_print_expr;
  opt_output_expr,       (bo_string arg_output_expr),      msg_output_expr;

  opt_print_norm_expr,   (bo_string arg_print_norm_expr),  msg_print_norm_expr;
  opt_output_norm_expr,  (bo_string arg_output_norm_expr), msg_output_norm_expr;
  
  opt_print_type,        (bo_string arg_print_type),       msg_print_type;
  opt_output_type,       (bo_string arg_output_type),      msg_output_type;

  opt_print_rewr_expr,   (bo_string arg_print_rewr_expr),  msg_print_rewr_expr;
  opt_output_rewr_expr,  (bo_string arg_output_rewr_expr), msg_output_rewr_expr;
  
  opt_print_fact_expr,   (bo_string arg_print_fact_expr),  msg_print_fact_expr;
  opt_output_fact_expr,  (bo_string arg_output_fact_expr), msg_output_fact_expr;

  opt_print_log_plan,    (bo_string arg_print_log_plan),   msg_print_log_plan;
  opt_output_log_plan,   (bo_string arg_output_log_plan),  msg_output_log_plan;

  opt_print_opt_plan,    (bo_string arg_print_opt_plan),   msg_print_opt_plan;
  opt_output_opt_plan,   (bo_string arg_output_opt_plan),  msg_output_opt_plan;

  opt_print_phys_plan,   (bo_string arg_print_phys_plan),  msg_print_phys_plan;
  opt_output_phys_plan,  (bo_string arg_output_phys_plan), msg_output_phys_plan;

  opt_print_xml,         (bo_string arg_print_xml),        msg_print_xml;
  opt_output_xml,        (bo_string arg_output_xml),       msg_output_xml;

  opt_print_error,        (bo_string arg_print_error),        msg_print_error;
  opt_print_plan,         (bo_string arg_print_plan),         msg_print_plan;
  opt_print_annotations,  (bo_string arg_print_annotations),  msg_print_annotations;
  opt_print_prolog,       (bo_string arg_print_prolog),       msg_print_prolog;
  opt_output_all,         (bo_string arg_output_all),         msg_output_all;
  opt_output_err,         (bo_string arg_output_err),         msg_output_err;
  opt_print_log_rewrites, (bo_string arg_print_log_rewrites), msg_print_log_rewrites;
  opt_output_log_rewrites,(bo_string arg_output_log_rewrites),msg_output_log_rewrites;
  opt_print_comp_annot,   (bo_string arg_print_comp_annot),   msg_print_comp_annot;
  opt_print_materialize,  (bo_string arg_print_materialize),  msg_print_materialize;
  opt_output_dfgraph,     (bo_string arg_output_dfgraph),     msg_output_dfgraph ^ title_optimization_options;

  (* Logical optimization options *)
  opt_aggressive,        (bo_string arg_aggressive),        msg_aggressive;
  opt_projection,        (bo_string arg_projection),        msg_projection;
  opt_treejoin_log,      (bo_string arg_treejoin_log),      msg_treejoin_log;
  opt_pruning,           (bo_string arg_pruning),           msg_pruning;
  opt_sbdo,              (bo_string arg_sbdo),              msg_sbdo ^ title_code_selection_options;

  (* Code selection options *)
  opt_treejoin_phys,     (bo_string arg_treejoin_phys),     msg_treejoin_phys;
  opt_nested_loop_join,  (bo_string arg_nested_loop_join),  msg_nested_loop_join;
  opt_force_materialized,(bo_string arg_force_materialized),msg_force_materialized;
  opt_descendant_hack,   (bo_string arg_descendant_hack),   msg_descendant_hack ^ title_runtime_options;

  (* Runtime options *)
  opt_buffer_chunks,     (bo_string arg_buffer_chunks),     msg_buffer_chunks;
  opt_buffer_csize,      (bo_string arg_buffer_csize),      msg_buffer_csize;
  opt_buffer_inc,        (bo_string arg_buffer_inc),        msg_buffer_inc;
]
    (fun arg -> args := arg :: !args) usage_galax_run;
*)
  let args =
    make_options
      proc_ctxt
      usage_galax_run
      [ Misc_Options;Monitoring_Options;Encoding_Options;Context_Options;DataModel_Options;Serialization_Options;Behavior_Options;ProcessingPhases_Options;Printing_Options;Optimization_Options;CodeSelection_Options;Runtime_Options]
  in
  match args with
  | [] -> failwith "Input file(s) not specified"
  | fnames ->
      List.rev fnames


(*********************************************************************************)
(* NOTO BENE!                                                                    *)
(*                                                                               *)
(* For accurate monitoring of execution time, Galax-run MUST delegate            *)
(* to the Galax O'Caml API, which ensures that every API function is monitored.  *)
(*                                                                               *)
(* Do NOT call the compiler directly, as this circumvents the monitor            *)
(* and will yield imprecise results.  If some functionality is missing           *)
(* from the API, then let Mary know.                                             *)
(*********************************************************************************)

let process_main_module proc_ctxt mod_ctxt mod_file =
  begin
    print_processing_file mod_file;
    let (ext_ctxt_item,ext_ctxt) = set_up_external_context proc_ctxt in
    (* Load main module *)
    let (mod_ctxt'', stmts) = 
      match !context_file with
      | None ->
	  compile_main_module_helper ext_ctxt_item mod_ctxt (Galax_io.File_Input mod_file) 
      | Some f ->
	  let compiled_prolog = Galax.import_prolog_only mod_ctxt ext_ctxt_item (Galax_io.File_Input f) in
	  compile_main_module_helper false compiled_prolog (Galax_io.File_Input mod_file) 
    in

    (* Evaluate all global variables *)
    let mod_ctxt''' = Galax.eval_prolog mod_ctxt'' (Some ext_ctxt) in

    (* Evaluate all statements in module and serialize, if necessary *)
      
      
    (********************************************************************************)
    (* Hack! Serializes the query result from a stream iff treejoins are turned on. *)
    (* - Michael                                                                    *)
    (********************************************************************************)
      
    let eval_serialize_stream stmt =
      let stream = Galax.eval_compiled_statement_as_sax mod_ctxt''' stmt in
	if !Conf.print_xml then
	  Galax.serialize_as_sax proc_ctxt (Galax_io.Channel_Output !Conf.xml_output) stream
	else
	  Streaming_ops.discard_typed_xml_stream stream
    in

      if (proc_ctxt.treejoin_log = TreeJoin && proc_ctxt.treejoin_phys = Streaming)
      then
	List.iter eval_serialize_stream stmts
      else
	(* proceed as usual *)
	List.iter (fun stmt -> 
		     let v = Galax.eval_compiled_statement mod_ctxt''' stmt in
		       if !Conf.print_xml then
			 Galax.serialize proc_ctxt (Galax_io.Channel_Output !Conf.xml_output) v
		       else ()) stmts;
  end


(*********************************************************************************)
(* NOTO BENE!                                                                    *)
(*                                                                               *)
(* For accurate monitoring of execution time, Galax-run MUST delegate            *)
(* to the Galax O'Caml API, which ensures that every API function is monitored.  *)
(*                                                                               *)
(* Do NOT call the compiler directly, as this circumvents the monitor            *)
(* and will yield imprecise results.  If some functionality is missing           *)
(* from the API, then let Mary know.                                             *)
(*********************************************************************************)

let main proc_ctxt module_files =
  let mod_ctxt = init_all proc_ctxt in

  Monitor.start_monitor_external_call proc_ctxt "Galax-run.main";
  List.iter (process_main_module proc_ctxt mod_ctxt) module_files;
  close_channel_ref(Conf.glx_stderr);
  close_channel_ref(Conf.xml_output);
  close_channel_ref(Conf.projection_output);
  close_channel_ref(Conf.projected_file_output);
  if not(!output_all) then
    begin
      close_channel_ref(Conf.expr_output);
      close_channel_ref(Conf.type_output);
      close_channel_ref(Conf.core_expr_output);
      close_channel_ref(Conf.optimized_expr_output); 
      close_channel_ref(Conf.factorized_expr_output); 
    end;
  Monitor.end_monitor_external_call proc_ctxt    


(*************)
(* Let's go! *)
(*************)

let go () =
  (* 1. First get the default processing context for galax-run *)
  let proc_ctxt = galax_run_proc_ctxt () in
  
  (* 2. Parses the command-line arguments *)
  let module_files = process_args proc_ctxt in

  (* 3. Execute the input queries *)
  let ret = exec main proc_ctxt module_files in

  (* 4. Call the close handlers *)
  let ()  = Register_handlers.call_close_handlers () in 
  begin
    if (!Conf.debugging)
    then
      begin
	Printf.printf "LOAD CALLED %i TIMES!\n" !Conf.countload;
	Printf.printf "EXPORT CALLED %i TIMES!\n" !Conf.countexpo;
	Printf.printf "NEXT CALLED %i TIMES!\n" !Conf.countnext;
      end;
    ret
  end

let _ =
  try
    low_exec go ()
  with
  | e ->
      begin
	eprintf_error "  " e;
	fprintf (!Conf.glx_err_formatter) "@."
      end

