open Error

type xquery_kind = 
    XQueryString
  | XQueryPlan

(**********************************************************************)
(* TODO                                                               *)
(* Needs to detect and handle recursion.                              *)
(* The xquery evaluation is inefficient, it creates an input file for *)
(*   galax, runs galax to produce an output file, then slurps in the  *)
(*   output file.                                                     *)
(* There are many other inefficiencies.                               *)
(* See other TODOs commented below.                                   *)
(**********************************************************************)

(* This is a hack... We should construct a real element here. *)
let server_error_msg e = 
  let errmsg = bprintf_error "" e in
  "<glx:error xmlns:glx=\"http://www.galaxquery.org\">"^errmsg^"</glx:error>"

(* modules Sim and Gui are copied from the galax gui and need to
   stay in sync *)
module Sim = struct
  (* For simulations.  Look in a directory with .xq files and
     deterministically assign each one a port.

     For example, if the starting port is 3000 and there are files

       a.xq b.xq bar.xq foo.xq

     in the directory, then we would have the port assignments

      a->3000
      b->3001
      bar->3002
      foo->3003

     and a, b, bar, and foo would be considered virtual hosts operating
     at those ports.
  *)
  let get_files_in_directory dir = 
    let dirh = Unix.opendir dir in
    let rec get_them l =
      match
        try Some(Unix.readdir dirh) with _ -> None
      with
      | None ->
          Unix.closedir dirh; l
      | Some x ->
          get_them (x::l)
    in
    Sort.list (<=) (get_them [])
      
  (* Convert a shell-style regular expression, using the special characters,
     ?*[], to a caml-style regular expression. *)
  let convert_regexp s =
    let s = Str.global_replace (Str.regexp "\\+") "\\+" s in
    let s = Str.global_replace (Str.regexp "\\^") "\\^" s in
    let s = Str.global_replace (Str.regexp "\\$") "\\$" s in
    let s = Str.global_replace (Str.regexp "\\.") "\\." s in
    let s = Str.global_replace (Str.regexp "\\?") "." s in
    let s = Str.global_replace (Str.regexp "\\*") ".*" s in
    s ^ "$"

  let ls dir pattern =
    let files = get_files_in_directory dir in
    let re = Str.regexp (convert_regexp pattern) in
    let rec filter l =
      match l with
        [] -> []
      | hd::tl ->
          if Str.string_match re hd 0 then hd::(filter tl)
          else filter tl
    in filter files

  (* Grab a file and stuff it into a string *)
  let string_of_file file =
    try
      let inchan = open_in file in
      let len = in_channel_length inchan in
      let buf = String.create len in
      let howmuch = input inchan buf 0 len in
      close_in inchan;
      (* For some reason, under Windows 95, the value reported
         by in_channel_length (above) is too high, so we get garbage
         characters in buf.  This gets rid of them. *)
      let buf =
        if howmuch <> len
        then String.sub buf 0 howmuch
        else buf in
      (* Windows 95 terminates lines with carraige return as well as
         newline; this screws up the Unix version when it reads in
         files created under Windows.  So, we get rid of the carraige
         returns, not the ideal solution. *)
      if Sys.os_type = "Win32" then buf
      else Str.global_replace (Str.regexp "\r\n") "\n" buf
    with
      Sys_error err ->
        Printf.printf
          "galaxd: could not read the file %s, got error Sys_error %s\n"
          file
          err;
        raise(Sys_error err)

  let portmap dir starting_port =
    let xqfiles = ls dir "*.xq" in
    let curr_port = ref starting_port in
    List.map
      (fun xqf ->
        let port = !curr_port in
        curr_port := !curr_port + 1;
        let host = Filename.chop_extension xqf in
        let xq_program = string_of_file (dir ^ "/" ^ xqf) in
        (host,port,xq_program))
      xqfiles

end (* module Sim *)

(* Executed once, in the main process *) 
let state_file = Printf.sprintf "/tmp/%d" (Unix.getpid())

module Gui = struct
  type nodeName = string

  type link =
      int * int * int * string (* line, char, length, link content *)

  type guiEvent =
      GuiMessage of nodeName * nodeName * string * string * link list
    | GuiNode of nodeName * string * string * link list
    | GuiNewNode of nodeName * string * link list
    | GuiNone
  let guievent_of_string s =
    let rec parts s i =
      try 
        let j = String.index_from s i '\000' in
        let part = String.sub s i (j-i) in
        part::(parts s (j+1))
      with Not_found -> [] in
    let p = parts s 0 in
    match p with
    | ["NewNode";node;s] -> GuiNewNode(node,s,[])
    | ["Node";node;s1;s2] -> GuiNode(node,s1,s2,[])
    | ["Message";src;dest;s1;s2] -> GuiMessage(src,dest,s1,s2,[])
    | _ -> GuiNone
  let string_of_guievent g =
    match g with
    | GuiNewNode(node,s,l) ->
        Printf.sprintf "NewNode\000%s\000%s\000" node s
    | GuiNode(node,s1,s2,l) ->
        Printf.sprintf "Node\000%s\000%s\000%s\000" node s1 s2
    | GuiMessage(src,dest,s1,s2,l) ->
        Printf.sprintf "Message\000%s\000%s\000%s\000%s\000" src dest s1 s2
    | GuiNone -> Printf.sprintf "None\000"

  (**********************************************************************)
  (* Sending reports to the gui                                         *)
  (**********************************************************************)
  let gui_report g =
    if Sys.file_exists state_file then begin
      try
        let inchan = open_in state_file in
        let host = input_line inchan in
        let port = int_of_string (input_line inchan) in
        close_in inchan;
        ignore (Http.HTTP.gen_request host port "/report" "POST"
                  (Some("",string_of_guievent g)))
      with _ -> ()
    end

end (* module Gui *)

module Server = struct
  (* Much of this is taken from QCM *)
  open Gui

  let debug = ref false

  let default_port = 3000

  let pm = ref [] (* portmap for simulation case *)

  exception MyError of string

  (**********************************************************************)
  (* Figure out local host name                                         *)
  (**********************************************************************)
  let full_hostname = Unix.gethostname() (* e.g., saul.cis.upenn.edu *)
  let short_hostname =                   (* e.g., saul *)
    try
      let i = String.rindex full_hostname '.' in
      String.sub full_hostname 0 i
    with Not_found -> full_hostname
  let my_host_port = ref ""

  open Http.HTTP

  (**********************************************************************)
  (* Evaluate a query at a remote principal                             *)
  (**********************************************************************)
  let do_query(host,port_opt,xquery_kind,query) =
    let report host_port =
      gui_report (GuiMessage(!my_host_port,
			     host_port,
			     "?",query,[])) in
    let (hostname,port_num) =
      try 
	(* See if we are simulating this host *)
        let (_,p,_) = List.find (fun (h,_,_) -> h=host) !pm in
	report host;
        ("localhost",p)
      with Not_found ->
	let port_num =
	  match port_opt with
	    None -> (report host; default_port)
	  | Some s ->
	      (report (host^":"^s);
	       try int_of_string s with _ ->
		 raise(MyError(Printf.sprintf "invalid port '%s'\n" s))) in
	(host,port_num) in
    (* TODO: tell the server our host_port, so it can report its reply to the gui
       Poll.write_packet socket !my_host_port;
     *)
    try
      let command = 
	match xquery_kind with
	| XQueryString -> "/query"
	| XQueryPlan -> "/plan"
      in
      http_query_response !my_host_port hostname port_num command "POST" query
    with e -> 
      begin
	let errmsg = server_error_msg e in 
	gui_report(GuiNode(!my_host_port,"x","Got EOF!",[]));
	errmsg
      end

  (**********************************************************************)
  (* Diagnostic log messages                                            *)
  (**********************************************************************)
  let message s =
    (Printf.eprintf "[%s] %s\n" !my_host_port s; flush stderr)

  let http_protect s =
    let s =
      Pcre.replace ~pat:"&" ~templ:"&amp;" s in
    let s =
      Pcre.replace ~pat:"<" ~templ:"&lt;" s in
    let s =
      Pcre.replace ~pat:">" ~templ:"&gt;" s in
    s

  (***********************************************************)
  (* We use the Connected-Host header to tell the server our *)
  (* name so that is can report it to the GUI on the reply   *)
  (***********************************************************)
  let get_connected_host headers =
    let rec loop x =
      match x with
        Header(x,y)::tl -> if x = "Connected-Host" then y else loop tl
      | _::tl -> loop tl 
      | [] -> ""
    in 
    loop headers

  (**********************************************************************)
  (* Request handling                                                   *)
  (**********************************************************************)
  let server eval xq_program inchan outchan =
    let eval = eval xq_program in
    let hosts =
      if !pm <> [] then
        "<strong>Hosts</strong>: "^
        String.concat " "
          (List.map (fun (host,port,_) ->
            Printf.sprintf "<a href='http://localhost:%d/'>%s</a>" port host)
             !pm)
      else "" in
    let header =
      "<strong>Galaxd "^
      !my_host_port^
      "</strong>:\n"^
      "<a href='/'>Home</a> <a href='/program'>Program</a>"^
      "<br />"^
      hosts^
      "<hr />\n" in
    let (meth,headers,payload) = get_http_request inchan in
    if (!debug) then 
      Echo.echo (meth,headers,payload); 
    match meth with
      Post("/query") -> (* A query from another server *)
        begin match payload with
          None -> 
            message "Error: received an empty query"
        | Some qs ->
            let connected_host = get_connected_host headers in
            let result =
              try eval XQueryString qs
              with e -> 
		begin
		  let errmsg = server_error_msg e in
                  message ("Error in evaluation: "^errmsg);
                  message (Printf.sprintf "Query was '%s'" qs);
		  errmsg
		end 
	    in
	    if (connected_host <> "") then
              gui_report (GuiMessage(!my_host_port,connected_host,"!",result,[]));
            send_http_response outchan result;
        end
    | Post("/plan") -> (* A query plan from a client or server *)
        begin match payload with
          None -> 
            message "Error: received an empty query"
        | Some qs ->
	    if (!debug) then message ("Query plan is:\n"^qs);
            let connected_host = get_connected_host headers in
            let result =
              try eval XQueryPlan qs
              with e -> 
		begin
		  let errmsg = server_error_msg  e in
                  message ("Error in evaluation: "^errmsg);
                  message (Printf.sprintf "Query was '%s'" qs);
		  errmsg
		end 
	    in
	    if (connected_host <> "") then
              gui_report (GuiMessage(!my_host_port,connected_host,"!",result,[]));
            send_http_response outchan result;
        end
    | Post("/job") -> (* A job submitted by gui *)
        begin match payload with
          None -> 
            message "Error: gui submitted an empty job request"
        | Some qs -> 
            gui_report (GuiNode(!my_host_port,"?",qs,[]));
            let result = eval XQueryString qs in
            gui_report (GuiNode(!my_host_port,"!",result,[]));
            send_http_response outchan ""
        end
    | Get("/") -> (* Produce a form for browser submit *) 
        let response =
          "<html>\n<head><title>Submit Xquery</title></head>\n<body>\n" ^
          header ^
          "<form method='POST' action='/submit'>\n" ^
          "<p>Query:</p>\n" ^
          "<p>\n"^
          "<textarea name='qs' rows='5' cols='60'></textarea>\n"^
          "</p>\n" ^
          "<p><input type='submit' value='Evaluate'></p>\n" ^
          "</form>\n" ^
          "</body>\n</html>\n" in
        send_http_response outchan response
    | Post("/submit") -> (* A job submitted by browser *)
        begin match payload with
          None -> 
            let response =
              "<html>\n<head>\n"^
              "<title>Error: empty query</title></head>\n<body>\n" ^
              "<h1>Error</h1>\n" ^
              "<p>You submitted an empty query</p>\n" ^
              "</body>\n</html>\n" in
            send_http_response outchan response
        | Some formsubmit -> 
            (* expect Content-Type: application/x-www-form-urlencoded *)
            let params = 
              Netencoding.Url.dest_url_encoded_parameters formsubmit in
            try
              let qs = List.assoc "qs" params in
              gui_report (GuiNode(!my_host_port,"?",qs,[]));
              let result = eval XQueryString qs in 
              gui_report (GuiNode(!my_host_port,"!",result,[]));
              let result = http_protect result in
              let response =
                "<html>\n<head><title>"^
                !my_host_port^
                ": evaluation result</title></head>\n<body>\n" ^
                header ^
                "<strong>Result:</strong><br />\n"^
                "<pre><code>\n"^
                result^
                "</code></pre>\n"^
                "</body>\n</html>\n" in
              send_http_response outchan response
            with _ ->
              message "Error: received a garbled form submission"
        end
    | Get("/program") -> (* Show the context *) 
        let xq_program = http_protect xq_program in
        let response =
          "<html>\n<head><title>"^
          !my_host_port^
          ": program</title></head>\n<body>\n" ^
          header ^
          "<pre><code>\n"^
          xq_program^
          "</code></pre>\n"^
          "</body>\n</html>\n" in
        send_http_response outchan response
    | Post("/guistart") -> (* Start reporting to the gui *)
        begin match payload with
          None -> 
            message "Error: empty gui start request"
        | Some formsubmit ->
            begin
              try
                let params = 
                  Netencoding.Url.dest_url_encoded_parameters formsubmit in
                let host = List.assoc "host" params in
                let port = List.assoc "port" params in
                let outchan2 = open_out state_file in
                Printf.fprintf outchan2 "%s\n" host;
                Printf.fprintf outchan2 "%s\n" port;
                close_out outchan2;
                let payload =
                  string_of_guievent(GuiNewNode(!my_host_port,xq_program,[])) in
                send_http_response outchan payload;
                message (Printf.sprintf "Starting gui reporting to %s:%s\n" host port)
              with _ ->
                message "Error: starting gui reporting"
            end
        end
    | Get("/favicon.ico") -> () (* common, ignore it *)
    | Post(s) ->
	(* A garbled request *)
        message
	  (Printf.sprintf "Error: received a garbled request (POST %s)" s)
    | Get(s) ->
	(* A garbled request *)
        message
	  (Printf.sprintf "Error: received a garbled request (GET %s)" s)

  let go portnum eval xq_program = (* never returns *)
    (try 
      let (h,_,_) = List.find (fun (h,p,_) -> p=portnum) !pm in
      my_host_port := h;
    with Not_found -> 
      my_host_port := Printf.sprintf "%s:%d" short_hostname portnum);
    
    (* A quick check to see if the network is up, useful in Windows 95 *)
    (try ignore (Unix.gethostbyname(Unix.gethostname()))
    with Not_found ->
      message "Error: the network is not up!\n";
      flush stderr;
      Pervasives.exit(33));

    if Sys.file_exists state_file then Sys.remove state_file;

    Unix.establish_server (server eval xq_program)
      (Unix.ADDR_INET(Unix.inet_addr_any,portnum))

end (* module Server *)

(* for remote queries by dxq urls in fn:doc *)
let dxq_back_end ctxt (host,port,local) =
  (* Do the query; ignore local until we figure out how to push selects *)
  let res1 = Server.do_query(host,port, XQueryString, ("local:main()")) in
  let (result_f,result_ch) = Filename.open_temp_file "res" ".xml" in
  output_string result_ch res1;
  close_out result_ch;
  let result = Fn_doc.lookup_doc_function_no_table result_f ctxt in
  if Sys.file_exists result_f then Sys.remove result_f;
  result

let glx_remote_query = (Namespace_builtin.glx_prefix, Namespace_builtin.glx_uri, "remote-query") 
let _glx_remote_query comp_ctxt alg_ctxt n =
  let proc_ctxt = Compile_context.processing_context_from_compile_context comp_ctxt in
  let (p1,p2,p3,p4) = Args.get_array_param4 n in
  let host = Physical_util.get_string p1 in
  let port = Physical_util.get_string p2 in 
  let fnname = Physical_util.get_string p3 in 
  let args = 
    match Cursor.cursor_peek p4 with
    | None -> raise (Query(Internal_Error("In glx:remote-query: argument count missing.")))
    | Some argct_item -> 
      begin
        ignore(Cursor.cursor_next p4);
	let argct = Decimal._int_of_integer((Physical_util.get_atomic argct_item)#getAtomicInteger()) in 
	if (argct > 0) then 
	  begin
	    try
             (* Need to unbox the arguments here, each of which is wrapped in an element <a></a> 
                This is ugly!
             *)
              let node_args = (Physical_util.get_node_list_of_item_list(Cursor.cursor_npeek argct p4)) in
	      List.map (fun n -> Physical_util._node_cursor(n#children(None))) node_args 
	    with
	    | Query(Cursor_Error msg) -> 
   raise (Query(Internal_Error("In glx:remote-query: argument count does not match number of actual arguments.")))
	  end
	else []
      end 
  in
  let serialize_arg arg = Serialization.bserialize_datamodel proc_ctxt arg in
  let argstring = String.concat "," (List.map (fun a -> "("^(serialize_arg a)^")") args) in 
  let querystring = fnname^"("^argstring^")" in
  let res = Server.do_query(host,Some port,XQueryString,querystring) in 
  (*
     We could return a token stream directly, but first we need a way
     to register the physical-type signature of the function:
     C.f. Code_builtin_fn.select_physical_op and
     Code_builtin_fn.build_builtin_fn_code.
  *)
  let (result_f,result_ch) = Filename.open_temp_file "res" ".xml" in
  output_string result_ch res;
  close_out result_ch;
  (* Write result to a file, register the file, and re-parse. *)
  let result = Fn_doc.lookup_doc_function_no_table result_f proc_ctxt in
  if Sys.file_exists result_f then   Sys.remove result_f;
  Cursor.cursor_of_list(result)
  
let _glx_remote_plan comp_ctxt alg_ctxt n =
  let proc_ctxt = Compile_context.processing_context_from_compile_context comp_ctxt in
  let (p1,p2,p3) = Args.get_array_param3 n in
  let host = Physical_util.get_string p1 in
  let port = Physical_util.get_string p2 in 
  let plan = Serialization.bserialize_datamodel proc_ctxt p3 in
  let res = Server.do_query(host,Some port, XQueryPlan, plan) in 
  (*
     We could return a token stream directly, but first we need a way
     to register the physical-type signature of the function:
     C.f. Code_builtin_fn.select_physical_op and
     Code_builtin_fn.build_builtin_fn_code.
  *)
  let (result_f,result_ch) = Filename.open_temp_file "res" ".xml" in
  output_string result_ch res;
  close_out result_ch;
  (* Write result to a file, register the file, and re-parse. *)
  let result = Fn_doc.lookup_doc_function_no_table result_f proc_ctxt in
  if Sys.file_exists result_f then   Sys.remove result_f;
  Cursor.cursor_of_list(result)
  
let _ = Fn_doc.register_back_end "dxq" dxq_back_end
let _ = Code_fn.add_bltin_fctn ((glx_remote_query, 4), _glx_remote_query)
let _ = Code_fn.add_bltin_fctn ((Namespace_builtin.glx_remote_plan, 3), _glx_remote_plan)

