(**********************************************************************)
(* Trevor Jim                                                         *)
(*                                                                    *)
(* Galaxd: stand-alone Galax server that communicates over HTTP.      *)
(**********************************************************************)

(* Typically, you will invoke galaxd

       galaxd a.xq b.xq ...

   This starts up a server listening on the default port, 3000.  The
   files a.xq, b.xq, ... are the context for evaluation, and they are
   optional.

   You can submit a query to the server by visiting
   http://localhost:3000/ with your web browser.  This will give you a
   form to submit an xquery expression for evaluation.

   The galaxd server can also be queried by a galax process running
   over the network.  For example, if you run galaxd on host
   alice.example.com, you can use the following xquery expression on
   another host bob.example.com:

       fn:doc("dxq://alice.example.com/")

   This will cause bob.example.com to evaluate the query

       local:main()

   using the galaxd server on alice.example.com.  In this case, the
   context files a.xq, b.xq, ... must define the function
   local:main().

   You may use the -port flag to start the server on a non-default
   port, e.g.,

       galaxd -port 8000

   will start the server on port 8000 instead of port 3000.

   It is sometimes useful to simulate a virtual network of galaxd
   servers.  Use the -d flag to do this:

       galaxd -s dir

   Here, dir must be a directory containing xquery files, on per host
   that you want to simulate.  For example, if dir contains files

       a.xq, b.xq, c.xq, d.xq, e.xq

   then galaxd will simulate a network of five hosts named a, b, c, d,
   and e on localhost.  Within the programs a.xq, b.xq, ...,

       fn:doc("dxq://a/")

   will evaluate local:main() on the server for a, and similarly

       fn:doc("dxq://b/")

   will evaluate local:main() on the server for b, etc.

   The server for a will actually be listening to port 3000 on
   localhost, b will be on port 3001, etc.  These ports can be used to
   submit queries to the servers using a web browser.

*)
(***********************************************************************)
(* taken from galax-run.ml                                             *)
(***********************************************************************)

open Format
open Print_top
open Error
open Dm

open Galax_server

open Processing_context
open Monitoring_context

open Top_util
open Top_config
open Top_options

let compile_library_module proc_ctxt xq_program = 
  (* prepare input file *)
    let (in_file,in_channel) = Filename.open_temp_file "in" ".xq" in
    output_string in_channel xq_program;
    output_string in_channel "\n";
    close_out in_channel;

    let mod_ctxt = init_all proc_ctxt in

    let compiled_prolog = Galax.import_library_module mod_ctxt (Galax_io.File_Input in_file) in 
(*    if Sys.file_exists in_file then Sys.remove in_file;  *)
    compiled_prolog 

let eval compiled_prolog library_module xquery_kind query =
  let result = 
    try
      let proc_ctxt = Galax.procctxt_from_compiled_prolog compiled_prolog in
      Monitor.start_monitor_external_call proc_ctxt "Galaxd.main";

      (* set output to temp file *)
      let out_file = Filename.temp_file "out" ".xml" in
      arg_output_xml proc_ctxt out_file;

      (* Mary & Doug: This isn't very efficient because the global
         prolog is re-evaluated every time a client query is
         evaluated.  We should evaluate the server module (aka prolog)
         once and use that context to evaluate the client query.  
      *)
      let v = 
	match xquery_kind with
	| XQueryString -> 
	    let prepared_prolog = Galax.eval_prolog compiled_prolog None in
	    Galax.eval_statement prepared_prolog (Galax_io.String_Input query) 
	| XQueryPlan ->
(*	We should support both kinds of compilation here: modules and statements
    let (compiled_prolog', stmt_list) = 
	      Galax.compile_serialized_logical_main_module compiled_prolog (Galax_io.String_Input query) in 
*)
	    let stmt_list = 
	      [Galax.compile_serialized_logical_statement compiled_prolog (Galax_io.String_Input query)] in
	    let prepared_prolog = Galax.eval_prolog compiled_prolog None in
	    begin
	      match stmt_list with
	      |	[stmt] -> Galax.eval_compiled_statement prepared_prolog (List.hd stmt_list)
	      |	_ -> raise (Query(Internal_Error("Invalid number of statements to execute: "^(string_of_int(List.length stmt_list)))))
	    end
      in
      if !Conf.print_xml then
	Galax.serialize proc_ctxt (Galax_io.Channel_Output !Conf.xml_output) v
      else ();

      Monitor.end_monitor_external_call proc_ctxt;
      Monitor.serialize_monitor proc_ctxt;

      let result = Sim.string_of_file out_file in
      if Sys.file_exists out_file then Sys.remove out_file; 
      result
    with e ->
      let errmsg = Galax_server.server_error_msg e in
      Gui.gui_report(Gui.GuiNode(!Server.my_host_port,"x",errmsg,[]));
      Printf.eprintf "%s: %s\n" !Server.my_host_port errmsg;
      flush stderr;
      errmsg 
  in
  Register_handlers.call_close_handlers ();
  result

let rec sim_loop proc_ctxt l =
  match l with
    [] -> ()
  | [(host,port,xq_program)] ->
      (Printf.eprintf "galaxd: starting server for %s on port %d\n" host port;
       flush stderr;
       let proc_ctxt' = (default_processing_context()) in 
       let _ = Processing_context.set_proc proc_ctxt' proc_ctxt in 
       let compiled_prolog = compile_library_module proc_ctxt' xq_program in 
       Server.go port (eval compiled_prolog) xq_program)
  | (host,port,xq_program)::tl ->
      let pid = Unix.fork() in
      if pid <> 0 then sim_loop proc_ctxt tl else
      (Printf.eprintf "galaxd: starting server for %s on port %d\n" host port;
       flush stderr;
       let proc_ctxt' = (default_processing_context()) in 
       let _ = Processing_context.set_proc proc_ctxt' proc_ctxt in 
       let compiled_prolog = compile_library_module proc_ctxt' xq_program in 
       Server.go port (eval compiled_prolog) xq_program)

(************************)
(* Command-line options *)
(************************)

let port = ref Server.default_port
let simdir = ref None (* default no simulation *) 
let files = ref []
;;
begin
try
   (* Mary: I think each server should have its own copy of the processing context. *)
   let proc_ctxt = galax_run_proc_ctxt () in

  (* specify xml output*)
    Processing_context.set_serialization_kind proc_ctxt
      Processing_context.Serialize_As_Well_Formed ;

  (* functions to build options *)
  let bo f = f proc_ctxt in
  let bo_string f = Arg.String (bo f) in
  Arg.parse
    [
     ("-port",(Arg.Set_int port),"Listen on the given port");
     ("-debug", (Arg.Unit (fun () -> Server.debug := true)), "Enable debugging output ");
     ("-s",
      Arg.String (fun d -> simdir := Some d),
      " <d>  Simulate directory <d> on localhost");
  (* Monitoring options *)
      opt_monitor, 	  (bo_string arg_monitor), 	      msg_monitor;
      opt_monitor_mem,    (bo_string arg_monitor_mem),    msg_monitor_mem;
      opt_monitor_time,   (bo_string arg_monitor_time),   msg_monitor_time;
      opt_output_monitor, (bo_string arg_output_monitor), msg_output_monitor ^ title_character_encoding_options;
   ]
    (fun x -> files := x :: !files)
    "Usage: galaxd [-port n] [-s dir] [-debug] files";

  match !simdir with
    None ->
      (* xq_program : XQuery Server Module that is exported by the Galax daemon *)
      let xq_program =
        String.concat "\n" (List.map Sim.string_of_file !files) in
      Printf.eprintf "galaxd: starting server on port %d\n" !port;
      flush stderr;
      let compiled_prolog = compile_library_module proc_ctxt xq_program in 
      Server.go !port (eval compiled_prolog) xq_program
  | Some d ->
      (* d : Directory containing set of XQuery Server Modules that
             are exported by the Galax daemon *)
      Server.pm := Sim.portmap d !port;
      sim_loop proc_ctxt !Server.pm
  with
  | e -> 
      begin
	eprintf_error "  " e;
	fprintf (!Conf.glx_err_formatter) "@."
      end
end
