(***********************************************************************)
(*                                                                     *)
(*                               Galax                                 *)
(*                            XQuery Engine                            *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: top_util.ml,v 1.17 2006/02/02 21:03:17 mff Exp $ *)

(* Module: Top_util
   Description:
     This module implements some support functions for Galax
     command-line tools.
 *)

open Format

open Conf
open Print_top
open Error

open Xquery_common_ast
open Xquery_ast
open Xquery_core_ast

open Processing_context

open Top_config


(****************)
(* I/O handling *)
(****************)

let init_output_refs f f_out f_formatter = 
  f_out := open_out(f); 
  f_formatter := Format.formatter_of_out_channel (!f_out)

let close_channel_ref chan =
  if (!chan != stdout) then
    begin
      close_out (!chan);
      chan := stdout;
    end


(***********************)
(* Top-level execution *)
(***********************)

(* A catch-all exception handler *)

let exec func proc_ctxt arg =
  try
    begin
      try
	func proc_ctxt arg;
      with
      | Fn_error.Xquery_error msg -> Fn_error.downgrade_error proc_ctxt msg
    end; 
    Monitor.serialize_monitor proc_ctxt;
  with
  | e ->
      begin
	Monitor.serialize_monitor proc_ctxt;
	eprintf_error "  " e;
	fprintf (!Conf.glx_err_formatter) "@."
      end

let low_exec func () =
  try
    func ()
  with
  | e ->
      begin
	eprintf_error "  " e;
	fprintf (!Conf.glx_err_formatter) "@."
      end

(************)
(* Printing *)
(************)

let separator =
  ref "\n\n------------------------------------------------------------------------------\n\n"

let print_processing_file file =
  if (!Conf.print_global && !Conf.verbose) then
    begin
      printf "---------------------------------------\n";
      printf "Processing file: %s\n" file;
      printf "---------------------------------------\n";
      printf "%s" !separator;
      flush stdout;
    end


(******************)
(* Initialization *)
(******************)

let galax_run_proc_ctxt () =
  let proc_ctxt = default_processing_context () in
  begin
    set_serialization_kind proc_ctxt Serialize_As_XQuery;
    Conf.print_xml := true;
    proc_ctxt
  end

let galax_compile_proc_ctxt () =
  let proc_ctxt = default_processing_context () in
  begin
    set_evaluation_phase_arg proc_ctxt "off";
    Conf.print_optimized_algebra := true;
    proc_ctxt
  end

let init_all proc_ctxt =
  begin
    (* Check the given user's configuration *)
    Processing_context.check_config proc_ctxt;
    Galax.load_standard_library proc_ctxt 
  end


(*******************************)
(* Set up the external context *)
(*******************************)

(* Set up external context *)
(* Load documents bound to global variables *)

let set_up_external_context proc_ctxt =
  let (ext_ctxt_item,opt_ctxt_item) =
    match !context_item with
    | None -> (false,None)
    | Some f ->
	let actual_item =
	  List.hd (Galax.load_document proc_ctxt (Galax_io.File_Input f))
	in (true,Some (Physical_value.Item_Node actual_item))
  in
  let var_vals =
    List.map
      (fun (v, f) -> (v, (Physical_value.Item_Atomic (new Dm_atomic.atomicUntyped(f))) :: []))
      !global_vars
  in
  let var_docs =
    List.map
      (fun (v, f) -> (v, List.map (fun n -> (Physical_value.Item_Node n)) (Galax.load_document proc_ctxt (Galax_io.File_Input f)))) 
      !global_docs
  in
  (* Build external context of global variables and values *)
  (* Note:
       What about the types of external variables?  Where are they
       defined?
   - Jerome *)
  (ext_ctxt_item,
   Galax.build_external_context
     proc_ctxt
     opt_ctxt_item
     (Some (new Dm_atomic.atomicDayTimeDuration(DateTime.local_timezone())))
     (var_vals @ var_docs))


(***********************)
(* Compilation helpers *)
(***********************)

let compile_main_module_helper opt_ext_ctxt mod_ctxt input =
  if !(Conf.execute_logical_plan)
  then
    Galax.compile_serialized_logical_main_module mod_ctxt input
  else if !(Conf.execute_optimized_plan)
  then
    Galax.compile_serialized_optimized_main_module mod_ctxt input
  else
    Galax.import_main_module opt_ext_ctxt mod_ctxt input

