(***********************************************************************)
(*                                                                     *)
(*                               Galax                                 *)
(*                            XQuery Engine                            *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)




(** 
  @(#)wsdl_load.mlp

  Takes a data model instance containing a wsdl file 
  and builds an instance of wsdl AST.

  @author Nicola Onose
*)


open Error

open Namespace_names
open Namespace_builtin
open Namespace_symbols
open Namespace_context

open Streaming_types
open Streaming_util
open Xquery_type_ast

open Wsdl_ast
open Wsdl_util



			      
(**
  extract the value of the  attribute name_attr (as a string) 
  and the list of values of the xmlns attributes from the
  list of attributes att_el
*)

let get_attr att_el name_attr =
  let rec get_nx n =
    function
      [] -> n
    | (rsym, t) :: rest ->
        match rattr_prefix rsym with
          NSDefaultElementPrefix ->
            begin match rattr_name rsym with
              _, _, name ->
                if name = name_attr then get_nx (Some t) rest
                else get_nx n rest
            end
        | p -> print_endline (string_of_prefix p); get_nx n rest
  in
  get_nx None att_el
	    
(**
  extract the value the attribute "name" and the 
  values of the xmlns attributes from the list of attributes
  att_el
*)

let get_name attl =
  match get_attr attl "name" with
    None -> raise (Query (Load_Error "missing name attribute"))
  | Some n -> n


let checked_uqname_of_string str =
  try uqname_element_of_string str with
    _ -> raise (Query (Load_Error ("Bad uqname: " ^ str ^ "\n")))

let build_rqname str delta =
  let (p, loc) = checked_uqname_of_string str in
  match p with
    NSPrefix _ | NSDefaultElementPrefix ->
      begin try p, List.assoc p delta, loc with
        Not_found ->
          raise (Query (Load_Error ("Bad prefix or URI in " ^ str ^ "\n")))
      end
  | _ -> raise (Query (Load_Error ("Bad prefix or URI in " ^ str ^ "\n")))


let mk_special_prefix base_name env =
  let rec new_var v i =
    function
      [] -> v
    | x :: rest ->
        if v = x then new_var (base_name ^ string_of_int (i + 1)) (i + 1) rest
        else new_var v i rest
  in
  new_var base_name 0
    (List.sort Pervasives.compare (List.map (fun (x, _) -> x) env))
      

(** 
  * generate a name for the SOAP envelope nms, needed for the 'return' part of the query 
  *
**)
let mk_soapenv_prefix = mk_special_prefix default_soapenv_prefix_name

let mk_xsi_prefix = mk_special_prefix default_xsi_prefix_name
let mk_xsd_prefix = mk_special_prefix default_xsd_prefix_name
      

	  
(**
  extract the value of the attribute name_attr (as an rqname) and the 
  values of the xmlns attributes from the list of attributes  att_el
*)
	  
let rec get_name_and_rqname delta attr_name n d =
  function
    [] -> n, d
  | (rel, t) :: rest when rattr_prefix rel = NSDefaultElementPrefix ->
      begin match rattr_name rel with
        _, _, name ->
          if name = "name" then
            get_name_and_rqname delta attr_name (Some t) d rest
          else if name = attr_name then
            get_name_and_rqname delta attr_name n
              (Some (build_rqname t delta)) rest
          else get_name_and_rqname delta attr_name n d rest
      end
  | _ :: rest -> get_name_and_rqname delta attr_name n d rest
	



(**
  create a wsdl_module record using the attributes attl
  and the name of the service
*)

let make_module attl serv_name env =
  let tn = get_attr attl "targetNamespace" in
  match tn with
    None -> raise (Query (Load_Error "missing targetNamespace"))
  | Some target ->
      {wsdl_name = serv_name; global_xmlns = env;
       targetNamespace = Datatypes_util.anyURI_of_untyped target;
       wsdl_services = []; wsdl_types = None; wsdl_bindings = [];
       wsdl_port_types = []; wsdl_messages = [];
       soapenv_prefix_name = "soapenv"; xsi_prefix_name = "xsi";
       xsd_prefix_name = "xsd"}


(**
  extract the value of the 'message' attribute as a uqname from a list of attributes
*)

let rec get_message_attr delta =
  function
    (rsym, t) :: _
    when
      rqname_equal (rattr_name rsym) (empty_prefix, empty_uri, "message") ->
      build_rqname t delta
  | _ :: rest -> get_message_attr delta rest
  | [] ->
      raise
        (Query
           (Load_Error "missing message attribute in operation subcomponent"))


(**
  create a wsdl_service record using the attributes attl
*)

let make_service attl =
  let n = get_name attl in {service_name = n; ports = []}
    
(**
  create a operation_in_port_type record using the attributes attl
*)

let make_operation_pt attl =
  let n = get_name attl in
  {operation_name = n; input = None; output = None; faults = []}

(**
  create a operation_in_binding record using the attributes attl
*)

let make_binding_operation attl =
  let n = get_name attl in
  {operation_in_binding_name = n; binding_input = None; binding_output = None;
   binding_faults = []; binding_soap_operation = None;
   http_relative_uri = None}

(**
  create a port_type record using the attributes attl
*)

let make_portType attl =
  let n = get_name attl in
  let ext =
    try
      let (_, content) =
        List.find
          (fun a ->
             match a with
               _, "extends" -> true
             | _ -> false)
          attl
      in
      Some content
    with
      Not_found -> None
  in
  {port_type_name = n; operations = []; extends = ext}
    (**
      create a message record using the attributes attl
    *)

let make_message attl =
  let n = get_name attl in {message_name = n; message_parts = []}

(**
  create a 'part' record using the attributes attl
*)

let make_part att_el nsenv =
  let rec get_ntx n d =
    function
      [] -> n, d
    | (rsym, t) :: rest when rattr_prefix rsym = NSDefaultElementPrefix ->
        begin match rattr_name rsym with
          _, _, name ->
            if name = "name" then get_ntx (Some t) d rest
            else if name = "element" or name = "type" then
              let qn = uqname_element_of_string t in
              match qn with
                p, s ->
                  let u = get_ns_of_prefix nsenv p in
                  if name = "element" then
                    get_ntx n (Some (PartElement (u, qn))) rest
                  else get_ntx n (Some (PartType (u, qn))) rest
            else get_ntx n d rest
        end
    | _ :: rest -> get_ntx n d rest
  in
  match get_ntx None None att_el with
    Some nm, Some typ -> {part_name = nm; part_type = typ}
  | None, _ ->
      raise
        (Query (Load_Error "Missing `name' attribute in `part' component"))
  | _, None ->
      raise
        (Query
           (Load_Error
              "Missing `element' or `type' attribute in 'part' component"))


(**
  create a service_binding record using the attributes attl
*)

let make_binding delta att_el =
  match get_name_and_rqname delta "type" None None att_el with
    Some nm, Some typ ->
      {binding_name = nm; binding_type = typ; bind_soap = None;
       binding_operations = []; http_method = "POST"}
  | None, _ ->
      raise
        (Query (Load_Error "Missing `name' attribute in `binding' component"))
  | _, None ->
      raise
        (Query (Load_Error "Missing `type' attribute in `binding' component"))


(**
  create a soap_binding record using the attributes attl
*)

let make_soap_binding att_el nms =
  let rec get_params transp stl enc =
    function
      [] -> transp, stl, enc
    | (rsym, t) :: rest ->
        match rattr_prefix rsym with
          NSDefaultElementPrefix ->
            begin match rattr_name rsym with
              _, _, "transport" ->
                get_params (Some (Datatypes_util.anyURI_of_untyped t)) stl enc
                  rest
            | _, _, "style" | _, _, "styleDefault" ->
                if t = "rpc" then get_params transp (Some Soap_RPC) enc rest
                else if t = "document" then
                  get_params transp (Some Soap_Document) enc rest
                else raise (Query (Load_Error ("Unknown soap style: " ^ t)))
            | _, _, encodingstyle_default ->
                get_params transp stl
                  (Some (Datatypes_util.anyURI_of_untyped t)) rest
            end
        | _ -> get_params transp stl enc rest
  in
  let (t, s, e) = get_params None None None att_el in
  {nms_prefix = nms; transport = t; soap_binding_style = s;
   encodingStyleDefault = e}
    
    
(**
  create a soap_operation record using the attributes attl
*)
    
let make_soap_operation att_el nms_prefix =
  let rec get_params act stl =
    function
      [] -> act, stl
    | (rsym, t) :: rest ->
        match rattr_prefix rsym with
          NSDefaultElementPrefix ->
            begin match rattr_name rsym with
              _, _, "soapAction" ->
                get_params (Some (Datatypes_util.anyURI_of_untyped t)) stl
                  rest
            | _, _, "style" ->
                if t = "rpc" then get_params act (Some Soap_RPC) rest
                else if t = "document" then
                  get_params act (Some Soap_Document) rest
                else raise (Query (Load_Error ("Unknown soap style: " ^ t)))
            | _ -> get_params act stl rest
            end
        | _ -> get_params act stl rest
  in
  match get_params None None att_el with
    Some sa, stl ->
      {nms_soap_operation = nms_prefix; soapAction = Some sa;
       soap_operation_style = stl}
  | None, _ -> raise (Query (Load_Error "Missing soapAction"))    
	  
	  
(**
  create a soap_body record using the attributes attl
*)

let make_soap_body nms_prefix att_el =
  let rec get_params es u nms =
    function
      [] -> es, u, nms
    | (rsym, t) :: rest ->
        match rattr_prefix rsym with
          NSDefaultElementPrefix ->
            begin match rattr_name rsym with
              _, _, "encodingStyle" ->
                get_params (Some (Datatypes_util.anyURI_of_untyped t)) u nms
                  rest
            | _, _, "use" ->
                if t = "literal" then get_params es (Some Literal) nms rest
                else if t = "encoded" then
                  get_params es (Some Encoded) nms rest
                else
                  raise (Query (Load_Error ("Unknown soap body use: " ^ t)))
            | _, _, "namespace" ->
                get_params es u (Some (Datatypes_util.anyURI_of_untyped t))
                  rest
            | _ -> get_params es u nms rest
            end
        | _ -> get_params es u nms rest
  in
  let (es, u, nms) = get_params None None None att_el in
  {soap_body_encodingStyle = es; use = u; soap_body_nms = nms}

let make_soap_header nms_prefix att_el =
  let rec get_params h =
    function
      [] -> h
    | (rsym, t) :: rest ->
        match rattr_prefix rsym with
          NSDefaultElementPrefix ->
            begin match rattr_name rsym with
              _, _, "element" ->
                get_params
                  {h with
                    soap_header_element = Some (checked_uqname_of_string t)}
                  rest
            | _, _, "type" ->
                get_params
                  {h with
                    soap_header_type = Some (checked_uqname_of_string t)}
                  rest
            | _, _, "localname" ->
                get_params {h with soap_header_localname = Some t} rest
            | _, _, "namespace" ->
                get_params
                  {h with
                    soap_header_nms =
                      Some (Datatypes_util.anyURI_of_untyped t)}
                  rest
            | _, _, "encodingStyle" ->
                get_params
                  {h with
                    soap_header_encodingStyle =
                      Some (Datatypes_util.anyURI_of_untyped t)}
                  rest
            | _, _, "role" ->
                get_params
                  {h with
                    soap_header_role =
                      Some (Datatypes_util.anyURI_of_untyped t)}
                  rest
            | _ -> get_params h rest
            end
        | _ -> get_params h rest
  in
  get_params
    {soap_header_element = None; soap_header_type = None;
     soap_header_localname = None; soap_header_nms = None;
     soap_header_encodingStyle = None; soap_header_role = None}
    att_el

      
let make_soap_headerfault nms_prefix att_el =
  let rec get_params h =
    function
      [] -> h
    | (rsym, t) :: rest ->
        match rattr_prefix rsym with
          NSDefaultElementPrefix ->
            begin match rattr_name rsym with
              _, _, "message" ->
                get_params
                  {h with
                    soap_headerfault_message = checked_uqname_of_string t}
                  rest
            | _, _, "part" ->
                get_params {h with soap_headerfault_part = t} rest
            | _, _, "namespace" ->
                get_params
                  {h with
                    soap_headerfault_nms =
                      Some (Datatypes_util.anyURI_of_untyped t)}
                  rest
            | _, _, "encodingStyle" ->
                get_params
                  {h with
                    soap_headerfault_encodingStyle =
                      Some (Datatypes_util.anyURI_of_untyped t)}
                  rest
            | _ -> get_params h rest
            end
        | _ -> get_params h rest
  in
  get_params
    {soap_headerfault_message = NSDefaultElementPrefix, "";
     soap_headerfault_part = ""; soap_headerfault_nms = None;
     soap_headerfault_encodingStyle = None}
    att_el
      

let make_soap_fault nms_prefix att_el =
  let rec get_params name nms enc =
    function
      [] -> name, nms, enc
    | (rsym, t) :: rest ->
        match rattr_prefix rsym with
          NSDefaultElementPrefix ->
            begin match rattr_name rsym with
              _, _, "name" -> get_params t nms enc rest
            | _, _, "namespace" ->
                get_params name (Some (Datatypes_util.anyURI_of_untyped t))
                  enc rest
            | _, _, "encodingStyle" ->
                get_params name nms
                  (Some (Datatypes_util.anyURI_of_untyped t)) rest
            | _ -> get_params name nms enc rest
            end
        | _ -> get_params name nms enc rest
  in
  let (faultname, nms, enc) = get_params "" None None att_el in
  {soap_fault_prefix = nms_prefix; soap_fault_name = faultname;
   soap_fault_nms = nms; soap_fault_encodingStyle = enc}
    
    
let void_soap_fault =
  {soap_fault_prefix = NSDefaultElementPrefix; soap_fault_name = "";
   soap_fault_nms = None; soap_fault_encodingStyle = None}


(**
  create a service_port record using the attributes attl
*)

let make_port delta att_el =
  match get_name_and_rqname delta "binding" None None att_el with
    Some nm, Some typ ->
      {service_port_name = nm; service_port_binding = typ;
       port_soap_address = None; http_base_uri = None}
  | None, _ ->
      raise (Query (Load_Error "Missing name attribute in port element"))
  | _, None ->
      raise (Query (Load_Error "Missing binding attribute in port element"))
	
	
(**
  create a soap_address record using the attributes attl
*)

let make_soap_address nms attl =
  let rec find_loc loc =
    function
      [] -> loc
    | (rsym, t) :: rest ->
        match rattr_prefix rsym with
          NSDefaultElementPrefix ->
            begin match rattr_name rsym with
              _, _, "location" -> Some (Datatypes_util.anyURI_of_untyped t)
            | _ -> find_loc loc rest
            end
        | _ -> find_loc loc rest
  in
  match find_loc None attl with
    Some addr -> {nms_soap_address = nms; location = addr}
  | None -> raise (Query (Load_Error "Missing "))	


(**
  parse till the end of an element and throw everything to garbage
*)

let rec parse_element_content (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some RSAX_endElement -> Stream.junk strm__; ()
  | Some (RSAX_startElement (_, att_el, _, _, _)) ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try parse_element_content strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some x ->
      Stream.junk strm__;
      begin try parse_element_content strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> raise Stream.Failure
	
(**
  main parser: parse the root element
*)
	
let rec parse_doc proc_ctxt serv_name (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startDocument (xml_dec, dtd_dec, base_uri)) ->
      Stream.junk strm__;
      let d =
        try parse_doc proc_ctxt serv_name strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endDocument -> Stream.junk strm__; d
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rel, att_el, has, _, env)) ->
      Stream.junk strm__;
      let d =
        try
          parse_component proc_ctxt
            (make_module att_el serv_name (active_bindings env)) strm__
        with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin match relem_name rel with
            _, uri, "definitions" when uri = wsdl_uri -> d
          | _ ->
              raise (Query (Load_Error "Root element should be definitions"))
          end
      | _ -> raise (Stream.Error "")
      end
  | _ -> raise Stream.Failure
and parse_component proc_ctxt doc (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_component proc_ctxt doc strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_component proc_ctxt doc strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "types") ->
      Stream.junk strm__;
      let t =
        try parse_types proc_ctxt doc strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        parse_component proc_ctxt {doc with wsdl_types = t} strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "portType") ->
      Stream.junk strm__;
      let pt =
        try parse_portType (make_portType att_el) strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_component proc_ctxt
              {doc with wsdl_port_types = pt :: doc.wsdl_port_types} strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "message") ->
      Stream.junk strm__;
      let m =
        try parse_message (make_message att_el) strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_component proc_ctxt
              {doc with wsdl_messages = m :: doc.wsdl_messages} strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "binding") ->
      Stream.junk strm__;
      let b =
        try
          parse_binding (make_binding (active_bindings env) att_el) strm__
        with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_component proc_ctxt
              {doc with wsdl_bindings = b :: doc.wsdl_bindings} strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "service") ->
      Stream.junk strm__;
      let s =
        try parse_service (make_service att_el) strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_component proc_ctxt
              {doc with wsdl_services = s :: doc.wsdl_services} strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (_, att_el, has, _, _)) ->
      Stream.junk strm__;
      let doc2 =
        try parse_component proc_ctxt doc strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_component proc_ctxt doc2 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | _ -> flush Pervasives.stdout; doc
and parse_types proc_ctxt wsdl_mod resolved_xml_stream =
  let wrap x =
    try Some (fmkrse_event (Stream.next resolved_xml_stream) Finfo.bogus) with
      Failure _ -> None
  in
  let resolved_xml_stream = Stream.from wrap in
  let xschema =
    Schema_import.import_schema_element proc_ctxt None
      (Cursor.cursor_of_stream resolved_xml_stream)
  in
  wsdl_mod.soapenv_prefix_name <-
    mk_soapenv_prefix xschema.xschema_namespace_declarations;
  wsdl_mod.xsi_prefix_name <-
    mk_xsi_prefix
      ((wsdl_mod.xsi_prefix_name, NSWildcardUri) ::
         xschema.xschema_namespace_declarations);
  wsdl_mod.xsd_prefix_name <-
    mk_xsd_prefix
      ((wsdl_mod.xsd_prefix_name, NSWildcardUri) ::
         xschema.xschema_namespace_declarations);
  Some xschema
and parse_portType pt (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, has, _, env))
    when
      rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "operation") ->
      Stream.junk strm__;
      let op1 =
        try
          parse_operation_in_port_type (make_operation_pt att_el) strm__
        with
          Stream.Failure -> raise (Stream.Error "")
      in
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        parse_portType {pt with operations = op1 :: pt.operations} strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let pt1 =
        try parse_portType pt strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_portType pt1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_portType pt strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_portType pt strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> pt
and parse_operation_in_port_type op (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, _, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "input") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        (let m = get_message_attr (active_bindings env) att_el in
         parse_operation_in_port_type {op with input = Some m})
          strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, _, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "output") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        (let m = get_message_attr (active_bindings env) att_el in
         parse_operation_in_port_type {op with output = Some m})
          strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, _, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "fault") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        (let m = get_message_attr (active_bindings env) att_el in
         parse_operation_in_port_type {op with faults = m :: op.faults})
          strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let op1 =
        try parse_operation_in_port_type op strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_operation_in_port_type op1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_operation_in_port_type op strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_operation_in_port_type op strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> op
and parse_message m (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, _, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "part") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      let m2 =
        try
          (let part1 =
             make_part att_el (add_all_ns empty_nsenv (active_bindings env))
           in
           parse_message {m with message_parts = part1 :: m.message_parts})
            strm__
        with
          Stream.Failure -> raise (Stream.Error "")
      in
      flush Pervasives.stdout; m2
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let m1 =
        try parse_message m strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_message m1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_message m strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_message m strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> m
and parse_binding b (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, _, _, _))
    when rqname_equal (relem_name rsym) (http_prefix, http_uri, "binding") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        (let m =
           let meth = get_attr att_el "verb" in
           match meth with
             None -> ""
           | Some s -> s
         in
         parse_binding {b with http_method = m})
          strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, _, _, _))
    when rqname_equal (relem_name rsym) (soap_prefix, soap_uri, "binding") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        (let sb = make_soap_binding att_el (NSPrefix "soap") in
         parse_binding {b with bind_soap = Some sb})
          strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, _, _, _))
    when
      rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "operation") ->
      Stream.junk strm__;
      let bo =
        try
          parse_binding_operation (make_binding_operation att_el) strm__
        with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_binding
              {b with binding_operations = bo :: b.binding_operations} strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let b1 =
        try parse_binding b strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_binding b1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_binding b strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_binding b strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> b
and parse_binding_operation bo (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, has, _, _))
    when
      rqname_equal (relem_name rsym) (http_prefix, http_uri, "operation") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        (let addr = get_attr att_el "location" in
         parse_binding_operation {bo with http_relative_uri = addr})
          strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, _))
    when
      rqname_equal (relem_name rsym) (soap_prefix, soap_uri, "operation") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        (let sb = make_soap_operation att_el (NSPrefix "soap") in
         parse_binding_operation {bo with binding_soap_operation = Some sb})
          strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, _))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "input") ->
      Stream.junk strm__;
      let sb =
        try parse_soap_body_and_headers None [] [] strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_binding_operation {bo with binding_input = Some sb} strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, _))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "output") ->
      Stream.junk strm__;
      let sb =
        try parse_soap_body_and_headers None [] [] strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_binding_operation {bo with binding_output = Some sb} strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, _))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "fault") ->
      Stream.junk strm__;
      let sf =
        try parse_soap_fault void_soap_fault strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_binding_operation
              {bo with binding_faults = sf :: bo.binding_faults} strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let b1 =
        try parse_binding_operation bo strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_binding_operation b1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_binding_operation bo strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_binding_operation bo strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> bo
and parse_soap_body_and_headers sb sh shf (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, has, _, _))
    when rqname_equal (relem_name rsym) (soap_prefix, soap_uri, "body") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        parse_soap_body_and_headers
          (Some (make_soap_body (NSPrefix "soap") att_el)) sh shf strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, _))
    when rqname_equal (relem_name rsym) (soap_prefix, soap_uri, "header") ->
      Stream.junk strm__;
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            parse_soap_body_and_headers sb
              (make_soap_header (NSPrefix "soap") att_el :: sh) shf strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let (b1, h1, hf1) =
        try parse_soap_body_and_headers sb sh shf strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_soap_body_and_headers b1 h1 hf1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_soap_body_and_headers sb sh shf strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_soap_body_and_headers sb sh shf strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> sb, sh, shf
and parse_soap_fault sf (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, has, _, _))
    when rqname_equal (relem_name rsym) (soap_prefix, soap_uri, "fault") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        parse_soap_fault (make_soap_fault (NSPrefix "soap") att_el) strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let f1 =
        try parse_soap_fault sf strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_soap_fault f1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_soap_fault sf strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_soap_fault sf strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> sf
and parse_service s (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, has, _, env))
    when rqname_equal (relem_name rsym) (wsdl_prefix, wsdl_uri, "port") ->
      Stream.junk strm__;
      let p =
        try parse_port (make_port (active_bindings env) att_el) strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_service {s with ports = p :: s.ports} strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let s1 =
        try parse_service s strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_service s1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_service s strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters cars) ->
      Stream.junk strm__;
      begin try parse_service s strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> s
and parse_port p (strm__ : _ Stream.t) =
  match Stream.peek strm__ with
    Some (RSAX_startElement (rsym, att_el, has, _, _))
    when rqname_equal (relem_name rsym) (http_prefix, http_uri, "address") ->
      Stream.junk strm__;
      let _ =
        try parse_element_content strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin try
        (let addr = get_attr att_el "location" in
         parse_port {p with http_base_uri = addr})
          strm__
      with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement (rsym, att_el, has, _, _))
    when rqname_equal (relem_name rsym) (soap_prefix, soap_uri, "address") ->
      Stream.junk strm__;
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try
            (let sa = make_soap_address (NSPrefix "soap") att_el in
             parse_port {p with port_soap_address = Some sa})
              strm__
          with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_startElement _) ->
      Stream.junk strm__;
      let p1 =
        try parse_port p strm__ with
          Stream.Failure -> raise (Stream.Error "")
      in
      begin match Stream.peek strm__ with
        Some RSAX_endElement ->
          Stream.junk strm__;
          begin try parse_port p1 strm__ with
            Stream.Failure -> raise (Stream.Error "")
          end
      | _ -> raise (Stream.Error "")
      end
  | Some (RSAX_comment c) ->
      Stream.junk strm__;
      begin try parse_port p strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | Some (RSAX_characters s) ->
      Stream.junk strm__;
      begin try parse_port p strm__ with
        Stream.Failure -> raise (Stream.Error "")
      end
  | _ -> p


(**
  produce a wsdl_ast from an xml_stream
*)
	
let xml_to_wsdl_ast proc_ctxt name xml_stream =
  try
    let stream = Streaming_ops.resolve_xml_stream xml_stream in
    let stream = Cursor.stream_of_cursor stream in
    let unwrap x =
      try Some (Stream.next stream).rse_desc with
        Failure _ -> None
    in
    let stream = Stream.from unwrap in parse_doc proc_ctxt name stream
  with
    Stream.Error q ->
      raise
        (Query (Load_Error ("Error in the syntax of the wsdl file. " ^ q)))
	
let test_parser wsdl_file =
  let proc_ctxt = Processing_context.default_processing_context () in
  let (_, wsdl_stream) =
    Streaming_parse.open_xml_stream_from_io (Galax_io.File_Input wsdl_file)
  in
  xml_to_wsdl_ast proc_ctxt wsdl_file wsdl_stream
      
